!=========================================================================
!  Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
! 
!  GbjGciInterface Class for GBJ 3.0
!
!       GbjGciInterface.gs  
!
!=========================================================================


doit
Object subclass: 'GbjGciInterface'
  instVarNames: #( verbose classOrganizer mappedClassRoots clientForwarders delay )
  classVars: #( )
  classInstVars: #()
  poolDictionaries: #()
  inDictionary: Globals
%

! Remove existing behavior from GbjGciInterface
doit
GbjGciInterface removeAllMethods.
GbjGciInterface class removeAllMethods.
%
! ------------------- Class methods for GbjGciInterface
category: 'Private'
classmethod: GbjGciInterface
excludeFromGbjCollectionMapping 
  "Exclude following and subclasses thereof from results of allSubclassOopsFor: .
   We want these classes mapped to GbjObject not GbjCollection , because
   they cannot be enumerated with GciFetchOops or are considered private"
^  {
  AbstractDictionary  .
  DepListTable .
  GsPipe  .
  RcQueue .
  "Repository ok"
  "SequenceableCollection ok (" 
  "  Array ok ("
      AbstractCollisionBucket  .
      BlockSorter .
      BtreeNode  .
      BtreePlusNode  .
      ClassHistory .
      ClusterBucketArray .
      DependencyList .
      DepListBucket .
      ErrorDescription  .
      GsAbstractIndex  .
      GsCommitList .
      GsPathTerm  .
      GsSingleRefPathResult .
      GsStackBuffer .
      GsX509CertificateChain .
      HtInternalNode  .
      HtLeafNode  .
      IdentityIndex  .
      IndexDictionaryEntryHolder .
      IndexList .
      "InstVarMappingArray ok" 
      "InvariantArray ok"
      MappingInfo .
      "NonPersistentArray ok"
      NscBuilder .
      PathEvaluator .
      PathSorter .
      PathTerm  .
      QueryExecuter .
      "RcArray ok"
      RcQueueRemovalSeqNumbers .
      RcQueueSessionComponent .
      SortNodeArray .
      "SymbolList ok"
     "end Array ) "
    ByteArray  .
    CharacterCollection  .
    Interval  .
    "OrderedCollection ok"
       Semaphore .
    Uint16Array   "byte format" .
    Uint32Array  "byte format" .
    Uint64Array  "byte format" .
    "end SequenceableCollection  )" 
  "UnorderedCollection ok ( "
    Bag .
    "IdentityBag ok "
      BucketValueBag .
      "IdentitySet ok"
      RcIdentityBag .
      "RcLowMaintenanceIdentityBag ok"
    Set .
    TreeSet .
   "end UnorderedCollection )"
}
%

category: 'Creation'
classmethod: GbjGciInterface
initialize
  "executed on demand from  GbjGciInterface class >> current "
    | interface |
    interface := self new.
    interface verbose: false.
    interface classOrganizer: (ClassOrganizer new).
    interface mappedClassRoots: IdentitySet new.
    interface clientForwarders: IdentitySet new.
    interface delay: (Delay forMilliseconds: 1).
    System __sessionStateAt: self sessionStateIndex put: interface.
    ^ interface
%

category: 'Operations'
classmethod: GbjGciInterface
verbose

    ^ self current verbose
%

category: 'Operations'
classmethod: GbjGciInterface
poll

"poll for errors by tickling process queue using a delay wait,
 which allows pending processes (such as exception handlers)
 to run. This also allows any forked Smalltalk processes to run. "

    self current delay wait.
    ^ nil
%

! 44674
category: 'Operations'
classmethod: GbjGciInterface
chkG2G: chkG2G chkObjChg: chkObjChg

"This is a varation of poll that after the delay checks for
 gem-to-gem and/or object change notifications if selected by the flags"

    | result g2gArray g2gSig |
    self current delay wait.
    result := Array new: 2.
    chkG2G ifTrue: [
      g2gArray := Array new.
      [ (g2gSig := System _signalFromGemStoneSession) isNil ] whileFalse: [
           g2gArray add: g2gSig ].
      result at: 1 put: g2gArray ].
    chkObjChg ifTrue: [
      result at: 2 put: (System signaledObjects)].
    ^ result                     
%

category: 'Operations'
classmethod: GbjGciInterface
current
    | current |
    current := System __sessionStateAt: self sessionStateIndex.
    current ifNil:[ current := self initialize ].
    ^ current
%

category: 'Operations'
classmethod: GbjGciInterface
allSubclassOopsFor: aClass
  "Return an Array of SmallIntegers , elements are oop values"
  | list allSubclassesFor|
  list := self allSubclassesFor: aClass .
  1 to: list size do:[:n | list at: n put:(list at:n) asOop ].
  ^ list
%

category: 'Operations'
classmethod: GbjGciInterface
allSubclassesFor: aClass
  "Return an Array of classes"
   | organizer roots results inst |
   (inst := self current) verbose ifTrue: [
      GsFile gciLogServer: 
         'GbjGciInterface: calling allSubclassOopsFor: ' , (aClass asString) ].
   organizer := inst classOrganizer .
   roots := inst mappedClassRoots .
   results := Array new.
   results add: aClass  . roots add: aClass .
   self _allSubclassesFor: aClass into: results from: organizer roots: roots.
   ^ results
%

category: 'Private'
classmethod: GbjGciInterface
_allSubclassesFor:  rootClass into: anArray from: organizer roots: roots
    | sclasses excludeSet |
    excludeSet := IdentitySet new .
    self excludeFromGbjCollectionMapping do:[:aClass | 
      excludeSet add: aClass .
      excludeSet addAll: (organizer allSubclassesOf: aClass)].
    sclasses := organizer allSubclassesOf: rootClass.
    sclasses do: [ :aClass |
      (excludeSet includes: aClass) ifFalse:[
        (roots includes: aClass) ifFalse: [ | hist |
            anArray add: aClass . 
            roots add: aClass .
            ((hist := aClass classHistory) size > 1) ifTrue:[ | otherVers |
              otherVers := IdentitySet withAll: hist .
              otherVers remove: aClass otherwise: nil .
              otherVers do:[:xClass | anArray add: xClass . roots add: xClass ].
            ].
        ].
      ].
    ].
%

category: 'Operations'
classmethod: GbjGciInterface
clientForwarder: id

   | cf |
   self verbose ifTrue: [
      GsFile gciLogServer: 
         'GbjGciInterface: setting up clientForwarder for ' , (id asString) ].
   cf := ClientForwarder new.

   cf clientObject: id.
   self current clientForwarders add: cf.
   ^ cf   
%
category: 'Operations'
classmethod: GbjGciInterface
sessionStateIndex

" By default, GBJ uses the GS/S session state index 8 for holding 
  GBJ session information.  This index was previously assigned to
  GemEnterprise.  Customers still using GemEnterprise should modify
  this method to use a free customer session state index of 20 or above. "

    ^ 8
%


category: 'Operations'
classmethod: GbjGciInterface
knownSubclassesArray
  "Return an Array of pairs,   
       class  ,  class to use in setupReport function in GbjGciInterface.c
   excluding classes of special objects , for non numeric oop format classes
   that are understood by GbjGciInterface.c
  "
  | ary localSet | 
  ary := { } .
  localSet := IdentitySet new .
  localSet add: SmallDateAndTime .
  "first list are leaf classes in terms of GBJ class mapping"
  { GbjLocalDateTime . GbjZonedDateTime . GbjUtilDate . GbjInstant . DateTime . Fraction } do:[:aClass |
    localSet add: aClass .
    aClass allSubclasses do:[:cls |
      localSet add: cls .
      ary add: cls ; add: aClass .
    ].
    ary add: aClass ; add: aClass .
  ].

  { DateAndTime } do:[:aClass | 
     aClass allSubclasses do:[:cls |
       (localSet includes: cls) ifFalse:[  "exclude classes in first list"
         ary add: cls ; add: aClass .
       ]
     ].
     ary add: aClass; add: aClass .
  ].
  ^ ary
%

! ------------------- Instance methods for GbjGciInterface

category:  'Accessing'
method: GbjGciInterface
verbose
    ^ verbose
%

category:  'Accessing'
method: GbjGciInterface
classOrganizer

    ^ classOrganizer
%

category:  'Accessing'
method: GbjGciInterface
mappedClassRoots

    ^ mappedClassRoots
%

category: 'Accessing'
method: GbjGciInterface
clientForwarders

    ^ clientForwarders
%

category: 'Accessing'
method: GbjGciInterface
delay

    ^ delay
%

category 'Updating'
method: GbjGciInterface
verbose: aBoolean

    verbose := aBoolean
%

category 'Updating'
method: GbjGciInterface
classOrganizer: organizer

    classOrganizer := organizer.
%

category 'Updating'
method: GbjGciInterface
mappedClassRoots:  aCollection

    mappedClassRoots := aCollection.
%

category 'Updating'
method: GbjGciInterface
clientForwarders:  aCollection

    clientForwarders := aCollection.
%

category: 'Updating'
method: GbjGciInterface
delay: aDelay

    delay := aDelay.
%

