/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjCollection.java  
 *
 *========================================================================
 */

package com.gemstone.gbj;
import java.util.Enumeration;

/**
 *  This subclass of GbjObject represents server instances of Collection or one
 *  of its subclasses, excluding subclasses of the server class AbstractDictionary.
 *  GbjCollection provides methods for manipulating collections
 *  in the server.  It can create an Enumeration that performs efficient caching
 *  of the server object's contents for minimal client/server traffic during
 *  the enumeration operation.  This class is not synchronized, if the application
 *  is accessing the same instance from multiple threads, the application is responsible
 *  for synchronization.
 * 
    <h2>GBJ 3.2 Changes</h2>

    Server classes that are subclasses of the server class Collection 
    and whose element cannot be directly enumerated via com.gemstone.gbjgci.GbjGciObject.getObjs
    are no longer mapped to GbjCollection.  The classes no longer mapped are 
    server classes AbstractDicitionary, Bag, Set   and subclasses thereof.
    If enumration of instances of those server classes is desired a new GBJ class would
    be needed to provide a correct implementation (ref bug 51203).
    <br>
    Removed methods  asSet() and asBag(), because the results would not be representable as a Collection . 
 * 
 * @version 3.2
 */
public class GbjCollection extends GbjObject {

    /**
     * Return an Enumeration of the server object's contents.  
     * If the receiver is not a sequenceableCollection in GemStone, 
     * the server object's contents will be sequenced in the server.
     * @return Enumeration result
     */
    public Enumeration elements() {
        return new GbjIterator(this);
    };


    /**
     * Return an array holding all of the server collection's contents.
     * @return result Gbjobject array
     */
    public GbjObject [] allElements() {
        return (new GbjIterator(this)).remainingElements();
    }


    private void refresh() {
       markChanged();
       checkValue();
    }

    /**
     * Add the object to the receiver, returning the object.
     * @param obj - a GbjObject instance or an object that the interface can
     * convert to a server object
     * @return  The receiver
     * @exception GbjException    thrown if there is a server exception
     */
    public Object add(Object obj) throws GbjException {
        sendMsg("add:", obj);
        refresh();
        return obj;
    }

    /**
     * Add the contents of the given collection to the receiver, returning the arg.
     * @param obj - a GbjCollection of server objects
     * @return  The collection of server objects added
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection addAll(GbjCollection obj) throws GbjException {
        sendMsg("addAll:", obj);
        this.refresh();
        return obj;
    }

    /**
     * Add the contents of the given collection to the receiver, returning the arg.
     * @param obj - a Java array of GbjObjects or objects that can be converted to server objects
     * @return  The array of objects added
     * @exception GbjException    thrown if there is a server exception
     */
    public Object addAll(Object obj[]) throws GbjException {
        int i;
        for (i=0; i<obj.length; i++) {
            sendMsg("add:", obj[i]);
        }
        this.refresh();
        return obj;
    }

    /**
     * Create an extensible array with the contents of the server object
     * @return  The resulting server array object
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection asArray() throws GbjException {
        return (GbjCollection)sendMsg("asArray");
    }

    public GbjObject asByteArray() throws GbjException {
        GbjObject o = sendMsg("asByteArray");
        return o;
    }

    /**
     * Create an identity bag with the contents of the server object
     * @return  The resulting server identity bag object
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection asIdentityBag() throws GbjException {
        return (GbjCollection)sendMsg("asIdentityBag");
    }

    /**
     * Create an identity set with the contents of the server object
     * @return  The resulting server identity set object
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection asIdentitySet() throws GbjException {
        return (GbjCollection)sendMsg("asIdentitySet");
    }

    /**
     * Create an ordered collection with the contents of the server object
     * @return  The resulting server ordered collection object
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection asOrderedCollection() throws GbjException {
        return (GbjCollection)sendMsg("asOrderedCollection");
    }

    /**
     * Sort the server object according to &lt;=
     * @return  The resulting sorted collection object
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection asSortedCollection() throws GbjException {
        return (GbjCollection)sendMsg("asSortedCollection");
    }

    /**
     * Use the given Smalltalk block to sort the server object.  The block must
     * be in the form [:a:b | booleanExpression],  for instance 
     * <pre>
     * [:a:b | a address street &lt;= b address street ]
     * </pre>
     * @param sortBlock  String containing sort block
     * @return  The resulting sorted collection object
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection asSortedCollection(String sortBlock) throws GbjException {
        return (GbjCollection)sendMsg("asSortedCollectionWithSortBlockString:", sortBlock);
    }

    /**
     * Create an ordered collection with the sorted contents of the server object
     * @return  The resulting sorted collection object
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection asSortedOrderedCollection() throws GbjException {
        return (GbjCollection)sendMsg("asSortedOrderedCollection");
    }

    /**
     * Evaluate the given block with each element of the server object returning
     * a collection, usually of the same kind, with the results. For instance, 
     * <pre>
     *  collect("[:each | each segment owner userId]")
     * </pre>
     * @param collectBlock  String containing collect block
     * @return  The resulting collection object
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection collect(String collectBlock) throws GbjException {
        return (GbjCollection)sendMsg("collectWithBlockString:", collectBlock);
    }

    /**
     * Find the first element in the server enumeration that satisfies the
     * given block.  For instance, 
     * <pre>
     *  detect("[:each | each segment owner userId = #DataCurator]")
     * </pre>
     * @param detectBlock  String containing detect block
     * @return  The detected object
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject detect(String detectBlock) throws GbjException {
        return sendMsg("detectWithBlockString:", detectBlock);
    }

    /**
     * Return the number of occurrences of the given object in the server object
     * @param anObject  Given object
     * @return int result
     * @exception GbjException    thrown if there is a server exception
     */
    public int identicalOccurrencesOf(GbjObject anObject) throws GbjException {
        return sendMsg("identicalOccurrencesOf:", anObject).intValue();
    }

    /**
     * Return whether an object equal to the argument is in the server object
     * @param anObject  Given object
     * @return boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean includes(GbjObject anObject) throws GbjException {
        return sendMsg("includes:", anObject).booleanValue();
    }

    /**
     * Return  whether the argument is in the server object
     * @param anObject  Given object
     * @return boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean includesIdentical(GbjObject anObject) throws GbjException {
        return sendMsg("includesIdentical:", anObject).booleanValue();
    }

    /**
     * Return  whether an object equal to the argument is in the server object
     * @param anObject  Given object
     * @return boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean includesValue(GbjObject anObject) throws GbjException {
        return sendMsg("includesValue:", anObject).booleanValue();
    }

    /**
     * Evaluate the given block with the given argument plus an element of
     * the server object.  Evaluate the block for remaining elements of the
     * server object using the result of the previous evaluation as the first
     * argument and the element as the second argument.  E.g.,
     * <pre>
     * coll.injectInto(GbjGetSession().GSOne, "[:v:e | v+e]").intValue()
     * </pre>
     * @param  arg    Object for first inject
     * @param  block  Block to use for inject
     * @return  The final value returned by the block
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject injectInto(GbjObject arg, String block) throws GbjException {
        return sendMsg("inject:", arg, "intoBlockString:", block);
    }

    /**
     * Does the server object hold anything?
     * @return  boolean
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean isEmpty() throws GbjException {
        return sendMsg("isEmpty").booleanValue();
    }

    /**
     * Return  the number of occurrences of the given object's value in the server object
     * @param  anObject  Given object
     * @return int result
     * @exception GbjException    thrown if there is a server exception
     */
    public int occurrencesOf(GbjObject anObject) throws GbjException {
        return sendMsg("occurrencesOf:", anObject).intValue();
    }

    /**
     * Return a collection of all the elements of the server object that do not
     * satisfy the given block
     * @param  rejectBlock   Block to reject on
     * @return GbjCollection result
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection reject(String rejectBlock) throws GbjException {
        return (GbjCollection)sendMsg("rejectWithBlockString:", rejectBlock);
    }

    /**
     * Remove the given object's value from the server object, throwing an error
     * if not found
     * @param  obj   Object to remove
     * @return The receiver
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject remove(GbjObject obj) throws GbjException {
        return sendMsg("remove:", obj);
    }

    /**
     * Remove the given object from the server object, throwing an error if not found
     * @param obj  Object to remove
     * @return The receiver
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject removeIdentical(GbjObject obj) throws GbjException {
        return sendMsg("removeIdentical:", obj);
    }

    /**
     * Remove the given object's value from the server object, throwing an error
     * if not found
     * @param obj  Object to remove
     * @return The receiver
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject removeValue(GbjObject obj) throws GbjException {
        return sendMsg("removeValue:", obj);
    }

    /**
     * Remove the given collection's values from the server object, throwing an error
     * if not found
     * @param obj  Object to remove
     * @return The receiver
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject removeAll(GbjCollection obj) throws GbjException {
        return sendMsg("removeAll:", obj);
    }

    /**
     * Remove the given objects from the server object, throwing an error if not found
     * @param obj  Object to remove
     * @return The receiver
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject removeAllIdentical(GbjCollection obj) throws GbjException {
        return sendMsg("removeAllIdentical:", obj);
    }

    /**
     * Return a collection of all the elements of the server object that
     * satisfy the given block
     * @param selectBlock  Select block to use
     * @return  Result collection
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection select(String selectBlock) throws GbjException {
        return (GbjCollection)sendMsg("selectWithBlockString:", selectBlock);
    }

    /**
     * Return an array of the server sorted values
     * @return GbjCollection result
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection sortAscending() throws GbjException {
        return (GbjCollection)sendMsg("sortAscending");
    }

    /**
     * Return an array of the server sorted values
     * @return GbjCollection result
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection sortDescending() throws GbjException {
        return (GbjCollection)sendMsg("sortDescending");
    }

    /**
     * Sort the server object on the given instance variable
     * @param  variableName  Name of instance variable
     * @return  The receiver
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection sortAscending( String variableName ) throws GbjException {
        return (GbjCollection)sendMsg("sortAscending:", variableName);
    }

    /**
     * Sort the server object on the given instance variable
     * @param variableName  Name of instance variable
     * @return  The receiver
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection sortDescending( String variableName ) throws GbjException {
        return (GbjCollection)sendMsg("sortDescending:", variableName);
    }

    /**
     * Sort the server object using the given GemStone sort specification
     * @param specs  Sort specification
     * @return  The receiver
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjCollection sortWith( String specs ) throws GbjException {
        return (GbjCollection)sendMsg("sortWith:", specs);  // should handle an array of strings
    }

    /**
     * Lock the contents of this collection.  Throws an exception with number
     * LOCK_ERR_INCOMPLETE if all locks cannot be obtained.
     * @exception GbjException    thrown if there is a server exception
     * @see com.gemstone.gbj.GbjGemStoneErrors#LOCK_ERR_INCOMPLETE
     */
    public void readLockContents() throws GbjException {
        getSession().kernel.System.sendMsg("readLockAll:", this);
    }

    /**
     * Lock the contents of this collection.  Throws an exception with number
     * LOCK_ERR_INCOMPLETE if all locks cannot be obtained.
     * @exception GbjException    thrown if there is a server exception
     * @see com.gemstone.gbj.GbjGemStoneErrors#LOCK_ERR_INCOMPLETE
     */
    public void writeLockContents() throws GbjException {
        getSession().kernel.System.sendMsg("writeLockAll:", this);
    }

    /**
     * Lock the contents of this collection.  Throws an exception with number
     * LOCK_ERR_INCOMPLETE if all locks cannot be obtained.
     * @exception GbjException    thrown if there is a server exception
     * @see com.gemstone.gbj.GbjGemStoneErrors#LOCK_ERR_INCOMPLETE
     */
    public void exclusiveLockContents() throws GbjException {
        getSession().kernel.System.sendMsg("exclusiveLockAll:", this);
    }

    /**
     * Release locks held on the contents of this collection.  Objects in
     * the collection that are not locked are ignored.
     * @exception GbjException    thrown if there is a server exception
     */
    public void removeLocksOnContents() throws GbjException {
        getSession().kernel.System.sendMsg("removeLockAll:", this);
    }

}

