/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjCompilerException.java  
 *
 *========================================================================
 */

package com.gemstone.gbj;

import com.gemstone.gbjgci.*;


/**
 *  This class represents a GemStone server compilation exception.  These most commonly
 *  occur during a GbjSession doit() message or GbjObject execute() message and indicate
 *  that there is a Smalltalk syntax error in the string passed to the method.
 *
 *  Instances of this class will discard their GemStone error message string and instead
 *  return the text of the first compilation error.  Additional compilation errors may
 *  exist in the compiland - the first argument of the exception is a GemStone Array of
 *  compilation error specifications.  Each of these specifications is an Array with
 *
 *     1) specific GemStoneError error number
 *     2) 1-based position of the error in the text
 *
 *   Thus, args[0].at(1) is the first error specification, and args[0].at(1).at(2) is the
 *   position of the first error.  The method errorPosition() is provided to access this
 *   value in a more convenient manner.
 * 
 * 
 * @version 3.0
 */

 public class GbjCompilerException extends GbjException {
    private int errorPos;
    private boolean particularsFetched;

     ///**
     //* internal method
     //*/
     //    protected GbjCompilerException(GbxInput input) {
     //
     //        super(input);
     //        errorPos = 0;
     //        particularsFetched = false;
     //    }

    public GbjCompilerException(GbjSession s, GbjGciException e) {
	super(s, e);
    }

    private void getParticulars() {
        try {
            message = ((GbjObject)args[0]).execute(
                "^(GemStoneError at: System myUserProfile nativeLanguage) at: ((self at: 1) at: 1)").stringValue();
            errorPos = ((GbjObject)args[0]).execute("^(self at: 1) at: 2").intValue();
        }
        catch (GbjException e) {
            // note that the execution context here is very likely system i/o.  Logging
            // error information may lock the system.
        }
        particularsFetched = true;
    }

    /**
     * returns the 1-based position of the first compilation error.  This will
     * be zero if there was a problem retrieving the error position.
     * @return int result
     */
    public int errorPosition() {
        if (!particularsFetched)
            getParticulars();
        return errorPos;
    }

     /**
      * Return the message associated with this compiler exception.
      * @return string result
      */
    public String getMessage() {
        if (!particularsFetched)
            getParticulars();
        return message;
    }

 }

