/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjException.java  
 *
 *========================================================================
 */

package com.gemstone.gbj;  // Basic Java Interface

import java.io.PrintWriter;
import java.io.PrintStream;
import com.gemstone.gbjgci.*;

/**
 *  This class represents a GemStone server exception.  The members of a GbjException
 *  are very similar to those in a GciErrSType structure in the "GemBuilder for C"
 *  product.
 * 
 * @version 3.0
 */
public class GbjException extends RuntimeException {
    /**
     * Holds an extensive list of error numbers for GbjExceptions of
     * category GemStoneError.
     * @see com.gemstone.gbj.GbjSession#kernel
     * @see com.gemstone.gbj.GbjKernelObjects
     * @see com.gemstone.gbj.GbjGemStoneErrors
     */

    public static final GbjGemStoneErrors kernel = new GbjGemStoneErrors();

    // flag to save/clear process on finalization
    private boolean saveProc;

    /**
     * Holds a displayable representation of the exception
     */
    public String	message;
    /**
     * Holds a reference to the category (signalDictionary) of the exception.
     * This can be compared with the GbjSession GemStoneError member to see if
     * the exception was thrown by GemStone or was a user-defined error.
     */
    public GbjObject	category;
    /**
     * Holds the number of the exception.  A full list of the errors that GemStone
     * may throw can be found in the file gcierr.ht in the GemStone installation
     * directory's "include" subdirectory.
     */
    public int		number;
    /**
     * An array of references to the arguments of the exception
     */
    public Object args[];
    /**
     * The size of args[]
     */
    public int		argCount;
    /**
     * Tells you whether the exception was a fatal error.  If it was a fatal
     * error, the session should be closed.
     */
    public boolean	fatal;
    /**
     * The suspended state of execution on the server at the time the exception
     * was thrown.  An exception's process object is released on finalization
     * of the exception unless the process variable is nulled or the <i>saveProcess</i>
     * message is sent to the exception.
     *
     * @see com.gemstone.gbj.GbjException#saveProcess
     */
    public GbjObject process;
    /**
     * The session in which the exception occurred.
     */
    public GbjSession session;

    /**
     * If stack traces have been turned on, this will hold a trace of GemStone
     * execution at the time of the exception.
     * @deprecated  use stackReport
     */
    @Deprecated
    public GbjObject stackTrace;

    /** Smalltalk stack or null,  requires 
     *   System gemConfigurationAt:#GemExceptionSignalCapturesStack put: true 
     * to have been executed, such as via  UserProfile >> loginHook:
     */
    public String stackReport;

    /**
     * Construct an instance that will be filled in by another
     */
    public GbjException() {
	super();
        message = "Undefined exception";
        argCount = 0;
        fatal = false;
        number = 2520; // RT_Err_GbjError from gcierr.ht
    }

    /**
     * Construct an instance with specified message.
     * @param msg String containing message
     */
    public GbjException(String msg) {
	this();
        message = msg;
    }

    public GbjException(GbjSession s, String msg) {
	this();
        message = msg;
        session = s;
    }

    //
    //
    /**
     * Construct an instance from information coming from the GbjGciInterface
     * @param e GbjGciException
     */
    public GbjException(GbjGciException e) {
	this(e.message);
        saveProc    = false;
        number		= e.number;
        fatal		= e.fatal;
        argCount	= e.argCount;
        stackReport = e.stackReport;
    }

    public GbjException(GbjSession s, GbjGciException e) {
	this(e);
        session = s;
	// fill in the other fields only if we're connected..
	if (number < 4000 && s.isConnected()) {
	    category	= GbjObject.withOop(e.category, s);
	    process		= GbjObject.withOop(e.context, s);
	    args		= new GbjObject[argCount];
	    for (int i=0; i < argCount; i++) {
		args[i] = GbjObject.withOop(e.args[i], s);
	    }
	}
    }

    /**
     * return the error message held by the receiver.  Except for local
     * internal errors, this message will be localized based on the account
     * configuration in GemStone.
     */
    public String getMessage() {
        return message;
    }

    /**
     * is this a Smalltalk compilation exception?
     * @return boolean
     */
    public boolean isCompilerException() {
        return number / 1000 == 1 ;
    }
    /**
     * is this a Smalltalk execution exception?
     * @return boolean
     */
    public boolean isRuntimeException() {
        return number / 1000 == 2 ;
    }
    /**
     * is this an Aborting exception - an exception associated with a rolled back
     * transaction in the server?
     * @return boolean
     */
    public boolean isAbortingException() {
        return number / 1000 == 3 ;
    }
    /**
     * is this a Fatal exception - one that terminated the connection or
     * otherwise rendered it unusable?
     * @return boolean
     */
    public boolean isFatalException() {
        return number / 1000 == 4 ;
    }
    /**
     * is this an Internal exception - one that signals an internal problem in
     * the GemStone server or interface?
     * @return boolean
     */
    public boolean isInternalException() {
        return number / 1000 == 5 ;
    }
    /**
     * is this an Event exception - a signal or debugging event?
     * @return boolean
     */
    public boolean isEventException() {
        return number / 1000 == 6 ;
    }
    /**
     * is this an Event exception that should be handled by the event notification
     * mechanism?
     * @return boolean
     */
    protected boolean isEventSignal() {
        return 6006 < number && number < 6011;
    }

    /**
     * Set the GemStone exception message text.
     * @param aString Message text
     */
    public void setMessage(String aString) {
        message = aString;
    }

    /**
     * causes the exception's Smalltalk process to be saved in the server's export
     * set even if this exception object is reclaimed as garbage.  This is useful
     * if the process will be held for debugging or rescheduling.  This method does
     * not protect the method from being removed from the server's export set by
     * other means.
     */
    // GBJ30: not clear if this is still useful/needed
    public void saveProcess() {
        saveProc = true;
    }

    public void printStackTrace() {
      if (stackReport != null)
        System.out.print(stackReport);
      super.printStackTrace();
    }

    public void printStackTrace(PrintStream s) {
      if (stackReport != null)
        s.print(stackReport);
      super.printStackTrace(s);
    }

    public void printStackTrace(PrintWriter s) {
      if (stackReport != null)
        s.print(stackReport);
      super.printStackTrace(s);
    }
}

