/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjIterator.java  
 *
 *========================================================================
 */

package com.gemstone.gbj;

import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.logging.*;
import com.gemstone.gbjgci.*;

/**
 * Instances of this class iterate over collections in the GemStone server.
 *
 * Contents of the collection are sequenced on the server and cached in batches
 * in Java for efficiency.
 * 
 * <h2> GBJ30 Changes </h2>
 *
 * <h3> Constructor Changes </h3>
 *
 * GbjIterator(GbjObject[] initialCache) no longer supported.
 * <p>
 * New constructor GbjIterator(GbjObject onCollection, long startingIdx) .
 * 
 * @version 3.0
 */
public class GbjIterator implements Enumeration {
    GbjObject cache[];
    GbjObject collection;
    GbjSession session;

    long collectionSize, collectionIdx;
    int cacheSize, cacheIdx;

    /** Specifies batch size.
	Default size is 1000, can be redefined by the application */
    public static int BatchSize = 1000;

    /** Logger for debug trace messages */
    private static Logger log;

    // static code for setting up logger
    static {
	log = GbjGciInterface.setupLogger(GbjIterator.class);
    }

    /** Constructor with starting index 
	@param onCollection collection to iterate over
	@param startingIdx  starting index for iteration
     */
    public GbjIterator(GbjObject onCollection, long startingIdx) {
	if (GbjGciSession.isDEBUG) log.fine("Setting up GbjIterator on " + onCollection);
        GbjObject batch, obj;
	session = onCollection.getSession();
        collection = onCollection;
	collectionSize = onCollection.getIndexedSize();
	collectionIdx = startingIdx;
	// skip most of the setup if collectionSize is 0
	if (collectionSize > 0) {
	    if (collectionIdx > collectionSize) 
	       throw new GbjException("Invalid startingIdx in GbjIterator()");
	    cacheSize = (int)Math.min(collectionSize, BatchSize);
	    cacheIdx = cacheSize;
	    cache = new GbjObject[cacheSize];
	}
    }

    /** Usual Constructor.
	@param onCollection  collection to iterate over
     */
    public GbjIterator(GbjObject onCollection) {
	this(onCollection, 1);
    }

    /** Usual Constructor 
	@param onCollection  collection to iterate over
    */
    public GbjIterator(GbjCollection onCollection) {
	this((GbjObject)onCollection);
    }

    /** Return true if there are more elements to return from this iterator.
	@return boolean result
     */
    public boolean hasMoreElements() {
	return collectionIdx <= collectionSize;
    }

    /** Return the next element from this iterator.
     *  If no more elements are available, throw an error.
	@return object result
     */
    public Object nextElement() {
	if (GbjGciSession.isDEBUG) log.fine("Getting next element at index " + collectionIdx + "/" + cacheIdx);
	if (!hasMoreElements()) 
            throw new NoSuchElementException("End of data");
	if (cacheIdx >= cacheSize) getNextBatch();
	GbjObject next = cache[cacheIdx];
	cacheIdx++;
	collectionIdx++;
	return next;
    }

    private void getNextBatch() {
	if (GbjGciSession.isDEBUG) log.fine("Getting batch at index " + collectionIdx);
	    // 47289: use batchSize instead of cacheSize in getObjs() call
	int batchSize = (int)Math.min(BatchSize, collectionSize - collectionIdx + 1);
	session.getObjs(collection, collectionIdx, batchSize, cache);
	cacheIdx = 0;
    }
	
    /** Return all of the remaining elements left in the iterator.
     *  Note: Maximum number of elements that can be returned is
     *  Integer.MAX_VALUE.
     *  @return ObjObject array with elements
     */
    public GbjObject [] remainingElements() {
	long lsize = collectionSize - collectionIdx + 1;
	if (lsize == 0) return new GbjObject[0];
	if (lsize > Integer.MAX_VALUE) {
	    throw new GbjException(
	       "Implementation limit on remainingElements()");
	}
	int size = (int)lsize;
	if (size > cacheSize) cache = new GbjObject[size];
	if (GbjGciSession.isDEBUG) log.fine("remainingElements: " + collectionIdx + " " + size);
	session.getObjs(collection, collectionIdx, size, cache);
	collectionIdx = collectionSize;
	cacheIdx = cacheSize;
	return cache;
    }

}


