/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjObject.java  
 *
 * Java representation for a GemStone object
 *
 *========================================================================
 */

package com.gemstone.gbj;
import java.util.Enumeration;
import java.util.Date;
import java.util.TimeZone;
import java.util.logging.*;
import java.time.*;

// GBJ imports
import com.gemstone.gbjgci.*;
import com.gemstone.gbjgci.GbjGciReport;

// new for 3.2.0:
import java.math.BigInteger;
import java.math.BigDecimal;
import org.apache.commons.math3.fraction.BigFraction; // For GS Fractions


/**
Instances of this class represent server objects in a GemStone server.  Each
GbjObject holds a reference to the remote object and may also hold the value
of the server object in various protected value fields.
<p>
GbjObject defines methods for fundamental operations on server objects.  
Other methods for specific objects will be available on the server and 
may be accessed through the GbjObject instance with the sendMsg() methods.

<h2>GBJ 3.2 Changes</h2>

Added constructors taking as arguments instances of 
BigInteger , BigDecimal, BigDecimalFloat, BigFraction, LocalDate, LocalTime,
LocalDateTime, ZonedDateTime, OffsetDateTime, Instant .

Removed constructors taking as arguments instances of DoubleByteString, Calendar .

<h3>Object Change Synchronization </h3>

The subclass GbjGciObject now provides support via the Java 
Observable/Observer API to notify customer application when the associated
GS server object has changed, either due to GS Smalltalk code execution or
committed transactions from other sessions.  Refer to the GbjGciObject
class documentation for details.

 * @version 3.2
 */
public class GbjObject extends GbjGciObject implements Cloneable {

    /** Logger for debug trace messages */
    protected static Logger log;

    // static code for setting up logger
    static {
	log = GbjGciInterface.setupLogger(GbjObject.class);
    }

    /**
     * Special constructor - should only be used for generating
     * a template object for java/GS class mapping operations
     */
    public GbjObject() { super(); }

    public GbjObject(GbjSession s) { super(s); }

    /**
     * Construct a GS object representing a boolean for this session.
     * Generates a GS Boolean object.
     * @param b boolean
     * @param s session
     */
    public GbjObject(boolean b, GbjSession s) {
	super(b, s);
    }

    /**
     * Construct a GS object representing a character for this session.
     * Generates a GS Character object.
     * @param c char
     * @param s session
     */
    public GbjObject(char c, GbjSession s) {
	super(c, s);
    }

    /**
     * Construct a GS object representing a byte for this session.
     * Generates a GS SmallInteger object.
     * @param b byte
     * @param s session
     */
    public GbjObject(byte b, GbjSession s) {
	super(b, s);
    }

    /**
     * Construct a GS object representing a short for this session.
     * Generates a GS SmallInteger object.
     * @param i short
     * @param s session
     */
    public GbjObject(short i, GbjSession s) {
	super(i, s);
    }

    /**
     * Construct a GS object representing an int for this session.
     * Generates a GS SmallInteger object.
     * @param i int
     * @param s session
     */
    public GbjObject(int i, GbjSession s) {
	super(i, s);
    }

    /**
     * Construct a GS object representing a long for this session.
     * Generates a GS SmallInteger object.
     * @param i long
     * @param s session
     */
    public GbjObject(long i, GbjSession s) {
	super(i, s);
    }

    /**
     * Construct a GS object representing a BigInteger for this session.
     * Generates a GS Integer or LargeInteger as appropriate
     * @param i BigInteger
     * @param s session
     */
    public GbjObject(BigInteger i, GbjSession s) {
	super(i, s);
    }

    /**
     * Construct a GS object representing a float for this session.
     * Generates a GS SmallDouble object.
     * @param f float
     * @param s session
     */
    public GbjObject(float f, GbjSession s) {
	super(f, s);
    }

    /**
     * Construct a GS object representing a double for this session.
     * Generates a GS Float object.
     * @param d double
     * @param s session
     */
    public GbjObject(double d, GbjSession s) {
	super(d, s);
    }

    /**
     * Construct a GS object representing a BigDecimal for this session.
     * Generates a GS ScaledDecimal object.
     * @param d double
     * @param s session
     */
    public GbjObject(BigDecimal d, GbjSession s) {
	super(d, s);
    }

    /**
     * Construct a GS object representing a BigDecimalFloat for this session.
     * Generates a GS DecimalFloat object.
     * @param df BigDecimalFloat
     * @param s session
     */
    public GbjObject(BigDecimalFloat df, GbjSession s) {
	super(df, s);
    }
    
    /**
     * Construct a GS object representing a BigFraction for this session.
     * Generates a GS Fraction object.
     * @param f BigFraction
     * @param s session
     */
    public GbjObject(BigFraction f, GbjSession s) {
	super(f, s);
    }

    /**
     * Construct a GS object representing a string for this session.
     * Generates either a GS String or DoubleByteString, as appropriate.
     * @param str String
     * @param s session
     */
    public GbjObject(String str, GbjSession s) {
	super(str, s);
    }

    /**
     * Construct a GS object representing a LocalDate for this session.
     * Generates a GS Date object.
     * @param d LocalDate
     * @param s session
     */
    public GbjObject(LocalDate d, GbjSession s) {
	super(d, s);
    }

    /**
     * Construct a GS object representing a LocalTime for this session.
     * Generates a GS Time object.
     * @param t LocalTime
     * @param s session
     */
    public GbjObject(LocalTime t, GbjSession s) {
	super(t, s);
    }

    /**
     * Construct a GS object representing a LocalDateTime for this session.
     * Generates a GS LocalDateTime object.
     * @param ldt LocalDateTime
     * @param s session
     */
    public GbjObject(LocalDateTime ldt, GbjSession s) {
	super(ldt, s);
    }

        /**
     * Construct a GS object representing a ZonedDateTime for this session.
     * Generates a GS DateTime object.
     * @param zdt LocalTime
     * @param s session
     */
    public GbjObject(ZonedDateTime zdt, GbjSession s) {
	super(zdt, s);
    }

    /**
     * Construct a GS object representing a OffsetDateTime for this session.
     * Generates a GS DateAndTime object.
     * @param odt OffsetDateTime
     * @param s session
     */
    public GbjObject(OffsetDateTime odt, GbjSession s) {
	super(odt, s);
    }

    /**
     * Construct a GS object representing a Date for this session.
     * Generates a GS DateAndTime object.
     * @param d Date
     * @param s session
     */
    public GbjObject(Date d, GbjSession s) {
	super(d, s);
    }

    /**
     * Construct a GS object representing an Instant for this session.
     * Generates a GS DateAndTime object.
     * @param i Instant
     * @param s session
     */
    public GbjObject(Instant i, GbjSession s) {
	super(i, s);
    }

    /**
     * Construct a GS object representing an object for this session.
     * Note that object must be an instance of boolean, character, byte,
     * short, int, long, float, double, date, or string.
     * (as of GBJ 3.2.0: also LocalTime, LocalDate)
     * @param o Object
     * @param s session
     */
    public GbjObject(Object o, GbjSession s) {
	// 40316: if a ClassCastException is returned, 
	// we setup a ClientForwarder on the server instead..
	super(o, s);
    }

    /**
     * Return the GS object with this designated oop.
     * This method replaces the use of GbjObject(long, GbjSession)
     * used prior to GBJ30.
     * @param oopValue  Designated oop
     * @param s         GbjSession
     * @return  The GS object with this designated oop.
     */
    public static GbjObject withOop(long oopValue, GbjSession s) {
	return (GbjObject)GbjGciObject.withOop(oopValue, s);
    }

    /** Convert a Java int to a GemStone SmallInteger
	@deprecated Use GbjObject(int, GbjSession) instead. 
	@param value      Java int
	@param session    GbjSession
	@return GbjObject
    */
    @Deprecated
    public static GbjObject fromInteger(int value, GbjSession session) {
	return new GbjObject(value, session);
    }

    /** Convert a Java long to a GemStone SmallInteger
	@deprecated Use GbjObject(long, GbjSession) instead.
	@param value     Java long
	@param session   GbjSession
	@return GbjObject
    */
    @Deprecated
    public static GbjObject fromLong(long value, GbjSession session) {
	return new GbjObject(value, session);
    }

    /** This method returns the cachedValue if the receiver represents a server object
        for which there is a direct mapping from the Server class to a Java class .  
        The direct mappings include the classes used by each of the constructors in GbjObject.
        See the *value() methods in GbjGciObject .
	@return java object , or null if there is not a direct mapping.
    */
    public Object cachedValue() {
	checkValue();
	switch (valueType) {
   	case VALTYPE_NULL:	      return null;
	case VALTYPE_BOOLEAN: 	      return booleanValue( );
	case VALTYPE_CHARACTER:	      return charValue( );
	case VALTYPE_LONG:	      return longValue( );
	case VALTYPE_DOUBLE:	      return doubleValue( );
	case VALTYPE_STRING:          return stringValue();

        case VALTYPE_CALENDAR:        return calendarValue();
	case VALTYPE_ZONED_DATETIME:  return zonedDateTimeValue( );
	case VALTYPE_DATE_AND_TIME:   return offsetDateTimeValue( );
	case VALTYPE_DAY: 	      return localDateValue( );
	case VALTYPE_TIME:            return localTimeValue( );
	case VALTYPE_LARGE_INTEGER:   return bigIntegerValue( );
	case VALTYPE_SCALED_DECIMAL:  return bigDecimalValue( );
	case VALTYPE_DECIMAL_FLOAT:   return bigDecimalFloatValue( );
	case VALTYPE_FRACTION:        return bigFractionValue( );
	case VALTYPE_LOCAL_DATETIME:  return localDateTimeValue( );
        case VALTYPE_utilDATE:        return dateValue();
        case VALTYPE_INSTANT:         return instantValue();
	}
	return null;
    }

    /**
       Return the GbjSession associated with this GbjObject
       @return GbjSession
    */
    public GbjSession getSession() {
	return (GbjSession)super.getSession();
    }

    /**
     * Return the cached String object held by the receiver as a new 
     * StringBuffer.  If the receiver does not hold a StringBuffer, this
     * throws a ClassCastException.
     * @return  StringBuffer
     * @exception ClassCastException on receiver not holding a StringBuffer
     */
    public StringBuffer stringBufferValue() throws ClassCastException {
        return new StringBuffer(stringValue());
    }


    /**
     * Return the cached Unicode string object held by the receiver.  If the
     * receiver does not represent a GemStone DoubleByteString object, this
     * throws a ClassCastException.
     * @return  The cached Unicode string object held by the receiver.
     * @exception ClassCastException on receiver not a DoubleByteString
     * @deprecated Use  stringValue() inherited from GbjGciObject
     */
    public String dbStringValue() throws ClassCastException {
	return stringValue();
    }

    /**
     * Return the cached Unicode string object held by the receiver as a new
     * StringBuffer object.  If the receiver does not represent a GemStone
     * DoubleByteString object, this throws a ClassCastException.
     * @return  The cached Unicode string object held by the receiver.
     * @exception ClassCastException on receiver not a DoubleByteString
     * @deprecated Use  stringBufferValue() 
     */
    public StringBuffer dbStringBufferValue() throws ClassCastException {
       return stringBufferValue();
    }

    /**
     * Transmit a string holding GemStone Smalltalk code to the server and
     * compile and execute it, returning the result of execution.  The code
     * may reference "self" as the server object represented by this GbjObject
     * instance, and may directly reference instance variables of the server
     * object.
     * @param code  GemStone Smalltalk source code to be compiled and executed
     * @return  Result of execution
     * @exception GbjException  thrown if there is a server exception
     */
    public GbjObject execute(String code) throws GbjException {
      GbjSession s = getSession();
      GbjObject codeObj = new GbjObject(code, s);
      GbjObject[] args = new GbjObject[2];
      args[0] = this;
      args[1] = s.kernel.Nil;
      return (GbjObject)codeObj.perform( "evaluateInContext:symbolList:", args, 2);
    }

    /**
     * Send a message to a GemStone object and expect a server object in reply.
     * In general it is more convenient to use the sendMsg() methods to send a
     * message to a server object.  The perform() method is a lower level 
     * alternative to those methods which can be used if there are more 
     * keywords in the message than sendMsg() allows, or if it is otherwise 
     * more convenient.
     * @param message   a message selector that is dynamically looked up 
     * on the server
     * @param args      an array of arguments that must be GbjObject 
     * instances or wrappers of fundamental types
     * @param numArgs   the number of entries in args that are significant
     * @return  Result of perform
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject perform(String message, Object args[], int numArgs) throws GbjException {
	GbjObject[] gciArgs = new GbjObject[numArgs];
	for (int i = 0; i < numArgs; i++) {
	    if (args[i] instanceof GbjObject)
		gciArgs[i] = (GbjObject)args[i];
	    else {
		// 40316: handle ClassCastException by setting up clientObject
		try {
		    gciArgs[i] = new GbjObject(args[i], getSession());
		} catch (ClassCastException e) {
		    gciArgs[i] = getSession().clientObject(args[i]);
		}
	    }
	}
	try {
	    return (GbjObject)getSession().perform(this, message, gciArgs, numArgs);
	} catch (GbjException e) {
	    return getSession().handleExecutionException(e);
	}
    }

    public GbjObject perform(String message) throws GbjException {
        return perform(message, null, 0);
    }

    /**
     * Send a unary message to a GemStone object and expect an Object reply.
     * E.g., sendMsg("size");
     * @param selector  the message selector
     * @return  Result of message
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject sendMsg(String selector) throws GbjException {
	return perform(selector, null, 0);
    }

    /**
     * Send a one-keyword or binary message to a GemStone object and 
     * expect an Object reply.
     * E.g., sendMsg("at:", new Integer(1));
     * @param selector  the message selector
     * @param arg       the argument to the message.
     * @return  Result of message
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject sendMsg(String selector, Object arg) throws GbjException {
	Object args[] = new Object[1];
	args[0] = arg;
	return perform(selector, args, 1);
    }

    /**
     * Send a two-keyword message to a GemStone object and expect 
     * an Object reply.
     * E.g., sendMsg("at:", new Integer(1), "put:", Boolean.TRUE);
     * @param keyword1  the first keyword of the message selector
     * @param arg1      the argument to the message for the first keyword.
     * @param keyword2  the second keyword of the message selector
     * @param arg2      the argument to the message for the second keyword.
     * @return  Result of message
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject sendMsg(String keyword1, Object arg1, 
			     String keyword2, Object arg2) 
	throws GbjException {
	Object args[] = new Object[2];
	String selector = keyword1 + keyword2;
	args[0] = arg1;
	args[1] = arg2;
	return perform(selector, args, 2);
    }

    /**
     * Send a three-keyword message to a GemStone object and expect 
     * an Object reply.
     * E.g., sendMsg("at:", new Integer(1), "put:", Boolean.TRUE, "with:", 
     * Boolean.FALSE);
     * @param keyword1  the first keyword of the message selector
     * @param arg1      the argument to the message for the first keyword.
     * @param keyword2  the second keyword of the message selector
     * @param arg2      the argument to the message for the second keyword.
     * @param keyword3  the third keyword of the message selector
     * @param arg3      the argument to the message for the third keyword.
     * @return  Result of message
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject sendMsg(String keyword1, Object arg1, 
			     String keyword2, Object arg2,
			     String keyword3, Object arg3) 
	throws GbjException {

	Object args[] = new Object[3];
	String selector = keyword1 + keyword2 + keyword3;
	args[0] = arg1;
	args[1] = arg2;
	args[2] = arg3;
	return perform(selector, args, 3);
    }

    /**
     * Send a four-keyword message to a GemStone object and expect
     * an Object reply.
     * @param keyword1  the first keyword of the message selector
     * @param arg1      the argument to the message for the first keyword.
     * @param keyword2  the second keyword of the message selector
     * @param arg2      the argument to the message for the second keyword.
     * @param keyword3  the third keyword of the message selector
     * @param arg3      the argument to the message for the third keyword.
     * @param keyword4  the third keyword of the message selector
     * @param arg4      the argument to the message for the third keyword.
     * @return  Result of message
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject sendMsg(String keyword1, Object arg1, 
			     String keyword2, Object arg2,
			     String keyword3, Object arg3, 
			     String keyword4, Object arg4) 
	throws GbjException {

	Object args[] = new Object[4];
	String selector = keyword1 + keyword2 + keyword3 + keyword4;
	args[0] = arg1;
	args[1] = arg2;
	args[2] = arg3;
	args[3] = arg4;
	return perform(selector, args, 4);
    }

    /**
     * Send a five-keyword message to a GemStone object and expect 
     * an Object reply.
     * @param keyword1  the first keyword of the message selector
     * @param arg1      the argument to the message for the first keyword.
     * @param keyword2  the second keyword of the message selector
     * @param arg2      the argument to the message for the second keyword.
     * @param keyword3  the third keyword of the message selector
     * @param arg3      the argument to the message for the third keyword.
     * @param keyword4  the third keyword of the message selector
     * @param arg4      the argument to the message for the third keyword.
     * @param keyword5  the third keyword of the message selector
     * @param arg5      the argument to the message for the third keyword.
     * @return  Result of message
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject sendMsg(String keyword1, Object arg1, 
			     String keyword2, Object arg2,
			     String keyword3, Object arg3, 
			     String keyword4, Object arg4,
			     String keyword5, Object arg5) 
	throws GbjException {

	Object args[] = new Object[5];
	String selector = keyword1 + keyword2 + keyword3 + keyword4 + keyword5;
	args[0] = arg1;
	args[1] = arg2;
	args[2] = arg3;
	args[3] = arg4;
	args[4] = arg5;
	return perform(selector, args, 5);
    }

    /**
     * Return  the named instance variables of this server object, in their
     * order of appearance (see GemStone Smalltalk 
     * Behavior&gt;&gt;allInstVarNames.
     * @return  The named instance variables of this server object.
     * @exception GbjException    thrown if the session is not connected or 
     * there is a server exception
     */
    public GbjObject [] namedInstanceVariables() throws GbjException {
	try {
	    return getSession().getNamedObjs(this, 1, this.namedSize);
	} catch (GbjGciException e) {
	    throw new GbjException(getSession(), e);
	}
    }

    /**
     * Fetch the server object's size.
     * @return  the integer size of the server object
     * @exception GbjException    thrown if there is a server exception
     */
    public int remoteSize() throws GbjException {
        return sendMsg("size").intValue();
    }

    /**
     * Fetch the server object's class.
     * @return  a reference to the server object's class
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject remoteClass() throws GbjException {
        return sendMsg("class");
    }

    /**
     * Fetch a printable representation of the server object.
     * @return  a String that can be displayed
     * @exception GbjException    thrown if there is a server exception
     */
    public String printString() throws GbjException {
        return sendMsg("printString").stringValue();
    }

    /**
     * Return  whether the server object and the argument have equal values.
     * @param arg  Argument to use
     * @return  Boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean remoteEquals( Object arg ) throws GbjException {
        return sendMsg("=", arg).booleanValue();
    }

    /**
     * Return   whether the server object is identical to the given object.
     * @param arg Argument to use
     * @return  Boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean remoteEqualsIdentical( Object arg ) {
        return equals(arg);
    }

    /**
     * Return  whether the server object is equal to the given object with
     * case-insensitive comparison.
     * @param arg Argument to use
     * @return Boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean remoteEqualsNoCase(Object arg) throws GbjException {
        return sendMsg("equalsNoCase:", arg).booleanValue();
    }

    /**
     * Return  the server object's identity hash code.
     * @return  int result
     */
    // Current design just returns low-order 32 bits of oop (same as hashCode)
    // For FORMAT32/FORMAT641, this is the same as the oop.
    public int remoteIdentityHash() {
        return (int)oop;
    }

    /**
     * Return  the server object's equality hash code.
     * @return int result
     * @exception GbjException    thrown if there is a server exception
     */
    public int remoteHash() throws GbjException {
        return sendMsg("hash").intValue();
    }

    /**
     * Return  a shallow copy of the server object.
     * @return GbjObject result
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject remoteCopy() throws GbjException {
        return sendMsg("copy");
    }

    /**
     * Return  a shallow copy of the server object.
     * @return GbjObject result
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject remoteClone() throws GbjException {
        return sendMsg("copy");
    }

    /**
     * Return  a deep copy of the server object.
     * @return GbjObject result
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject deepCopy() throws GbjException {
        return sendMsg("deepCopy");
    }

    /**
     * Make the server object invariant (immutable).
     * @exception GbjException    thrown if there is a server exception
     */
    public void immediateInvariant() throws GbjException {
        sendMsg("immediateInvariant");
    }

    /**
     * Is the server object invariant (immutable)?
     * @return boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean isInvariant() throws GbjException {
        return sendMsg("isInvariant").booleanValue();
    }

    /**
     * Is the server object in the given remote collection?
     * @param collection  Collection to use
     * @return  boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean in(GbjCollection collection) throws GbjException {
        return sendMsg("in:", collection).booleanValue();
    }

    /**
     * Is the server object an instance of the given class or one of
     * its superclasses?
     * @param remoteClass  Class to use
     * @return boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean isKindOf(GbjObject remoteClass) throws GbjException {
        return sendMsg("isKindOf:", remoteClass).booleanValue();
    }


    /**
     * Is the server object an instance of the given class?
     * @param remoteClass  Class to use
     * @return boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean isMemberOf(GbjObject remoteClass) throws GbjException {
        return sendMsg("isMemberOf:", remoteClass).booleanValue();
    }

    /**
     * Does the user have write access to the server object?
     * @return boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean canBeWritten() throws GbjException {
        return sendMsg("canBeWritten").booleanValue();
    }

    /**
     * Change the server object's segment and do the same with its contents.
     * @param segment  Segment to use
     * @exception GbjException    thrown if there is a server exception
     */
    public void changeToSegment(GbjObject segment) throws GbjException {
        sendMsg("changeToSegment:", segment);
    }

    /**
     * Assign the server object, but not its contents, to the given segment.
     * @param segment  Segment to use
     * @exception GbjException    thrown if there is a server exception
     */
    public void assignToSegment(GbjObject segment) throws GbjException {
        sendMsg("assignToSegment:", segment);
    }

    /**
     * Access one of the server object's anonymous instance variables.
     * @param index  Index to use
     * @return  GbjObject result
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject at(int index) throws GbjException {
        return sendMsg("at:", Integer.valueOf(index));
    }

    /**
     * Update one of the server object's anonymous instance variables.
     * @param index  Index to use
     * @param value  Value to use for update
     * @return  value
     * @exception GbjException    thrown if there is a server exception
     */
    public Object atPut(int index, Object value) throws GbjException {
        sendMsg("at:", Integer.valueOf(index), "put:", value);
        return value;
    }

    /**
     * Answer the segment that the server object is currently assigned to.
     * @return the segment
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject segment() throws GbjException {
        return sendMsg("segment");
    }

    /**
     * Alter the size of the server object's anonymous part.
     * @param newSize   Size to use
     * @exception GbjException    thrown if there is a server exception
     */
    public void size(int newSize) throws GbjException {
        sendMsg("size:", Integer.valueOf(newSize));
    }

    /**
     * Is the server object equivalent to the given object?
     * This test is oriented toward string comparisons and is not
     * generally useful for other types of objects.
     * @param otherObj  given object
     * @return boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean remoteIsEquivalent(GbjObject otherObj) throws GbjException {
        return sendMsg("isEquivalent:", otherObj).booleanValue();
    }

    /**
     * Is the receiver not a nil value?
     * Does not currently require a message send to the server.
     * @return boolean result
     */
    public boolean notNil() {
        return ! isNil();
    }

    /**
     * Was the server object in existence before the current transaction?
     * @return boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean isCommitted() throws GbjException {
        return sendMsg("isCommitted").booleanValue();
    }

    /**
     * @return boolean result , always false
     * @deprecated    no longer supported by the server
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean isConnected() throws GbjException {
       return false;
    }

    /**
     * Is the server object new in this transaction or 
     * been modified in this transaction
     * @return boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean isWritten() throws GbjException {
        return sendMsg("isWritten").booleanValue();
    }

    /**
     * Is the server object a class or metaclass object
     * @return boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean isBehavior() throws GbjException {
        return sendMsg("isBehavior").booleanValue();
    }

    /**
     * Convert the server object to a string
     * @return  String result
     * @exception GbjException    thrown if there is a server exception
     */
    public String asString() throws GbjException {
        if (session == null) {
          return javaValue().toString();
        }
        return sendMsg("asString").stringValue();
    }

    /**
     * The page number of the server object, or nil if a special or
     * unconnected object
     * @return result
     * @exception GbjException    thrown if there is a server exception
     */
    public int page() throws GbjException {
        GbjObject result = sendMsg("page");
        if (result.isNil())
            return 0;
        else
            return result.intValue();
    }

    /**
     * The page creation time of the server object
     * @return Date result
     * @exception GbjException    thrown if there is a server exception
     */
    public Date pageCreationTime() throws GbjException {
        return sendMsg("pageCreationTime").dateValue();
    }

    /**
     * Force the server object to disk if not already there
     * @exception GbjException    thrown if there is a server exception
     */
    public void moveToDisk() throws GbjException {
        sendMsg("moveToDisk");
    }

    /**
     * Cluster the server object
     * @return  boolean whether the server object has already been clustered 
     * in this transaction
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean cluster() throws GbjException {
        return sendMsg("cluster").booleanValue();
    }


    /**
     * Cluster the server object and its contents doing a depth-first 
     * traversal.
     * @return  boolean whether the server object has already been clustered 
     * in this transaction
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean clusterDepthFirst() throws GbjException {
        return sendMsg("clusterDepthFirst").booleanValue();
    }

    /**
     * The cluster id that the server object was assigned during 
     * the current clustering operation
     * @return int result
     * @exception GbjException    thrown if there is a server exception
     */
    public int clusterBucket() throws GbjException {
        return sendMsg("clusterBucket").intValue();
    }

    /**
     * nil the instance variables of the server object
     * @exception GbjException    thrown if there is a server exception
     */
    public void nilFields() throws GbjException {
        sendMsg("nilFields");
    }

    /**
     * Does the server object responds to the given message selector?
     * @param selector message selector to check
     * @return  boolean result
     * @exception GbjException    thrown if there is a server exception
     */
    public boolean respondsTo(String selector) throws GbjException {
        return sendMsg("respondsTo:", selector).booleanValue();
    }

    /**
     * The server object's class for cloning operations
     * @return result GbjObject
     * @exception GbjException    thrown if there is a server exception
     */
    public GbjObject remoteSpecies() throws GbjException {
        return sendMsg("species");
    }

    /**
     * Lock this database object.  Throws an exception if another session
     * holds a conflicting lock on this object.  Also throws an exception
     * if another session has committed a change to this object 
     * (LOCK_ERR_OBJ_HAS_CHANGED).
     * You may not read-lock an object that you already have a write or 
     * exclusive lock upon.
     * @exception GbjException    thrown if there is a server exception
     * @see com.gemstone.gbj.GbjGemStoneErrors#LOCK_ERR_OBJ_HAS_CHANGED
     */
    public void readLock() throws GbjException {
        getSession().kernel.System.sendMsg("readLock:", this);
    }

    /**
     * Lock this database object.  Throws an exception if another session
     * holds a conflicting lock on this object.  Also throws an exception
     * if another session has committed a change to this object 
     * (LOCK_ERR_OBJ_HAS_CHANGED).
     * @exception GbjException    thrown if there is a server exception
     * @see com.gemstone.gbj.GbjGemStoneErrors#LOCK_ERR_OBJ_HAS_CHANGED
     */
    public void writeLock() throws GbjException {
        getSession().kernel.System.sendMsg("writeLock:", this);
    }


    /**
     * Lock this database object.  Throws an exception if another session
     * holds a conflicting lock on this object.  Also throws an exception
     * if another session has committed a change to this object 
     * (LOCK_ERR_OBJ_HAS_CHANGED).
     * @exception GbjException    thrown if there is a server exception
     * @see com.gemstone.gbj.GbjGemStoneErrors#LOCK_ERR_OBJ_HAS_CHANGED
     */
    public void exclusiveLock() throws GbjException {
        getSession().kernel.System.sendMsg("exclusiveLock:", this);
    }

    /**
     * Release any lock held on this object by this application.  If the object
     * is not locked by this session, a GbjException with number 
     * LOCK_ERR_REMOVE is thrown.
     * @exception GbjException    thrown if there is a server exception
     * @see com.gemstone.gbj.GbjGemStoneErrors#LOCK_ERR_REMOVE
     */
    public void removeLock() throws GbjException {
        getSession().kernel.System.sendMsg("removeLock:", this);
    }

}


