/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjPoller.java  
 *
 *========================================================================
 */

package com.gemstone.gbj;  // Basic Java Interface

import com.gemstone.gbjgci.*;
import com.gemstone.gbj.*;
import java.util.logging.*;


/**
 * GbjPoller provides support for a polling thread that keeps track
 * of asynchronous errors and triggers an appropriate method when they
 * occur.
 * 
 * @version 3.0
 */
public class GbjPoller extends Thread {

    /** Logger for debug trace messages */
    private static Logger log;

    // static code for setting up logger
    {
	log = Logger.getLogger(GbjGciSession.class.getName());
	Level logLevel = GbjGciPreferences.logLevel("GbjPoller");
	log.setLevel(logLevel);
	Handler[] handlers = Logger.getLogger("").getHandlers();
	for (int i=0; i<handlers.length; i++) {
	    handlers[i].setLevel(logLevel);
	}
    }

    /** Session being tracked by this poller */
    public GbjSession session;
    private boolean connected;
    private boolean running;
    private boolean stopRequested;

    // 44674: remove pollInterval from GbjPoller - retrieve from session

    public GbjPoller(GbjSession s) {
	session = s;
        connected = true;
        running = true;
        stopRequested = false;
	this.start();
    }

    public synchronized void stopPoller() {  
      stopRequested = true; 
      while (running) { 
	try {
          Thread.sleep(2/*ms*/);
	} catch(InterruptedException e){};
      }
    }

    // 44674: move checks for gem-to-gem and object change signals here..
    public void run( ) {
	log.fine("GbjPoller thread is running!");
	while (connected && !stopRequested) {
	    try {
		session.checkForSignals(true);
		} catch (GbjGciException e) {
		    System.out.println("[GBJPOLL]: Unexpected GbjGciException " + 
				       e.number);
		    e.printStackTrace();
		} catch (GbjException e) {
		    System.out.println("[GBJPOLL]: GbjException " + e.number);
		    e.printStackTrace();
		} catch (java.lang.NullPointerException e) {
		    // This can happen if we poll after disconnect..
		    connected = false;
		} catch (Throwable e) {
		    System.out.println("[GBJPOLL] Unknown error " + e);
		    e.printStackTrace();
		} finally {
		    if (connected) {
			try {
			    Thread.sleep(session.pollInterval);
			} catch(InterruptedException e){};
			connected = (session.getSessionId( ) > 0);
		    }
                }
        }
        running = false;
	log.fine("GbjPoller thread is stopping!");
    }

}



