/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjSession.java  
 *
 *========================================================================
 */

package com.gemstone.gbj;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import java.util.HashMap;
import java.io.IOException;
import java.util.Observer;
import java.util.Set;      // 44674
import java.util.HashSet;  // 44674
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import org.apache.commons.math3.fraction.BigFraction;

import java.util.logging.*;
import com.gemstone.gbjgci.*;
import com.gemstone.gbj.*;

/**
   A GbjSession is used to communicate with a GemStone server process.  It
   must be configured with a GbjParameters instance to be useful, and must
   be connected to a server peer with the connect() method.  GbjSession
   maintains translation information for Java and GemStone as well as
   export sets and other low-level information. 
   <br>
   In this GBJ jar , multiple instances of GbjSession are supported within a Java VM.
   Access to server state associated with a GbjSession is through
   synchronized instance methods in com.gemstone.gbjgci.GbjGciSession .
   Multiple java threads can simultaneously access server state if each thread
   is using a different instance of GbjSession .

<h2> New in GBJ 3.2 </h2>

   The method commit() was added.

<h3> Deprecated Methods </h3>
Methods that are still available but deprecated due to changes in design:
<ul>
<li> public GbjObject putInServer()
putInServer</ul>

 * @version 3.2
 */
public class GbjSession extends GbjGciSession implements GbjObservable {

    class ExceptionHandlerDetails {
      Object handlerObject;
      Method handlerMethod;
      public ExceptionHandlerDetails(Object o, Method m) {
        handlerObject = o;
        handlerMethod = m;
      }
    }    

    /** Default setting for initialize parameter useLinked (false) */
    private static final boolean DEFAULT_INIT_USE_LINKED = false;

    /** Default setting for initialize parameter appName (GbjApplication) */
    private static final String DEFAULT_INIT_APP_NAME = "GbjApplication";

    /** Default setting for initialize parameter bufSize (100000) */
    private static final int DEFAULT_INIT_BUF_SIZE = 100000;

    /** Default setting for initialize parameter preloadSize (1000) */
    private static final int DEFAULT_INIT_PRELOAD_SIZE = 1000;

    /** Default setting for initialize parameter maxSessions (1000) */
    private static final int DEFAULT_INIT_MAX_SESSIONS = 1000;

    /** Default setting for GbjPoller interval in milliseconds (1000) */
    private static final int DEFAULT_GBJ_POLL_INTERVAL = 1000; 

    private static boolean initialized = false;

    private String gemHost;
    private String gciHost;

    /**
     * The adapter registry maintains a correspondence between Java classes
     * and message adapters.  If a Java class implements GMsgAdapter, it need
     * not be registered to have messages forwarded correctly.
     */
    private Hashtable adapterRegistry;
    
    /**
     * isConnected indicates whether the session is able to communicate with 
     * its server.  Using a session that is not connected will throw an Error.
     * @return boolean result
     */
    public boolean isConnected() { return isLoggedIn(); }
    
    /**
     * A collection of frequently accessed objects from the server
     * Globals dictionary.
     * GBJ30: due to design changes for GbjGciInterface requiring these
     * to be fetched from the server, now limited to just the essentials,
     * rather than *all* Globals.
     */

    public GbjKernelObjects kernel;

    // 40316: restored. Now used only for GbjClientAdapters based on instances
    /**
     * A list of objects exported to the GemStone server 
     * (not ints/floats/strings/etc.)
     */
    private Vector exportedClientObjects;

    /**
     * lastError is deprecated.  It is initialized to null .
     */

    public RuntimeException lastError;
    
    /**
     * A name may be assigned to the session to make it easier to distinguish
     * between sessions.  A name is generated using the broker machine name
     * and a unique session identifier at connect time if no name is given
     * by the application.
     */
    private String name;
    
    /**
     * dependents of the session who will be notified via update() of changes
     */
    private Vector observers;
    
    /**
     * A session's parameters object is used to configure the session at
     * connect time.  It is made private so that the session may be passed
     * to an untrusted object without revealing configuration information.
     */
    private GbjParameters parameters;
    
    /**
     * Server profiling request - must be set before logging it.  This is not
     * a tested feature of the interface.
     */
    protected boolean profileServer;

    /**
     * The static adapter registry holds message translation adapters listed
     * by Java class.
     */
    private static Vector staticAdapterRegistry;
    
    /**
     * The static stub registry holds GemStone/Java class correspondences
     * that can be set up at static initialization time.
     */
    private static Vector staticStubRegistry;

    /**
     *
     * For polling for asynchronous GbjExceptions
     *
     */
    private GbjPoller poller;

    /**
     * Method map for exception handlers
     *
     */
    private HashMap<Integer, ExceptionHandlerDetails> exceptionMap;
    
    /**
     * toolData is for GemStone internal use
     */
    private Object toolData;

    public void setToolData(Object o) { toolData = o; }
    public Object toolData_() { return toolData; }
    
    /**
     * socket operations
     */
    
    /**
     * the server GbjConnection object
     */
    protected GbjObject connection;
    
    /**
     * userData is for customer use.  Any session-related information may be 
     * placed here for convenient access.
     */
    public Object userData;
    
    /**
     * debug flags for troubleshooting session connection problems
     */
    private boolean verboseClient;
    private boolean verboseServer;

    // 40660 
    /** Flag used to control if session signals should be monitored */
    public boolean monitorSessionSignals = false;

    // 44674
    /** Flag used to control if object change notification should be active */
    public boolean objChgNotify = false;

    /** Time in milliseconds since last async signal check */
    protected long lastSignalCheck = 0;

    /** Flag indicating when setting up session is completed */
    private boolean setupComplete = false;
      
    /**
     * The version of the interface
     */
    public static String version = GbxConstants.VERSION_STRING;

    // 41755
    // 44674: remove static definition so can be changed by session
    /**
     * The polling interval used to check for asynchronous signals
     * (sigAbort, gem-to-gem signals, changed objects).
     */
    public int pollInterval = DEFAULT_GBJ_POLL_INTERVAL;
    
    /**
     * communications compatibility level
     */
    private static final int commCompatibility = 0x03000000; 
                                               // <major><major><minor><minor>

    private static Class gbjObjectClass;

    public static Class gbjObjClass() { return gbjObjectClass; }
    
    /** Logger for debug trace messages */
    private static Logger log;

    // static code for setting up logger
    static {
	log = GbjGciInterface.setupLogger(GbjSession.class);
    }

    // method to generate stack trace string
    public static String stackTraceString(Throwable t) {
        String errMsg;
	if (t.getMessage() == null)
	    errMsg = t.getClass().getName();
	else
	    errMsg = t.getClass().getName() + ": " + t.getMessage();
	StackTraceElement[] stack = t.getStackTrace();
	StringBuilder sbStackTrace = new StringBuilder();
	for (int i=0; i< stack.length; i++) {
	    sbStackTrace.append(stack[i]);
	    sbStackTrace.append("\n");
	}
	return sbStackTrace.toString();
    }
    
    //========================================================================
    // PACKAGE methods
    //========================================================================


    // 40316: restored
    /**
     * hold on to a client object that is being exported to the server so
     * it won't be garbage collected, and so it can be found if returned
     * as a reference from the server
     */
    protected void addClientObject(Object clientObj) {
        exportedClientObjects.addElement(clientObj);
    }

    // 40316: restored
    /**
     * find a client object that was retained by addClientObject
     */
    protected Object findClientObject(int hash) {
        int i;
        Enumeration e = exportedClientObjects.elements();
        Object element;
        
        while (e.hasMoreElements()) {
            element = e.nextElement();
            if (element.hashCode() == hash)
                return element;
        }
        return null;
    }

    /**
     * When a connected session is finalized it closes its connection to the
     * GemStone server
     */
    protected void finalize() {
	if (isConnected()) logout();
    }
    
    /**
     * during the close operation a seperate thread will need to notify 
     * observers.  Since the close operation is already synchronized, 
     * the thread would block if it had to go through the public 
     * notification interface.
     */
    protected void privNotifyObservers(String notificationMessage, 
				       Object argument) {
        int i, sz;
        Vector copy;

        if (observers != null) {
            copy = (Vector)observers.clone();
            sz = copy.size();
            for (i=0; i<sz; i++) {
                ((GbjObserver)copy.elementAt(i)).update(
                      (GbjObservable)this, notificationMessage, argument);
            }
        }
    }
    
    /**
     * logging operation for verboseness
     */
    protected void privNotifyLog(String logMessage) {
        if (verboseClient) {
            log.info(logMessage);
            privNotifyObservers("message", logMessage);
        }
    }
    
    /**
     * logging operation for errors
     */
    protected void privNotifyErr(String logMessage) {
        log.severe(logMessage);
        privNotifyObservers("message", logMessage);
    }

    // 40316: restored
    /**
     * find an adapter for the given client object
     * @return the adapter, or null if one can't be found
     */
    protected GbjClientAdapter findClientAdapter(Object obj) {
        Class objClass;
        GbjClientAdapter adapter;
        
        objClass = obj.getClass();
        while (objClass != null) {
            if (adapterRegistry.containsKey(objClass))
                return (GbjClientAdapter)adapterRegistry.get(objClass);
            objClass = objClass.getSuperclass();
        }
        return null;
    }
    

    /**
     * Handle details of processing a GbjGciException from a
     * perform or execute call. If its a 2336 (client forwarder call)
     * trigger execution of the appropriate java code and return control
     * back to server.  fix 40316
     */
    protected GbjObject handleExecutionException(GbjException ex) {
        if (GbjGciSession.isDEBUG) {
          System.out.println("handleExecutionException: " + ex );
          System.out.println( stackTraceString(ex));
        }
	if (ex.number == GbjGemStoneErrors.RT_ERR_CLIENT_FWD_SEND) {
	    // we have a client forwarder send
	    int clientHash = ((GbjObject)ex.args[1]).intValue();
	                                // 45749: loosen typecast
	    Object clientObject = findClientObject(clientHash);
	    String selector = ((GbjObject)ex.args[2]).stringValue();
	    GbjGciObject argCollection = (GbjGciObject)ex.args[3];
	    int size = (int)argCollection.getIndexedSize();
	    GbjObject gbjArgs[] = new GbjObject[size];
	    argCollection.getObjs(1, size, gbjArgs);
	    Object args[] = new Object[size];
	    for (int i=0; i<size; i++) {
		if (gbjArgs[i].isString())   // 45749: handle strings specially
		    args[i] = gbjArgs[i].stringValue();
		else
		    args[i] = gbjArgs[i].cachedValue();
	    }
	    GbxClientInterface clientInterface = 
		new GbxClientInterface(this, clientObject, selector, args);
	    Object result = null;
	    GbjObject stResult = null;
	    GbjObject jvResult = new GbjObject();
	    try {
		result = clientInterface.execute();
		stResult = new GbjObject(result, this); // 45770
	    } catch (Throwable t) {
		String errMsg;
		if (t.getMessage() == null)
		    errMsg = t.getClass().getName();
		else
		    errMsg = t.getClass().getName() + ": " + t.getMessage();
		String stackTrace = stackTraceString(t);
		try {
		    serverContinue(ex.process, true,
			stResult, jvResult, errMsg, stackTrace);
		    return jvResult;
		} catch (GbjGciException e2) {
		    throw new GbjException(e2);
		} 
	    }
	    // now need to return result to the server.. 
	    try {
		return (GbjObject) serverContinue(ex.process, false, stResult, jvResult, null, null);
	    } catch (GbjGciException e2) {
		throw new GbjException(e2);
	    } 
	}

        if (exceptionMap == null) {
          throw ex; // error during login or session init
        }
        ExceptionHandlerDetails d = exceptionMap.get(ex.number);
        if (d == null) {
          // no exception handler found -- throw original error
          throw ex;
        }

        if (GbjGciSession.isDEBUG) log.fine("Exception handler found for error " + ex.number);
        Object h = d.handlerObject;
        Method m = d.handlerMethod;
        GbjObject jvResult;
        try {
          jvResult = (GbjObject)m.invoke(h, this, ex); 
        } catch (IllegalArgumentException x) {
            log.warning("Problem with exception handler set for " + ex.number);
            throw ex;
        } catch (IllegalAccessException x) {
            log.warning("Don't have public access to exception handler " + ex.number);
            throw ex;
        } catch (InvocationTargetException x) {
            log.warning("Exception thrown by exception handler " + ex.number);
            Throwable targetException = x.getTargetException();
            System.out.println("[ERROR] Error in exception handler: ");
	    System.out.println(stackTraceString(x));
            System.out.println("[ERROR] Throwing original error.");
            throw ex;
        }
        return jvResult;
    }

    //========================================================================
    // PRIVATE methods
    //========================================================================
    
    /**
     * builds a stub lookup table based on the static table and
     * the current superclass lookup table.  This also considers
     * the well known classes defined in GbjSession.
     */
    private void initStubLookup() {
        int i, sz;
        GbjObject cls, o;
        GbjCollection c;

	// no mapping for CharacterCollection, they are all VALTYPE_STRING
	mapGsToJavaClass(objectNamed("Collection"), new GbjCollection());
        try {
          if (staticStubRegistry != null) { // register class-name/stub pairs
            sz = staticStubRegistry.size();
            for (i=0; i<sz; i+=2) {
              cls = objectNamed((String)staticStubRegistry.elementAt(i));
              if (cls == null) {
                throw new Error("Unable to find class " + (String)staticStubRegistry.elementAt(i) + " for stub registration");
              } else {
		mapGsToJavaClass( cls, (GbjObject)staticStubRegistry.elementAt(i+1));
              }
            }
          }
        } catch (GbjException e) {
            // the app can't continue at this point, so stop
            throw new Error("Unable to initialize stub registry: " + e.getMessage());
        }
    }
    
    /**
     * initialize the interface lookup table for dispatching client messages 
     * from the server
     */
    private void initAdapters() {
        int i;
        
        adapterRegistry = new Hashtable();
        // registerAdapter("java.util.Vector", new GVectorAdapter());
        // registerAdapter("java.util.Hashtable", new GHashtableAdapter());
        // registerAdapter("java.lang.Class", new GClassAdapter());
        if (staticAdapterRegistry != null) {
            for (i=0; i<staticAdapterRegistry.size(); i+=2) {
                adapterRegistry.put(staticAdapterRegistry.elementAt(i),
                staticAdapterRegistry.elementAt(i+1));
            }
        }
    }

    // 40660
    /**
     * Used by GbjPoller to test for gem-to-gem and changed object notification
     * signals.  
     *
     * Format of pollData: a two-element array:
     * #1:  Array of gem-to-gem signal arrays:
     *      #1:  sender session serial number
     *      #2:  message number
     *      #3:  message string
     * #2:  Array of objects that have triggered object change notification
     * 
     */
    protected void handleSignalData(GbjCollection pollData) {
	// 44659: add check for kernel != null
	GbjObject pollC[] = ((GbjCollection)pollData).allElements();
	GbjObject g2gData = pollC[0];
	GbjObject ocData  = pollC[1];
	if (monitorSessionSignals && (kernel != null)) {
	    GbjObject g2gDataC[] = ((GbjCollection)g2gData).allElements();
	    for (int i=0; i<g2gDataC.length; i++) {

		// process signal
		GbjObject sig[] = ((GbjCollection)g2gDataC[i]).allElements();
		GbjException e = new GbjException(sig[2].stringValue());
		e.category = kernel.GemStoneError;
		e.number = GbjGemStoneErrors.RT_ERR_SIGNAL_GEMSTONE_SESSION;
		e.args = sig;
		e.argCount = sig.length;
		privNotifyObservers(sig[2].stringValue(), e);
	    }
	}

	// 44674: check for signaled objects in the object notify set
	if (objChgNotify && (kernel != null)) {
	    GbjObject notifyObjs[] = ((GbjCollection)ocData).allElements();
	    if (notifyObjs.length > 0) {
		notifyObservers("ObjChg", notifyObjs);
	    }
	}
    }

    protected void checkForSignals(Boolean poller) {    // 44674
	if (!setupComplete) return;
	long currentTime = System.currentTimeMillis();
	if ((lastSignalCheck + pollInterval) < currentTime) {
	    lastSignalCheck = currentTime;
	    if (monitorSessionSignals || objChgNotify) {
                GbjObject args[] = new GbjObject[2];
                args[0] = new GbjObject(monitorSessionSignals, this);
                args[1] = new GbjObject(objChgNotify, this);
                // System.out.println("checkForSignals " + poller + " chkG2G");
                GbjCollection pollData = 
                 (GbjCollection)perform(kernel.GbjGciInterface, "chkG2G:chkObjChg:", args, 2);
		handleSignalData(pollData);
	    } else if (poller) {
		// Use the simpler poll method
                // System.out.println("checkForSignals " + poller + " poll");
		perform(kernel.GbjGciInterface, "poll", null, 0);
	    }
	} else {
            // System.out.println("checkForSignals " + poller + " do nothing");
        }
    }




    //========================================================================
    // PUBLIC methods
    //========================================================================

    // placeholder
    public GbjSession( ) { }
    
    /**
     * GbjSession() requires a sessionParameters object, which tells the 
     * session the host and port for the connection as well as the type 
     * of session requested
     * @param sessionParameters     a configured instance of GbjParameters
     */
    public GbjSession( GbjParameters sessionParameters) {
        parameters = sessionParameters;
        verboseClient = false;
        verboseServer = false;
        profileServer = false;
        initAdapters();
        exportedClientObjects = new Vector();  // 40316: restored
    }

    /**
     * registerStaticStub allows building of a GemStone@lt-@gtJava class 
     * correspondence at static initialization time.  The stub instance 
     * is used to clone new proxies to represent instances of the given 
     * GemStone class at run time.
     * @param stubInstance     an instance of the stub class 
     * (must currently be a GbjObject)
     * @param gemstoneClassName the name of the GemStone class, 
     * which is found at session initialization time
     */
    public static void registerStaticStub(GbjObject stubInstance, 
					  String gemstoneClassName) {
        if (staticStubRegistry == null) {
            staticStubRegistry = new Vector();
        }
        staticStubRegistry.addElement(gemstoneClassName);
        staticStubRegistry.addElement(stubInstance);
    }
    /**
     * registerStub allows building of a GemStone@lt-@gtJava class 
     * correspondence
     * at run time.  The stub instance is used to clone new proxies to 
     * represent instances of the given GemStone class.
     * @param stubInstance     an instance of the stub class 
     * (must currently be a GbjObject)
     * @param gemstoneClassName the name of the GemStone class
     * @exception GbjException    thrown if the session is not connected 
     * or there is a server exception
     */
    public void registerStub(GbjObject stubInstance, String gemstoneClassName) throws GbjException {
	registerStub(stubInstance, objectNamed(gemstoneClassName));
    }
    
    
    /**
     * registerStub allows building of a GemStone&lt;-&gt;Java class 
     * correspondence
     * at run time.  The stub instance is used to clone new proxies to 
     * represent instances of the given GemStone class.
     * @param stubInstance     an instance of the stub class 
     * (must currently be a GbjObject)
     * @param classStub     a GbjObject stub for the server class object
     * @exception GbjException    thrown if the session is not connected 
     * or there is a server exception
     */
    // GBJ30: superceeded by mapGsToJavaClass
    public void registerStub(GbjObject stubInstance, 
			     GbjObject classStub) throws GbjException {

	mapGsToJavaClass(classStub, stubInstance);
    }
    
    /**
     * registerStaticAdapter specifies an object to be used in forwarding 
     * messages for the given Java class.  The adapter is sent the receiver, 
     * method name, and a Vector holding the arguments and must forward the 
     * message to the object.  This part of the API will be removed when 
     * Java has enough reflective access to allow the GemStone interface 
     * to dispatch messages automatically.
     *
     * The static form of registration establishes the adapter for all 
     * subsequently instantiated sessions.
     *
     * An class can function as its own adapter without need of registration by
     * implementing the interface GMsgAdapter.
     *
     * See documentation on the GMsgAdapter interface for more information on
     * implementing an adapter.
     * @param className     A fully qualified class name.  E.g., java.util.Stack
     * @param adapter       The object that will forward messages
     * @see com.gemstone.gbj.GbjSession#registerAdapter
     */
    public synchronized static void registerStaticAdapter(String className, GbjClientAdapter adapter) {

        Class classObj;
     
        try {
            classObj = Class.forName(className);
            if (staticAdapterRegistry == null)
                staticAdapterRegistry = new Vector(2);
            staticAdapterRegistry.addElement(classObj);
            staticAdapterRegistry.addElement(adapter);
        }
        catch (ClassNotFoundException c) {
            log.warning("Warning: unable to find " + 
			className + 
			" - GemStone adapter not installed");
        }
    }
    
    
    /**
     * registerAdapter specifies an object to be used in forwarding messages
     * for the given Java class.  The adapter is sent the receiver, method 
     * name, and a Vector holding the arguments and must forward the message
     * to the object.  This part of the API will be removed when Java has 
     * enough reflective access to allow the GemStone interface to dispatch 
     * messages automatically.
     *
     * An class can function as its own adapter without need of registration by
     * implementing the interface GMsgAdapter.
     *
     * See documentation on the GMsgAdapter interface for more information on
     * implementing an adapter.
     * @param className     A fully qualified class name.  E.g., java.util.Stack
     * @param adapter       The object that will forward messages
     * @see com.gemstone.gbj.GbjSession#registerStaticAdapter
     */
    public synchronized void registerAdapter(String className, GbjClientAdapter adapter) {

        Class classObj;
        
        try {
            classObj = Class.forName(className);
            adapterRegistry.put(classObj, adapter);
        }
        catch (ClassNotFoundException c) {
            log.warning("Warning: unable to find " + 
			className + 
			" - GemStone adapter not installed");
        }
    }
    
/**
Initializes the GBJ environment.  
<p>
If used, this method should be called before the first GbjSession.connect( ).
If not explicitly called first by the application, this method will be
automatically called during the first GbjSession.connect( ), 
using the default values specified below.
<p>
@param useLinked    If true, first session login is linked.  Must be false .

This GBJ jar does not support linked sessions.  

@param appName      Optional application name.

The name that is displayed in the GS/S session cache statistics
and statmonitor/VSD output for this gem process.  
Default is "GbjApplication".

@param bufSize      Maximum size (in bytes) for internal byte/oop buffers.

The maximum number of bytes that can be transfered in a single
getOops()/getNamedOops()/getBytes() operation.  Exceeding this value will 
trigger a GbjGciException.  The value is used to allocate internal buffers
for storage during transfer.
Default is 100000.

@param preloadSize  Maximum size (in bytes) for initial preload of objects

When an object is initially fetched from the server, this 
parameter specifies how many bytes are preloaded with the initial fetch.
These are cached, and subsequent getOops()/getNamedOops()/getBytes() 
calls can retrieve their results from this cache if available.  Application
developers can tune performance by sizing this as appropriate -- larger
sizes will improve the performance of subsequent getX operations at the 
expense of a slower initial object fetch.  1000 bytes is a good starting value.
Default is 1000.

@param maxSessions  Maximum number of sessions that will be run on this Stone.

This should be equal to or greater than the setting of the stone configuration
file parameter STN_MAX_SESSIONS.
Default is 1000.

@param objectClass  Defines what class of object is returned from the server
(unless overriden by registerStub()).  

If the Application extends GbjObject functionality, 
supply the appropriate application class here so that results are
returned as the appropriate class.  Defaults to GbjObject.

*/
    public static synchronized void initialize(boolean useLinked, 
                                  String appName, 
                                  int bufSize, 
                                  int preloadSize, 
                                  int maxSessions,
                                  Class objectClass) {
        int status = 0; 
        if (initialized) 
            throw new GbjException("GbjSession already initialized");
        try {
            status = GbjGciInterface.initialize(useLinked, 
                                                 appName, 
                                                 bufSize, 
                                                 preloadSize,
                                                 maxSessions);
          gbjObjectClass = objectClass;
        } catch (GbjGciException e) { 
            throw new GbjException(e);
        }
        catch (Throwable t1) {
            t1.printStackTrace();
            throw new GbjException("Problem loading GbjGciLibrary");
        }
        initialized = true;
        // ensure apache jar is in CLASSPATH .
        BigFraction bf = new BigFraction(1,3);
    }    

   private static synchronized void initializeIfNeeded() {
	if (! initialized) {
           if (GbjGciSession.isDEBUG)  log.fine("GbjGciInterface not initialized: defaults used.");
	   initialize(DEFAULT_INIT_USE_LINKED, 
			  DEFAULT_INIT_APP_NAME, 
			  DEFAULT_INIT_BUF_SIZE, 
			  DEFAULT_INIT_PRELOAD_SIZE,
                          DEFAULT_INIT_MAX_SESSIONS);
        }
   }

/**
Initializes the GBJ environment.  
<p>
See initialize(useLinked, appName, bufSize, preloadSize, maxSession, objectClass) for parameter details.
@param useLinked      See initialize( ... objectClass)  must be false
@param appName        See initialize( ... objectClass)
@param bufSize        See initialize( ... objectClass)
@param preloadSize    See initialize( ... objectClass)
@param maxSessions    See initialize( ... objectClass)
*/

    public static void initialize(boolean useLinked, 
                                  String appName, 
				  int bufSize, 
                                  int preloadSize, 
                                  int maxSessions) {
        initialize(useLinked, appName, bufSize, 
                   preloadSize, maxSessions, GbjObject.class);
    }

/**
Retrieve the designated bytes from this byte-based GS object.
@param obj         The GS object to retrieve the bytes from.
@param startIndex  The starting index (1-based)
@param numBytes    The number of bytes to retrieve
@return the retrieved bytes
@exception com.gemstone.gbjgci.GbjGciException  thrown on any errors
*/

    public byte[] getBytes(GbjObject obj, long startIndex, int numBytes) {
	try {
	    return super.getBytes(obj, startIndex, numBytes);
	} catch (GbjGciException e) {
	    throw new GbjException(this, e);
	}
    }

/**
Retrieve the designated GS objects from this indexable GS object
@param obj         The GS object to retrieve the GS objects from.
@param startIndex  The starting index (1-based)
@param numOops     The number of objects to retrieve
@return An array of GS objects containing the results
@exception com.gemstone.gbjgci.GbjGciException  thrown on any errors
*/
    public GbjObject[] getObjs(GbjObject obj, long startIndex, int numOops) {
	if (GbjGciSession.isDEBUG) log.fine("getObjs: startIndex " + startIndex + " numOops " + numOops);
	try {
            GbjObject[] res = new GbjObject[numOops];
            super.getObjs(obj, startIndex, numOops, res);
            return res;
	} catch (GbjGciException e) {
	    throw new GbjException(this, e);
	}
    }

/**
Retrieve the designated GS objects from the named instvars for this GS object.
@param obj         The GS object to retrieve the GS objects from.
@param startIndex  The starting index (1-based)
@param numOops     The number of objects to retrieve
@return An array of GS objects containing the results
@exception com.gemstone.gbjgci.GbjGciException  thrown on any errors
*/
    public GbjObject[] getNamedObjs(GbjObject obj, long startIndex, int numOops) {
	try {
          GbjObject[] res = new GbjObject[numOops];
	  super.getNamedObjs(obj, startIndex, numOops, res);
          return res;
	} catch (GbjGciException e) {
	    throw new GbjException(this, e);
	}
    }

    /**
     * Set debugging options.  These debugging options can be used to diagnose
     * problems in the interface.  While debugging options may be set at any
     * time, they are most useful if turned on before connecting to the server.
     * <br>
     * Setting verboseClient to true will cause the interface to write
     * messages to Java's System.out stream.  Messages are also sent to
     * observers of the session with a "message" designator and the string
     * as an argument.
     * <br>
     * Setting verboseServer to true will cause the server to write
     * messages to the session's log file on the server.
     * @param verboseClient   display debugging information on System.out, 
     * and notify observers
     * @param verboseServer   write debugging information to 
     * gbj&lt;sessionId&gt;.log on server
     */
    public synchronized void setDebugOptions(boolean verboseClient, boolean verboseServer) {
        if (verboseClient) log.setLevel((Level)Level.FINE);
        else log.setLevel((Level)Level.INFO);
        if (this.verboseServer != verboseServer) {
            if (isConnected()) {
                try {
		    // 40306: talk to GbjGciInterface now..
                    doit("GbjGciInterface verbose: " + verboseServer);
                    this.verboseServer = verboseServer;
                } catch (GbjException e) {
                    privNotifyErr("Unable to set server verbose option :" + 
				  e.getMessage());
                }
            }
            else
                this.verboseServer = verboseServer;
        }
    }
    
    /** connect() is usually sent soon after instantiating a session to connect
     * to the server.  The session is not very usable until it's been connected.
     * @see com.gemstone.gbj.GbjSession#close
     * @exception GbjException thrown if there is a server exception 
     * while connecting
     */
    public synchronized void connect() throws GbjException {
        if (GbjGciSession.isDEBUG) log.fine("GbjSession: connect()..");
        initializeIfNeeded();
        if (isConnected()) {
            GbjException error;
            error = new GbjRuntimeException();
            error.number = GbjGemStoneErrors.GCI_ERR_ALREADY_LOGGED_IN;
            error.message = "An attempt was made to connect after a connection was already established";
            error.category = kernel.GemStoneError;
            error.argCount = 0;
            error.process = kernel.Nil;
            throw error;
        }

	// No stone name supplied
        if (parameters.serverName == null) {
	    // 44659 - kernel oops not setup yet
	    throw new GbjException("No server name supplied");
        }

	try {
	    login(parameters.serverName,
		  parameters.gemnetName,
		  parameters.userName,
		  parameters.password,
		  parameters.serverOSUserName,
		  parameters.serverOSPassword);
	} catch (GbjGciException e) {
	    throw new GbjException(e);
	}
	// setup transaction mode
	switch (parameters.transactionMode) {
	case GbjParameters.ManualTransactions:
	    setManualTransactionMode();
	    break;
	case GbjParameters.AutomaticTransactions:
	    // get this by default
	    break;
	case GbjParameters.Transactionless:
	    setTransactionlessMode();
	    break;
	}
	
        exportedClientObjects = new Vector();  // 40316: restored
	kernel = new GbjKernelObjects(this, GbjGciInterface.kernelOops);
        initStubLookup();

	gemHost = this.doit("(System descriptionOfSession: (System session)) at: 3").stringValue();

	gciHost = this.doit("(System descriptionOfSession: (System session)) at: 11").stringValue();

	poller = new GbjPoller(this); 
        exceptionMap = new HashMap<Integer, ExceptionHandlerDetails>();

       // This should be last..
       setupComplete = true;

    }
    
    
    /**
     * close() closes the connection.
     * @see com.gemstone.gbj.GbjSession#connect
     */
    public void close() {
      poller.stopPoller();
      close_();
    }
    private synchronized void close_() {
        if (GbjGciSession.isDEBUG)  log.fine("Closing connection..");
        if (false) { // uncomment (change to true) to get trace of logouts
          try { 
            throw new GbjGciException("logout trace");
          } catch (GbjGciException e) {
            System.out.println("--x GbjSession logout trace:");
            System.out.print(stackTraceString(e));
            System.out.println("--x ");
          }
        }
        if (! isConnected())
            return;
        try {
	    logout();
        } catch (GbjGciException ignore) {
        }
	setupComplete = false;
    }
    
    //========================================================================
    // GemStone server's export set
    //========================================================================
    /// GBJ30 - export set now managed by GbjGciSession internally..


    //========================================================================
    // client's export set
    //========================================================================
    // GBJ30 - not currently supported


    // 40316: new
    /* 
     * Return a server clientObject for this java object
     */
    public GbjObject clientObject(Object o) {
        GbjObject result = 
	    doit("GbjGciInterface clientForwarder: " + o.hashCode());
	addClientObject(o);
	return result;
    }

    
    //========================================================================
    // transaction control
    //========================================================================
    
    /**
     * abortTransaction() rolls back the current transaction in the database.
     * @see com.gemstone.gbj.GbjSession#beginTransaction
     * @see com.gemstone.gbj.GbjSession#commitTransaction
     * @see com.gemstone.gbj.GbjSession#commitAndReleaseLocks
     * @exception GbjException    thrown if the session is not connected 
     * or there is a server exception
     */
    public void abortTransaction() throws GbjException {
        // doit("System abortTransaction");
        kernel.System.perform("abortTransaction", null, 0);
        notifyObservers("abort", null);
    }
    
    /**
     * rollbackTransaction() rolls back the current transaction in the 
     * database.  This is the same as abortTransaction().
     * @exception GbjException    thrown if the session is not connected 
     * or there is a server exception
     * @see com.gemstone.gbj.GbjSession#abortTransaction
     * @see com.gemstone.gbj.GbjSession#beginTransaction
     * @see com.gemstone.gbj.GbjSession#commitTransaction
     * @see com.gemstone.gbj.GbjSession#commitAndReleaseLocks
     */
    public void rollbackTransaction() throws GbjException {
        abortTransaction();
    }
    
    /**
     * beginTransaction() starts a new transaction in the database.  
     * This will throw a GbjException with number RT_ERR_BEGIN_TRANS 
     * if a transaction is already in progress.  You must specify 
     * ExplicitTransactions in the transactionMode parameter for a session,
     * or set manual transaction mode with System/transactionMode: before 
     * using this method.  The default transaction mode for Java sessions
     * explicit transactions.
     * @exception GbjException    thrown if the session is not connected 
     * or there is a server exception
     * @see com.gemstone.gbj.GbjSession#abortTransaction
     * @see com.gemstone.gbj.GbjSession#commitTransaction
     * @see com.gemstone.gbj.GbjSession#commitAndReleaseLocks
     */
    public void beginTransaction() throws GbjException {
        // doit("System beginTransaction");
        kernel.System.perform("beginTransaction", null, 0);
        notifyObservers("begin", null);
    }
    
    /**
     * commitTransaction() commits the current transaction in the database.
     * <br>
     * This will throw a GbjException with number RT_ERR_PRIM_OUTSIDE_TRANS if
     * a transaction is not in progress.<br>
     * In GemStone, a commit may fail due to read-write/write-read conflicts
     * or write/write conflicts.  If the commit operation should fail due
     * to one of these conditions, this method returns false.  If the
     * commit is successful, this method returns true.
     * @return boolean result
     * @exception GbjException    thrown if the session is not connected 
     * or there is a server exception
     * @deprecated use commit() for new code to ensure that conflicts signal a GbjException.
     * @see com.gemstone.gbj.GbjSession#abortTransaction
     * @see com.gemstone.gbj.GbjSession#beginTransaction
     * @see com.gemstone.gbj.GbjSession#commitAndReleaseLocks
     */
    public boolean commitTransaction() throws GbjException {
        // result = doit("System commitTransaction").booleanValue();
        boolean result = kernel.System.perform("commitTransaction", null, 0).booleanValue();
        if (result)
            notifyObservers("commit", null);
        return result;
    }

    /**
     * commit() commits the current transaction in the database.
     * <br>
     * @return boolean true , or throws a GbjException if commit failed
     * @exception GbjException    thrown if the session is not connected ,
     * if the session is not in a transcation, 
     * or there is a server exception, or if there are  conflicts
     **/ 
    public boolean commit() throws GbjException {
        boolean result;
        result = kernel.System.perform("commit", null, 0).booleanValue();
        if (result)
            notifyObservers("commit", null);
        return result;
    }
    
    /**
     * commitAndReleaseLocks() is like commitTransaction but any locks held
     * are released if the commit succeeds.  This is equivalent to the 
     * ODMG commit operation while the commitTransaction() method is 
     * equivalent to the ODMG checkpoint operation.
     *<br>
     * This will throw a GbjException with number RT_ERR_PRIM_OUTSIDE_TRANS if
     * a transaction is not in progress.<br>
     * In GemStone, a commit may fail due to read-write/write-read conflicts
     * or write/write conflicts.  If the commit operation should fail due
     * to one of these conditions, this method returns false.  If the
     * commit is successful, this method returns true.
     * @return boolean result
     * @exception GbjException    thrown if the session is not connected 
     * or there is a server exception
     * @see com.gemstone.gbj.GbjSession#abortTransaction
     * @see com.gemstone.gbj.GbjSession#beginTransaction
     * @see com.gemstone.gbj.GbjSession#commitTransaction
     */
    public boolean commitAndReleaseLocks() throws GbjException {
        // boolean result = doit("System commitAndReleaseLocks").booleanValue();
        boolean result = kernel.System.perform("commitAndReleaseLocks", null, 0).booleanValue();
        if (result)
            notifyObservers("commit", null);
        return result;
    }
    
    /**
     * Returns true if the GemStone server has started a transaction
     * and false if not.
     * @see com.gemstone.gbj.GbjSession#beginTransaction
     * @see com.gemstone.gbj.GbjSession#setAutomaticTransactionMode
     * @see com.gemstone.gbj.GbjSession#setManualTransactionMode
     * @return boolean result
     * @exception GbjException    thrown if the session is not connected
     * or there is a server exception
     */
    public boolean inTransaction() throws GbjException {
        // return doit("System inTransaction").booleanValue();
        return kernel.System.perform("inTransaction", null, 0).booleanValue();
    }
    
    /**
     * setManualTransactionMode() will set the GemStone transaction mode 
     * to  manual (explicit <i>beginTransaction</i> required).  This is 
     * the default transaction mode for connections from Java clients.
     * @see com.gemstone.gbj.GbjParameters#transactionMode
     * @see com.gemstone.gbj.GbjSession#setAutomaticTransactionMode
     * @see com.gemstone.gbj.GbjSession#setTransactionlessMode
     * @exception GbjException    thrown if the session is not connected
     * or there is a server exception
     */
    public void setManualTransactionMode() throws GbjException {
        
        doit("System transactionMode: #manualBegin");
        notifyObservers("abort", null);
        
    }
    
    /**
     * setAutomaticTransactionMode() will set the GemStone transaction mode
     * to automatic (chained transactions, with an implicit begin after a 
     * commit or abort).
     * @see com.gemstone.gbj.GbjParameters#transactionMode
     * @see com.gemstone.gbj.GbjSession#setManualTransactionMode
     * @see com.gemstone.gbj.GbjSession#setTransactionlessMode
     * @exception GbjException    thrown if the session is not connected
     * or there is a server exception
     */
    public void setAutomaticTransactionMode() throws GbjException {
        doit("System transactionMode: #autoBegin");
        notifyObservers("abort", null);
    }
    
    
    /**
     * setTransactionlessMode() will set the GemStone transaction mode to
     * transactionless transaction mode.  This mode is intended for idle
     * sessions.  Applications using transactionless mode may see an 
     * inconsistent view of the database.
     * @see com.gemstone.gbj.GbjParameters#transactionMode
     * @see com.gemstone.gbj.GbjSession#setAutomaticTransactionMode
     * @see com.gemstone.gbj.GbjSession#setManualTransactionMode
     * @exception GbjException    thrown if the session is not connected
     * or there is a server exception
     */
    public void setTransactionlessMode() throws GbjException {
        doit("System transactionMode: #transactionless");
        notifyObservers("abort", null);
    }
    
    
    /**
     * isAutomaticTransactionModeSet() will return true if the server is 
     * in automatic transaction mode and false if not.
     * @return boolean result
     * @exception GbjException   thrown if the session is not connected
     * or there is a server exception
     */
    public boolean isAutomaticTransactionModeSet() throws GbjException {
        return doit("System transactionMode == #autoBegin").booleanValue();
    }
    
    /** 
     * getTransactionMode() returns a string describing the current 
     * transaction mode.  The string will be one of "autoBegin", 
     * "manualBegin", or "transactionless".
     * @return string result
     * @exception GbjException  thrown if the session is not connected 
     * or there is a server exception
     */
    public String getTransactionMode() throws GbjException {
        // return doit("System transactionMode").stringValue();
        return kernel.System.perform("transactionMode", null, 0).stringValue();
    }
    
    /** 
     * setTransactionMode() sets the transaction mode of the session. 
     * Acceptable values are "autoBegin", "manualBegin" or "tranactionless".
     * @param mode String
     * @throws GbjException thrown if passed an invalid transaction mode, 
     * if the session is not connected or there is a server exception.
     */
    public void setTransactionMode(String mode) throws GbjException {
        doit("System transactionMode: #" + mode);
        notifyObservers("abort", null);
    }
    //========================================================================
    // object lookup and ad-hoc execution
    //========================================================================
    
    /**
     * objectNamed() is a name lookup service.  
     * The @param smalltalkObjectName must refer to a name in the 
     * symbolList of the GemStone user-account on the server.  If the name is found,
     * a stub is returned for the object, otherwise null is returned.
     * @exception GbjException    thrown if the session is not connected
     * or there is a server exception
     */
  public GbjObject objectNamed(String smalltalkObjectName) throws GbjException {
      GbjObject result = (GbjObject)super.objectNamedOrNil(smalltalkObjectName);
      if (result.isNil())
          return null;
      return result;
  }

    /**
     * Execute smalltalk code on the server.
     * @param smalltalkCode String containing code
     *
     */
    public GbjObject execute(String smalltalkCode) throws GbjException {
	checkForSignals(false);  // 44674
	try {
	    return (GbjObject)super.execute(smalltalkCode);
	} catch (GbjException e) {
	    return handleExecutionException(e); 
	}
    }


    /**
     * Perform a particular smalltalk method against a particular object.
     * @param receiver GbjObject to send message against
     * @param selector string with method name
     * @param args  GbjObject array containing arguments
     * @param argCount number of arguments in args array
     * @return GbjObject result
     */
    public GbjObject perform(GbjObject receiver, String selector, GbjObject[] args, int argCount) {
	checkForSignals(false);  // 44674
        try {
	  return (GbjObject)super.perform( receiver, selector, args, argCount);
        } catch (GbjException e) {
            return handleExecutionException(e); 
        }
    }

    /**
     * doit() sends the given @param smalltalkCode string to the server for 
     * compilation and execution.  The code may have temporaries and return 
     * ("^") statements, but may not refer to "self", or "super".  The result
     * of execution (a stub for a remote object) is returned.
     * @param smalltalkCode String containing code
     * @return GbjObject result
     * @exception GbjException    thrown if the session is not connected 
     * or there is a server exception
     */
    public GbjObject doit(String smalltalkCode) throws GbjException {
	return execute(smalltalkCode);
    }
    
    /**
     * printit() is the same as doit(), but the message "printString" is 
     * sent to the result and the answer of this message is returned.  The
     * answer will always hold a string value (see GbjObject.stringValue()).
     * @param smalltalkCode String containg code
     * @return String result
     * @exception GbjException    thrown if the session is not connected 
     * or there is a server exception
     */
    public String printit(String smalltalkCode) throws GbjException {
        GbjObject result;
        try {
	  result = execute(smalltalkCode);
	  result = (GbjObject)super.perform(result, "printString", null, 0);
	  return result.asString();
	} catch (GbjException e) {
	    return handleExecutionException(e).asString(); 
	}
    }
    

    //========================================================================
    // miscellaneous methods
    //========================================================================
    
    /**
     * putInServer() puts the given object in the server and returns a 
     * GbjObject stub for it.  This does not make the object persistent - 
     * to make it persistent it must be named, or reachable from a named 
     * object.
     * @param obj Object to put on server
     * @return GbjObject result
     * @exception GbjException    thrown if the session is not connected 
     * or there is a server exception
     * @deprecated use GbjObject constructor
     */
    @Deprecated
    public GbjObject putInServer( Object obj ) throws GbjException {
	return new GbjObject(obj, this);
    }

    /**
     * getName() will return an identifying description of a session.  The
     * default name is the session broker's machine, an underscore, and the 
     * session id assigned at connect time (nil if not connected).
     * @return String value
     * @see com.gemstone.gbj.GbjSession#setName
     */
    public String getName() {
        if (name != null)
            return name;
        else
            return "gem:" + gemHost + "/gci:" + gciHost + "/sid:" + getSessionId();
    }
    
    /**
     * getShortName() will return a short identifying description of a 
     * session.  The short name includes the string "Session: " followed
     * by the sessionId.
     * @return String result
     */    
    public String getShortName(){
        return ("Session: " + getSessionId());
    }
    
    /**
     * setName() will assign an identifying description to a session.  
     * If a name is not assigned, a system-generated name is used.
     * @param newName String for name
     * @see com.gemstone.gbj.GbjSession#getName
     */
    public void setName(String newName) {
        name = newName;
    }
    
    /**
     * return a description of this session.
     * @return String containing description.
     */
    public String toString() {
        return "GemStone Session(" + getName() + ")";
    }
    
    /**
     * checkConnected() will throw a GbjException if the session is not in 
     * a connected state.  It checks the state of the connected flag to 
     * determine if the session is connected.
     * @exception GbjException    thrown if this session is not connected
     */
    public void checkConnected() throws GbjException {
        if (!isConnected()) {
            GbjException error;
            error = new GbjFatalException();
            error.number = GbjGemStoneErrors.GCI_ERR_BAD_SESSION_ID; // 4100
            error.message = "An attempt was made to use a session that was not connected";
            error.category = kernel.GemStoneError;
            error.argCount = 0;
            error.process = kernel.Nil;
            throw error;
        }
    }
    
    //========================================================================
    // session observation methods
    //========================================================================
    
    /**
     * add an observer to the receiver
     */
    public void addObserver(GbjObserver obs) {
        if (observers == null)
            observers = new Vector();
        if (!observers.contains(obs))
            observers.addElement(obs);
    }
    
    /**
     * remove an observer
     */
    public void removeObserver(GbjObserver obs) {
        if (observers != null && observers.contains(obs))
            observers.removeElement(obs);
    }
    
    /**
     * notify observers of a change
     */
    public void notifyObservers(String notificationMessage, Object argument) {
        privNotifyObservers(notificationMessage, argument);
    }


    // 41755
    /**

     Specify an object and method to execute when a particular
     GbjException is received.  This method provides a way of
     establishing client-side exception handlers for any exception,
     but is particularly useful for the following GemStone error codes:
     <ul>
     <li> 6008 - Changed object notification
     <li> 6009 - SigAbort handling
     <li> 6010 - Gem-to-Gem signals
     </ul>
     Arguments are:
     <ul>
     <li>    1.  The GS error code
     <li>    2.  The object to send the message to.
     <li>    3.  The method to execute.
     </ul>
     Method must be recognized by the object, and have two arguments:
     <ul>
     <li>    1.  The GbjSession receiving the exception
     <li>    2.  The GbjException that was triggered
     </ul>
     Method must return a GbjObject as a result
     @param errCode  The int error code to be handled
     @param handlerObject   The designated object to handle the exception
     @param handlerMethodName  The name of the method to be called when
                               the exception is raised
     @exception NoSuchMethodException thrown if method doesn't exist
     */
    public void setupExceptionHandler(int errCode, 
                                      Object handlerObject,
                                      String handlerMethodName)
           throws NoSuchMethodException {

       log.info(handlerObject.getClass().getName() + "." +
                handlerMethodName + " will be used to handle exception " +
                errCode);
       
       Method handlerMethod = handlerObject.getClass().getMethod(
            handlerMethodName,
            new Class[]{GbjSession.class, GbjException.class});
       exceptionMap.put(
         errCode, 
         new ExceptionHandlerDetails(handlerObject, handlerMethod));
    }

}

