/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjGciException.java  
 *
 *========================================================================
 */

package com.gemstone.gbjgci;
import java.io.PrintStream;
import java.io.PrintWriter;

/** 
GbjGciException provides the mechanism for passing errors occuring
in the GemStone GCI and JNI layers up to java.
<p>
In package GBJ, the class GbjException does not inherit from this class, but
includes code to accept a GbjGciException as an argument and processes it 
further.
<p>
Most of the fields in GbjGciException represent data passed from
the GemStone GCI.  Refer to appropriate GemStone documentation for
information on these fields.
<p>
The field "number" indicates a GCI error number.  For non-GCI errors
originating in the JNI, the GCI error 2318 (RT_ERR_GENERIC_ERROR)
is used, and information about the non-GCI error contained in the
message field.
<p>

@version 3.0

*/

public class GbjGciException extends RuntimeException {

    /** GCI error number. <br>
	2318/RT_ERR_GENERIC_ERROR used for non-GCI errors */
    public int number;
    /** Text message describing error */
    public String message;
    /** GS oop of the error dictionary categorizing this error */
    public long category;
    /** GS oop of the context GsProcess that the error occurred in */
    public long context;
    /** Boolean flag indicating if this is a fatal error or not */
    public boolean fatal;
    /** Number of additional arguments associated with this error */
    public int argCount;
    /** List of additional arguments associated with this error */
    public long args[];
    /** Smalltalk stack or null */
    public String stackReport;

    /** Create an instance of GbjGciException */
    public GbjGciException( ) { 
	this("GBJ Error");
    };

    /** Create an instance of GbjGciException 
	@param msg String
     */
    public GbjGciException(String msg) { 
	number = 2318;    // RT_ERR_GENERIC_ERROR
	message = msg;
	category = GbjGciInterface.OOP_NIL;
	context = GbjGciInterface.OOP_NIL;
	fatal = false;
	argCount = 0;
    };

    /** Create an instance of GbjGciException 
	@param num int number of GS error
	@param msg String text of error message
     */
    public GbjGciException(int num, String msg) { 
	this(msg);
	number = num;
    };

    /** Make internal adjustments to this exception based on the
	GbjGciInterface status code */
    protected GbjGciException  setup(int status) {
	this.fillInStackTrace();
	// JNI_ALLOCATION_ERRORS require a little extra clean-up,
	// since we weren't able to set this up in the JNI code
	if (status == GbjGciInterface.JNI_ALLOCATION_ERROR) {
	    number = 2318;   // RT_ERR_GENERIC_ERROR
	    message = "JNI:  Memory allocation error";
	    category = 0;
	    context = 0;
	    fatal = false;
	    argCount = 0;
	}
	return this;
    }

    /** Set the message field */
    protected void setMessage(String msg) {
	message = msg;
    }

    /** Return a composite of both the Exception and GCI error messages */
    public String getMessage() {
	return (super.getMessage() + ": " + message);
    }

    /** Utility/debug method for printing fields */
    public void print( ) {
	System.out.println("GbjGciException:");
	System.out.println("  number:   " + number);
	System.out.println("  message:  " + message);
	System.out.println("  category: " + category);
	System.out.println("  context:  " + context);
	System.out.println("  fatal:    " + fatal);
	System.out.println("  argCount: " + argCount);
	if (args != null) {
	    for (int i = 0; i < argCount; i++) {
		System.out.println("  Arg:      " + args[i]);
	    }
	}
	printStackTrace();
    }

    public void printStackTrace() {
      if (stackReport != null) 
        System.out.print(stackReport);
      super.printStackTrace(); 
    }

    public void printStackTrace(PrintStream s) {
      if (stackReport != null) 
        s.print(stackReport);
      super.printStackTrace(s); 
    }

    public void printStackTrace(PrintWriter s) {
      if (stackReport != null) 
        s.print(stackReport);
      super.printStackTrace(s); 
    }
}

