/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *  GbjGciInterface.java 
 *
 *========================================================================
 */

package com.gemstone.gbjgci;

import java.io.FileReader;
import java.io.BufferedReader;
import java.util.Properties;
import java.util.logging.*;
import static com.gemstone.gbjgci.GbjGciReport.*;
import com.gemstone.gbj.GbjSession;
import com.gemstone.gbj.GbjException;

// javah -jni com.gemstone.gbjgci.GbjGciInterface

/**

  GbjGciInterface provides the low-level interface to the GemStone GCI.
  <p>
  The only public methods that customer applications should directly call are:
  <ul>
  <li>    getFormat() 
  <li>    gciVersion()
  </ul>

  <p>

  All other functionality is either protected and is intended to be accessed 
  though GbjGciSession, or a call to the GCSI layer and should be accessed
  through package gbjstats classes.  See documentation for those classes 
  for details.

  <p>

  @version 3.1
 */

public class GbjGciInterface {

  /** Indicates GS/S product oop format.
    <p>
    GS/64 2.X and later:  642
   */
  public static int format = 642;  

  /** String representing the GbjGci version */
  public static final String GBJGCI_VERSION = "3.2.0";

  // Operation return codes
  // Would have been nice to just use a boolean,
  // but for some reason the JNI doesn't like boolean
  // args and return values

  /** Value returned (0) from JNI methods indicating normal behavior */
  protected static final int OK                    =  0;

  /** Value returned (-1) from JNI methods indicating an error has occurred */
  protected static final int ERROR                 = -1;

  /** Value returned (-2) from JNI methods indicating a 
    memory allocation problem */
  // allocation errors need to be cleaned up after return from JNI
  protected static final int JNI_ALLOCATION_ERROR  = -2;

  // size for cache holding downloaded byte/oop info

  /** Size of transport cache in bytes */
  private static int CacheSizeBytes = 0;
  private static int CacheSizeOops  = 0;

  /** Maximum number of sessions */
  private static int MaxSessions = 0;

  private static boolean isLinked = false;

  /** Class of GS object instances to generate */
  private static Class gbjObjectClass;

  /** Used to force early loading of GbjGciInterface class */
  private static GbjGciInterface ignore = new GbjGciInterface();

  /** Logger/Level for debug trace messages */
  private static Logger log;
  private static Level logLevel;

  // 40657
  private static int KERNEL_OOPS_SIZE = 16;

  /** Boolean flag indicating if we're running on Android */
  public static final boolean isAndroid = false;

  // 46283
  /** Boolean flag indicating if host is big-endian */
  public static final boolean isBigEndian = false;

  /** oop of nil  */
  public static final long OOP_NIL = 20;

  /** Kernel oops. 
    Used internally to initialize GbjKernelObjects */
  public static long[] kernelOops = null;

  /** Setup Logging for gbjgci/gbj/gbjstat classes 
    @param loggerClass Class
    @return instance of Logger
   */
  public static Logger setupLogger(Class loggerClass) {
    String fullName = loggerClass.getName();
    String shortName = loggerClass.getSimpleName();
    Logger log = Logger.getLogger(fullName);
    logLevel =  GbjGciPreferences.logLevel(shortName);
    log.setLevel(logLevel);
    Handler logHandler = new ConsoleHandler();
    logHandler.setFormatter(new GbjGciLogFormatter());
    logHandler.setLevel(logLevel);
    log.addHandler(logHandler);
    return log;
  }

  // static code for setting up prefs, logger, 
  // and loading GbjGciInterface library
  static {
    // setup logging
    log = GbjGciInterface.setupLogger(GbjGciInterface.class);

    String gemstoneEnvVal = System.getenv("GEMSTONE");
    String ver = gsVersion(gemstoneEnvVal);
    loadLibrary(gemstoneEnvVal, ver);
  }

  protected static String gsVersion(String gemstoneEnvVal ) {
    Properties props = java.lang.System.getProperties();
    String cmdLineVer = props.getProperty("gemstone.version");
    String gsVer = null;
    if (gemstoneEnvVal != null) {
      gsVer = versionFromVersionTxt(gemstoneEnvVal);
      if (cmdLineVer != null && gsVer != null &&
          ! cmdLineVer.equals(gsVer)) {
        libLoadError("-Dgemstone.version=" + cmdLineVer + "  disagrees with  " + 
                               gsVer + "  from $GEMSTONE/version.txt");
      }
    } 
    if (gsVer == null) {  
      gsVer = cmdLineVer;
    }
    if (gsVer == null) {
      libLoadError("Either -Dgemstone.version must be on command line or GEMSTONE must be defined in environment");
    }
    return gsVer; /* lint*/
  }

    protected static String versionFromVersionTxt(String gs) {
	// extract the version string from $GEMSTONE/version.txt
	if (gs.length() == 0) {
	  libLoadError("$GEMSTONE value is of size zero.");
	}
	System.out.println("[Info] $GEMSTONE = \"" + gs + "\"");
	BufferedReader r = null;
	try { 
	  r = new BufferedReader(
	      new FileReader(gs + "/version.txt")); 
	} catch (Throwable e) {
	  libLoadError("$GEMSTONE/version.txt not found");
	}
	String line = null;
	try { 
	  // skip the first line
	  r.readLine();
	  // second line of form: "X.X.X Build: XXXXXXXXX"
	  line = r.readLine();
	  r.close();
	} catch (Throwable e) {
	  libLoadError("Problem reading $GEMSTONE/version.txt");
	}
        String ver = "unknown";
	try {        // 46417: handle special parsing of version.txt
	  int i = line.indexOf(' ');
	  if (i < 0) 
	    ver = line;
	  else
	    ver = line.substring(0, line.indexOf(' '));
	} catch (Throwable e) {
	  libLoadError("Problem parsing $GEMSTONE/version.txt");
	}
	System.out.println("[Info] GemStone Version: " + ver);
        return ver;
    }

  /** string used within libgbjlnk library name */
  protected static final String GBJGCI_SHORT_VERSION = "320";

  /** Try loading different name variations of the GbjGci library */
  private static void loadLibrary(String gemstoneEnvVal, String gsVer/* server version*/) {
    String libname = "gbjlnk" + GBJGCI_SHORT_VERSION + "-" + gsVer + "-64" ;
    Properties props = java.lang.System.getProperties();
    // props.list(System.out); //  uncomment to get a printout of all of the properties
    if (gemstoneEnvVal != null) {
      String sfx = "";
      if (props.getProperty("os.name").equals("Linux")) sfx = ".so" ;
      if (props.getProperty("os.name").equals("Mac OS X")) sfx = ".dylib";
      String fName = gemstoneEnvVal + "/lib/lib" + libname + sfx ;
      System.out.println("[Info] Loading library from $GEMSTONE/lib");
      System.load(fName);
    } else {
      // Use inherited LD_LIBRARY_PATH or -Djava.library.path 
      // For a Java VM started with $jdk/bin/java , the default java.library.path contains $jdk/lib  
      // If LD_LIBRARY_PATH defined in the environment,  
      //  java.library.path includes value of LD_LIBRARY_PATH .
      // If java command line contains -Djava.library.path, 
      // value of -Djava.library.path is used by Java and  LD_LIBRARY_PATH is ignored.
      System.out.println("LD_LIBRARY_PATH = " + System.getenv("LD_LIBRARY_PATH"));
      System.out.println("Loading library using java.library.path = " + props.getProperty("java.library.path"));
      System.loadLibrary(libname);
    }
    System.out.println("[Info] Loaded Library lib" + libname );
  }

    /** Error message handling for loadLibrary */
    private static void libLoadError(String msg) {
	System.out.println("ERROR " + msg );
	throw new GbjGciException(msg);
    }
    
    /** convenience method for generating new GS Object of appropriate class */
    private static GbjGciReport newObject() {
	try {
	    return (GbjGciReport)gbjObjectClass.newInstance();
	} catch (Exception e) {
	    if (GbjGciSession.isDEBUG)  log.fine("newObject exception: " + e.getMessage());
	    throw new GbjGciException("Problem generating Object");
	}
    }

    // Following method help avoid creating too many extra GbjGciExceptions

    private static GbjGciException nextErrObj;

     /**
     Customer applications should not create instances of GbjGciInterface, 
     as all functionality is static.
    */
    private GbjGciInterface() {};


    // NATIVE JNI FUNCTIONS (preceeded by underscore)

    private static synchronized native int _initialize(
	                 int useLinked,   // because boolean args dont work ;-(
			String appName,  // optional application name
			int bufSize,     // size for byte/oop buffer in bytes
			int preloadSize, // size for preload cache
			// int maxSessions, // 49019: MaxSessions needed C layer
			int debugC,      // 0 = no debug, 1 = info, 2 = debug
                        int quietMode,   // printing by linked GCI , 0 = normal , 1 = quiet
			long[] kernelOops, // 40657
			GbjGciException errReport);

    /** JNI function to return the GCI version string
	@return int sessionId if successful, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int _gciVersion(String[] version,
                        GbjGciException errReport);

    /** JNI function to Login a session.
	@return int sessionId if successful, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int _login(String serverName,
		  String gemnetName,
		  String userName,
		  String password,
		  String hostUserName,
		  String hostPassword,
		  GbjGciException errReport);

    /** JNI function to Logout a session
	@return 0 if successful, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int _logout(int sessionId,
		   GbjGciException errReport);

    /** JNI function to check for asynchronous errors
	@return 0 if successful, <br>
	int negative errorCode if error found (details in errReport)
    */
    // 41755
    private static synchronized native int _checkErrors(int sessionId,
		         GbjGciException errReport);

    /** JNI function to finalize setup of a new GS object report
	@return 0 if sucessful, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int _setupReport(int sessionId,
		   GbjGciReport result,
		   GbjGciException errReport);

    private static synchronized native int _executeString(int sessionId,
			String execStr, 
                        GbjGciReport result,
			GbjGciException errReport);

    private static synchronized native int _executeObject(int sessionId,
			GbjGciReport execStrObj,
                      GbjGciReport result,
			GbjGciException errReport);

    private static synchronized native int _perform(int sessionId,
		    GbjGciReport receiver,
		    String selector,
		    GbjGciReport args[],
		    int numArgs,
                    GbjGciReport result,
		    GbjGciException errReport);

    /** JNI function to retrieve bytes from a byte-based GS object.
	@return number of bytes retrieved if successful, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int _getBytes(int sessionId,
		      GbjGciReport obj,
		      long startIndex,
		      int numBytes,
		      byte[] bytes,
		      GbjGciException errReport);

    /** JNI function to retrieve GS objects from an indexable 
	oop-based GS object.
	@return number of objects retrieved if successful, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int _getObjs(int sessionId,
		     GbjGciReport obj,
		     long startIndex,
		     int numObjs,
		     GbjGciReport[] objs,
		     GbjGciException errReport);

    /** JNI function to retrieve GS objects from the named instvars
	of an oop-based GS object.
	@return number of objects retrieved if successful, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int _getNamedObjs(int sessionId,
			  GbjGciReport obj,
			  long startIndex,
			  int numObjs,
			  GbjGciReport[] objs,
			  GbjGciException errReport);

    /** JNI function to write bytes to a byte-based GS object.
	@return number of bytes written if successful, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int _putBytes(int sessionId,
		      GbjGciReport obj,
		      long startIndex,
		      int numBytes,
		      byte[] bytes,
		      GbjGciException errReport);

    /** JNI function to write objects to an indexable oop-based GS object.
	@return number of objects written if successful, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int _putObjs(int sessionId,
		     GbjGciReport obj,
		     long startIndex,
		     int numObjs,
		     GbjGciReport[] objs,
		     GbjGciException errReport);

    /** JNI function to write objects to the named instvars 
	of an oop-based GS object.
	@return number of objects written if successful, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int _putNamedObjs(int sessionId,
			  GbjGciReport obj,
			  long startIndex,
			  int numObjs,
			  GbjGciReport[] objs,
			  GbjGciException errReport);

    /** JNI function to determine if a designated GS object isKindOf
	a specified GS class or any of its subclasses.
	@return 0/1 for false/true, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int _isKindOf(int sessionId,
		      GbjGciReport obj,
		      GbjGciReport cls,
		      GbjGciException errReport);

    /** JNI function to release a batch of oops from the GS session's
	export set.
	@return 0 if successful, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int _releaseOops(int sessionId,
			 long[] oopsToRemove,
			 int numOops,
			 GbjGciException errReport);

    /** JNI function to return a batch of changed objects.
	@return 0 if successful, <br>
	int negative errorCode if problems (details in errReport)
    */
	// returns number of oops returned in changedOops buffer
    private static synchronized native int _changedOops(int sessionId,
			 long[] oopBuffer,
			 int numOops,
			 GbjGciException errReport);

     /** JNI function to return control back to server.
       @return 0 if sucessful, <br>
       int negative errorCode if problems (details in errReport)
     */
     private static synchronized native int _continue(int sessionId,
                     GbjGciReport process,
		     boolean errFlag,         // 45662
		     GbjGciReport stResult,   // 46468
		     GbjGciReport jvResult,
		     String errMsg,
		     String errStack,
                     GbjGciException errReport);

 // GBJ 3.1 JNI functions to support GBJMonitor

     /** JNI function to initialize GCSI Interface
       @return 0 if sucessful, <br>
       int negative errorCode if problems (details in errReport)
     */
     private static synchronized native int _GCSIinitialize(GbjGciException errReport);

     /** JNI function to return an array of array of Strings
         containing the names for the various stats returned 
         in _GCSIgetStats().
         @return 0 if sucessful, <br>
         int negative errorCode if problems (details in errReport)
     */
     private static synchronized native int _GCSIstatNames(String[][] statNames, GbjGciException errReport);

     /** JNI function to attach to a SPC for the GCSI Interface
         @return 0 if sucessful, <br>
         int negative errorCode if problems (details in errReport)
     */
     private static synchronized native int _GCSIattach(String stoneName, GbjGciException errReport);

     /** JNI function to detach from a SPC via the GCSI Interface
       @return 0 if sucessful, <br>
       int negative errorCode if problems (details in errReport)
     */
     private static synchronized native int _GCSIdetach(GbjGciException errReport);

     /** JNI function to sample a batch of stats from a SPC via the 
	 GCSI Interface, and return the count.  Use _GCSIgetStats()
	 to return stats from an individual process after calling this..
	 @return statistic count <br>
	 int negative statistic count if problems (details in errReport)
     */
     private static synchronized native int _GCSIsampleStats(long[] processIds, GbjGciException errReport);

     /** JNI function to retrive a set of stats from a previously
	 acquired batch from _GCSIsampleStats().
       @return statistic count <br>
       int negative statistic count if problems (details in errReport)
     */
     private static synchronized native int _GCSIgetStats(long pid, 
			   GbjGciCacheStats stats,
			   GbjGciException errReport);

    private static synchronized native int _resolveSymbol(int sessionId,
			String objName, 
                        GbjGciReport result,
			GbjGciException errReport);

    private static synchronized native int _resolveSymbolOrNil(int sessionId,
			String objName, 
                        GbjGciReport result,
			GbjGciException errReport);
   

// PROTECTED API


/**

Login a GS session, returning the session ID.
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param serverName   The name of the GS stone process.
@param gemnetName   The name of the Gem Service.  If null, the default "gemnetobject" is used.
@param userName     The GS username for the session.
@param password     The GS password for this username.
@param hostUserName  The OS username for this session (if needed).
@param hostPassword  The OS password for this OS username (if needed).
@return the GS session ID.
@exception com.gemstone.gbjgci.GbjGciException
*/
   
   protected static int login(String serverName,
			   String gemnetName,
			   String userName,
			   String password,
			   String hostUserName,
			   String hostPassword) {
        int sessionId = 0;
	log.fine("GbjGciInterface: calling _login.." + GbjGciSession.isDEBUG);
        GbjGciException errObj = nextErrObj;
        sessionId = _login(serverName,
		           gemnetName,
			   userName,
			   password,
			   hostUserName,
			   hostPassword,
			   errObj);
	if (sessionId <= 0/*GCI_INVALID_SESSION_ID*/) {
          nextErrObj = new GbjGciException();
          throw errObj;
	}
        if (false) {    // uncomment (change to true) to get stack traces of logins
          try {
            throw new GbjGciException(" =====  trace login");
          } catch (GbjGciException e) {
            e.printStackTrace(System.out);
          }
        }
	return sessionId;
    }

/**
Logout a GS session.
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param sessionId   The session ID for the GS session to logout.
@exception com.gemstone.gbjgci.GbjGciException
*/
    protected static void logout(int sessionId) {
	int status;
        GbjGciException errObj = nextErrObj;
        status = _logout(sessionId, errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw errObj;
        }
    }

/**
Check for asynchronous errors on a GS session.
<p>
@param sessionId   The session ID for the GS session to logout.
@exception com.gemstone.gbjgci.GbjGciException
*/
    // 41755
    protected static void checkErrors(int sessionId) {
	int status;
        GbjGciException errObj = nextErrObj;
        status = _checkErrors(sessionId, errObj);
	if (status < 0) {
          nextErrObj = new GbjGciException();
          throw errObj;
        }
    }


/**
Execute a GS/S Smalltalk String (represented as a java String)
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param sessionId   The session ID of the GS session.
@param execStr     The Smalltalk string to execute.
@return a GS object representing the results of execution.
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/

    // Execute smalltalk code
    protected static GbjGciReport execute(GbjGciSession sess, int sessionId, String execStr) {
	GbjGciReport result = newObject();
        GbjGciException errObj = nextErrObj;
        int status = _executeString(sessionId, execStr, result, errObj);
	if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)sess, errObj);
        }
	return result;
    }

/**
Execute a GS/S Smalltak String (represented as a GS Object)
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param sessionId   The session ID of the GS session.
@param execStrObj  The Smalltalk string to execute.
@return a GS object representing the results of execution.
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static GbjGciReport execute(GbjGciSession sess, int sessionId, GbjGciReport execStrObj) {

	GbjGciReport result = newObject();
        GbjGciException errObj = nextErrObj;
        int   status = _executeObject(sessionId, execStrObj, result, errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)sess, errObj);
        }
	return result;
    }

/**
Execute a Smalltalk method against a designated receiver using these arguments.
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param sessionId   The session ID of the GS session.
@param receiver    The GS object to execute this method against.
@param selector    The method to execute.
@param args        The list of arguments.
@param numArgs     The number of arguments in the list.
@return a GS object representing the results of execution.
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static GbjGciReport perform(GbjGciSession sess, int sessionId,
				       GbjGciReport receiver,
				       String selector,
				       GbjGciReport args[],
				       int numArgs) {
	GbjGciReport result = newObject();
        GbjGciException errObj = nextErrObj;
	int status = _perform(sessionId, receiver, selector, args, numArgs, result, errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)sess, errObj);
        }
	return result;
    }


/**
Retrieve the designated bytes from this byte-based GS object.
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param sessionId   The session ID of the GS session.
@param obj         The GS object to retrieve the bytes from.
@param startIndex  The starting index (1-based)
@param numBytes    The number of bytes to retrieve
@return the retrieved bytes
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static byte[] getBytes(GbjGciSession sess, int sessionId,
				  GbjGciReport obj,
				  long startIndex,
				  int numBytes) {
	int status;
	if (numBytes == 0) return null;
	if (obj.objType_() != OBJTYPE_BYTE) 
	    throw new GbjException("Invalid obj type for getBytes()");
	if ((startIndex <= 0) ||
	    ((startIndex + numBytes - 1) > obj.indexedSize))
	    throw new GbjException("startIndex/numBytes out of range in getBytes()");
	byte[] bytes = new byte[numBytes];
          //  native method handles bytes already fetched into preload cache
        GbjGciException errObj = nextErrObj;
	status = _getBytes(sessionId, obj, startIndex, numBytes, bytes, errObj);
	if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)sess, errObj);
        }
	if (status != numBytes)
	  throw new GbjException("getBytes(): number returned not same as requested");
	return bytes;
    }

/**
Retrieve the designated GS objects from this indexable GS object
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param sessionId   The session ID of the GS session.
@param obj         The GS object to retrieve the GS objects from.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to retrieve
@param objs        An array of GS objects to return the results
@return the number of GS objects actually returned (may be less than numObjs)
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static int getObjs(GbjGciSession sess, int sessionId,
			      GbjGciReport obj,
			      long startIndex,
			      int numObjs,
			      GbjGciReport[] objs) {
        boolean isDEBUG = GbjGciSession.isDEBUG;
	if (isDEBUG) {
          log.fine("GbjGciInterface.getObjs.");
	  log.fine("startIndex: " + startIndex);
	  log.fine("numObjs: " + numObjs);
        }
	int status = 0;
	if (numObjs <= 0) return 0;
        int objType = obj.objType_();
	if (objType != OBJTYPE_OOP && objType != OBJTYPE_NSC)
	    throw new GbjException("Invalid obj type for getObjs()");
	if ((startIndex <= 0) ||
	    ((startIndex + numObjs - 1) > obj.indexedSize))
	    throw new GbjException("startIndex/numObjs out of range in getObjs()");
	if (numObjs > objs.length)
	    throw new GbjException("numObjs out of range for objs in getObjs()");
	if ((obj.namedSize + startIndex + numObjs) < CacheSizeOops) {
	    // oops available from preload cache
	    int j = obj.namedSize + (int)startIndex - 1;
	    if (isDEBUG) {
              log.fine("getObjs: out of cache starting at " + j);
	      log.fine("getObjs: " + numObjs + " to transfer");
            }
            long[] valOops = obj.valueOops();
	    for (int i=0; i<numObjs; i++,j++) {
		if (isDEBUG) log.fine("getObjs: element " + i);
		objs[i] = objWithOop(sess, sessionId, valOops[j]);
	    }
	} else {
	    if (isDEBUG) log.fine("getObjs: from GCI starting at " + startIndex);
	    for (int i=0; i<numObjs; i++) 
		objs[i] = newObject();
            GbjGciException errObj = nextErrObj;
            status =  _getObjs(sessionId, obj, startIndex, numObjs, objs, errObj);
	    if (status < 0) {
              nextErrObj = new GbjGciException();
              throw new GbjException((GbjSession)sess, errObj);
            }
	    if (status != numObjs)
		throw new GbjException((GbjSession)sess,
                      "getObjs(): number returned not same as requested");
	    // 47084
	    for (int i=0; i<numObjs; i++) 
	    	setupObj(sess, sessionId, objs[i]);
	}
	return status;
    }

/**
Retrieve the designated GS objects from the named instvars for this GS object.
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param sessionId   The session ID of the GS session.
@param obj         The GS object to retrieve the GS objects from.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to retrieve
@param objs        An array of GS objects to return the results
@return the number of GS objects actually returned (may be less than numObjs)
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static int getNamedObjs(GbjGciSession sess, int sessionId,
				   GbjGciReport obj,
				   long startIndex,
				   int numObjs,
				   GbjGciReport[] objs) {
	int status = 0;
	if (numObjs <= 0) return 0;
        int objType = obj.objType_();
	if (objType != OBJTYPE_OOP && objType != OBJTYPE_NSC)
	    throw new GbjException((GbjSession)sess, "Invalid obj type for getNamedObjs()");
	if ((startIndex < 0) ||
	    ((startIndex + numObjs - 1) > obj.namedSize)) {
	    throw new GbjException((GbjSession)sess, "startIndex/numObjs out of range in getNamedObjs()");
	}
	if (numObjs > objs.length)
	    throw new GbjException("numObjs out of range for objs in getNamedObjs()");
	if ((startIndex + numObjs) < CacheSizeOops) {
	    // oops available from preload cache
	    int j = (int)startIndex - 1;
            long[] valOops = obj.valueOops();
	    for (int i=0; i<numObjs; i++,j++) {
		objs[i] = objWithOop(sess, sessionId, valOops[j]);
	    }
	} else {
	    for (int i=0; i<numObjs; i++) 
		objs[i] = newObject();
            GbjGciException errObj = nextErrObj;
            status = _getNamedObjs(sessionId, obj, startIndex, numObjs, objs, errObj);
	    if (status < 0) {
              nextErrObj = new GbjGciException();
              throw new GbjException((GbjSession)sess, errObj);
            }
	    if (status != numObjs)
		throw new GbjException((GbjSession)sess,
                      "getNamedObjs(): number returned not same as requested");
	}
	return status;
    }

/**
Write the designated bytes to this byte-based GS object.
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param sessionId   The session ID of the GS session.
@param obj         The GS object to write the bytes to.
@param startIndex  The starting index (1-based)
@param numBytes    The number of bytes to write
@param bytes       The bytes to write
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static void putBytes(GbjGciSession sess, int sessionId,
				   GbjGciReport obj,
				   long startIndex,
				   int numBytes,
				   byte[] bytes) {
	int status;
	if (numBytes == 0) return;
	if (numBytes > bytes.length)
	    throw new GbjException("numBytes too large in putBytes()");
	if (obj.objType_() != OBJTYPE_BYTE) 
	    throw new GbjException("Invalid obj type for putBytes()");
	if ((startIndex <= 0) ||
	    ((startIndex - 1) > obj.indexedSize))
	    throw new GbjException("startIndex/numBytes out of range in putBytes()");
        GbjGciException errObj = nextErrObj;
        status = _putBytes(sessionId, obj, startIndex, numBytes, bytes, errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)sess, errObj);
        }
	return;
    }

/**
Write some designated GS objects to the indexed component for this GS object.
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param sessionId   The session ID of the GS session.
@param obj         The GS object to write the GS objects to.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to write
@param objs        An array of GS objects to write out
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static void putObjs(GbjGciSession sess, int sessionId,
				  GbjGciReport obj,
				  long startIndex,
				  int numObjs,
				  GbjGciReport[] objs) {
	int status = 0;
	if (numObjs <= 0) return;
        int objType = obj.objType_();
	if (objType != OBJTYPE_OOP && objType != OBJTYPE_NSC)
	    throw new GbjException("Invalid obj type for putObjs()");
	if ((startIndex < 0) ||
	    ((startIndex - 1) > obj.indexedSize))
	    throw new GbjException("startIndex/numObjs out of range in putNamedObjs()");
        GbjGciException errObj = nextErrObj;
        status = _putObjs(sessionId, obj, startIndex, numObjs, objs, errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)sess, errObj);
        }
	return;
    }

/**
Write some designated GS object to the named instvars for this GS object.
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param sessionId   The session ID of the GS session.
@param obj         The GS object to write the GS objects to.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to write
@param objs        An array of GS objects to write out
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static void putNamedObjs(GbjGciSession sess, int sessionId,
				       GbjGciReport obj,
				       long startIndex,
				       int numObjs,
				       GbjGciReport[] objs) {
	if (numObjs <= 0) return;
        int objType = obj.objType_();
	if (objType != OBJTYPE_OOP && objType != OBJTYPE_NSC)
	    throw new GbjException("Invalid obj type for putNamedObjs()");
	if ((startIndex < 0) ||
	    ((startIndex + numObjs - 1) > obj.namedSize))
	    throw new GbjException("startIndex/numObjs out of range in putNamedObjs()");
        GbjGciException errObj = nextErrObj;
	int status = _putNamedObjs(sessionId, obj, startIndex, numObjs, objs, errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)sess, errObj);
        }
	return;
    }

/**
Return the GS object with the designed oop.
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param sessionId   The session ID of the GS session.
@param oopValue    The oop to search for.
@return the GS object with this oop (if it exists)
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static GbjGciReport objWithOop(GbjGciSession sess, int sessionId, long oopValue) {
	GbjGciReport obj = newObject();
	obj.oop = oopValue;
        
	setupObj(sess, sessionId, obj); 
	return obj;
    }

/**
Return true if this GS object is a member of the designated GS class 
or one of it's subclasses.
<p>
NOTE: customer applications should call the Gbj[Gci]Object equivalent.
<p>
@param sessionId   The session ID of the GS session.
@param obj         The GS object to check.
@param cls         The GS class to check against.
@return true/false
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static boolean isKindOf(GbjGciSession sess, int sessionId, 
				   GbjGciReport obj,
				   GbjGciReport cls) {
        GbjGciException errObj = nextErrObj;
	int status = _isKindOf(sessionId, obj, cls, errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)sess, errObj);
        }
	return ( status == 0 ) ? false : true;
    }

/**
Return true if the GS object is a type of String
<p>
NOTE: customer applications should call the Gbj[Gci]Object equivalent.
<p>
@param obj         The GS object to check.
@return true/false
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static boolean isString(GbjGciReport obj) {
	return obj.isString();
    }

/**
Given a partially configured object, complete its setup.
<p>
Initially generated objects either know their oop and need the other
fields defined, or have the other fields defined and need to determine
(or generate) the oop.  This method handles both tasks.
<p>
NOTE: customer applications should not need to call this method.
<p>
@param sessionId   The session ID of the GS session.
@param obj         The GS object to setup
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static void setupObj(GbjGciSession sess, int sessionId, GbjGciReport obj) {
        GbjGciException errObj = nextErrObj;
        int   status =  _setupReport(sessionId, obj, errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)sess, errObj);
        }
	return;
    }

/**
Release a batch of oops from the GS export set.
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param sessionId     The session ID of the GS session
@param oopsToRemove  The list of oops to remove from the export set
@param numOops       The number of oops in the list
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static void releaseOops(GbjGciSession sess, int sessionId,
				   long[] oopsToRemove, 
				   int numOops) {
        GbjGciException errObj = nextErrObj;
	int status = _releaseOops(sessionId, oopsToRemove, numOops, errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)sess, errObj);
        }
	return;
    }

/**
Return a batch of oops associated with changed objects.
<p>
Objects can become changed as a result of GS/S smalltalk code execution,
or when the session commits/aborts and the changes made by other sessions
become visible.  This method returns back a batch of the oops of these
changed objects.  The method should be called repeatedly until it returns
zero to insure that all changed object oops have been retrieved.
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param sessionId     The session ID of the GS session
@param oopBuffer    A buffer to hold the oops of changed objects
@param numOops       The size of the oopBuffer
@return The number of oops returned in the oopBuffer
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    protected static int changedOops(GbjGciSession sess, int sessionId,
				  long[] oopBuffer,
				  int numOops) {
        GbjGciException errObj = nextErrObj;
	int count =  _changedOops(sessionId, oopBuffer, numOops, errObj);
        if (count < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)sess, errObj);
        }
	return count;
    }

// PUBLIC API

/**
Initialize the GbjGciInterface, returning 32/641/642 depending on the
underlying GS product oop format.  
<p>
Applications should call GbjSession.initialize( ).
See GbjSession.initialize( ) for details and parameter usage.
@param useLinked       See GbjSession.initialize( ) , must be true
@param appName         See GbjSession.initialize( )
@param bufSize         See GbjSession.initialize( )
@param preloadSize     See GbjSession.initialize( )
@param maxSessions     See GbjSession.initialize( )
@param objectClass     See GbjSession.initialize( )
<p>
@return    The integer 32/641/642 depending on the type of GS product.
32 for GS/S, 641 for GS/64 version 1.X, and 642 for GS/64 versions 2.X 
and later.
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    public static int initialize(boolean useLinked,
		       String appName,  // optional application name
		       int bufSize,     // size for byte/oop buffer in bytes
		       int preloadSize, // size for preload cache in bytes
                       int maxSessions, // maximum number of sessions
		       Class objectClass) {

        if (! useLinked) {
          throw new GbjGciException("only linked configuration supported with gbjlnk jar");
        }
	isLinked = useLinked ;
	int debugC = GbjGciPreferences.debugC();
	CacheSizeBytes = preloadSize;
	CacheSizeOops  = preloadSize / 8;
        MaxSessions = maxSessions;
	gbjObjectClass = objectClass;
	nextErrObj = new GbjGciException();
        kernelOops = new long[KERNEL_OOPS_SIZE];
        int quiet = 1;  // TODO make configurable ?
        GbjGciException errObj = nextErrObj;
	int status = _initialize(isLinked ? 1 : 0, appName, bufSize, 
			     preloadSize, debugC, quiet, kernelOops, errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException( errObj);
        }
	if (kernelOops[KERNEL_OOPS_SIZE - 1] > 0) {
          throw new GbjException("big endian server not supported");
        }

	return status;
    }

/**
Return 32/641/642, depending on the associated GS/S product oop format.
@return 642
GS/64 versions 2.X and later: 642 <br>
*/
    public static int getFormat() { 
        return 642;
    }

/**

Return the GCI version string.
@return the GCI version string.
@exception com.gemstone.gbj.GbjException    if there is an error in server execution
*/

   public static String gciVersion( ) {
    
      String[] version = new String[1];
      GbjGciException errObj = nextErrObj;
      int status = _gciVersion(version, errObj);
      if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException(errObj);
      }
      return version[0];
   }


/**
Return control back to the server after handling a client forwarder error.
<p>
This method should only be used by GbjSession.
*/
    public static void serverContinue(GbjSession sess, int sessionId, 
			    GbjGciReport process, 
			    boolean errFlag,          // 45662
			    GbjGciReport stResult,    // 46468
			    GbjGciReport jvResult,
			    String errMsg,
			    String errStack) {
        GbjGciException errObj = nextErrObj;
	int status = _continue(sessionId, process, errFlag, stResult, jvResult, 
			   errMsg, errStack, errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException(sess, errObj);
        }
	return;
    }

// Public GBJ 3.1 GCSI FUNCTIONS

/**

Initialize the GCSI (GCI Cache Statistics Interface).
<p>
This method should only be used by package gbjstats classes.
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/

     public static void GCSIinitialize( ) {
	int status;
        GbjGciException errObj = nextErrObj;
        status =  _GCSIinitialize(errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException(errObj);
        }
	return;
     }

/**

Return a String[][] containing cache statistic names.
<p>
This method should only be used by package gbjstats classes.
<p>
First index is by type, second is by particular cache stat.
@return    A String[][] containing cache stat names
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/

    public static String[][] GCSIstatNames( ) {
	String[][] statNames = new String[20][];
        GbjGciException errObj = nextErrObj;
        int status =  _GCSIstatNames(statNames, errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException(errObj);
        }
	return statNames;
     }

/**

Attach to a particular shared page cache on this host.
<p>
This method should only be used by package gbjstats classes.

@param stoneName  Name of the stone 
                  (or the stone associated with this remote SPC)

@return        Maximum number of processes that can attach to this cache 
               (can be used to size arrays used to hold pid data returned
               by GCSIsampleStats( ) )

@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/

     public static int GCSIattach(String stoneName) {
        GbjGciException errObj = nextErrObj;
        int maxProcesses =  _GCSIattach(stoneName, errObj);
        if (maxProcesses < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException(errObj);
        }
	return maxProcesses;
     }

/**

Detach from an attached shared page cache.
<p>
This method should only be used by package gbjstats classes.

@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/

     public static void GCSIdetach( ) {
        GbjGciException errObj = nextErrObj;
        int status =  _GCSIdetach(errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException(errObj);
        }
	return;
     }

/**

Take a sample of cache stats for all processes on this cache. 
<p>
This method should only be used by package gbjstats classes.

<p>

Note that this only takes the sample -- use GCSIgetStats( ) to return
the actual sampled statistics.

@param processIds  A long[] to hold the returned process ID's of the 
                   sampled sessions.
@return            The actual number of sessions that were sampled.
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    public static int GCSIsampleStats(long[] processIds) {
        GbjGciException errObj = nextErrObj;
        int count = _GCSIsampleStats(processIds, errObj);
        if (count < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException(errObj);
        }
	return count;
    }

/**

Return back the cache statistics sampled earlier for a particular process.
<p>
This method should only be used by package gbjstats classes.

@param pid    The ID of the process of interest.
@param stats  A GbjGciCacheStats object to hold the returned cache stats.
@return       The GbjGciCacheStats object passed in above.
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    public static GbjGciCacheStats GCSIgetStats(long pid, GbjGciCacheStats stats) {
        GbjGciException errObj = nextErrObj;
        int status =  _GCSIgetStats(pid, stats, errObj);
        if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException(errObj);
        }
	return stats;
     }

/**

Return back the cache statistics sampled earlier for a particular process.
<p>
This method should only be used by package gbjstats classes.

@param index  A numeric index for indexing through a batch of sampled stats.
              Should be less than the max value returned by GCSIsampleStats( ).
@return       A GbjGciCacheStats object containg the cache stats.
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/
    public static GbjGciCacheStats GCSIgetStats(int index) {
	GbjGciCacheStats stats = null;
        stats = new GbjGciCacheStats( );
	return GCSIgetStats(index, stats);
    }

/**
Resolve a name in the session's current symbolList , calls GciResolveSymbol.
<p>
@param sessionId   The session ID of the GS session.
@param objName     The string to resolve
@return a GS object representing the result .
@exception com.gemstone.gbj.GbjException  if there is an error in server execution
*/

   protected static GbjGciReport resolveSymbol(int sessionId, String objName) {
      if (GbjGciSession.isDEBUG) log.fine("resolveSymbol " + objName);
      GbjGciReport result = newObject();
      GbjGciException errObj = nextErrObj;
      int status = _resolveSymbol(sessionId, objName, result, errObj);
      if (status < 0) {
	nextErrObj = new GbjGciException();
	throw new GbjException(errObj);
      }
      return result;
   }

   protected static GbjGciReport resolveSymbolOrNil(int sessionId, String objName) {
      if (GbjGciSession.isDEBUG) log.fine("resolveSymbol " + objName);
      GbjGciReport result = newObject();
      GbjGciException errObj = nextErrObj;
      int status = _resolveSymbolOrNil(sessionId, objName, result, errObj);
      if (status < 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException(errObj);
      }
      return result;
   }

}




