/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *  GbjGciInterface.java 
 *
 *========================================================================
 */

package com.gemstone.gbjgci;

import java.io.FileReader;
import java.io.BufferedReader;
import java.util.Properties;
import java.util.logging.*;
import static com.gemstone.gbjgci.GbjGciReport.*;

// javah -jni com.gemstone.gbjgci.GbjGciInterface

/**

  GbjGciInterface provides the low-level interface to the GemStone GCI.
  <p>
  The only public methods that customer applications should directly call are:
  <ul>
  <li>    getFormat() 
  <li>    gciVersion()
  </ul>

  <p>

  All other functionality is either protected and is intended to be accessed 
  though GbjGciSession, or a call to the GCSI layer and should be accessed
  through package gbjstats classes.  See documentation for those classes 
  for details.

  <p>

  @version 3.1
 */

public class GbjGciInterface {

  /** Indicates GS/S product oop format.
    <p>
    GS/64 2.X and later:  642
   */
  public static int format = 642;  

  /** String representing the GbjGci version */
  public static final String GBJGCI_VERSION = "3.2.0";

  // Operation return codes
  // Would have been nice to just use a boolean,
  // but for some reason the JNI doesn't like boolean
  // args and return values

  /** Value returned (0) from JNI methods indicating normal behavior */
  protected static final int OK                    =  0;

  /** Value returned (-1) from JNI methods indicating an error has occurred */
  protected static final int ERROR                 = -1;

  /** Value returned (-2) from JNI methods indicating a 
    memory allocation problem */
  // allocation errors need to be cleaned up after return from JNI
  protected static final int JNI_ALLOCATION_ERROR  = -2;

  // size for cache holding downloaded byte/oop info
  protected static int CacheSizeBytes = 0;
  protected static int CacheSizeOops  = 0;

  private static final boolean isLinked = false;

  /** Used to force early loading of GbjGciInterface class */
  private static GbjGciInterface ignore = new GbjGciInterface();

  /** Logger/Level for debug trace messages */
  private static Logger log;
  private static Level logLevel;

  // 40657
  private static int KERNEL_OOPS_SIZE = 16;

  /** Boolean flag indicating if we're running on Android */
  public static final boolean isAndroid = false;

  // 46283
  /** Boolean flag indicating if host is big-endian */
  public static final boolean isBigEndian = false;

  /** oop of nil  */
  public static final long OOP_NIL = 20;

  /** Kernel oops. 
    Used internally to initialize GbjKernelObjects */
  public static long[] kernelOops = null;

  /** Setup Logging for gbjgci/gbj/gbjstat classes 
    @param loggerClass Class
    @return instance of Logger
   */
  public static Logger setupLogger(Class loggerClass) {
    String fullName = loggerClass.getName();
    String shortName = loggerClass.getSimpleName();
    Logger log = Logger.getLogger(fullName);
    logLevel =  GbjGciPreferences.logLevel(shortName);
    log.setLevel(logLevel);
    Handler logHandler = new ConsoleHandler();
    logHandler.setFormatter(new GbjGciLogFormatter());
    logHandler.setLevel(logLevel);
    log.addHandler(logHandler);
    return log;
  }

  // static code for setting up prefs, logger, 
  // and loading GbjGciInterface library
  static {
    // setup logging
    log = GbjGciInterface.setupLogger(GbjGciInterface.class);

    String gemstoneEnvVal = System.getenv("GEMSTONE");
    String ver = gsVersion(gemstoneEnvVal);
    loadLibrary(gemstoneEnvVal, ver);
  }

  protected static String gsVersion(String gemstoneEnvVal ) {
    Properties props = java.lang.System.getProperties();
    String cmdLineVer = props.getProperty("gemstone.version");
    String gsVer = null;
    if (gemstoneEnvVal != null) {
      gsVer = versionFromVersionTxt(gemstoneEnvVal);
      if (cmdLineVer != null && gsVer != null &&
          ! cmdLineVer.equals(gsVer)) {
        libLoadError("-Dgemstone.version=" + cmdLineVer + "  disagrees with  " + 
                               gsVer + "  from $GEMSTONE/version.txt");
      }
    } 
    if (gsVer == null) {  
      gsVer = cmdLineVer;
    }
    if (gsVer == null) {
      libLoadError("Either -Dgemstone.version must be on command line or GEMSTONE must be defined in environment");
    }
    return gsVer; /* lint*/
  }

    protected static String versionFromVersionTxt(String gs) {
	// extract the version string from $GEMSTONE/version.txt
	if (gs.length() == 0) {
	  libLoadError("$GEMSTONE value is of size zero.");
	}
	System.out.println("[Info] $GEMSTONE = \"" + gs + "\"");
	BufferedReader r = null;
	try { 
	  r = new BufferedReader(
	      new FileReader(gs + "/version.txt")); 
	} catch (Throwable e) {
	  libLoadError("$GEMSTONE/version.txt not found");
	}
	String line = null;
	try { 
	  // skip the first line
	  r.readLine();
	  // second line of form: "X.X.X Build: XXXXXXXXX"
	  line = r.readLine();
	  r.close();
	} catch (Throwable e) {
	  libLoadError("Problem reading $GEMSTONE/version.txt");
	}
        String ver = "unknown";
	try {        // 46417: handle special parsing of version.txt
	  int i = line.indexOf(' ');
	  if (i < 0) 
	    ver = line;
	  else
	    ver = line.substring(0, line.indexOf(' '));
	} catch (Throwable e) {
	  libLoadError("Problem parsing $GEMSTONE/version.txt");
	}
	System.out.println("[Info] GemStone Version: " + ver);
        return ver;
    }

  /** string used within libgbjts library name */
  protected static final String GBJGCI_SHORT_VERSION = "320";

  /** Load the Gbj C library which is built as part of the GemStone64 server build */
  private static void loadLibrary(String gemstoneEnvVal, String gsVer/* server version*/) {
    String libname = "gbjts" + GBJGCI_SHORT_VERSION + "-" + gsVer + "-64" ;
    Properties props = java.lang.System.getProperties();
    // props.list(System.out); //  uncomment to get a printout of all of the properties
    if (gemstoneEnvVal != null) {
      String sfx = ".dll"; // Windows
      if (props.getProperty("os.name").equals("Linux")) sfx = ".so" ;
      if (props.getProperty("os.name").equals("Mac OS X")) sfx = ".dylib";
      String fName = gemstoneEnvVal + "/lib/lib" + libname + sfx ;
      System.out.println("[Info] Loading library from $GEMSTONE/lib");
      System.load(fName);
    } else {
      // Use inherited LD_LIBRARY_PATH or -Djava.library.path 
      // For a Java VM started with $jdk/bin/java , the default java.library.path contains $jdk/lib  
      // If LD_LIBRARY_PATH defined in the environment,  
      //  java.library.path includes value of LD_LIBRARY_PATH .
      // If java command line contains -Djava.library.path, 
      // value of -Djava.library.path is used by Java and  LD_LIBRARY_PATH is ignored.
      System.out.println("LD_LIBRARY_PATH = " + System.getenv("LD_LIBRARY_PATH"));
      System.out.println("Loading library using java.library.path = " + props.getProperty("java.library.path"));
      System.loadLibrary(libname);
    }
    System.out.println("[Info] Loaded Library lib" + libname );
  }

    /** Error message handling for loadLibrary */
    private static void libLoadError(String msg) {
	System.out.println("ERROR " + msg );
	throw new GbjGciException(msg);
    }
    
    /** convenience method for generating GbjGciExceptions */
    private static GbjGciException newException() {
	GbjGciException ex = new GbjGciException();
	return ex;
    }

    /** convenience method for generating GbjGciExceptions */
    private static GbjGciException newException(String msg) {
	if (GbjGciSession.isDEBUG)  log.fine("newException: " + msg);
	GbjGciException ex = new GbjGciException(msg);
	ex.printStackTrace();
	return ex;
    }

     /**
     Customer applications should not create instances of GbjGciInterface, 
     as all functionality is static.
    */
    private GbjGciInterface() {};

    // NATIVE JNI FUNCTIONS (preceeded by underscore)

    /** JNI function to initialize interface
	@return int format code (32/641/642) if successful, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int initialize(
			int bufSize,     // size for byte/oop buffer in bytes
			int preloadSize, // size for preload cache
			// int maxSessions, // 49019: MaxSessions needed C layer
			int debugC,      // 0 = no debug, 1 = info, 2 = debug
                        int quietMode,   // printing by linked GCI , 0 = normal , 1 = quiet
			long[] kernelOops, // 40657
			GbjGciException errReport);

    /** JNI function to return the GCI version string
	@return int sessionId if successful, <br>
	int negative errorCode if problems (details in errReport)
    */
    private static synchronized native int _gciVersion(String[] version,
                        GbjGciException errReport);

    /** JNI function to Login a session.
	@return int sessionId if successful, <br>
	int negative errorCode if problems (details in errReport)
    */

    private static native long login(String serverName,
                           String gemnetName,
                           String userName,
                           String password,
                           String hostUserName,
                           String hostPassword, GbjGciException errObj);

/**

Login a GS session, returning the session ID.
<p>
NOTE: customer applications should call the Gbj[Gci]Session equivalent.
<p>
@param serverName   The name of the GS stone process.
@param gemnetName   The name of the Gem Service.  If null, the default "gemnetobject" is used.
@param userName     The GS username for the session.
@param password     The GS password for this username.
@param hostUserName  The OS username for this session (if needed).
@param hostPassword  The OS password for this OS username (if needed).
@return the GS session ID.
@exception com.gemstone.gbjgci.GbjGciException
*/
   
   protected static synchronized long login(String serverName,
			   String gemnetArg,
			   String userName,
			   String password,
			   String hostUserName,
			   String hostPassword) {
	log.fine("GbjGciInterface: calling _login.." + GbjGciSession.isDEBUG);
        GbjGciException errObj = new GbjGciException();
        String gemnet = gemnetArg;
        if (gemnet == null || gemnet.length() == 0) {
          gemnet = "gemnetobject";
        }
        long sessionPtr = login(serverName,
		           gemnet,
			   userName,
			   password,
			   hostUserName,
			   hostPassword,
			   errObj);
	if (sessionPtr == 0) {
          throw errObj;
	}
        if (false) {    // uncomment (change to true) to get stack traces of logins
          try {
            throw new GbjGciException(" =====  trace login");
          } catch (GbjGciException e) {
            e.printStackTrace(System.out);
          }
        }
	return sessionPtr;
    }


// PUBLIC API

/**
Initialize the GbjGciInterface, 
<p>
Applications should call GbjSession.initialize( ).
See GbjSession.initialize( ) for details and parameter usage.
@param useLinked      See GbjSession.initialize( ) , must be false 
@param appName        ignored See GbjSession.initialize( )
@param bufSize         See GbjSession.initialize( )
@param preloadSize     See GbjSession.initialize( )
@param maxSessions     See GbjSession.initialize( )
<p>
@return    The integer 642 
@exception com.gemstone.gbjgci.GbjGciException  thrown for any error
*/
    public static int initialize(boolean useLinked,
		       String appName,  // optional application name
		       int bufSize,     // size for byte/oop buffer in bytes
		       int preloadSize, // size for preload cache in bytes
                       int maxSessions  /* maximum number of sessions*/) {

        if (useLinked)
          throw new GbjGciException("linked configuration not supported with gbjts jar");

	int debugC = GbjGciPreferences.debugC();
	CacheSizeBytes = preloadSize;
	CacheSizeOops  = preloadSize / 8;
	GbjGciException errReport = new GbjGciException();
        kernelOops = new long[KERNEL_OOPS_SIZE];
        int quiet = 1;  // TODO make configurable ?
	int status = initialize(bufSize, 
			     preloadSize, debugC, quiet, kernelOops, errReport);
	if (status < 0) throw errReport;

	if (kernelOops[KERNEL_OOPS_SIZE - 1] > 0) {
          throw new GbjGciException("big endian server not supported");
        }

	return status;
    }

/**
Return 32/641/642, depending on the associated GS/S product oop format.
Should be run after doing GbjGciInitialize(), othewise will trigger
a GbjGciException.
@return 32, 641, or 642
@exception com.gemstone.gbjgci.GbjGciException
<p>
GS/S: 32 <br>
GS/64 versions 1.X: 641 <br>
GS/64 versions 2.X and later: 642 <br>
*/
    public static int getFormat() { 
        return 642;
    }

/**

Return the GCI version string.
@return the GCI version string.
@exception com.gemstone.gbjgci.GbjGciException  thrown for any error
*/

   public static String gciVersion( ) {
    
      String[] version = new String[1];
      GbjGciException errObj = new GbjGciException();
      int status = _gciVersion(version, errObj);
      if (status < 0) {
         throw errObj;
      }
      return version[0];
   }

     /** JNI function to initialize GCSI Interface
       @return 0 if sucessful, <br>
       int negative errorCode if problems (details in errReport)
     */
     private static native int _GCSIinitialize(GbjGciException errReport);

     /** JNI function to return an array of array of Strings
         containing the names for the various stats returned
         in _GCSIgetStats().
         @return 0 if sucessful, <br>
         int negative errorCode if problems (details in errReport)
     */
     private static native int _GCSIstatNames(String[][] statNames, GbjGciException errReport);

     /** JNI function to attach to a SPC for the GCSI Interface
         @return 0 if sucessful, <br>
         int negative errorCode if problems (details in errReport)
     */
     private static native int _GCSIattach(String stoneName, GbjGciException errReport);

     /** JNI function to detach from a SPC via the GCSI Interface
       @return 0 if sucessful, <br>
       int negative errorCode if problems (details in errReport)
     */
     private static native int _GCSIdetach(GbjGciException errReport);

     /** JNI function to sample a batch of stats from a SPC via the
         GCSI Interface, and return the count.  Use _GCSIgetStats()
         to return stats from an individual process after calling this..
         @return statistic count <br>
         int negative statistic count if problems (details in errReport)
     */
     private static native int _GCSIsampleStats(long[] processIds, GbjGciException errReport);

     /** JNI function to retrive a set of stats from a previously
         acquired batch from _GCSIsampleStats().
       @return statistic count <br>
       int negative statistic count if problems (details in errReport)
     */

     private static native int _GCSIgetStats(long pid,
                           GbjGciCacheStats stats,
                           GbjGciException errReport);


// Public GBJ 3.1 GCSI FUNCTIONS

/**

Initialize the GCSI (GCI Cache Statistics Interface).
<p>
This method should only be used by package gbjstats classes.
@exception com.gemstone.gbjgci.GbjGciException  thrown for any error
*/

     public static synchronized void GCSIinitialize( ) {
	int status;
        GbjGciException err = new GbjGciException();
        status =  _GCSIinitialize(err);
	if (status < 0) throw err;
	return;
     }

/**

Return a String[][] containing cache statistic names.
<p>
This method should only be used by package gbjstats classes.
<p>
First index is by type, second is by particular cache stat.
@return    A String[][] containing cache stat names
@exception com.gemstone.gbjgci.GbjGciException  thrown for any error
*/

    public static synchronized String[][] GCSIstatNames( ) {
	int status;
	String[][] statNames = new String[20][];
        GbjGciException err = new GbjGciException();
        status =  _GCSIstatNames(statNames, err);
	if (status < 0) throw err;
	return statNames;
     }

/**

Attach to a particular shared page cache on this host.
<p>
This method should only be used by package gbjstats classes.

@param stoneName  Name of the stone 
                  (or the stone associated with this remote SPC)

@return        Maximum number of processes that can attach to this cache 
               (can be used to size arrays used to hold pid data returned
               by GCSIsampleStats( ) )

@exception com.gemstone.gbjgci.GbjGciException  thrown for any error
*/

     public static synchronized int GCSIattach(String stoneName) {
	int maxProcesses;
        GbjGciException err = new GbjGciException();
        maxProcesses =  _GCSIattach(stoneName, err);
	if (maxProcesses < 0) throw err;
	return maxProcesses;
     }

/**

Detach from an attached shared page cache.
<p>
This method should only be used by package gbjstats classes.

@exception com.gemstone.gbjgci.GbjGciException  thrown for any error
*/

     public static synchronized void GCSIdetach( ) {
	int status;
        GbjGciException err = new GbjGciException();
        status =  _GCSIdetach(err);
	if (status < 0) throw err;
	return;
     }

/**

Take a sample of cache stats for all processes on this cache. 
<p>
This method should only be used by package gbjstats classes.

<p>

Note that this only takes the sample -- use GCSIgetStats( ) to return
the actual sampled statistics.

@param processIds  A long[] to hold the returned process ID's of the 
                   sampled sessions.
@return            The actual number of sessions that were sampled.
@exception com.gemstone.gbjgci.GbjGciException  thrown for any error
*/
    public static synchronized int GCSIsampleStats(long[] processIds) {
	int count;
        GbjGciException err = new GbjGciException();
        count = _GCSIsampleStats(processIds, err);
	if (count < 0) throw err;
	return count;
    }

/**

Return back the cache statistics sampled earlier for a particular process.
<p>
This method should only be used by package gbjstats classes.

@param pid    The ID of the process of interest.
@param stats  A GbjGciCacheStats object to hold the returned cache stats.
@return       The GbjGciCacheStats object passed in above.
@exception com.gemstone.gbjgci.GbjGciException  thrown for any error
*/
    public static synchronized GbjGciCacheStats GCSIgetStats(long pid, 
						GbjGciCacheStats stats) {
	int status;
        GbjGciException err = new GbjGciException();
        status =  _GCSIgetStats(pid, stats, err);
	if (status < 0) throw err;
	return stats;
     }

/**

Return back the cache statistics sampled earlier for a particular process.
<p>
This method should only be used by package gbjstats classes.

@param index  A numeric index for indexing through a batch of sampled stats.
              Should be less than the max value returned by GCSIsampleStats( ).
@return       A GbjGciCacheStats object containg the cache stats.
@exception com.gemstone.gbjgci.GbjGciException  thrown for any error
*/
    public static GbjGciCacheStats GCSIgetStats(int index) {
	GbjGciCacheStats stats = null;
        stats = new GbjGciCacheStats( );
	return GCSIgetStats(index, stats);
    }

}




