/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjGciObject.java  
 *
 *========================================================================
 */

package com.gemstone.gbjgci;

import java.io.PrintStream;
import java.util.Date;
import java.util.logging.*;
import java.math.BigInteger;
import java.math.BigDecimal;
import java.time.*;
import org.apache.commons.math3.fraction.BigFraction;
import com.gemstone.gbj.BigDecimalFloat;

/**

GbjGciObject is the primary class for representing GS Objects in the
GbjGci package.  It extends GbjGciReport, which provides the basic
mechanism for communicating with the GemStone GCI layer, and provides
additional functionality, including object canonicalization, automatic
export set management, and object change synchronization.
<p>
Object canonicalization: a weak hash dictionary is used to insure that there
is only one copy of a given GS object with a particular oop.
<p>
Export set management:  using object canonicalization and the java
finalize mechanism, we can automatically track GS objects when brought
over to java (putting them into the export set), and recognize when
java GC's them (removing them from the export set).
<p>
Object change synchronization:  using the Java Observable/Observer
and the GCI dirty object tracking mechanisms, we can now track when
a GS object changes and have java update the associated java GS object.
<p>
Much of the core functionality of the GBJ class GbjObject has been moved
to this class, providing sufficient functionality for a "GBJ lite".
<p>
To support object canonicalization, constructors for GbjGciObject are 
different from the original GbjObject class.

<h3>Object Change Synchronization </h3>

GbjGciObject now provides 2 different mechanisms that allow a customer
application to have the java-side GbjGciObject (or superclass) instances
automatically updated when changes occur to the associated GS-server object,
either due to GS Smalltalk code execution or committed transactions from 
other sessions.
<p>
There is some overhead associated with these mechanisms, so by default
they are turned off.  To activate them for a given session, set the 
GbjSession field <b>monitorChangedObjects</b> to true.
<p>
The first approach uses the method updateClient().  Customer applications
should define their own versions of this method in superclasses of
GbjGciObject to update local fields.  Methods getBytes(), getNamedObjs(),
getObjs(), and related methods will be useful here.  Remember to also
map your superclasses in the session using GbjGciSession.mapGsToJavaClass().
<p>
The second approach uses the Java Observable/Observer API to notify 
customer application when the associated GS server object has changed.
Define an application class that implements the java Observer Interface.
This class should define the method <b>update(Observable o, Object arg)</b>,
to handle the changed object (returned as the Observable o) as appropriate
for your application.  The second argument <b>arg</b> is returned as nil.
<p>
Next, for a GbjGciObject that you wish to monitor, call the method
<b>addObserver(observer)</b>, where observer is an instance of your
application Observer class defined above.
<p>
When a change to the object occurs, the application Observer method 
<b>update(GbjObject obj, nil)</b> will be called, indicating
that the object has changed.
<p>
Note that both updateClient() and Observer objects can be used together.

@version 3.0
*/

public class GbjGciObject extends GbjGciReport implements Cloneable {

    /** Logger for debug trace messages */
    private static Logger log;

    // static code for setting up logger
    static {
        log = GbjGciInterface.setupLogger(GbjGciObject.class);
    }

    /** The GbjGciSession associated with this GS object */
    protected GbjGciSession session;


    // Protected Protocol

    // Note that for many of these methods, care must be taken to
    // preserve canonicalization in a given session.  "Loose" GbjGciObjects
    // (not associated with a session) should be handled carefully.


    // Protected Constructors

    /** Construct new GbjGciObject.
	Use with caution: GbjGciObjects fields need to be setup according
	to the type of object, and must be canonicalized in the session.
    */
    protected GbjGciObject(GbjGciSession aSession) {
	super();
        session = aSession;
    }

    protected GbjGciObject() {
	super();
        session = null;
    }

    /** Construct new GbjGciObject on a boolean.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(boolean b) {
	super(b);
    }

    /** Construct new GbjGciObject on a character.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(char c) {
	super(c);
    }

    /** Construct new GbjGciObject on a byte.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(byte b) {
	super(b);
    }

    /** Construct new GbjGciObject on a short.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(short i) {
	super(i);
    }

    /** Construct new GbjGciObject on a int.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(int i) {
	super(i);
    }

    /** Construct new GbjGciObject on a long.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(long i) {
	super(i);
    }

    /** Construct new GbjGciObject on a BigInteger.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(BigInteger i) {
	super(i);
    }

    /** Construct new GbjGciObject on a float.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(float f) {
	super(f);
    }

    /** Construct new GbjGciObject on a double.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(double d) {
	super(d);
    }

    /** Construct new GbjGciObject on a BigDecimal.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(BigDecimal d) {
	super(d);
    }
    
    /** Construct new GbjGciObject on a BigDecimalFloat.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(BigDecimalFloat df) {
	super(df);
    }
    
    /** Construct new GbjGciObject on a BigFraction.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(BigFraction f) {
	super(f);
    }
    
    /** Construct new GbjGciObject on a boolean.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(String str) {
	super(str);
    }

    /** Construct new GbjGciObject on a LocalDate.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(LocalDate d) {
	super(d);
    }
    
    /** Construct new GbjGciObject on a LocalTime.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(LocalTime t) {
	super(t);
    }

    /** Construct new GbjGciObject on a LocalDateTime.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(LocalDateTime dt) {
	super(dt);
    }

    /** Construct new GbjGciObject on a ZonedDateTime.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(ZonedDateTime zdt) {
	super(zdt);
    }

    /** Construct new GbjGciObject on a OffsetDateTime.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(OffsetDateTime odt) {
	super(odt);
    }

    /** Construct new GbjGciObject on a Date.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(Date d) {
	super(d);
    }

    /** Construct new GbjGciObject on an Instant.
	Use with caution: GbjGciObjects must be canonicalized in the session.
    */
    protected GbjGciObject(Instant i) {
	super(i);
    }

    /** Construct new GbjGciObject on an object.
	Use with caution: GbjGciObjects must be canonicalized in the session.
	The object must be an instance of a class recognized by other
	Constructor, otherwise a ClassCastException is generated.
    */
    protected GbjGciObject(Object o) {
	super(o);
    }

    /** Finalize this GbjGciObject when GC'ed.
	Make sure we clean up the object table weak reference and 
	add object to export set in the session
    */
    protected void finalize() {
	try { super.finalize(); } catch (Throwable x) {};
	if (session != null) session.releaseObject(this);
    }

    /** Set the GbjGciSession object associated with this GS object.
	This can only be done once when we initially setup the object.
	GbjGciObjects cannot be shared or transfered between sessions.
    */
    protected void setSession(GbjGciSession s) {
	// can only be done once
	if (s == null) 
	    throw new GbjGciException("Attempt to reset session");
	if (session != null) 
	    throw new GbjGciException("Attempt to reassign session");
	session = s;
    }

    /** Add additional info to the toString() printout for this object.
	For this class, add info on the session ID.
    */
    protected void toStringAdditions(StringBuffer sb) {
	super.toStringAdditions(sb);
	// sb.append("|serial:"); sb.append(serialNum_());
	sb.append("|sid:");
	if (session == null)
	    sb.append("0");
	else
	    sb.append(session.serialNum());
    }

    /** Check the valueType field, and if unknown, set it up based
	on GS server information on the object.  We need to check this
	periodically because the field is cleared when the object changes
	on the server, requiring us to update the java client object.
    */
    protected void checkValue() {
	// if valueType is UNKNOWN, may have been reset due to dirty object changes -- regenerate
	if ((valueType == VALTYPE_UNKNOWN) && (session != null)) {
	  session.setupObj(this);
	}
	return;
    }

    /** Make a copy of the GbjGciObject.
	Use with caution: GbjGciObjects must be canonicalized in the session.
	This method used by GbjGciSession.remapObject().
    */
    protected GbjGciObject copyFrom(GbjGciObject obj) {
	session = obj.session;
	super.copyFrom(obj);
	return this;
    }

    /** Mark this object as changed by setting the valueType and objType
	fields to unknown.
    */
    protected void markChanged() {
        int objType = objType_();
        if (objType == OBJTYPE_SPECIAL) {
          throw new GbjGciException("cannot change value of a special object");
        }
	valueType = VALTYPE_UNKNOWN;
	setObjType(OBJTYPE_UNKNOWN);
    }

    /** Notify observers that the state of this object has changed */
    protected void notifyChanged() {
	// clear the value setting so we'll update object state
	// the next time we look at this object
	markChanged();
	setChanged();
	updateClient();
	notifyObservers();
    }

    /** Return valueString, making sure it's set if needed.
	Note that this variant returns the full string.
     */
    protected String getValueString() {
	if (valueCache == null && valueType == VALTYPE_STRING) {
	   throw new GbjGciException("missing String value");
	}
	return (String)valueCache;
    }

    /** Update state of java client object when the GS server object changes.
	This method should be overridden in superclasses to define desired
	behavior for those superclasses. 
    */
    protected void updateClient() {
	// no behavior for the base class
    }


    // Public Protocol

    // Public Constructors

     /** Construct new GbjGciObject for a session based on a boolean.
 	 This generates a GS Boolean object.
	@param b Boolean
	@param s GbjGciSession
     */
    public GbjGciObject(Boolean b, GbjGciSession s) {
	this(b);
	session = s;
	s.setupObj(this);
    }

    /** Construct new GbjGciObject for a session based on a char.
 	 This generates a GS Character object.

	@param c char
	@param s GbjGciSession
     */
    public GbjGciObject(char c, GbjGciSession s) {
	this(c);
	session = s;
	s.setupObj(this);
    }

    /** Construct new GbjGciObject for a session based on a byte.
        This generates a GS SmallInteger object.
	@param b byte
	@param s GbjGciSession
     */
    public GbjGciObject(byte b, GbjGciSession s) {
	this(b);
	session = s;
	s.setupObj(this);
    }

    /** Construct new GbjGciObject for a session based on a short 
        This generates a GS SmallInteger object.
	@param i short
	@param s GbjGciSession
     */
    public GbjGciObject(short i, GbjGciSession s) {
	this(i);
	session = s;
	s.setupObj(this);
    }

    /** Construct new GbjGciObject for a session based on a int 
        This generates a GS SmallInteger object.
     @param i int
     @param s GbjGciSession
    */
    public GbjGciObject(int i, GbjGciSession s) {
	this(i);
	session = s;
	s.setupObj(this);
    }

    /** Construct new GbjGciObject for a session based on a long 
        This generates a GS SmallInteger object.
	@param i long
	@param s GbjGciSession
     */
    public GbjGciObject(long i, GbjGciSession s) {
	this(i);
	session = s;
	s.setupObj(this);
    }

    /** Construct new GbjGciObject for a session based on a float 
        This generates a GS SmallDouble object.
	@param f float
	@param s GbjGciSession
     */
    public GbjGciObject(float f, GbjGciSession s) {
	this(f);
	session = s;
	s.setupObj(this);
    }

    /** Construct new GbjGciObject for a session based on a double 
        This generates a GS Float object.
	@param d double
	@param s GbjGciSession
     */
    public GbjGciObject(double d, GbjGciSession s) {
	this(d);
	session = s;
	s.setupObj(this);
    }

    /** Construct new GbjGciObject for a session based on a string
	This will generate either a GS String or DoubleByteString
	as appropriate.
	@param str String
	@param s GbjGciSession
     */
    public GbjGciObject(String str, GbjGciSession s) {
	this(str);
	session = s;
	s.setupObj(this);
    }

    /** Construct new GbjGciObject for a session based on a LocalDate.
	This generates a GS Date object.
	@param d LocalDate
	@param s GbjGciSession
     */
    public GbjGciObject(LocalDate d, GbjGciSession s) {
	this(d);
	session = s;
	s.setupObj(this);
    }


    /** Construct new GbjGciObject for a session based on a LocalTime.
	This generates a GS Time object.
	@param t LocalTime
	@param s GbjGciSession
     */
    public GbjGciObject(LocalTime t, GbjGciSession s) {
	this(t);
	session = s;
	s.setupObj(this);
    }

    /** Construct new GbjGciObject for a session based on a LocalDateTime.
	This generates a GS LocalDateTime object.
	@param ldt LocalDateTime
	@param s GbjGciSession
     */
    public GbjGciObject(LocalDateTime ldt, GbjGciSession s) {
	this(ldt);
	session = s;
	s.setupObj(this);
    }

    /** Construct new GbjGciObject for a session based on a ZonedDateTime.
	This generates a GS DateTime object.
	@param zdt ZonedDateTime
	@param s GbjGciSession
     */
    public GbjGciObject(ZonedDateTime zdt, GbjGciSession s) {
	this(zdt);
	session = s;
	s.setupObj(this);
    }

    /** Construct new GbjGciObject for a session based on a OffsetDateTime.
	This generates a GS DateAndTime object.
	@param odt LocalDateTime
	@param s GbjGciSession
     */
    public GbjGciObject(OffsetDateTime odt, GbjGciSession s) {
	this(odt);
	session = s;
	s.setupObj(this);
    }
    
    /** Construct new GbjGciObject for a session based on a Date.
	This generates a GS DateAndTime object.
	@param d Date
	@param s GbjGciSession
    */
    public GbjGciObject(Date d, GbjGciSession s) {
	this(d);
	session = s;
	s.setupObj(this);
    }
    
    /** Construct new GbjGciObject for a session based on a Instant.
	This generates a GS DateAndTime object.
	@param i Instant
	@param s GbjGciSession
    */
    public GbjGciObject(Instant i, GbjGciSession s) {
	this(i);
	session = s;
	s.setupObj(this);
    }

    /** Construct new GbjGciObject for a session based on an object.
	The object must be an instance of a class recognized by other
	Constructor, otherwise a ClassCastException is generated.
	@param o Object
	@param s GbjGciSession
    */
    public GbjGciObject(Object o, GbjGciSession s) {
	this(o);
	session = s;
	s.setupObj(this);
    }

    // Various accessor methods need to double-check the valueType
    // field, in case object changes have reset it to UNKNOWN and
    // we need to update the cache fields

    /** Returns an integer value representing the type of java object
      	cached on the client.  See GbjGciReport for details.
	@return int result
     */
    public int getValType() {
	checkValue();
	return super.getValType();
    }

    /** Returns an integer value representing the type of GS server
     	object.  See GbjGciReport for details.
        @return int result
     */
    public int getObjType() {
	checkValue();
	return super.getObjType();
    }

    /** Return the GbjGciSession object associated with this GS object 
        @return GbjGciSession result
     */
    public GbjGciSession getSession() { 
	return session; };

    /** Return the size of named instance variables
     	of this oop-based GS object 
        @return int result
     */
    public int getNamedSize() { 
	checkValue();
	return namedSize; 
    }

    /** Return the size of the indexed component
        of this oop-based GS object 
        @return long result
     */
    public long getIndexedSize() { 
	checkValue();
	return indexedSize; 
    }

    /** Return the total size (including both instance variables and
        indexed components) of this oop-based GS object
        @return long result
     */
    public long getTotalSize() { 
	checkValue();
	return namedSize + indexedSize; 
    }

    /** Return the GS object represented by this oop 
	@param oopValue  oop 
	@param s GbjGciSession
        @return GbjGciObject result
     */
    public static GbjGciObject withOop(long oopValue,
				       GbjGciSession s) {
	return s.objWithOop(oopValue);
    }

    /** Return a hash code value for this object 
        @return int result
     */
    public int hashCode() {
	checkValue();
	int hash = super.hashCode();
	if (session != null)
	    hash = hash ^ session.hashCode();
	return hash;
    }


    /** Equals comparison against another java object 
        @return boolean result
     */
    public boolean equals(Object o) {
	if (!(o instanceof GbjGciObject)) return false;
	GbjGciObject obj = (GbjGciObject)o;
	if (session != obj.session) return false;
	checkValue();
	return super.equals(obj);
    }

    /** Return the GS object with this name in the DB.
     	On the server, looks through the symbol list for this session
     	looking for entry with this name, scanning symbol dictionaries
     	in the order listed in the symbol list.
	@param name  Name to use
	@param s GbjGciSession
        @return GbjGciObject result
     */
    public static GbjGciObject named(String name, GbjGciSession s) {
	return s.objectNamed(name);
    }

    /** Return a short string representation of this GbjGciReport 
        @return String result
     */
    public String toString() {
	checkValue();
	return super.toString();
    }

    /** Write to printstream a detailed multi-line string describing
     	internal structure of this GbjGciObject.  
     	Intended for diagnostic purposes.
     */
    public void debugPrint(PrintStream out) {
	checkValue();
	super.debugPrint(out);
	if (session == null) {
	    out.println("  session: [null]");
	} else {
	  GbjGciObject res = session.execute("((Object _objectForOop: " + oop + ") ifNil:['nil'])  asString");
	  out.println("  server class name " + res.stringValue());
	  out.println("  session: " + session.getSessionId());
        }
    }

    /** Return true if this GS object is a "string".
     	Strings include any instance of a subclass of GS class
     	CharacterCollection.
        @return boolean result
     */
    public boolean isString() {
	checkValue();
	return super.isString();
    }

    /** If this GS object is representable as a boolean, return that value.
        @return boolean result
     	@exception ClassCastException thrown if GS object not a boolean
     */
    public boolean booleanValue() {
	checkValue();
	return super.booleanValue();
    }

    /** If this GS object is representable as a character, return that value.
        @return char result
     	@exception ClassCastException thrown if GS object not a character
     */
    public char charValue() {
	checkValue();
	return super.charValue();
    }

    /** If this GS object is representable as a byte, return that value.
	@return byte result
     	@exception ClassCastException thrown if GS object not a byte
     */
    public byte byteValue() {
	checkValue();
	return super.byteValue();
    }

    /** If this GS object is representable as a short, return that value.
	@return short result
     	@exception ClassCastException thrown if GS object not a short
     */
    public short shortValue() {
	checkValue();
	return super.shortValue();
    }

    /** If this GS object is representable as an int, return that value.
	@return int result
     	@exception ClassCastException thrown if GS object not a int
     */
    public int intValue() {
	checkValue();
	return super.intValue();
    }

    /** If this GS object is representable as a long, return that value.
     	If the value is too large, raises an ArithmeticException.
	(use bigIntegerValue( ) in that case).
	@return long result
     	@exception ClassCastException thrown if GS object not a long
     */
    public long longValue() {
	checkValue();
	return super.longValue();
    }

    /** If this GS object is representable as a BigInteger, return that value.
	@return BigInteger result
     	@exception ClassCastException thrown if GS object not a BigInteger
     */
    public BigInteger bigIntegerValue() {
	checkValue();
	return super.bigIntegerValue();
    }

    /** If this GS object is representable as a float, return that value.
	@return float result
     	@exception ClassCastException thrown if GS object not a float
     */
    public float floatValue() {
	checkValue();
	return super.floatValue();
    }

    /** If this GS object is representable as a double, return that value.
	@return double result
     	@exception ClassCastException thrown if GS object not a double
     */
    public double doubleValue() {
	checkValue();
	return super.doubleValue();
    }

    /** If this GS object is representable as a BigDecimal, return that value.
	@return BigDecimal result
     	@exception ClassCastException thrown if GS object not a BigDecimal
     */
    public BigDecimal bigDecimalValue() {
	checkValue();
	return super.bigDecimalValue();
    }

    /** If this GS object is representable as a BigDecimalFloat, return that value.
	@return BigDecimalFloat result
     	@exception ClassCastException thrown if GS object not a BigDecimalFlat
     */
    public BigDecimalFloat bigDecimalFloatValue() {
	checkValue();
	return super.bigDecimalFloatValue();
    }
    
    /** If this GS object is representable as a BigFraction, return that value.
	@return BigFraction result
     	@exception ClassCastException thrown if GS object not a BigFraction
     */
    public BigFraction bigFractionValue() {
	checkValue();
	return super.bigFractionValue();
    }

    /** If this GS object is representable as an LocalDate, return that value.
	@return LocalDate result
	@exception ClassCastException thrown if GS object not a (Local)Date
    */
    public LocalDate localDateValue() {
	checkValue();
	return super.localDateValue();
    }

    /** If this GS object is representable as a LocalTime, return that value.
	@return LocalTime result
	@exception ClassCastException thrown if GS object not a LocalTime
    */
    public LocalTime localTimeValue() {
	checkValue();
	return super.localTimeValue();
    }

    /** If this GS object is representable as a LocalDateTime, return that value.
	@return LocalDateTime result
	@exception ClassCastException thrown if GS object not a LocalDateTime
    */
    public LocalDateTime localDateTimeValue() {
	checkValue();
	return super.localDateTimeValue();
    }

    /** If this GS object is representable as a ZonedDateTime, return that value.
	@return ZonedDateTime result
	@exception ClassCastException thrown if GS object not a ZonedDateTime
    */
    public ZonedDateTime zonedDateTimeValue() {
	checkValue();
	return super.zonedDateTimeValue();
    }

    /** If this GS object is representable as a OffsetDateTime, return that value.
	@return OffsetDateTime result
	@exception ClassCastException thrown if GS object not a OffsetDateTime
    */
    public OffsetDateTime offsetDateTimeValue() {
	checkValue();
	return super.offsetDateTimeValue();
    }

    /** If this GS object is representable as a Date, return that value.
	@return Date result
	@exception ClassCastException thrown if GS object not a Date
    */
    public Date dateValue() {
	checkValue();
	return super.dateValue();
    }

    /** If this GS object is representable as an Instant, return that value.
	@return Instant result
	@exception ClassCastException thrown if GS object not a Instant
    */
    public Instant instantValue() {
	checkValue();
	return super.instantValue();
    }

    /** If this GS object is reprentable in bytes (other than Strings),
	return a byte array represention that value.
	<p>
        WARNING: only the cached portion of this value is returned,
	limited by the GbjGciInterface.initialize preloadSize.  
	To insure getting the complete object, use the method getBytes().
	@return byte array result
	@exception ClassCastException thrown if GS object not bytes
    */
    public byte[] bytesValue() {
	checkValue();
	return super.bytesValue();
    }

    /** If this GS object is representable as a String, return that value.
	If server object is neither a String, MultiByteString, Utf8 , nor Utf16 ,
        returns result of sending   asString   to the server object.
	<p>
	@return String result
	@exception ClassCastException thrown if session is not logged in
    */
    public String stringValue() { 
	checkValue();
	if (valueType != VALTYPE_STRING) {
	  if (session.isLoggedIn()) {
	    return session.perform(this, "asString", null, 0).stringValue();
          } else {
	    throw new ClassCastException("session not connected");
          }
	}
	return (String)valueCache;
    }

    /** If this GS object is representable as a String, return that value.
	@return String result
	@exception ClassCastException thrown if server object is neither a String,
         MultiByteString, Utf8 , nor Utf16 .
    */
    public String stringValue_() { 
	checkValue();
	if (valueType != VALTYPE_STRING) {
	  throw new ClassCastException("not a String");
	}
	return (String)valueCache;
    }


/**
Retrieve the designated bytes from this byte-based GS object.
@param startIndex  The starting index (1-based)
@param numBytes    The number of bytes to retrieve
@return the retrieved bytes
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public byte[] getBytes(int startIndex,
			   int numBytes) {
	return session.getBytes(this, startIndex, numBytes);
    }

/** 
Retrieve the designated byte from this byte-based GS object.
@param index  The index of the byte to retrieve (1-based)
@return the retrieved byte
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public byte getByte(int index) {
	byte[] bytes = getBytes(index, 1);
	return bytes[0];
    }

/**
Retrieve the designated GS objects from this indexable GS object.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to retrieve
@param objs        An array of GS objects to hold the results
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public void getObjs(int startIndex, 
			int numObjs, 
			GbjGciObject[] objs) {
	session.getObjs(this, startIndex, numObjs, objs);
    }


/**
Retrieve the designated GS objects from this indexable GS object.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to retrieve
@return An array of GS objects to hold the results
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public GbjGciObject[] getObjs(int startIndex, 
				  int numObjs) {
	GbjGciObject[] objs = new GbjGciObject[numObjs];
	getObjs(startIndex, numObjs, objs);
	return objs;
    }

/**
Retrieve the designated GS object from this indexable GS object
@param index  The starting index (1-based)
@return The GS object at the designated index 
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public GbjGciObject getObj(int index) {
	GbjGciObject[] objs = new GbjGciObject[1];
	getObjs(index, 1, objs);
	return objs[0];
    }

/**
Retrieve the designated GS objects from the named instvars for this GS object.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to retrieve
@param objs        An array of GS objects to hold the results
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public void getNamedObjs(int startIndex, 
			     int numObjs, 
			     GbjGciObject[] objs) {
	session.getNamedObjs(this, startIndex, numObjs, objs);
    }

/**
Retrieve the designated GS object from the named instvar for this GS object.
@param index  The index of the desired instvar (1-based)
@return       The GS object from the designated instvar
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public GbjGciObject getNamedObj(int index) {
	GbjGciObject[] objs = new GbjGciObject[1];
	getNamedObjs(index, 1, objs);
	return objs[0];
    }

/**
Write the designated bytes to this byte-based GS object.
@param startIndex  The starting index (1-based)
@param numBytes    The number of bytes to write
@param bytes       The bytes to write out
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public void putBytes(int startIndex,
			 int numBytes,
			 byte[] bytes) {
	session.putBytes(this, startIndex, numBytes, bytes);
    }

/**
Write the designated bytes to this byte-based GS object.
@param startIndex  The starting index (1-based)
@param bytes       The bytes to write out
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public void putBytes(int startIndex,
			 byte[] bytes) {
	putBytes(startIndex, bytes.length, bytes);
    }

/**
Write the designated byte to this byte-based GS object.
@param index  The index (1-based)
@param b      The byte to write out
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public void putByte(int index, byte b) {
	byte[] bytes = new byte[1];
	bytes[0] = b;
	putBytes(index, 1, bytes);
    }

/**
Write the designated GS objects to this indexable GS object.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to write
@param objs        An array of GS objects to write out
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public void putObjs(int startIndex,
			int numObjs, 
			GbjGciObject[] objs) {
	session.putObjs(this, startIndex, numObjs, objs);
    }

/**
Write the designated GS objects to this indexable GS object.
@param startIndex  The starting index (1-based)
@param objs        An array of GS objects to write out
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public void putObjs(int startIndex,
			GbjGciObject[] objs) {
	putObjs(startIndex, objs.length, objs);
    }

/**
Write the designated GS object to this indexable GS object.
@param index  The index of the instvar to write out (1-based)
@param obj    The GS object to write out
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public void putObj(int index, GbjGciObject obj) {
	GbjGciObject[] objs = new GbjGciObject[1];
	objs[0] = obj;
	putObjs(index, 1, objs);
    }

/**
Write the designated GS objects to the named instvars for this GS object.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to write
@param objs        An array of GS objects to write out
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public void putNamedObjs(int startIndex,
			     int numObjs, 
			     GbjGciObject[] objs) {
	session.putNamedObjs(this, startIndex, numObjs, objs);
    }

/**
Write the designated GS objects to the named instvars for this GS object.
@param startIndex  The starting index (1-based)
@param objs        An array of GS objects to write out
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public void putNamedObjs(int startIndex,
			     GbjGciObject[] objs) {
	putNamedObjs(startIndex, objs.length, objs);
    }

/**
Write the designated GS object to the named instvars for this GS object.
@param index  The index of the instvar to write out (1-based)
@param obj    The GS object to write out
@exception com.gemstone.gbjgci.GbjGciException thrown on any errors
*/
    public void putNamedObj(int index, GbjGciObject obj) {
	GbjGciObject[] objs = new GbjGciObject[1];
	objs[0] = obj;
	putNamedObjs(index, 1, objs);
    }

}

