/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjGciPreferences.java  
 *
 *========================================================================
 */

package com.gemstone.gbjgci;

import java.io.*;
import java.util.prefs.*;
import java.util.logging.*;

/** 

GbjGciPreferences uses java Preferences (java System.getProperties on Android)
to determine debug trace logging levels for the various Gbj/GbjGci modules.
Customer applications should not need to change these unless requested by 
GemTalk Technical Support to assist in debugging a Gbj/GbjGci problem.
<p>
Run by executing:
<p>
prompt&gt; java com.gemstone.gbjgci.GbjPreferences [command] [arg1] [arg2]
<p>
Where valid commands are:
<ul>
<li> help  - print out help text 
<li> print - print out current settings of all parameters
<li> clear - remove all parameters
<li> set [parameter] [value] - set parameter
<li> remove [parameter]      - remove parameter
</ul>

<p>
Valid parameters and their values are:
<ul>
<li> GbjGciInterface.CLogLevel <br>
     Enable/disable trace logging of JNI C code. <br>
     Possible values:  OFF INFO DEBUG <br>
     Default value: OFF
<li> GBJ.LogLevel
<li> &lt;GBJ Class&gt;.LogLevel <br>
     Specify level of trace logging for this class. <br>
     GBJ.LogLevel applies if class level is not set. <br>
     Possible values: OFF SEVERE WARNING INFO CONFIG FINE FINER FINEST ALL <br>
     Default value: WARNING
</ul>

*/

public class GbjGciPreferences {

    private static Preferences prefs = 
	Preferences.userRoot().node(GbjGciPreferences.class.getName());

    /** Constructor is not really needed as all behavior is static */
    public GbjGciPreferences() { }

    /** Return a preference string value 
	@param pref String
	@param deflt String
	@return String result
     */
    public static String get(String pref, String deflt) {
	return prefs.get(pref, deflt);
    }

    /** Return a preference string value 
	@param pref String
	@return String result
     */
    public static String get(String pref) {
	return get(pref, "");
    }

    /** Return 0/1/2 if off/info/debug trace logging of 
	GbjGciInterface JNI code should be enabled 
	@return int result
    */
    public static int debugC() {
	String debugCString = get("GbjGciInterface.CLogLevel", "OFF");
	if (debugCString.equals("OFF")) return 0;
	if (debugCString.equals("INFO")) return 1;
	if (debugCString.equals("DEBUG")) return 2;
	System.out.println("Invalid CLogLevel for GbjGciInterface: " + 
			   debugCString);
	System.out.println("Default level OFF will be used");
	return 0;
    }

    /** Return trace debug logging level for designated module 
	@param moduleName String
	@return Level result
     */
    public static Level logLevel(String moduleName) {
	String name = moduleName + ".LogLevel";
	String level = null;
	try {
	    level = get(name);
	    if (level.equals("")) {
		level = get("GBJ.LogLevel", "WARNING");
	    }
	    return Level.parse(level);
	} catch (Throwable e) {
	    // if any problems, just use WARNING
	    return Level.WARNING;
	}
    }

    // FOLLOWING ARE FOR THE COMMAND-LINE INTERFACE

    private static void error(String msg) {
	System.out.println("[ERROR] " + msg);
    }

    private static void badCommand(String msg) {
	error(msg);
	System.out.println(
           "Valid commands are: help, print, clear, set, remove");
    }

    private static void print() {
	String[] keys;
	try {
	    keys = prefs.keys();
	} catch (BackingStoreException e) {
	    error("Cant access preference backing store");
	    return;
	}
	for (int i=0; i<keys.length; i++) {
	    System.out.println(keys[i] + ": " + 
			       prefs.get(keys[i], "[unknown]"));
	}
    }

    private static void clear() {
	String[] keys;
	try {
	    keys = prefs.keys();
	} catch (BackingStoreException e) {
	    error("Cant access preference backing store");
	    return;
	}
	for (int i=0; i<keys.length; i++) remove(keys[i]);
    }

    private static void set(String key, String value) {
	prefs.put(key, value);
	try { 
	    prefs.flush(); 
	} catch (BackingStoreException e) {
	    error("Cant access preference backing store");
	}
	System.out.println("Parameter " + key + " set to " + value);
    }

    private static void remove(String key) {
	try { 
	    prefs.remove(key) ;
	}
	catch (NullPointerException e) {
	    error("No such key " + key);
	}
    }

    private static void help() {
    System.out.println("");
    System.out.println("Valid commands are:");
    System.out.println("");
    System.out.println("  help  - print out this help text");
    System.out.println("  print - print out current settings");
    System.out.println("  clear - remove all parameters");
    System.out.println("  set <parameter> <value> - set parameter");
    System.out.println("  remove <parameters>     - remove parameter");
    System.out.println("");
    System.out.println("Valid parameters and their values are:");
    System.out.println("");
    System.out.println("  GbjGciInterface.CLogLevel");
    System.out.println("     Enable/disable trace logging of JNI C code");
    System.out.println("     Possible values:  OFF INFO DEBUG");
    System.out.println("     Default value: OFF");
    System.out.println("");
    System.out.println("  GBJ.LogLevel");
    System.out.println("  <GBJ Class>.LogLevel");
    System.out.println("    Specify level of trace logging for this class.");
    System.out.println("    GBJ.LogLevel applies if class level is not set.");
    System.out.println("    Possible values:");
    System.out.println(
        "      OFF SEVERE WARNING INFO CONFIG FINE FINER FINEST ALL");
    System.out.println("    Default value: WARNING");
    System.out.println("");
    }

    /** Main program for viewing/setting/removing 
	Gbj/GbjGci debug trace logging levels 
	@param args  String array of input arguments
    */
    public static void main(String[] args) {
	if (args.length < 1) {
	    badCommand("No command specified");
	} else {
	    String command = args[0];
	    if (command.equals("help")) {
		help();
	    } else if (command.equals("set")) {
		if (args.length < 2) {
		    error("Missing key/value arguments in set command");
		} else if (args.length < 3) {
		    error("Missing value argument in set command");
		} else {
		    set(args[1], args[2]);
		}
	    } else if (command.equals("remove")) {
		if (args.length < 2) {
		    error("Missing key argument in remove command");
		} else {
		    remove(args[1]);
		}
	    } else if (command.equals("print")) {
		print();
	    } else if (command.equals("clear")) {
		clear();
	    } else {
		badCommand("Invalid command " + command);
	    }
	}
    }
}




