/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *  GbjGciReport.java 
 *
 *========================================================================
 */

package com.gemstone.gbjgci;

// javah -jni com.gemstone.gbjgci.GbjGciReport

import java.io.PrintStream;
import java.util.Observable;
import java.util.Calendar;
import java.util.Date;
import java.text.DateFormat;
import java.util.TimeZone;
import java.util.logging.*;
import java.time.*;
import java.time.format.*;
import java.math.BigInteger;
import java.math.BigDecimal;
import org.apache.commons.math3.fraction.BigFraction;
import com.gemstone.gbj.BigDecimalFloat;


/**
GbjGciReport provides the low-level GS Object representation that
is used in communication with the GS/S GCI layer.  The class extends
java.util.Observable to provide support for client/server object 
synchronization, although support for this capability is not present at 
this level (see class GbjGciObject for client/server synchronization support).
<p>
Customer applications should not directly use this class, but instead
use GbjGciObject or GbjObject as appropriate.  Note that for many of
the public methods, we don't confirm that the client fields are up-to-date
if there have been server-side object changes.  Many of these methods have
versions in GbjGciObject that do provide the update before calling the 
super-version.
<p>
The main public methods available from this class include:
<ul>
<li> public int getObjType() <br>
     Return an integer value representing the type of object
     as represented on the GS server.  Possible values include: <br>
     <pre>
     int  Constant
     ---  --------------
     -1:  OBJTYPE_UNKNOWN
      0:  OBJTYPE_OOP
      1:  OBJTYPE_BYTE
      2:  OBJTYPE_NSC
      3:  OBJTYPE_SPECIAL
     </pre>
     
<li> public int getValType() <br>
     Return an integer value representing the type of java object
     cached on the client as representative of the GS server object.
     Possible values include: <br>
     <pre>
Num ValType                    Java Type / Class       GemStone Classes
--- ----------------------     ---------------------   -----------------------
-1  VALTYPE_UNKNOWN
 0  VALTYPE_NULL               null                    UndefinedObject (one instance nil)
 1  VALTYPE_BOOLEAN            boolean                 Boolean
 2  VALTYPE_CHARACTER          char                    Character
 3  VALTYPE_LONG               byte/short/int/long     SmallInteger(61 bits) or LargeInteger
 4  VALTYPE_DOUBLE             float/double            SmallDouble/Float
 5  VALTYPE_BYTES              byte[]                  ByteArray other than Utf8, Utf16
 6  VALTYPE_OOPS
 7  VALTYPE_STRING             String                  String/DoubleByteString/QuadByteString
                                                       Unicode7/Unicode16/Unicode32
                                                       Symbol/DoubleByteSymbol/QuadByteSymbol
                                                       Utf8 , Utf16

 9  VALTYPE_CALENDAR           Calendar                DateTime   (legacy support, deprecated)

10  VALTYPE_ZONED_DATETIME     ZonedDateTime           GbjZonedDateTime (subclass of DateAndTime)
11  VALTYPE_DATE_AND_TIME      OffsetDateTime          DateAndTime / SmallDateAndTime                              
12  VALTYPE_DAY                LocalDate               SmallDate/Date
13  VALTYPE_TIME               LocalTime               SmallTime/Time
14  VALTYPE_LARGE_INTEGER      BigInteger              LargeInteger
15  VALTYPE_SCALED_DECIMAL     BigDecimal              SmallScaledDecimal/ScaledDecimal
16  VALTYPE_DECIMAL_FLOAT      BigDecimalFloat         DecimalFloat
17  VALTYPE_FRACTION           BigFraction             SmallFraction/Fraction      
18  VALTYPE_LOCAL_DATETIME     LocalDateTime           GbjLocalDateTime (subclass of DateAndTime) 
19  VALTYPE_utilDATE           Date                    GbjUtilDate (subclass of DateAndTime)
20  VALTYPE_INSTANT            Instant                 GbjInstant  (subclass of DateAndTime)

     </pre>
</ul>

@version 3.2.0

*/

public class GbjGciReport extends Observable implements Cloneable {

    /** Logger for debug trace messages */
    private static Logger log;

    // static code for setting up logger
    static {
	log = GbjGciInterface.setupLogger(GbjGciReport.class);
    }

    // code used for objType

       /** Constant value representing object type UNKNOWN */
    protected static final byte OBJTYPE_UNKNOWN   = -1;
       /** Constant value representing object type OOP */
    protected static final byte OBJTYPE_OOP       = 0;
       /** Constant value representing object type BYTE */
    protected static final byte OBJTYPE_BYTE      = 1;
       /** Constant value representing object type NSC */
    protected static final byte OBJTYPE_NSC       = 2;
       /** Constant value representing object type SPECIAL */
    protected static final byte OBJTYPE_SPECIAL   = 3;

    // code used for valueType
    // if these change, also update list in GbjGciInterface.c
    // (would have been nice to use an enumerated type,
    // but the jni generated .h file for the C-code doesn't
    // propagate them over..)

       /** Constant value representing value type UNKNOWN.
        Used when the local java value is unknown. */
    protected static final int VALTYPE_UNKNOWN   = -1;
       /** Constant value representing value type NULL.
        Used for representing java null / GemStone nil. */
    protected static final int VALTYPE_NULL       = 0;
       /** Constant value representing a java boolean / GemStone Boolean. */
    protected static final int VALTYPE_BOOLEAN    = 1;
       /** Constant value representing a java char / GemStone Character. */
    protected static final int VALTYPE_CHARACTER  = 2;
       /** Constant value representing a java byte/short/int/long / GemStone SmallInteger.*/
    protected static final int VALTYPE_LONG       = 3;
       /** Constant value representing a java float/double / GemStone SmallDouble/Float.*/
    protected static final int VALTYPE_DOUBLE     = 4;
       /** Constant value representing BYTES.
	Used for representing byte-based GemStone objects
	not otherwise covered by other value types. */
    protected static final int VALTYPE_BYTES      = 5;
       /** Constant value representing OOPS.
	Used for representing oop-based GemStone objects 
	not otherwise covered by other value types. */
    protected static final int VALTYPE_OOPS       = 6;
       /** Constant value representing java String / GemStone String/DoubleByteString.*/
    protected static final int VALTYPE_STRING     = 7;

      // VALTYPE_DBSTRING   = 8  no longer used

    /** Constant value representing value type CALENDAR.
        Legacy support for java objects of class Calendar, instances of GemStone  DateTime  */
    protected static final int VALTYPE_CALENDAR   = 9;

        /** Constant value representing a java ZonedDateTime / GemStone GbjZonedDateTime */
    protected static final int VALTYPE_ZONED_DATETIME          = 10;
        /** Constant value representing a java OffsetDateTime / GS DateAndTime. */
    protected static final int VALTYPE_DATE_AND_TIME     = 11;
        /** Constant value representing a java LocalDate / GemStone Date */
    protected static final int VALTYPE_DAY               = 12;
        /** Constant value representing a java LocalTime / GemStone Time */
    protected static final int VALTYPE_TIME              = 13;
        /** Constant value representing a java BigInteger / GemStone LargeInteger */
    protected static final int VALTYPE_LARGE_INTEGER     = 14;
        /** Constant value representing a java BigDecimal / GemStone ScaledDecimal */
    protected static final int VALTYPE_SCALED_DECIMAL    = 15;
        /** Constant value representing a java BigDecimalFloat / GemStone DecimalFloat */
    protected static final int VALTYPE_DECIMAL_FLOAT     = 16;
        /** Constant value representing a java BigFraction / GemStone Fraction */
    protected static final int VALTYPE_FRACTION          = 17;

      /** Constant value representing a java LocalDateTime / GemStone LocalDateTime. */
    protected static final int VALTYPE_LOCAL_DATETIME    = 18;

    /** Constant value representing a java.util.Date  /  GemStone GbjUtilDate   */
    protected static final int VALTYPE_utilDATE    = 19;

    /** Constant value representing a java.time.Instant  /  GemStone GbjInstant   */
    protected static final int VALTYPE_INSTANT    = 20;

      /** GemStone object identifier (oop) for this object */
    protected transient long oop;   // value assigned from C 
      /** The object identifier for the class of this object */
    protected long classOop;        // value assigned from C 

      /** An integer representing the GS object implementation of this object.
	See static int constants with name prefix OBJTYPE_* for possible values. */
    private byte objType; 
       /** An integer representing the type of java object cached locally
	to represent this object (if any). 
	See static int constants with name prefix VALTYPE_* for possible values. */
    protected byte    valueType;  // VALTYPE_* 
        /** The number of names instvars in this object */
    protected int namedSize; 
        /** If a collection, the number of bytes/oops in the collection */
    protected long indexedSize; 

    protected int     valueInt1;   // year in DateTime types
    protected long    valueLong;    // used in numeric types
    protected long    valueLong2;  // used in numeric types
    protected double  valueDouble;   // used in floating point types

    /** A cache used for holding bytes in byte-based objects */
    protected Object  valuesAry;  // usually  byte[] or long[]  cache of instVars 
    protected Object  secondValue;  // denominatorBytes or denominatorString for Fraction or other numeric string

    /** A cache used for holding oops in oop-based objects */
    protected long[]  valueOops() { return (long[])valuesAry; }
    protected byte[]  valueBytes() { return (byte[])valuesAry; }

    /** A cache for holding a complex java object best representing this object
	(other than boolean, char, long, double) */
    protected Object  valueCache;

    // private int serialNum; // uncomment for debugging  
    // protected int serialNum_() { return serialNum; }
    //
    // private static int staticSerialCounter = 0;
    //
    // private static synchronized int incrementSerial() { // do not checkin
    //    staticSerialCounter = staticSerialCounter + 1;
    //   return staticSerialCounter;
    // }
    // private void setSerialNum() {
    //   serialNum = incrementSerial();
    // }

    /** Construct a GbjGciReport for an object of unknown type.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport( ) {
	// most of the fields are setup in GbjGciInterface C layer
        // setSerialNum();
	objType   = OBJTYPE_UNKNOWN;
	valueType = VALTYPE_UNKNOWN;
    }

    /** Construct a GbjGciReport configured to represent a boolean.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(boolean b) {
	this();
	setup(b);
    }

    /** Construct a GbjGciReport configured to represent a character.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(char c) {
	this();
	setup(c);
    }

    /** Construct a GbjGciReport configured to represent a byte.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(byte b) {
	// use LONG field
	this();
	setup(b);
    }

    /** Construct a GbjGciReport configured to represent a short.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(short i) {
	// use LONG field
	this();
	setup(i);
    }

    /** Construct a GbjGciReport configured to represent an int.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(int i) {
	// use LONG field
	this();
	setup(i);
    }

    /** Construct a GbjGciReport configured to represent a long.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(long i) {
	// use LONG field
	this();
	setup(i);
    }

    /** Construct a GbjGciReport configured to represent a BigInteger.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(BigInteger i) {
	this( );
	setup(i);
    }

    /** Construct a GbjGciReport configured to represent a float.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(float f) {
	// use DOUBLE field
	this();
	setup(f);
    }

    /** Construct a GbjGciReport configured to represent a double.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(double d) {
	this();
	setup(d);
    }

    /** Construct a GbjGciReport configured to represent a BigDecimal.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(BigDecimal d) {
	this();
	setup(d);
    }
    
    /** Construct a GbjGciReport configured to represent a BigDecimalFloat.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(BigDecimalFloat df) {
	this();
	setup(df);
    }

    /** Construct a GbjGciReport configured to represent a BigFraction.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(BigFraction f) {
	this();
	setup(f);
    }
    
    /** Construct a GbjGciReport configured to represent a string.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(String str) {
	this();
	setup(str);
    }

    /** Construct a GbjGciReport configured to represent a LocalDate.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(LocalDate d) {
	this( );
	setup(d);
    }
    /** Construct a GbjGciReport configured to represent a LocalTime.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(LocalTime t) {
	this( );
	setup(t);
    }

    /** Construct a GbjGciReport configured to represent a LocalDateTime.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(LocalDateTime dt) {
	this( );
	setup(dt);
    }
    
    /** Construct a GbjGciReport configured to represent a ZonedDateTime.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(ZonedDateTime zdt) {
	this( );
	setup(zdt);
    }

    /** Construct a GbjGciReport configured to represent a OffsetDateTime.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(OffsetDateTime odt) {
	this( );
	setup(odt);
    }

    /** Construct a GbjGciReport configured to represent a Date.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(Date d) {
	this();
	setup(d);
    }

    /** Construct a GbjGciReport configured to represent an Instant.
	Note: representation is incomplete until associated with a session */
    protected GbjGciReport(Instant i) {
	this();
	setup(i);
    }
 
    /** Construct a GbjGciReport configured to represent an object.
	Note: The object must be an instance of a class already covered
	by another GbjGciReport constructor.
	Also note: representation is incomplete until associated with 
	a session 
    */
    protected GbjGciReport(Object o) {
	this();
	if (o == null)                         setupNil();
	else if (o instanceof Long)            setup((Long)     o);
	else if (o instanceof String)          setup((String)o);
	else if (o instanceof Double)          setup((Double)   o);
	else if (o instanceof Float)           setup((Float)    o);
	else if (o instanceof Integer)         setup((Integer)  o);
	else if (o instanceof BigInteger)      setup((BigInteger  )   o);
	else if (o instanceof LocalDate)       setup((LocalDate)      o);
	else if (o instanceof LocalTime)       setup((LocalTime)      o);
	else if (o instanceof LocalDateTime)   setup((LocalDateTime)  o);
	else if (o instanceof ZonedDateTime)   setup((ZonedDateTime)  o);
	else if (o instanceof OffsetDateTime)  setup((OffsetDateTime) o);
	else if (o instanceof Date)            setup((Date)           o);
	else if (o instanceof Instant)         setup((Instant)        o);
	else if (o instanceof BigFraction)     setup((BigFraction)    o);
	else if (o instanceof BigDecimalFloat) setup((BigDecimalFloat) o);
	else if (o instanceof BigDecimal)      setup((BigDecimal)     o);
	else if (o instanceof Boolean)         setup((Boolean)o);
	else if (o instanceof Character)       setup((Character)o);
	else if (o instanceof Byte)            setup((Byte)o);
	else if (o instanceof Short)           setup((Short)    o);
	else if (o instanceof Calendar)         setup((Calendar)  o);
	else { throw new ClassCastException(
	                 "Can't cast object " + o +
	                 " of class " + o.getClass().getSimpleName()); };	
    }

    protected int objType_() { return objType; }

    protected void setObjType(byte v) { objType = v; }

    /** Setup client-side fields for a GbjGciReport of type null */
    protected void setupNil() {
	valueType = VALTYPE_NULL;
    }

    /** Setup client-side fields for a GbjGciReport of type boolean */
    protected void setup(boolean b) {
	valueType = VALTYPE_BOOLEAN;
	valueLong = b ? 1 : 0;
	valueCache = null;
    }

    /** Setup client-side fields for a GbjGciReport of type character */
    protected void setup(char c) {
	valueType = VALTYPE_CHARACTER;
	valueLong = c;
	valueCache = c;
    }

    /** Setup client-side fields for a GbjGciReport of type byte */
    protected void setup(byte b) {
	// use LONG field
	valueType = VALTYPE_LONG;
	valueLong = b;
	valueCache = b;
    }

    /** Setup client-side fields for a GbjGciReport of type short */
    protected void setup(short i) {
	// use LONG field
	valueType = VALTYPE_LONG;
	valueLong = i;
	valueCache = (new Short(i)).longValue();
    }

    /** Setup client-side fields for a GbjGciReport of type int */
    protected void setup(int i) {
	// use LONG field
	valueType = VALTYPE_LONG;
	valueLong = i;
	valueCache = (new Integer(i)).longValue();
    }

    /** Setup client-side fields for a GbjGciReport of type long */
    protected void setup(long i) {
	// use LONG field
	valueType = VALTYPE_LONG;
	valueLong = i;
        valueCache = null;
    }

    /** Setup client-side fields for a GbjGciReport of type BigInteger */
    protected void setup(BigInteger i) {
        if (i.bitLength() <= 60) {
	  valueType = VALTYPE_LONG; // fits in a SmallInteger
          valueLong = i.longValueExact();
          valueCache = valueLong;
          secondValue = null;
        } else {
	  valueType = VALTYPE_LARGE_INTEGER;
	  secondValue = i.toString( );
	  valueCache = i;	
        }
    }

    /** Setup client-side fields for a GbjGciReport of type float */
    protected void setup(float f) {
	// use DOUBLE field
	valueType = VALTYPE_DOUBLE;
	valueDouble = f;
	valueCache = f;	
    }

    /** Setup client-side fields for a GbjGciReport of type double */
    protected void setup(double d) {
	valueType = VALTYPE_DOUBLE;
	valueDouble = d;
	valueCache = d;	
    }

    /** Setup client-side fields for a GbjGciReport of type BigDecimal */
    protected void setup(BigDecimal d) {
	valueType = VALTYPE_SCALED_DECIMAL;
	valueCache = d;	
	valueInt1 = d.scale();
        BigInteger mant = d.unscaledValue();
        if (mant.bitLength() <- 60) {
          valueLong = mant.longValueExact();
          secondValue = null;
        } else {
	  secondValue = mant.toString(); 
        }
    }

    /** Setup client-side fields for a GbjGciReport of type BigDecimalFloat */
    protected void setup(BigDecimalFloat d) {
	valueType = VALTYPE_DECIMAL_FLOAT;
	valueCache = d;	
	secondValue = d.toString( );
    }

    /** Setup client-side fields for a GbjGciReport of type BigFraction */
    protected void setup(BigFraction f) {
        valueCache = f;
	valueType = VALTYPE_FRACTION;
        BigInteger numer = f.getNumerator();
        BigInteger denom = f.getDenominator();
        if (numer.bitLength() <= 60) {
          valueLong = numer.longValueExact();
          valuesAry = null;
        } else {
          valueLong = 0;
          valuesAry = numer.toString(); 
        }
        if (denom.bitLength() <= 60) {
          valueLong2 = denom.longValueExact();
          secondValue = null;
        } else {
          secondValue = denom.toString(); 
          valueLong2 = 0;
        }
    }

    /** Setup client-side fields for a GbjGciReport of type String */
    protected void setup(String str) {
	valueType = VALTYPE_STRING;
	valueLong = str.length();  // use the valueLong field for str length
	valueCache = str;	
    }

    /** Setup client-side fields for a GbjGciReport of type LocalTime */
    protected void setup(LocalTime t) {
	valueType = VALTYPE_TIME;
	valueLong = t.toNanoOfDay( ) / 1000; // convert to microseconds
	valueCache = t;	
    }

    /** Setup client-side fields for a GbjGciReport of type LocalDate */
    protected void setup(LocalDate d) {
	valueType = VALTYPE_DAY;
	valueInt1 = d.getYear(); // year
	valueLong2 = d.getDayOfYear(); // day in year
	valueCache = d;	
    }

    protected void setup(LocalDateTime dt) {
	valueType = VALTYPE_LOCAL_DATETIME;
        ZoneOffset zofs = ZoneOffset.ofHours(0);
	valueDouble = (double)dt.toEpochSecond(zofs);  // secs since 1970
        valueDouble += 1.0e-6 * (dt.getNano() / 1000) ;
	valueCache = dt;
    }

    protected void setup(Instant t) {
	valueType = VALTYPE_INSTANT;
	valueDouble = (double)t.getEpochSecond(); // posixSeconds
        valueDouble += 1.0e-6 * (t.getNano() / 1000) ;
	valueCache = t;
    }

    protected void setup(ZonedDateTime dt) {
	valueType = VALTYPE_ZONED_DATETIME;
	OffsetDateTime t = dt.toOffsetDateTime();
	valueLong2 = dt.getOffset().getTotalSeconds( );  // time zone offset
	valueDouble = (double)dt.toEpochSecond();  // secs since 1970
        valueDouble += 1.0e-6 * (dt.getNano() / 1000) ;
	valueCache = dt;	
    }

    protected void setup(OffsetDateTime dt) {
	valueType = VALTYPE_DATE_AND_TIME;
	valueLong2 = dt.getOffset().getTotalSeconds( );  // time zone offset
	valueDouble = (double)dt.toEpochSecond();  // secs since 1970
        valueDouble += 1.0e-6 * (dt.getNano() / 1000) ;
	valueCache = dt;	
    }

    /** Setup client-side fields for a GbjGciReport of type Date , deprecated */
    protected void setup(Date d) {
	valueType = VALTYPE_utilDATE;  // GbjUtilDate
	// only need epoch microseconds and offset
	valueLong2 = d.getTimezoneOffset( ) * 60;
        long epochMs = d.getTime();
	valueDouble = (double)epochMs / 1000.0;
	valueCache = d;
    }

    /** Setup client-side fields for a GbjGciReport of type calendar */
    protected void setup(Calendar c) {
        Calendar cc = (Calendar)c.clone();
        cc.setTimeZone(TimeZone.getTimeZone("GMT"));
        int year = cc.get(Calendar.YEAR);
        int dayOfYear = cc.get(Calendar.DAY_OF_YEAR);
        int hr = cc.get(Calendar.HOUR_OF_DAY);
        int mn = cc.get(Calendar.MINUTE);
        int sc = cc.get(Calendar.SECOND);
        int ms = cc.get(Calendar.MILLISECOND);
        long msInDay = ms + (1000 * (sc + (60 * (mn + (60 * hr)))));
        // now setup report fields
        valueType = VALTYPE_CALENDAR;
        valueInt1 = year;
        valueLong2 = dayOfYear;
        valueLong = msInDay;
	valueCache = c;	
    }


    /** Copy fields from one GbjGciReport to this one */
    protected GbjGciReport copyFrom(GbjGciReport old) {
	oop             = old.oop;
	classOop        = old.classOop;
	objType         = old.objType;
	namedSize       = old.namedSize;
	indexedSize     = old.indexedSize;
	valueType       = old.valueType;
	valueInt1       = old.valueInt1;
	valueLong       = old.valueLong;
	valueLong2      = old.valueLong2;
	valueDouble     = old.valueDouble;
	// following assumes we're OK just linking in original arrays
	secondValue     = old.secondValue;
	valuesAry      = old.valuesAry;
	valueCache      = old.valueCache;
	return this;
    }

    /** Return valueString
     */
//    protected String getValueString() {
//	return valueString;
//    }

    /** Equals comparison against another java object. 
	@return boolean result
     */
    public boolean equals(Object o) {
	if (super.equals(o)) return true;
	if (o instanceof GbjGciReport) {
	    GbjGciReport obj = (GbjGciReport)o;
	    if (objType != OBJTYPE_UNKNOWN) return (oop == obj.oop);
	    if (oop             != obj.oop)             return false;
	    if (classOop        != obj.classOop)        return false;
	    // if (classHistoryOop != obj.classHistoryOop) return false;
	    if (objType         != obj.objType)         return false;
	    if (namedSize       != obj.namedSize)       return false;
	    if (indexedSize     != obj.indexedSize)     return false;
	    // if (isCollection    != obj.isCollection)    return false;
	    // if (isString        != obj.isString)        return false;
	    // if (isDBString      != obj.isDBString)      return false;
	    if (valueType       != obj.valueType)       return false;
	    if (valueInt1       != obj.valueInt1)       return false;
	    if (valueLong       != obj.valueLong)       return false;
	    if (valueLong2      != obj.valueLong2)      return false;
	    if (valueDouble     != obj.valueDouble)     return false;
	    if (!secondValue.equals(obj.secondValue))   return false;
	    if (!valuesAry.equals(obj.valuesAry))     return false;
	    return true;
	    }
	return false;
    }

    /** Return a hash code value for this object.  Note that at this level
	we dont update the client if the server object changes.  See the
	GbjGciObject version of hashCode for the version that does the
	update.
	@return int result
     */
    public int hashCode() {
	if (objType == OBJTYPE_UNKNOWN)
	    throw new GbjGciException("hashCode on unknown GbjGciReport");
	return (int)(oop % Integer.MAX_VALUE);
    }

    /** Return true if this GS object is NIL. 
	@return boolean result
     */
    public boolean isNil() {
	return (valueType == VALTYPE_NULL);
    }


    /** Return true if this GS object is a "special".
	Specials include: null, boolean, characters, small integers, 
	small doubles. 
	@return boolean result
    */
    public boolean isSpecial() {
	return (objType == OBJTYPE_SPECIAL);
    }

    /** Return true if this GS object is a "string".
	Strings include any instance of a subclass of GS class
	CharacterCollection. 
	@return boolean result
    */
    protected boolean isString() {
	return objType == VALTYPE_STRING;
    }

    /** Return the oop of this GS object. 
	@return long result
     */
    public long getOop() { return oop; }

    /** Return the size of named instances variables 
	of this oop-based GS object. 
	@return int result
    */
    protected int getNamedSize() { return namedSize; }

    /** Return the size of the indexed component
	of this oop-based GS object. 
	@return long result
    */
    protected long getIndexedSize() { return indexedSize; }

    /** Return the total size (including both instance variables and
	indexed components) of this oop-based GS object. 
	@return long result
    */
    protected long getTotalSize() { return namedSize + indexedSize; }

    /** Generate a clone of this GS object. 
	@return Object result
     */
    public Object clone() throws CloneNotSupportedException {
	GbjGciReport res = (GbjGciReport)super.clone();
        // res.setSerialNum();
        // serialNum = serialNum + 1000000; 
        return res;
    }


    // Support method for following xxxxValue( ) methods
    
    // Methods for returning java objects from the GbjGciReport

    /** If this GS object represents a boolean, return it's value.
	@return boolean result
	@exception ClassCastException thrown on class problems
    */
    protected boolean booleanValue() {
	if (valueType != VALTYPE_BOOLEAN) {
	    throw new ClassCastException("valueType = " + valueType);
	}
	if (valueCache != null) return (Boolean)valueCache;
	valueCache = (valueLong == 1);
	return (Boolean)valueCache;
    }

    /** If this GS object represents a character, return it's value.
	@return char result
	@exception ClassCastException thrown on class problems
    */
    protected char charValue() {
	if (valueType != VALTYPE_CHARACTER) {
	    throw new ClassCastException("valueType = " + valueType);
	}
	if (valueCache != null) return (char)valueCache;
	char c  = (char)Math.toIntExact((long)valueLong);
	valueCache = c;
	return c;
    }

    /** If this GS object represents a byte, return it's value.
	@return byte result
	@exception ClassCastException thrown on class problems
    */
    protected byte byteValue() {
	// note: this will fail if out of range
	return (byte)longValue( );
    }

    /** If this GS object represents a short, return it's value.
	@return short result
	@exception ClassCastException thrown on class problems
    */
    protected short shortValue() {
	// note: this will fail if out of range
	return (short)longValue( );
    }

    /** If this GS object represents an int, return it's value.
	@return int result
	@exception ClassCastException thrown on class problems
    */
    protected int intValue() {
	// note: this will fail if out of range
	return (int)longValue( );
    }

    // BigInteger version for *really* large integers
    private BigInteger largeInt64X(byte[] theBytes) {
	long sign = 1;
	BigInteger factor = BigInteger.ONE;
	BigInteger total = BigInteger.ZERO;
	BigInteger factorMultiplier = BigInteger.valueOf(4294967296L);
	// 
	int signByte = 0;
	// if (false /*GbjGciInterface.isBigEndian*/) signByte = 3;
	if (theBytes[signByte] > 0) sign = -1;
	for (int i=4; i<theBytes.length; i += 4) {
	    if (i > 4) factor = factor.multiply(factorMultiplier);
	    long b1 = Byte.toUnsignedLong(theBytes[i]);
	    long b2 = Byte.toUnsignedLong(theBytes[i+1]);
	    long b3 = Byte.toUnsignedLong(theBytes[i+2]);
	    long b4 = Byte.toUnsignedLong(theBytes[i+3]);
	    long digit;
	    //if (false /*GbjGciInterface.isBigEndian*/)
            //   digit = (b4 + (256 * (b3 + (256 * (b2 + (256 * b1))))));
	    digit = (b1 + (256 * (b2 + (256 * (b3 + (256 * b4))))));
	    BigInteger subtotal = factor.multiply(BigInteger.valueOf(digit));
	    total = total.add(subtotal);
	}
	return total.multiply(BigInteger.valueOf(sign));
    }

    /** If this GS object represents a long, return it's value.
	If the value is too large/small, raises an ArithmeticException.
        (use bigIntegerValue( ) in that case).
	@return long result
	@exception ClassCastException ArithmeticException thrown on problems
    */
    protected long longValue() {
	if (valueType == VALTYPE_LARGE_INTEGER) {
	    return bigIntegerValue().longValueExact();
	} else if (valueType != VALTYPE_LONG) {
	    throw new ClassCastException("valueType = " + valueType);
	}
	return  valueLong;
    }

    /** If this GS object represents a LargeInteger return a BigInteger.
	@return BigInteger result
	@exception ClassCastException ArithmeticException thrown on problems
    */
    protected BigInteger bigIntegerValue() {
	if (valueType != VALTYPE_LARGE_INTEGER) {
	    throw new ClassCastException("valueType = " + valueType);
	}
	if (valueCache != null) return (BigInteger)valueCache;
	BigInteger big =  largeInt64X(valueBytes());
	valueCache = big;
	return big;
    }

    /** If this GS object represents a float, return it's value.
	@return float result
	@exception ClassCastException thrown on class problems
    */
    protected float floatValue() {
	// primary class is double
	return (float)doubleValue( );
    }

    /** If this GS object represents a double, return it's value.
	@return double result
	@exception ClassCastException thrown on class problems
    */
    protected double doubleValue() {
	if (valueType != VALTYPE_DOUBLE) {
	    throw new ClassCastException("valueType = " + valueType);
	}
	if (valueCache != null) return (double)valueCache;
	valueCache = valueDouble;
	return (double)valueCache;
    }

    /** If this GS object represents a BigDecimal, return it's value.
	@return BigDecimal result
	@exception ClassCastException thrown on class problems
    */
    protected BigDecimal bigDecimalValue() {
	if (valueType != VALTYPE_SCALED_DECIMAL) {
	    throw new ClassCastException("valueType = " + valueType);
	}
	if (valueCache != null) return (BigDecimal)valueCache;
        BigInteger mantissa;
        if (valuesAry != null) 
	  mantissa = largeInt64X(valueBytes());
        else
          mantissa = BigInteger.valueOf(valueLong);
	BigDecimal bd = new BigDecimal(mantissa, valueInt1);
	valueCache = bd;
	return bd;
    }

    /** If this GS object represents a BigDecimalFloat, return it's value.
	@return BigDecimalFloat result
	@exception ClassCastException thrown on class problems
    */
    protected BigDecimalFloat bigDecimalFloatValue() {
	if (valueType != VALTYPE_DECIMAL_FLOAT) {
	    throw new ClassCastException("valueType = " + valueType);
	}
	if (valueCache != null) return (BigDecimalFloat)valueCache;
	BigDecimalFloat dfp = new BigDecimalFloat((String)secondValue);
	valueCache = dfp;
	return dfp;
    }

    /** If this GS object represents a BigFraction, return it's value.
	@return BigFraction result
	@exception ClassCastException thrown on class problems
    */
    protected BigFraction bigFractionValue() {
	if (valueType != VALTYPE_FRACTION) {
	    throw new ClassCastException("valueType = " + valueType);
	}
	if (valueCache != null) return (BigFraction)valueCache;
        BigInteger numer;
        if (valuesAry != null) 
          numer = largeInt64X(valueBytes());
        else
          numer = BigInteger.valueOf(valueLong);
        BigInteger denom;
        if (secondValue != null) 
          denom = largeInt64X((byte[])secondValue);
        else
          denom = BigInteger.valueOf(valueLong2);
	BigFraction f =  new BigFraction(numer, denom);
	valueCache = f;
	return f;
    }

    /** If this GS object represents a string, return it's cached value.
	@return String result
	@exception ClassCastException thrown on class problems
    */
    protected String stringValue() { 
	if (valueType != VALTYPE_STRING) { 
	   throw new ClassCastException("valueType = " + valueType);
	}
	return (String)valueCache;
    }

    /** If this GS object represents a day, return it's value.
	@return LocalDate result
	@exception ClassCastException thrown on class problems
    */
    protected LocalDate localDateValue() {
	LocalDate d;
	if (valueCache != null) return (LocalDate)valueCache;
	if (valueType == VALTYPE_STRING) {
	    try {
		d = LocalDate.parse((String)valueCache);
		valueCache = d;
		return d;
	    } catch (Exception e) {
		throw new ClassCastException("valueType = " + valueType);
	    }
	}
	if (valueType != VALTYPE_DAY) {
	    throw new ClassCastException("valueType = " + valueType);
	}

	d =  LocalDate.ofYearDay(valueInt1, (int)valueLong2);
	valueCache = d;
	return d;
    }

    /** If this GS object represents a time return it's value.
	@return LocalTime result
	@exception ClassCastException thrown on class problems
    */
    protected LocalTime localTimeValue() {
	LocalTime t;
	if (valueCache != null) return (LocalTime)valueCache;
	if (valueType == VALTYPE_STRING) {
	    try {
		t = LocalTime.parse((String)valueCache);
		valueCache = t;
		return t;
	    } catch (Exception e) { 
		throw new ClassCastException("valueType = " + valueType);
	    }
	}
	if (valueType != VALTYPE_TIME) {
	    throw new ClassCastException("valueType = " + valueType);
	}

	// convert microseconds to nanoseconds
	t =  LocalTime.ofNanoOfDay(valueLong * 1000);
	valueCache = t;
	return t;
    }

    /** If this GS object represents a LocalDateTime return it's value.
	@return LocalDateTime result
	@exception ClassCastException thrown on class problems
    */
    protected LocalDateTime localDateTimeValue( ) {
	if (valueType != VALTYPE_LOCAL_DATETIME) {
	    throw new ClassCastException("valueType = " + valueType);
	}
	if (valueCache != null)
	    return (LocalDateTime)valueCache;
        double posixSecs = valueDouble;
        long intSecs = (long)posixSecs ; 
	int nanosec =  (int)((posixSecs - (double)intSecs) * 1.0e9) ;
	LocalDateTime ldt =  LocalDateTime.ofEpochSecond(intSecs, nanosec, ZoneOffset.ofHours(0));
	valueCache = ldt;
	return ldt;
    }

    /** If this GS object represents a ZonedDateTime return it's value.
	@return ZonedDateTime result
	@exception ClassCastException thrown on class problems
    */
    protected ZonedDateTime zonedDateTimeValue( ) {
	if (valueType != VALTYPE_ZONED_DATETIME) {     
	    throw new ClassCastException("valueType = " + valueType);
	}
	if (valueCache != null) return (ZonedDateTime)valueCache;
	ZoneOffset zoff = ZoneOffset.ofTotalSeconds((int)valueLong2);
	long eMilliSec = (long)(valueDouble * 1000);  // epoch milliseconds
	Instant i = Instant.ofEpochMilli(eMilliSec);
	OffsetDateTime odt = i.atOffset(zoff);
        ZonedDateTime zt = odt.toZonedDateTime();
        valueCache = zt;
	return zt;
    }

    /** If this GS object represents a OffsetDateTime return it's value.
	@return OffsetDateTime result
	@exception ClassCastException thrown on class problems
    */
    protected OffsetDateTime offsetDateTimeValue( ) {
	if (valueType != VALTYPE_DATE_AND_TIME) {     
	    throw new ClassCastException("valueType = " + valueType);
	}
	if (valueCache != null) return (OffsetDateTime)valueCache;
	OffsetDateTime odt;
	ZoneOffset zoff = ZoneOffset.ofTotalSeconds((int)valueLong2);
	long eMilliSec = (long)(valueDouble * 1000);  // epoch milliseconds
	Instant i = Instant.ofEpochMilli(eMilliSec);
	odt = i.atOffset(zoff);
	valueCache = odt;
	return odt;
    }

    /** If this GS object represents a calendar, return it's value.
        @return Calendar result
        @exception ClassCastException  thrown on class problems
    */
    protected Calendar calendarValue() {
      if (valueType != VALTYPE_CALENDAR) {
        throw new ClassCastException("valueType = " + valueType);
      }
      if (valueCache != null) return (Calendar)valueCache;
      int year = valueInt1;
      int dayOfYear = (int)valueLong2;
      long ms = valueLong;
      int hr = (int)(ms / 3600000);
      ms -= hr * 3600000L;
      int mn = (int)(ms / 60000);
      ms -= mn * 60000L;
      int sc = (int)(ms / 1000);
      ms -= sc * 1000L;
      Calendar c = Calendar.getInstance(TimeZone.getTimeZone("GMT"));
      c.set(Calendar.YEAR, year);
      c.set(Calendar.DAY_OF_YEAR, dayOfYear);
      c.set(Calendar.HOUR_OF_DAY, hr);
      c.set(Calendar.MINUTE, mn);
      c.set(Calendar.SECOND, sc);
      c.set(Calendar.MILLISECOND, (int)ms);
      valueCache = c;
      return c;
    }


    private boolean canBeInstant() {
      return valueType == VALTYPE_ZONED_DATETIME ||
           valueType == VALTYPE_DATE_AND_TIME ||
           valueType == VALTYPE_utilDATE ||
           valueType == VALTYPE_CALENDAR ||
           valueType == VALTYPE_LOCAL_DATETIME ;
    }

    /** If this GS object represents an Instant return it's value.
	@return Instant result
	@exception ClassCastException thrown on class problems
    */
    protected Instant instantValue() {
        if (valueType == VALTYPE_INSTANT)
          return (Instant) valueCache;
	if (! canBeInstant()) {
	    throw new ClassCastException("valueType = " + valueType);
	}
	long millisec = (long)(valueDouble * 1000);
	return Instant.ofEpochMilli(millisec);
    }
		       
    /** If this GS object represents a date, return it's value.
	@return Date result
	@exception ClassCastException thrown on class problems
    **/
    protected Date dateValue() {
        if (valueType == VALTYPE_utilDATE)
          return (Date) valueCache;
	if (! canBeInstant()) {
	    throw new ClassCastException("valueType = " + valueType);
	}
	long millisec = (long)(valueDouble * 1000);
	return new Date(millisec);
    }

    /** If this GS object is byte-based, return it's cached value.
	Note that this version only returns the contents of the
	valueBytes cache field, which may not be complete or up to date.  
	See GbjGciObject for a version that can return the correct value.
	@return byte array result
	@exception ClassCastException thrown on class problems
    */
    protected byte[] bytesValue() {
	if (valueType != VALTYPE_BYTES) {
	    throw new ClassCastException("valueType = " + valueType);
	}
	return valueBytes();
    }

    /** If this GS object is oop-based, return it's cached value.
	Note that this version only returns the contents of the
	valueOops cache field, which may not be complete or up to date.  
	See GbjGciObject for a version that can return the correct value.
	@return byte array result
	@exception ClassCastException thrown on class problems
    */
    protected long[] oopsValue() {
	if (valueType != VALTYPE_OOPS) {
	    throw new ClassCastException("valueType = " + valueType);
	}
	return valueOops();
    }

    /** Return the java cached value according to the valueType of this GbjGciReport.
        @return Object
        @exception ClassCastException ArithmeticException
    */
    public Object javaValue( ) {
      switch ( valueType ) {
	case VALTYPE_UNKNOWN:         return null;
	case VALTYPE_NULL:            return null;  
	case VALTYPE_BOOLEAN:         return booleanValue( );
	case VALTYPE_CHARACTER:       return charValue( );
	case VALTYPE_LONG:            return longValue( );
	case VALTYPE_LARGE_INTEGER:   return bigIntegerValue( );
	case VALTYPE_DOUBLE:          return doubleValue( );
	case VALTYPE_SCALED_DECIMAL:  return bigDecimalValue( );
	case VALTYPE_DECIMAL_FLOAT:   return bigDecimalFloatValue( );
	case VALTYPE_FRACTION:        return bigFractionValue( );
	case VALTYPE_STRING:          return stringValue( );
	case VALTYPE_DAY:             return localDateValue( );
	case VALTYPE_TIME:            return localTimeValue( );
	case VALTYPE_LOCAL_DATETIME:  return localDateTimeValue( );
	case VALTYPE_ZONED_DATETIME:   return zonedDateTimeValue( );
	case VALTYPE_DATE_AND_TIME:   return offsetDateTimeValue( );
	case VALTYPE_BYTES:           return valuesAry;
	case VALTYPE_OOPS:            return valuesAry;
        case VALTYPE_CALENDAR:        return calendarValue();
      }
      return null;
    }

    // Support methods for printing out data

    private String printLocalDateTime(LocalDateTime dt) {
	DateTimeFormatter dtf = DateTimeFormatter.ofPattern( "MM/dd/YYY HH:mm:ss");
	return dt.format(dtf);
    }
    
    private String printOffsetDateTime(OffsetDateTime dt) {
	DateTimeFormatter dtf = DateTimeFormatter.ofPattern( "MM/dd/YYY HH:mm:ss.SSS O");
	return dt.format(dtf);
    }
    private String printZonedDateTime(ZonedDateTime dt) {
	DateTimeFormatter dtf = DateTimeFormatter.ofPattern( "MM/dd/YYY HH:mm:ss.SSS z");
	return dt.format(dtf);
    }

    // returns objtype identifier
    protected String objTypeString( ) {
	String ids[] = {"UNKNOWN" , "OOP" , "BYTE" , "NSC" , "SPECIAL"};
	return ids[objType+1];  // because unknown = -1
    }

    // returns valtype identifier
    private String valTypeString() {
	String ids[] = {
	    "UNKNOWN" ,
	    "NULL" ,
	    "BOOLEAN" ,
	    "CHARACTER" ,
	    "LONG" ,
	    "DOUBLE" ,
	    "BYTES" ,
	    "OOPS" ,
	    "STRING" ,
	    "DBSTRING[OBSOLETE]" ,
	    "LOCAL_DATETIME" ,
	    "DATETIME" ,
	    "DATE_AND_TIME" ,
	    "DAY" ,
	    "TIME" ,
	    "LARGE_INTEGER" ,
	    "SCALED_DECIMAL" ,
	    "DECIMAL_FLOAT" ,
	    "FRACTION" 
	};
	return ids[valueType+1];   // because unknown = -1
    }

    /** Write to printstream a detailed multi-line string describing
	the internal structure of this GbjGciReport.  
	Intended for diagnostic purposes.
	@param out  PrintStream to write to
    */
    public void debugPrint(PrintStream out) {
	long max;
	out.println(getClass().getSimpleName() + ":");
	out.println("  oop:     " + Long.toUnsignedString(oop));
	out.println("  objtype: " + objType   + " / " + objTypeString());
	out.println("  valtype: " + valueType + " / " + valTypeString());
	out.println("  class:   " + classOop);
	// out.println("  clshst:  " + classHistoryOop);
	// out.println("  isClc:   " + isCollection);
	// out.println("  isStr:   " + isString);
	// out.println("  isDBStr: " + isDBString);
	out.println("  nsize:   " + namedSize);
	out.println("  xsize:   " + indexedSize);
	out.println("  vints:   " + valueInt1 );
	out.println("  vlongs:  " + valueLong + " " + valueLong2);
	out.println("  vdouble: " + valueDouble);
	out.println("  valuesAry:    \"" + valuesAry + "\"");
	out.println("  secondValue:    \"" + secondValue + "\"");
  	out.println("  valueCache:  " + valueCache);
	if (valueCache != null)
	    out.println("          (" + valueCache.getClass() + ")");
  	out.print(  "  value:   ");
	switch (valueType) {
	case VALTYPE_NULL:      out.println("null");           break;
	case VALTYPE_BOOLEAN:   out.println(booleanValue());   break;
	case VALTYPE_CHARACTER: out.print("char: ");
	                        out.println(charValue());      break;
	case VALTYPE_LONG:      out.print("long: ");
	                        out.println(valueLong);        break;
	case VALTYPE_LARGE_INTEGER:
	                        out.print("lint: ");
	                        out.println(bigIntegerValue()); break;
	case VALTYPE_DOUBLE:    out.print("db: ");
	                        out.println(valueDouble);       break;
	case VALTYPE_LOCAL_DATETIME:
	                        out.println();
				out.print("lcdv: ");
				out.println(localDateTimeValue());
				out.println(printLocalDateTime(
				       localDateTimeValue()));
	 			                                break;
	case VALTYPE_ZONED_DATETIME:
	                        out.println();
				out.print("zdtv: ");
				out.println(zonedDateTimeValue());
				out.println(printZonedDateTime(
				       zonedDateTimeValue()));
				break;
	case VALTYPE_DATE_AND_TIME:
	                        out.println();
				out.print("odtv: ");
				out.println(offsetDateTimeValue());
				out.println(printOffsetDateTime(
					offsetDateTimeValue()));
				break;
	case VALTYPE_DAY:       out.print("day: ");
	                        out.println(localDateValue());      break;
	case VALTYPE_TIME:      out.print("time: ");
	                        out.println(localTimeValue());      break;

		    
	case VALTYPE_SCALED_DECIMAL:
	               out.print("sd: ");
	               out.println(bigDecimalValue());      break;
	case VALTYPE_FRACTION:  out.print("frc: ");
	                        out.println(bigFractionValue());  break;

	case VALTYPE_DECIMAL_FLOAT:
   	                        out.print("df: ");
				out.println(bigDecimalFloatValue());
				break;

	case VALTYPE_STRING:    out.print("str: ");
				out.println( (String)valueCache);
				break;

	case VALTYPE_BYTES:     // handled below
	                        break;
	default:
	    out.println("[unknown]");
	}
	switch (objType) {
	case OBJTYPE_BYTE:
	    out.print("  bytes: ");
	    if (valueBytes() == null) {
		out.print("[null]");
	    } else {
		max = Math.min(valueBytes().length, 100);
		for (int i=0; i<max; i++) {
		    int b = valueBytes()[i];
		    if (b < 0) b += 256;
		    out.print(" " + Integer.toHexString(b));
		}
		if (valueBytes().length > 100) out.print("...");
	    }
	    out.println("");
	    break;

	case OBJTYPE_OOP:
	    out.print("  oops: ");
	    max = Math.min(valueOops().length, 10);
	    for (int i=0; i<max; i++)
		System.out.print(" " + valueOops()[i]);
	    if (valueOops().length > 10) out.print("...");
	    out.println("");
	    break;
	}
    }

    /**
     * Method to add components to toString() strings.
     * Override in superclasses, first entry should be
     * call to super.toStringAdditions(). Example:
     * <pre>
     * protected void toStringAdditions(StringBuffer sb) {
     *    super.toStringAdditions(sb);
     *    sb.append("My changes here..");
     * }
     * </pre>
     */
    protected void toStringAdditions(StringBuffer sb) {
	sb.append(getClass().getSimpleName());
	sb.append("(oop:");
	sb.append(Long.toUnsignedString(getOop()));
	sb.append("|");
	switch (valueType) {
	case VALTYPE_NULL:      sb.append("null");                 break;
	case VALTYPE_BOOLEAN:   sb.append(booleanValue());         break;
	case VALTYPE_CHARACTER: sb.append("'");
	                        sb.append(charValue());
				sb.append("'");                    break;
	case VALTYPE_LONG:      sb.append(valueLong);              break;
	case VALTYPE_DOUBLE:    sb.append(valueDouble);            break;
	case VALTYPE_STRING:   {
            if (valueCache == null) {
              sb.append("(valueCache==null)"); 
            } else {
              sb.append("\""); 
              String s = (String)valueCache ;
              int len = s.length();
              if (len > 40) len = 40;
	      sb.append( s.substring(0, len));
              sb.append("\"");                   
            }
           }
           break;

	case VALTYPE_LOCAL_DATETIME:
	    sb.append("LocalDateTime :" +
		      printLocalDateTime(localDateTimeValue()));
	                                                           break;
	case VALTYPE_ZONED_DATETIME:
	    sb.append("ZonedDateTime :" +
		      printZonedDateTime(zonedDateTimeValue()));
	                                                           break;
	case VALTYPE_DATE_AND_TIME:
	    sb.append("OffsetDateTime:" +
		      printOffsetDateTime(offsetDateTimeValue()));
	                                                           break;
	case VALTYPE_DAY:       sb.append("day:" +
					  localDateValue().toString());  break;
	case VALTYPE_TIME:      sb.append("time:" +
				          localTimeValue().toString()); break;
	case VALTYPE_LARGE_INTEGER:
	                        sb.append("BigInteger:" +
					  bigIntegerValue().toString());
					                           break;
	case VALTYPE_SCALED_DECIMAL:
	    sb.append("ScaledDecimal:" + bigDecimalValue().toString());    break;
	case VALTYPE_FRACTION:  
	    sb.append("Fraction:" + bigFractionValue().toString());        break;
	case VALTYPE_DECIMAL_FLOAT:
	                        sb.append("DecimalFloat:");
				sb.append(bigDecimalFloatValue());    break;

	case VALTYPE_BYTES:       sb.append("bytes");    break;
	case VALTYPE_OOPS:        sb.append("oops");   break;

	}
	sb.append(")");
    }


    /** Return a short string representing this GbjGciReport 
	@return String result
     */
    public String toString() {
	// note: toString() must not attempt to re-initialize
	// cached values using checkValue(), as this can result 
	// in infinite recursion
	StringBuffer sb = new StringBuffer();
        toStringAdditions(sb);
	return sb.toString();
    }


    /** Returns an integer representing the type of object
	as represented on the GS server.
	<p>
        Possible values include:
    <pre>
     int  Constant
     ---  ---------------
     -1:  OJTYPE_UNKNOWN 
      0:  OBJTYPE_OOP    
      1:  OBJTYPE_BYTE   
      2:  OBJTYPE_NSC    
      3:  OBJTYPE_SPECIAL
    </pre>
    @return int result type
    */
    public int getObjType() { return objType; }

    /** Returns an integer representing the type of java object
	cached on the client as representative of the GS server object.
        This integer can be useful in determining which xValue() method to
        use to retrieve the java object.
	<p>
	Possible values include:
    <pre>
     int  Constant         Value retrieval methods
     ---  ---------------- --------------------------------------------------
     -1:  VALTYPE_UNKNOWN   
      0:  VALTYPE_NULL      null
      1:  VALTYPE_BOOLEAN   booleanValue()
      2:  VALTYPE_CHARACTER charValue()
      3:  VALTYPE_LONG      byteValue(), shortValue(), intValue(), longValue()
      4:  VALTYPE_DOUBLE    floatValue(), doubleValue()
      5:  VALTYPE_BYTES     getBytes()
      6:  VALTYPE_OOPS      getNamedOops(), getOops()
      7:  VALTYPE_STRING    stringValue()
      
      9:  VALTYPE_LOCAL_DATETIME  localDateTimeValue()
     10:  VALTYPE_ZONED_DATETIME  dateTimeValue() yyyy
     11:  VALTYPE_DATE_AND_TIME   dateAndTimeValue()
     12:  VALTYPE_DAY             dayValue()
     13:  VALTYPE_TIME            timeValue()
     14:  VALTYPE_LARGE_INTEGER   bigIntegerValue()
     15:  VALTYPE_SCALED_DECIMAL  bigDecimalValue()
     16:  VALTYPE_DECIMAL_FLOAT   bigDecimalFloatValue()
     17:  VALTYPE_FRACTION        bigFractionValue()
    </pre>
    Note that most of the retrieval methods can also be used on objects
    with object type VALTYPE_STRING, if the associated string happens to 
    be formatted correctly for the given type.
    @return int result
    */
    public int getValType() { return valueType; }

    /** Return the name of the GS class primarily associated with this valueType
        @return String valueType
    */
    public String getValTypeString( ) {
      switch ( valueType ) {
      case VALTYPE_UNKNOWN:         return "Unknown";
      case VALTYPE_NULL:            return "Null";
      case VALTYPE_BOOLEAN:         return "Boolean";
      case VALTYPE_CHARACTER:       return "Character";
      case VALTYPE_STRING:          return "String";
      case VALTYPE_LONG:            return "Long";
      case VALTYPE_LARGE_INTEGER:   return "LargeInteger";
      case VALTYPE_DOUBLE:          return "Double";
      case VALTYPE_DECIMAL_FLOAT:   return "DecimalFloat";
      case VALTYPE_SCALED_DECIMAL:  return "ScaledDecimal";
      case VALTYPE_FRACTION:        return "Fraction";
      case VALTYPE_DAY:             return "Day";
      case VALTYPE_TIME:            return "Time";
      case VALTYPE_LOCAL_DATETIME:  return "GbjLocalDateTime";
      case VALTYPE_ZONED_DATETIME:  return "GbjZonedDateTime";
      case VALTYPE_DATE_AND_TIME:   return "DateAndTime";
      case VALTYPE_BYTES:           return "Bytes";
      case VALTYPE_OOPS:            return "Oops";
      }
      return null;
    }

    /** Return the GS class primarily associated with this valueType
	@return Class name
    */
    public Class javaClass( ) {
        try {
	switch ( valueType ) {
	case VALTYPE_UNKNOWN:         return null;
	case VALTYPE_NULL:            return null;
	case VALTYPE_BOOLEAN:         return Class.forName("Boolean");
	case VALTYPE_CHARACTER:       return Class.forName("Character");
	case VALTYPE_STRING:          return Class.forName("String");
	case VALTYPE_LONG:            return Class.forName("Long");
	case VALTYPE_LARGE_INTEGER:   return Class.forName("BigInteger");
	case VALTYPE_DOUBLE:          return Class.forName("Double");
	case VALTYPE_DECIMAL_FLOAT:   return Class.forName("BigDecimalFloat");
	case VALTYPE_SCALED_DECIMAL:  return Class.forName("BigDecimal");
	case VALTYPE_FRACTION:        return Class.forName("BigFraction");
	case VALTYPE_DAY:             return Class.forName("LocalDate");
	case VALTYPE_TIME:            return Class.forName("LocalTime");
	case VALTYPE_LOCAL_DATETIME:  return Class.forName("LocalDateTime");
	case VALTYPE_ZONED_DATETIME:  return Class.forName("ZonedDateTime");
	case VALTYPE_DATE_AND_TIME:   return Class.forName("OffsetDateTime");
	case VALTYPE_BYTES:           return null;
	case VALTYPE_OOPS:            return null;
	}
	} catch (ClassNotFoundException e) { return null; }
	return null;
    }

}


