/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *========================================================================
 */

package com.gemstone.gbjgci;

import java.util.HashMap;
import java.lang.ref.WeakReference;
import java.lang.ref.ReferenceQueue;
import java.util.logging.*;
import java.time.LocalDate;
import java.time.LocalTime;
import java.time.LocalDateTime;
import static com.gemstone.gbjgci.GbjGciInterface.*;
import static com.gemstone.gbjgci.GbjGciObject.*;
import com.gemstone.gbj.GbjSession;
import com.gemstone.gbj.GbjException;

/**
GbjGciSessions represent GemStone sessions.
<p>
Much of the core functionality of the GBJ class GbjSession has been moved
to this class, providing sufficient functionality for a "GBJ lite".

@version 3.0

*/

public class GbjGciSession 
{

    /** Enables log.fine* messages */
    public static boolean isDEBUG = System.getenv("GBJ_DEBUG") != null ; 

    /** Logger for debug trace messages */
    private static Logger log;

    // static code for setting up logger
    static {
	log = GbjGciInterface.setupLogger(GbjGciSession.class);
    }

    /** GciTs session pointer 
      0 indicates session is not logged in  */
    private long sessionPtr;

    private static int serial_num = 0;
    private int serialNum;

    /** Flag used to control if changed objects should be monitored */
    public boolean monitorChangedObjects = false;

    private HashMap<Long, WeakReference<GbjGciObject>> objectTable;

    // Maps the oop of a GemStone class to a particular GbjGciObject subclass  
    private HashMap<Long, GbjGciObject> gsToJavaClassMap;

    // Structures to log batches of oops to remove from the export set

    // How many oops to batch before we bother to call GciReleaseOops
    private static final int EXPORT_SET_BATCH_SIZE = 500;
    private long[] oopsToRemoveFromExportSet;
    private int removeFromExportSetCount;
    private int setupObjEntryCount;

    private GbjGciException  nextErrObj;

    private GbjGciObject  GbjGciInterface_class;

    // support methods for objectTable management
    private void objectTableInitialize() {
	objectTable = new HashMap<Long, WeakReference<GbjGciObject>>();
    }

    private GbjGciObject objectTableGet(GbjGciObject obj) {
	return objectTableGet(obj.getOop());
    }

    private GbjGciObject objectTableGet(long oopValue) {
        // boolean isDEBUG = GbjGciSession.isDEBUG;
	// if (isDEBUG) log.finest("TableGet for " + oopValue);
	WeakReference<GbjGciObject> ref = objectTable.get(oopValue);
	if (ref == null) {
	    // if (isDEBUG) log.finest("      no entry");
	    return null;
	}
	GbjGciObject obj = ref.get();
        //if (isDEBUG) {
	//  if (obj == null) log.finest("      null entry");
	//  else             log.finest("      entry found");
        //}
	return obj;
    }

    private void objectTablePut(GbjGciObject obj, boolean overWrite) {
	long oopValue = obj.getOop();
        boolean isDEBUG = GbjGciSession.isDEBUG;
	if (isDEBUG) log.finest("TablePut for " + oopValue);
	if (oopValue == 0) 
	    throw new GbjGciException("Object table put on oop zero");
	WeakReference<GbjGciObject> ref;
	ref = objectTable.get(oopValue);
	if (ref != null) { 
          Object oldVal = ref.get();
	  if (oldVal != null) {
            if (overWrite) {
              // ok
              // System.out.println("objectTablePut OVERWRITE oop " + oopValue);
              // waitForDebug();
            } else {
	      throw new GbjGciException( "Object table already contains entry");
            }
          }
	}
	ref = new WeakReference<GbjGciObject>(obj);
	objectTable.put(oopValue, ref);
    }

    private void objectTablePut(GbjGciObject obj) {
	objectTablePut(obj, false);
    }

    private native int releaseOops(long sess, long[] objIds, int numOops);

    /** Release an un-used object from the weak reference used in
	the session's object table */
    protected synchronized void releaseObject(GbjGciObject obj) {
	// remove from object table (if it's recorded there)
	// Note: if called from obj.finalize(), a new version
	// of the obj might have already been loaded into the
	// table, so leave that one there..

        if (obj.objType_() == OBJTYPE_SPECIAL)
          return; // special objs never in object table nor export set

	long oop = obj.getOop();
	GbjGciObject tableObj = objectTableGet(oop);
	if (tableObj == obj /* explicit remove from app*/ ) { 
	   objectTable.put(oop, null);
        } else { 
	  // tableObj == null    was removed by Java finalize 
        }
	if (oop == 0) return;  // fix 42081

	oopsToRemoveFromExportSet[removeFromExportSetCount] = oop;
	if (++removeFromExportSetCount >= EXPORT_SET_BATCH_SIZE) {
	  releaseOops(sessionPtr, oopsToRemoveFromExportSet, removeFromExportSetCount);
	  removeFromExportSetCount = 0;
	}
    }


    private static synchronized int nextSerialNum() {
        serial_num = serial_num + 1 ;
        return serial_num;
    }

    // internal methods
    private void initialize() {
	sessionPtr = 0;
        setupObjEntryCount = 0;
	objectTableInitialize();
	classMapInitialize();
	oopsToRemoveFromExportSet = new long[EXPORT_SET_BATCH_SIZE];
	removeFromExportSetCount = 0;
        nextErrObj = new GbjGciException();
        GbjGciInterface_class = null;
        serialNum = nextSerialNum();;
    }

    /** Add/replace an entry in the object table */
    protected synchronized void updateObjectTable(GbjGciObject obj) {
	objectTablePut(obj);
    }

    // Methods for managing the GS-to-Java Class mapping functions

    private void classMapInitialize() {
	gsToJavaClassMap = new HashMap<Long, GbjGciObject>();
    }

    /** Map instances of a GemStone class to a designated java class,
	represented by an example instance.
	Defines a mapping such that any instances of the GemStone class 
	(and its subclasses) that are returned to java, will be represented
	as instances of the desired java class (which must be a subclass 
	of GbjGciObject).  The java class is indicated by an example 
	instance which is used as a template.
	@param gsClassName  Name of the GemStone class
	@param javaExample  An instance of the java class
    */
    public synchronized  void mapGsToJavaClass(String gsClassName, 
				 GbjGciObject javaExample) {
	mapGsToJavaClass(objectNamed(gsClassName), javaExample);
    }

    private GbjGciObject getGciInterfaceServerClass() {
      if (GbjGciInterface_class == null)
        GbjGciInterface_class = objectNamed("GbjGciInterface");
      return GbjGciInterface_class;
    } 

    /** Map instances of a GemStone class to a designated java class,
	represented by an example instance.
	Defines a mapping such that any instances of the GemStone class 
	(and its subclasses) that are returned to java, will be represented
	as instances of the desired java class (which must be a subclass 
	of GbjGciObject).  The java class is indicated by an example 
	instance which is used as a template.
	@param gsClass      GS object of the GemStone class
	@param javaExample  An instance of the java class
    */
    public synchronized void mapGsToJavaClass(GbjGciObject gsClass, GbjGciObject javaExample) {
        boolean isDEBUG = GbjGciSession.isDEBUG;
	if (isDEBUG) {
          System.out.println("Mapping: " + gsClass + " to:" + javaExample.getClass());
        }
	GbjGciObject args[] = new GbjGciObject[1];
	args[0] = gsClass;
        // allSubclassOopsFor:  returns Array of SmallIntegers containing oops ...
        GbjGciObject rcvr = getGciInterfaceServerClass();
	GbjGciObject subclassOops = perform(rcvr, "allSubclassOopsFor:", args, 1);
	GbjGciObject[] subclassOopArray = getObjs(subclassOops, 1, (int)subclassOops.indexedSize);
	if (isDEBUG) log.finest("mapGsToJavaClass:  " + subclassOops.indexedSize + " oops to transfer");
	for (int i=0; i< subclassOopArray.length; i++) {
            GbjGciObject anObj = subclassOopArray[i];
            long key = anObj.longValue();  // anObj is a SmallInteger (Java long)
	    if (isDEBUG) {
              String clsName = execute("(Object objectForOop: " + key + ") name").stringValue();
              System.out.println("   gsToJavaClassMap.put key " + key + " (" + clsName + ") -> " + javaExample );
            }
	    gsToJavaClassMap.put(key, javaExample);
	}
        releaseObject(subclassOops);
    }

    public static void waitForDebug() {
      return; // place to set jdb breakpoint
    }

    //protected int waitForDebug() {
    //  GbjGciSession.waitForDebug();
    //  sess.execute("System _sleep: 1");
    //  return 1;
    //}

    // check that the GbjGciObject is consistant for this session
    private void checkObject(GbjGciObject obj) {
      if (obj.objType_() == OBJTYPE_SPECIAL)
        return;   // special objects never in object table
      GbjGciObject otVal = objectTableGet(obj);
      if (otVal != obj) {
        if (otVal != null) {
          System.out.println("Defunct stub otVal oop " + otVal.getOop() + " obj " + obj.getOop());
        }
        waitForDebug();
	throw new GbjGciException("Defunct Stub");
      }
    }

    // notify objects which have been changed
    private static final int NOTIFY_CHANGE_BATCH_SIZE = 200;
    private long[] changedObjectBuffer;

    private native int changedOops(long sess, long[] buf, int bufSize);

    private synchronized void notifyChangedObjects() {

	// skip if we're not monitoring changed objects
	if (!monitorChangedObjects) return;

	// Notify observers of objects that have changed..
        boolean isDEBUG = GbjGciSession.isDEBUG;
	if (isDEBUG) log.finer("Changed oops: ");
	changedObjectBuffer = new long[NOTIFY_CHANGE_BATCH_SIZE];
	int count;
	do {
	    count = changedOops(sessionPtr, changedObjectBuffer, NOTIFY_CHANGE_BATCH_SIZE);
	    for (int i=0; i<count; i++) {
		if (isDEBUG) log.finer("  " + changedObjectBuffer[i]);
		GbjGciObject obj = objectTableGet(changedObjectBuffer[i]);
		if (obj == null)
		    throw new GbjGciException("notifyChangedObjects(): logic error");
		obj.notifyChanged();
	    }
	} while (count > 0);
    }


    // public methods

    /** Create an instance of GbjGciSession */
    public GbjGciSession() {
	initialize();
    }

    private GbjGciObject remapObject(GbjGciObject obj) {
        if (obj.valueType == VALTYPE_OOPS) {
	  long oop = obj.getOop();
	  GbjGciObject template  = gsToJavaClassMap.get(obj.classOop);
	  if (template != null && template.getClass() != obj.getClass()) {
	    GbjGciObject newObj;
	    try {
	      newObj = (GbjGciObject)template.clone();
	    } catch (CloneNotSupportedException ex) {
	      throw new GbjGciException(ex.getMessage());
	    }
	    newObj.copyFrom(obj);
	    obj = newObj;  // obj is now a GbjCollection
	  }
        }
	objectTablePut(obj, true);
	return obj;
    }

    private synchronized GbjGciObject addToObjTable(GbjGciObject result) {
	// first, check and see if we already have this guy in the objectTable
        int objType = result.objType_();
        if (objType == OBJTYPE_SPECIAL) {  // special objects are not in obj table
          return result;
        }
	GbjGciObject  obj = objectTableGet(result);
	if (obj == null) {
	    // remap class of result if needed
	    // this may generate a new result object
	    result = remapObject(result);
	    obj = result;
	}
	// in case the session hasn't been set..
	if (obj.getSession() == null) obj.setSession(this);

	obj.updateClient();

        boolean isDEBUG = GbjGciSession.isDEBUG;
	if (isDEBUG) log.finest("addToObjTable finish: " + obj );
	return obj;
    }

    private native int setupReport(long sess, GbjGciObject obj, GbjGciException errObj);

    /** Complete setup of an object newly returned from the JNI layer */
    protected synchronized GbjGciObject setupObj(GbjGciObject result) {
        boolean isDEBUG = GbjGciSession.isDEBUG;
        setupObjEntryCount += 1;
        if (setupObjEntryCount > 1) {
          throw new GbjException("recursion in setupObj");
        }
        GbjGciObject obj;
        try {
	  if (result.objType_() == OBJTYPE_UNKNOWN || result.oop == 0) {
            GbjGciException errObj = nextErrObj;
	    if (setupReport(sessionPtr, result, errObj) != 0) {
              nextErrObj = new GbjGciException();
              throw new GbjException((GbjSession)this, errObj);
	    }
          }
	  obj = addToObjTable(result);
	  if (isDEBUG) log.finest("setupObj finish: " + obj);
        } finally {
          setupObjEntryCount -= 1;
        }
	return obj;
    }


/**
Login a GS session using this GbjGciSession.
<p>
@param serverName   The name of the GS stone process.
@param gemnetName   The name of the Gem Service.  If null, the default "gemnetobject" is used.
@param userName     The GS username for the session.
@param password     The GS password for this username.
@param hostUserName  The OS username for this session (if needed).
@param hostPassword  The OS password for this OS username (if needed).
@exception com.gemstone.gbj.GbjException thrown on any error
*/

    public synchronized void login(String serverName,
		      String gemnetName,
		      String userName,
		      String password,
		      String hostUserName,
		      String hostPassword) {

	if (sessionPtr != 0)
	    throw new GbjException("Session already logged in");
	sessionPtr = GbjGciInterface.login(serverName, gemnetName,
					  userName, password,
					  hostUserName, hostPassword);
    }


    private native int logout(long sess);
 
/** Logout the GS session associated with this GbjGciSession
@exception com.gemstone.gbj.GbjException thrown on any error
*/
    public synchronized void logout() {
	logout(sessionPtr);
	initialize();
    }

    private GbjGciObject newGbjGciObject() {
      try {
        Class cls = GbjSession.gbjObjClass();
        GbjGciObject result = (GbjGciObject)cls.newInstance();
        result.setSession(this);
        return result;
      } catch (Exception e) {
        System.out.println(GbjSession.stackTraceString(e));
        if (GbjGciSession.isDEBUG)  log.fine("newObject exception: " + e.getMessage());
        throw new GbjException("Problem generating Object");
      }
    }

  
    private native int execute(long sessPtr, String execStr, GbjGciObject result, GbjGciException errObj);

/**
Execute a GS/S Smalltalk String (represented as a java String)
@param execStr     The Smalltalk string to execute.
@return a GS object representing the results of execution.
@exception com.gemstone.gbj.GbjException thrown on any error
*/
    public synchronized GbjGciObject execute(String execStr) {
	GbjGciObject result = newGbjGciObject();
        GbjGciException errObj = nextErrObj;
        if (execute(sessionPtr, execStr, result, nextErrObj) != 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)this, errObj);
        }
	notifyChangedObjects();
	return setupObj(result); 
    }

 private native int serverContinue(long sess, GbjGciObject gsProcess, boolean errFlag, 
       GbjGciObject stResult, GbjGciObject javaRes, String errMsg, String errStack, GbjGciException errObj);

/**
Return control back to the server after handling a client forwarder error.
<p>
This method should only be used by GbjSession.
*/
 public synchronized GbjGciObject serverContinue(GbjGciObject gsProcess, boolean errFlag, 
       GbjGciObject stResult, GbjGciObject javaRes, String errMsg, String errStack) {
    GbjGciException errObj = nextErrObj;
    if (serverContinue(sessionPtr, gsProcess, errFlag, stResult, javaRes, 
                        errMsg, errStack, errObj) != 0) {
       nextErrObj = new GbjGciException();
       throw new GbjException((GbjSession)this, errObj);
    }
    return setupObj(javaRes);
 }

/**
Execute a GS/S Smalltalk String (represented as a GS Object)
@param execStrObj  The Smalltalk string to execute.
@return a GS object representing the results of execution.
@exception com.gemstone.gbj.GbjException thrown on any error
*/
    public GbjGciObject execute(GbjGciObject execStrObj) {
       return execute( execStrObj.stringValue());
    }

   private native int perform(long sessPtr, GbjGciObject receiver, String selector, GbjGciObject args[], int numArgs,
                GbjGciObject result, GbjGciException errObj);

/**
Execute a Smalltalk method against a designated receiver using these arguments.
@param receiver    The GS object to execute this method against.
@param selector    The method to execute.
@param args        The list of arguments.
@param numArgs     The number of arguments in the list.
@return a GS object representing the results of execution.
@exception com.gemstone.gbj.GbjException thrown on any error
*/
    public synchronized GbjGciObject perform(GbjGciObject receiver,
				String selector,
				GbjGciObject args[],
				int numArgs) {
	checkObject(receiver);
	receiver.checkValue();
	int i;
	for (i=0;i<numArgs;i++) {
	  if ((args[i].getObjType() == OBJTYPE_UNKNOWN) || (args[i].getValType() == VALTYPE_UNKNOWN)) {
	     setupObj(args[i]);
	  }
	}
	GbjGciObject result = newGbjGciObject();
        GbjGciException errObj = nextErrObj;
        if (perform(sessionPtr, receiver, selector, args, numArgs, result, errObj) != 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)this, errObj);
        }
	notifyChangedObjects();
	return setupObj(result);
    }

   private native int getBytes(long sess, GbjGciObject obj, long startIndex, int numBytes, 
                                    byte[] result, GbjGciException errObj);

/**
Retrieve the designated bytes from this byte-based GS object.
@param obj         The GS object to retrieve the bytes from.
@param startIndex  The starting index (1-based)
@param numBytes    The number of bytes to retrieve
@return the retrieved bytes
@exception com.gemstone.gbj.GbjException thrown on any error
*/
    public synchronized byte[] getBytes(GbjGciObject obj,
			   long startIndex,
			   int numBytes) {
	checkObject(obj);
        byte[] bytes = new byte[numBytes];
        GbjGciException errObj = nextErrObj;
        if (getBytes(sessionPtr, obj, startIndex, numBytes, bytes, errObj) != 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)this, errObj);
        }
	return bytes;
    }


    private synchronized GbjGciObject objWithOop_(long oopValue, GbjGciException errObj) {
        GbjGciObject obj = newGbjGciObject();
        obj.oop = oopValue;
        if (setupReport(sessionPtr, obj, errObj) != 0) {
          nextErrObj = new GbjGciException();
          throw new GbjException((GbjSession)this, errObj);
        }
        return obj;
    }

/**
Return the GS object with the designed oop.
@param oopValue    The oop to search for.
@return the GS object with this oop (if it exists)
@exception com.gemstone.gbj.GbjException thrown on any error
*/
    public synchronized GbjGciObject objWithOop(long oopValue) {
	GbjGciObject  obj = objectTableGet(oopValue);
	if (obj != null) 
          return obj;

	obj = objWithOop_(oopValue, nextErrObj);
	objectTablePut(obj);
        return obj; 
    }

    private native int getNamedIvs(long sess, GbjGciObject obj, long startIndex, int numObjs,
                                GbjGciObject[] results, GbjGciException errObj);

    private native int getVaryingIvs(long sess, GbjGciObject obj, long startIndex, int numObjs,
                                GbjGciObject[] results, GbjGciException errObj);
   
    private synchronized void getVaryingIvs(GbjGciObject obj, long startIndex, int numObjs,
                                                    GbjGciObject[] objs) {
        if (numObjs <= 0) return;
        int objType = obj.objType_();
        if (objType != OBJTYPE_OOP && objType != OBJTYPE_NSC)
	  throw new GbjException("Invalid obj type ");
        if (startIndex - 1 + numObjs > obj.indexedSize)
	  throw new GbjException("numObjs out of range ");
        GbjGciException errObj = nextErrObj;
        if ((obj.namedSize + startIndex + numObjs) < GbjGciInterface.CacheSizeOops) {
          int startOfs = (int)startIndex - 1;
	  long[] valOops = obj.valueOops();
	  for (int i = 0; i < numObjs; i++) {
	    objs[i] = objWithOop_(valOops[startOfs + i], errObj);
	  }
        } else {
          for (int i=0; i < numObjs; i++) {
            objs[i] = newGbjGciObject();
          }
          int nRet = getVaryingIvs(sessionPtr, obj, startIndex, numObjs, objs, errObj);
          if (nRet != numObjs) {
            nextErrObj = new GbjGciException();
            throw new GbjException((GbjSession)this, errObj);
          }
        }
	for (int i=0; i < numObjs; i++) {
	  objs[i] = setupObj(objs[i]);
	}
   }

    private synchronized void getNamedIvs(GbjGciObject obj, long startIndex, int numObjs,
                                                    GbjGciObject[] objs) {
        if (numObjs <= 0) return;
        int objType = obj.objType_();
        if (objType != OBJTYPE_OOP && objType != OBJTYPE_NSC)
	  throw new GbjException("Invalid obj type ");
        if (startIndex - 1 + numObjs > obj.namedSize)
	  throw new GbjException("numObjs out of range ");
        GbjGciException errObj = nextErrObj;
        if ((startIndex - 1 + numObjs) < GbjGciInterface.CacheSizeOops) {
          int startOfs = (int)startIndex - 1;
	  long[] valOops = obj.valueOops();
	  for (int i = 0; i < numObjs; i++) {
	    objs[i] = objWithOop_(valOops[startOfs + i], errObj);
	  }
        } else {
          for (int i=0; i < numObjs; i++) {
            objs[i] = newGbjGciObject();
          }
          if (getNamedIvs(sessionPtr, obj, startIndex, numObjs, objs, errObj) != 0) {
            nextErrObj = new GbjGciException();
            throw new GbjException((GbjSession)this, errObj);
          }
        }
	for (int i=0; i < numObjs; i++) {
	  objs[i] = setupObj(objs[i]);
	}
   }

/**
Retrieve the designated GS objects from this indexable GS object
@param obj         The GS object to retrieve the GS objects from.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to retrieve
@return An array of GS objects containing the results
@exception com.gemstone.gbj.GbjException thrown on any error
*/
    public synchronized GbjGciObject[] getObjs(GbjGciObject obj, long startIndex, int numObjs) {
      if ( startIndex - 1 + obj.namedSize + numObjs > obj.indexedSize)
         throw new GbjException("startIndex/numObjs out of range");
      GbjGciObject[] objs = new GbjGciObject[numObjs];
      getVaryingIvs(obj, startIndex, numObjs, objs);
      return objs;
    }

    public synchronized void getObjs(GbjGciObject obj, long startIndex, int numObjs, GbjGciObject[] objs) {
      if ( startIndex - 1 + obj.namedSize + numObjs > obj.indexedSize)
         throw new GbjException("startIndex/numObjs out of range");
      getVaryingIvs(obj, startIndex, numObjs, objs);
    }

/**
Retrieve the designated GS objects from the named instvars for this GS object.
@param obj         The GS object to retrieve the GS objects from.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to retrieve
@return An array of GS objects containing the results
@exception com.gemstone.gbj.GbjException thrown on any error
*/
    public synchronized GbjGciObject[] getNamedObjs(GbjGciObject obj, long startIndex, int numObjs) {
      if ( startIndex - 1 + numObjs > obj.namedSize)
          throw new GbjException("startIndex/numObjs out of range");
      GbjGciObject[] objs = new GbjGciObject[numObjs];
      getNamedIvs(obj, startIndex, numObjs, objs);
      return objs;
    }

    public synchronized void getNamedObjs(GbjGciObject obj, long startIndex, int numObjs, GbjGciObject[] objs) {
      if ( startIndex - 1 + numObjs > obj.namedSize)
          throw new GbjException("startIndex/numObjs out of range");
      getNamedIvs(obj, startIndex, numObjs, objs);
  }

    private native int putBytes(long sess, GbjGciObject obj, long startIndex, int numBytes, byte[] bytes, 
                    GbjGciException errObj);

/**
Write the designated bytes to this byte-based GS object.
@param obj         The GS object to write the bytes to.
@param startIndex  The starting index (1-based)
@param numBytes    The number of bytes to write
@param bytes       The bytes to write out
@exception com.gemstone.gbj.GbjException thrown on any error
*/
   public synchronized void putBytes(GbjGciObject obj, long startIndex, int numBytes, byte[] bytes) {
      GbjGciException errObj = nextErrObj;
      if (putBytes(sessionPtr, obj, startIndex, numBytes, bytes, errObj) != 0) {
        nextErrObj = new GbjGciException();
        throw new GbjException((GbjSession)this, errObj);
      } 
      obj.notifyChanged();
      return;
    }

    private native int putNamedInstVars(long sess, GbjGciObject obj, long startIndex, int numObjs, 
                         GbjGciObject[] objs, GbjGciException errObj);

    private native int putVaryingInstVars(long sess, GbjGciObject obj, long startIndex, int numObjs, 
                         GbjGciObject[] objs, GbjGciException errObj);

/**
Write the designated GS objects to this indexable GS object.
@param obj         The GS object to write the GS objects to.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to write
@param objs        An array of GS objects to write out
@exception com.gemstone.gbj.GbjException thrown on any error
*/
    public synchronized void putObjs(GbjGciObject obj,
				     long startIndex,
				     int numObjs,
				     GbjGciObject[] objs) {
      long ofs = startIndex + obj.namedSize;
      GbjGciException errObj = nextErrObj;
      if (putVaryingInstVars(sessionPtr, obj, ofs, numObjs, objs, errObj) != 0) {
        nextErrObj = new GbjGciException();
        throw new GbjException((GbjSession)this, errObj);
      }
      obj.notifyChanged();
    }

/**
Write the designated GS objects to the named instvars for this GS object.
@param obj         The GS object to write the GS objects to.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to write
@param objs        An array of GS objects to write out
@exception com.gemstone.gbj.GbjException thrown on any error
*/
    public synchronized void putNamedObjs(GbjGciObject obj,
					  long startIndex,
					  int numObjs,
					  GbjGciObject[] objs) {
      if ( startIndex - 1 + numObjs > obj.namedSize)
        throw new GbjException("startIndex/numObjs out of range");
      GbjGciException errObj = nextErrObj;
      if (putNamedInstVars(sessionPtr, obj, startIndex, numObjs, objs, errObj) != 0) {
        nextErrObj = new GbjGciException();
        throw new GbjException(errObj);
      }
      obj.notifyChanged();
    }

   private native int isKindOf(long sess, GbjGciObject obj, GbjGciObject cls, GbjGciException errObj);

/**
Return true if this GS object is a member of the designated GS class 
or one of it's subclasses.
@param obj         The GS object to check.
@param cls         The GS class to check against.
@return true/false
@exception com.gemstone.gbj.GbjException thrown on any error
*/
    public synchronized boolean isKindOf(GbjGciObject obj, GbjGciObject cls) {
      GbjGciException errObj = nextErrObj; 
      int status = isKindOf(sessionPtr, obj, cls, errObj);
      if (status < 0) {
        nextErrObj = new GbjGciException();
        throw new GbjException((GbjSession)this, errObj);
      }
      return status == 1;
    }


    private native int resolveSymbol(long sess, String name, GbjGciObject result, GbjGciException errObj);

    /** Return the GS object with this name in the DB.
	On the server, looks through the symbol list for this session
	looking for entry with this name, scanning symbol dictionaries
	in the order listed in the symbol list
	@param name String
	@return GbjGciObject result
    */
    public synchronized GbjGciObject objectNamed(String name) {
      GbjGciObject result = newGbjGciObject();
      GbjGciException errObj = nextErrObj; 
      if (resolveSymbol(sessionPtr, name, result, errObj) != 0) {
        nextErrObj = new GbjGciException();
        throw new GbjException((GbjSession)this, errObj);
      }
      GbjGciObject res = setupObj( result);
      return res;
    }

    private native int resolveSymbolOrNil(long sess, String name, GbjGciObject result, GbjGciException errObj);

    public synchronized GbjGciObject objectNamedOrNil(String name) {
      GbjGciObject result = newGbjGciObject();
      GbjGciException errObj = nextErrObj; 
      if (resolveSymbolOrNil(sessionPtr, name, result, errObj) != 0) {
        nextErrObj = new GbjGciException();
        throw new GbjException((GbjSession)this, errObj);
      }
      GbjGciObject res = setupObj( result);
      return res;
    }

    private native int objExists(long sess, long oop, GbjGciException errObj);

    /** Return true if this session is logged in 
	@return boolean result
     */
    public synchronized boolean isLoggedIn() {
      if (sessionPtr == 0) return false;
      GbjGciException errObj = nextErrObj; 
      int status = objExists(sessionPtr, 207361/*OOP_GLOBALS_DICTIONARY*/, errObj);
      if (status < 0) {
        nextErrObj = new GbjGciException();
        throw new GbjException((GbjSession)this, errObj);
      }
      return status == 1;
    }

    /** Return the session ID for this session .
        Result not usable in thread safe GBJ
        @return int result    
     */
    public int getSessionId() { return -1; }

    public int serialNum() { return serialNum; }

    /** Check for asynchronous errors on this session .  
        has no effect when using the thread safe library */
    public void checkErrors( ) {
	return; 
    }

  // yyy todo  releaseOops
  // changedOops
  //   continue

}

