/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *========================================================================
 */

package com.gemstone.gbjgci;

import java.util.HashMap;
import java.lang.ref.WeakReference;
import java.lang.ref.ReferenceQueue;
import java.util.logging.*;
import java.time.LocalDate;
import java.time.LocalTime;
import java.time.LocalDateTime;
import static com.gemstone.gbjgci.GbjGciInterface.*;
import static com.gemstone.gbjgci.GbjGciObject.*;
import com.gemstone.gbj.GbjException;

/**
GbjGciSessions represent GemStone sessions.
<p>
Much of the core functionality of the GBJ class GbjSession has been moved
to this class, providing sufficient functionality for a "GBJ lite".

@version 3.0

*/

public class GbjGciSession {

    /** Enables log.fine* messages */
    public static boolean isDEBUG = System.getenv("GBJ_DEBUG") != null ; 

    /** Logger for debug trace messages */
    private static Logger log;

    // static code for setting up logger
    static {
	log = GbjGciInterface.setupLogger(GbjGciSession.class);
    }

    /** GCI session ID .
      sessionId <= 0 indicates session is not logged in  */
    private int sessionId;

    /** Flag used to control if changed objects should be monitored */
    public boolean monitorChangedObjects = false;

    private HashMap<Long, WeakReference<GbjGciObject>> objectTable;

    // Maps the oop of a GemStone class to a particular GbjGciObject subclass  
    private HashMap<Long, GbjGciObject> gsToJavaClassMap;

    // Structures to log batches of oops to remove from the export set

    // How many oops to batch before we bother to call GciReleaseOops
    private static final int EXPORT_SET_BATCH_SIZE = 500;
    private long[] oopsToRemoveFromExportSet;
    private int removeFromExportSetCount;
    private int setupObjEntryCount;

    /** The server object representing the GbjGciInterface class. */
    public GbjGciObject gbjGciInterface = null;


    // support methods for objectTable management
    private void objectTableInitialize() {
	objectTable = new HashMap<Long, WeakReference<GbjGciObject>>();
    }

    private GbjGciObject objectTableGet(GbjGciObject obj) {
	return objectTableGet(obj.getOop());
    }

    private GbjGciObject objectTableGet(long oopValue) {
        // boolean isDEBUG = GbjGciSession.isDEBUG;
	// if (isDEBUG) log.finest("TableGet for " + oopValue);
	WeakReference<GbjGciObject> ref = objectTable.get(oopValue);
	if (ref == null) {
	    // if (isDEBUG) log.finest("      no entry");
	    return null;
	}
	GbjGciObject obj = ref.get();
        //if (isDEBUG) {
	//  if (obj == null) log.finest("      null entry");
	//  else             log.finest("      entry found");
        //}
	return obj;
    }

    private void objectTablePut(GbjGciObject obj, boolean overWrite) {
	long oopValue = obj.getOop();
        boolean isDEBUG = GbjGciSession.isDEBUG;
	if (isDEBUG) log.finest("TablePut for " + oopValue);
	if (oopValue == 0) 
	    throw new GbjGciException("Object table put on oop zero");
	WeakReference<GbjGciObject> ref;
	ref = objectTable.get(oopValue);
	if (ref != null) { 
          Object oldVal = ref.get();
	  if (oldVal != null) {
            if (overWrite) {
              // ok
              // System.out.println("objectTablePut OVERWRITE oop " + oopValue);
              // waitForDebug();
            } else {
	      throw new GbjGciException( "Object table already contains entry");
            }
          }
	}
	ref = new WeakReference<GbjGciObject>(obj);
	objectTable.put(oopValue, ref);
    }

    private void objectTablePut(GbjGciObject obj) {
	objectTablePut(obj, false);
    }

    /** Release an un-used object from the weak reference used in
	the session's object table */
    protected synchronized void releaseObject(GbjGciObject obj) {
	// remove from object table (if it's recorded there)
	// Note: if called from obj.finalize(), a new version
	// of the obj might have already been loaded into the
	// table, so leave that one there..

	// For current design, we'll include specials in the table..
	//   skip specials -- we might need them later.
	//   if (obj.isSpecial()) return;       // TODO Fix

        if (obj.objType_() == OBJTYPE_SPECIAL)
          return; // special objs never in object table nor export set

	long oop = obj.getOop();
        boolean isDEBUG = GbjGciSession.isDEBUG;
        if (isDEBUG) log.finest("Releasing object " + oop);
	GbjGciObject tableObj = objectTableGet(oop);
	if (tableObj == obj /* explicit remove from app*/ ) { 
	   objectTable.put(oop, null);
        } else { 
	  // tableObj == null   means called by Java finalize 
        }
	if (oop == 0) return;  // fix 42081
	if (sessionId == 0) return;

	oopsToRemoveFromExportSet[removeFromExportSetCount] = oop;
	if (++removeFromExportSetCount >= EXPORT_SET_BATCH_SIZE) {
	   if (isDEBUG) log.finest("Releasing batch of oops..");
	   GbjGciInterface.releaseOops(this, sessionId, oopsToRemoveFromExportSet, removeFromExportSetCount);
	  removeFromExportSetCount = 0;
	}
    }


    // internal methods
    private void initialize() {
	sessionId = 0;
        setupObjEntryCount = 0;
	objectTableInitialize();
	classMapInitialize();
	oopsToRemoveFromExportSet = new long[EXPORT_SET_BATCH_SIZE];
	removeFromExportSetCount = 0;
    }

    /** Add/replace an entry in the object table */
    protected synchronized void updateObjectTable(GbjGciObject obj) {
	objectTablePut(obj);
    }

    // Methods for managing the GS-to-Java Class mapping functions

    private void classMapInitialize() {
	gsToJavaClassMap = new HashMap<Long, GbjGciObject>();
    }

    /** Map instances of a GemStone class to a designated java class,
	represented by an example instance.
	Defines a mapping such that any instances of the GemStone class 
	(and its subclasses) that are returned to java, will be represented
	as instances of the desired java class (which must be a subclass 
	of GbjGciObject).  The java class is indicated by an example 
	instance which is used as a template.
	@param gsClassName  Name of the GemStone class
	@param javaExample  An instance of the java class
    */
    public synchronized  void mapGsToJavaClass(String gsClassName, 
				 GbjGciObject javaExample) {
	mapGsToJavaClass(objectNamed(gsClassName), javaExample);
    }

    /** Map instances of a GemStone class to a designated java class,
	represented by an example instance.
	Defines a mapping such that any instances of the GemStone class 
	(and its subclasses) that are returned to java, will be represented
	as instances of the desired java class (which must be a subclass 
	of GbjGciObject).  The java class is indicated by an example 
	instance which is used as a template.
	@param gsClass      GS object of the GemStone class
	@param javaExample  An instance of the java class
    */
    public synchronized void mapGsToJavaClass(GbjGciObject gsClass, GbjGciObject javaExample) {
        boolean isDEBUG = GbjGciSession.isDEBUG;
	if (isDEBUG) {
          System.out.println("Mapping: " + gsClass + " to:" + javaExample.getClass());
        }
	GbjGciObject args[] = new GbjGciObject[1];
	args[0] = gsClass;
        // allSubclassOopsFor:  returns Array of SmallIntegers containing oops ...
	GbjGciObject subclassOops = perform(gbjGciInterface, "allSubclassOopsFor:", args, 1);
	GbjGciObject[] subclassOopArray = getObjs(subclassOops, 1, (int)subclassOops.indexedSize);
	if (isDEBUG) log.finest("mapGsToJavaClass:  " + subclassOops.indexedSize + " oops to transfer");
	for (int i=0; i< subclassOopArray.length; i++) {
            GbjGciObject anObj = subclassOopArray[i];
            long key = anObj.longValue();  // anObj is a SmallInteger (Java long)
	    if (isDEBUG) {
              String clsName = execute("(Object objectForOop: " + key + ") name").stringValue();
              System.out.println("   gsToJavaClassMap.put key " + key + " (" + clsName + ") -> " + javaExample );
            }
	    gsToJavaClassMap.put(key, javaExample);
	}
        releaseObject(subclassOops);
    }

    public static void waitForDebug() {
      return; // place to set jdb breakpoint
    }

    //protected int waitForDebug() {
    //  GbjGciSession.waitForDebug();
    //  sess.execute("System _sleep: 1");
    //  return 1;
    //}

    // check that the GbjGciObject is consistant for this session
    private void checkObject(GbjGciObject obj) {
      if (obj.objType_() == OBJTYPE_SPECIAL)
        return;   // special objects never in object table
      GbjGciObject otVal = objectTableGet(obj);
      if (otVal != obj) {
        if (otVal != null) {
          System.out.println("Defunct stub otVal oop " + otVal.getOop() + " obj " + obj.getOop());
        }
        waitForDebug();
	throw new GbjGciException("Defunct Stub");
      }
    }

    // notify objects which have been changed
    private static final int NOTIFY_CHANGE_BATCH_SIZE = 200;
    private long[] changedObjectBuffer;

    private void notifyChangedObjects() {

	// skip if we're not monitoring changed objects
	if (!monitorChangedObjects) return;

	// Notify observers of objects that have changed..
        boolean isDEBUG = GbjGciSession.isDEBUG;
	if (isDEBUG) log.finer("Changed oops: ");
	changedObjectBuffer = new long[NOTIFY_CHANGE_BATCH_SIZE];
	int count;
	do {
	    count = GbjGciInterface.changedOops(this, sessionId, changedObjectBuffer, NOTIFY_CHANGE_BATCH_SIZE);
	    for (int i=0; i<count; i++) {
		if (isDEBUG) log.finer("  " + changedObjectBuffer[i]);
		GbjGciObject obj = objectTableGet(changedObjectBuffer[i]);
		if (obj == null)
		    throw new GbjGciException( "notifyChangedObjects(): logic error");
		obj.notifyChanged();
	    }
	} while (count > 0);
    }


    // public methods

    /** Create an instance of GbjGciSession */
    public GbjGciSession() {
	initialize();
    }

    private GbjGciObject remapObject(GbjGciObject obj) {
        if (obj.valueType == VALTYPE_OOPS) {
	  long oop = obj.getOop();
	  GbjGciObject template  = gsToJavaClassMap.get(obj.classOop);
	  if (template != null && template.getClass() != obj.getClass()) {
	    GbjGciObject newObj;
	    try {
	      newObj = (GbjGciObject)template.clone();
	    } catch (CloneNotSupportedException ex) {
	      throw new GbjGciException(ex.getMessage());
	    }
	    newObj.copyFrom(obj);
	    obj = newObj;  // obj is now a GbjCollection
	  }
        }
	objectTablePut(obj, true);
	return obj;
    }

    private synchronized GbjGciObject addToObjTable(GbjGciObject result) {
	// first, check and see if we already have this guy in the objectTable
        int objType = result.objType_();
        if (objType == OBJTYPE_SPECIAL) {  // special objects are not in obj table
          return result;
        }
	GbjGciObject  obj = objectTableGet(result);
	if (obj == null) {
	    // remap class of result if needed
	    // this may generate a new result object
	    result = remapObject(result);
	    obj = result;
	}
	// in case the session hasn't been set..
	if (obj.getSession() == null) obj.setSession(this);

	obj.updateClient();

        boolean isDEBUG = GbjGciSession.isDEBUG;
	if (isDEBUG) log.finest("addToObjTable finish: " + obj );
	return obj;
    }

    /** Complete setup of an object newly returned from the JNI layer */
    protected synchronized GbjGciObject setupObj(GbjGciObject result) {
        boolean isDEBUG = GbjGciSession.isDEBUG;
        setupObjEntryCount += 1;
        if (setupObjEntryCount > 1) {
          throw new GbjGciException("recursion in setupObj");
        }
        GbjGciObject obj;
        try {
	  if (result.objType_() == OBJTYPE_UNKNOWN || result.oop == 0) {
	    GbjGciInterface.setupObj(this, sessionId, result);
	  }
	  obj = addToObjTable(result);
	  if (isDEBUG) log.finest("setupObj finish: " + obj);
        } finally {
          setupObjEntryCount -= 1;
        }
	return obj;
    }


/**

Login a GS session using this GbjGciSession.
<p>
@param serverName   The name of the GS stone process.
@param gemnetName   The name of the Gem Service.  If null, the default "gemnetobject" is used.
@param userName     The GS username for the session.
@param password     The GS password for this username.
@param hostUserName  The OS username for this session (if needed).
@param hostPassword  The OS password for this OS username (if needed).
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/

    public synchronized void login(String serverName,
		      String gemnetName,
		      String userName,
		      String password,
		      String hostUserName,
		      String hostPassword) {

	if (sessionId > 0)
	    throw new GbjGciException("Session already logged in");
	sessionId = GbjGciInterface.login(serverName,
					  gemnetName,
					  userName,
					  password,
					  hostUserName,
				  hostPassword);

        Exception ex = null;
        GbjGciReport cls  = null;
        try { 
	    cls = GbjGciInterface.resolveSymbol(sessionId, "GbjGciInterface");
	} catch (Exception e) {
           ex = e;
        }
        if (ex != null || cls == null) {
           String errmsg = "GbjGciInterface not found, GBJ classes not loaded in GS/S Server";
           log.severe(errmsg);
	   ex.printStackTrace();
           GbjGciInterface.logout(sessionId);
           throw new GbjGciException(errmsg);
        }
        gbjGciInterface = addToObjTable( (GbjGciObject)cls); 
    }

/** Logout the GS session associated with this GbjGciSession
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized void logout() {
	GbjGciInterface.logout(sessionId);
	initialize();
    }

/**
Execute a GS/S Smalltalk String (represented as a java String)
@param execStr     The Smalltalk string to execute.
@return a GS object representing the results of execution.
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized GbjGciObject execute(String execStr) {
	GbjGciObject result = (GbjGciObject)GbjGciInterface.execute(this, sessionId, execStr);
	notifyChangedObjects();
	return addToObjTable(result); 
    }

/**
Execute a GS/S Smalltalk String (represented as a GS Object)
@param execStrObj  The Smalltalk string to execute.
@return a GS object representing the results of execution.
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized GbjGciObject execute(GbjGciObject execStrObj) {
	checkObject(execStrObj);
	GbjGciObject result = (GbjGciObject)GbjGciInterface.execute(this, sessionId, execStrObj);
	notifyChangedObjects();
	return addToObjTable(result); 
    }

/**
Execute a Smalltalk method against a designated receiver using these arguments.
@param receiver    The GS object to execute this method against.
@param selector    The method to execute.
@param args        The list of arguments.
@param numArgs     The number of arguments in the list.
@return a GS object representing the results of execution.
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized GbjGciObject perform(GbjGciObject receiver,
				String selector,
				GbjGciObject args[],
				int numArgs) {
	checkObject(receiver);
	receiver.checkValue();
	int i;
	for (i=0;i<numArgs;i++) {
	  if ((args[i].getObjType() == OBJTYPE_UNKNOWN) || (args[i].getValType() == VALTYPE_UNKNOWN)) {
	     setupObj(args[i]);
	  }
	}
	GbjGciObject result = (GbjGciObject)GbjGciInterface.perform(this, sessionId, receiver, 
					          selector, args, numArgs);
	notifyChangedObjects();
	return addToObjTable(result);
    }


/**
Retrieve the designated bytes from this byte-based GS object.
@param obj         The GS object to retrieve the bytes from.
@param startIndex  The starting index (1-based)
@param numBytes    The number of bytes to retrieve
@return the retrieved bytes
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized byte[] getBytes(GbjGciObject obj,
			   long startIndex,
			   int numBytes) {
	checkObject(obj);
	byte[] results = GbjGciInterface.getBytes(this, sessionId, obj, startIndex, numBytes);
	return results;
    }

/**
Retrieve the designated GS objects from this indexable GS object
@param obj         The GS object to retrieve the GS objects from.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to retrieve
@return An array of GS objects containing the results
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized GbjGciObject[] getObjs(GbjGciObject obj,
				  long startIndex,
				  int numObjs) {
	GbjGciObject[] objs = new GbjGciObject[numObjs];
	getObjs(obj, startIndex, numObjs, objs);
	return objs;
    }

/**
Retrieve the designated GS objects from this indexable GS object, 
returning the results in the supplied array.
@param obj         The GS object to retrieve the GS objects from.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to retrieve
@param objs        An array of GS objects to return the results
@return The array of GS objects containing the results
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized GbjGciObject[] getObjs(GbjGciObject obj,
				  long startIndex,
				  int numObjs,
				  GbjGciObject[] objs) {
        boolean isDEBUG = GbjGciSession.isDEBUG;
	if (isDEBUG) log.fine("GbjGciSession.getObjs..");
	checkObject(obj);
	GbjGciInterface.getObjs(this, sessionId, obj, startIndex, numObjs, objs);
	if (isDEBUG) log.fine("DEBUG getObjs on " + obj);
	for (int i=0; i<numObjs; i++) {
	    objs[i] = addToObjTable(objs[i]);
            if (isDEBUG)  {
	      if (log.getLevel().intValue() <= Level.FINE.intValue()) {
		  if (i < 20) log.fine("   " + objs[i]);
              }
	    }
	}
	return objs;
    }

/**
Retrieve the designated GS objects from the named instvars for this GS object.
@param obj         The GS object to retrieve the GS objects from.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to retrieve
@return An array of GS objects containing the results
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized GbjGciObject[] getNamedObjs(GbjGciObject obj,
				       long startIndex,
				       int numObjs) {
	GbjGciObject[] objs = new GbjGciObject[numObjs];
	getNamedObjs(obj, startIndex, numObjs, objs);
	return objs;
    }

/**
Retrieve the designated GS objects from the named instvars for this GS object.
@param obj         The GS object to retrieve the GS objects from.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to retrieve
@param objs        An array of GS objects to return the results
@return The array of GS objects containing the results
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized GbjGciObject[] getNamedObjs(GbjGciObject obj,
				       long startIndex,
				       int numObjs,
				       GbjGciObject[] objs) {
	checkObject(obj);
	GbjGciInterface.getNamedObjs(this, sessionId, obj, startIndex, numObjs, objs);
	for (int i=0; i<numObjs; i++) {
          objs[i] = addToObjTable(objs[i]);
        }
	return objs;
    }

/**
Write the designated bytes to this byte-based GS object.
@param obj         The GS object to write the bytes to.
@param startIndex  The starting index (1-based)
@param numBytes    The number of bytes to write
@param bytes       The bytes to write out
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized void putBytes(GbjGciObject obj,
				      long startIndex,
				      int numBytes,
				      byte[] bytes) {
	checkObject(obj);
	GbjGciInterface.putBytes(this, sessionId, obj, startIndex, numBytes, bytes);
	// notifyChangedObjects(); - doesn't work?
	obj.notifyChanged();
	return;
    }

/**
Write the designated GS objects to this indexable GS object.
@param obj         The GS object to write the GS objects to.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to write
@param objs        An array of GS objects to write out
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized void putObjs(GbjGciObject obj,
				     long startIndex,
				     int numObjs,
				     GbjGciObject[] objs) {
	checkObject(obj);
	GbjGciInterface.putObjs(this, sessionId, obj, startIndex, numObjs, objs);
	// notifyChangedObjects(); - doesn't work?
	obj.notifyChanged();
	return;
    }

/**
Write the designated GS objects to the named instvars for this GS object.
@param obj         The GS object to write the GS objects to.
@param startIndex  The starting index (1-based)
@param numObjs     The number of objects to write
@param objs        An array of GS objects to write out
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized void putNamedObjs(GbjGciObject obj,
					  long startIndex,
					  int numObjs,
					  GbjGciObject[] objs) {
	checkObject(obj);
	GbjGciInterface.putNamedObjs(this, sessionId, obj, startIndex, numObjs, objs);
	// notifyChangedObjects(); - doesn't work?
	obj.notifyChanged();
	return;
    }

/**
Return the GS object with the designed oop.
@param oopValue    The oop to search for.
@return the GS object with this oop (if it exists)
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized GbjGciObject objWithOop(long oopValue) {
        boolean isDEBUG = GbjGciSession.isDEBUG;
	GbjGciObject  obj = objectTableGet(oopValue);
	if (obj == null) {
	    obj = (GbjGciObject)GbjGciInterface.objWithOop(this, sessionId, oopValue);
	    obj.setSession(this);
	    objectTablePut(obj);
	    if (isDEBUG) log.finer("objWithOop new: " + obj.getOop());
	} else {
	    if (isDEBUG) log.finer("objWithOop found in table: " + obj.getOop());
	}
	return obj;
    }

/**
Return true if this GS object is a member of the designated GS class 
or one of it's subclasses.
@param obj         The GS object to check.
@param cls         The GS class to check against.
@return true/false
@exception com.gemstone.gbjgci.GbjGciException thrown on any error
*/
    public synchronized boolean isKindOf(GbjGciObject obj,
			    GbjGciObject cls) {
	checkObject(obj);
	checkObject(cls);
        boolean isDEBUG = GbjGciSession.isDEBUG;
	if (isDEBUG) log.fine("isKindOf: " + obj.oop + " / " + cls.oop);
	boolean result;
	result = GbjGciInterface.isKindOf(this, sessionId, obj, cls);
	return result;
    }

    /** Return the GS object with this name in the DB.
	On the server, looks through the symbol list for this session
	looking for entry with this name, scanning symbol dictionaries
	in the order listed in the symbol list
	@param name String
	@return GbjGciObject result
    */
    public GbjGciObject objectNamed(String name) {
      boolean isDEBUG = GbjGciSession.isDEBUG;
      if (isDEBUG) log.info("objectNamed " + name );
      GbjGciReport val = GbjGciInterface.resolveSymbol(sessionId, name);
      GbjGciObject res = addToObjTable( (GbjGciObject)val);
      if (isDEBUG) log.info("objectNamed " + name + " = " + res );
      return res;
    }

    public GbjGciObject objectNamedOrNil(String name) {
      boolean isDEBUG = GbjGciSession.isDEBUG;
      if (isDEBUG) log.info("objectNamedOrNil " + name );
      GbjGciReport val = GbjGciInterface.resolveSymbolOrNil(sessionId, name);
      GbjGciObject res = addToObjTable( (GbjGciObject)val);
      if (isDEBUG) log.info("objectNamedOrNil " + name + " = " + res );
      return res;
    }

    /** Return true if this session is logged in 
	@return boolean result
     */
    public boolean isLoggedIn() {
	if (sessionId <= 0) return false;
        try {
          GbjGciReport val = GbjGciInterface.resolveSymbolOrNil(sessionId, "Globals");
          return val.getOop() == 207361; // OOP_GLOBALS_DICTIONARY
        } catch (GbjException e) {
          return false;
        }
    }

    /** Return the session ID for this session 
	@return int result
     */
    public int getSessionId() { return sessionId; }

    /** Return a hash code value for this object 
	@return int result
     */
    public int hashCode() {
	return ((sessionId * 77 + 13) % Integer.MAX_VALUE);
    }

    /** Check for asynchronous errors on this session */
    public void checkErrors( ) {
	GbjGciInterface.checkErrors(sessionId);
    }

}

