/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjCacheStats.java  
 *
 *========================================================================
 */

package com.gemstone.gbjstats;
import java.util.*;
import java.util.logging.*;
import com.gemstone.gbjgci.*;
import com.gemstone.gbj.*;

/** 

GbjCacheStats provides an application-friendly level of access to
shared page cache statistics. The class represents a batch of cache stats
associated with a particular GS process, and extends the low-level 
GbjGciCacheStats class, which provides the basic fields for representing
the stat info, including:

<table>
<caption> " " </caption>
<tr><td> &#149;&nbsp;&nbsp; 
    <td> name <td> The name of the process generating these stats
<tr><td> &#149;&nbsp;&nbsp; 
    <td> pid <td> The process ID
<tr><td> &#149;&nbsp;&nbsp; 
    <td> sid <td> The session ID
<tr><td> &#149;&nbsp;&nbsp; 
    <td> type <td> An integer representing the type of GS process
<tr><td> &#149;&nbsp;&nbsp; 
    <td> stats32 <td> An int[] array for holding the 32-bit stats
<tr><td> &#149;&nbsp;&nbsp; 
    <td> stats64 &nbsp; &nbsp; <td> A long[] array for holding the 64-bit stats
</table>

<p>

Useful behavior defined in GbjCacheStats includes:

<table> 
<caption> " " </caption>
<tr><td> &#149;&nbsp;&nbsp; 
    <td> attach <td> Attach to a particular shared page cache
<tr><td> &#149;&nbsp;&nbsp; 
    <td> sample <td> Take a sample of all cache stats for all processes
                     associated with this shared page cache
<tr><td> &#149;&nbsp;&nbsp; 
    <td> fetch  <td> Fetch the batch of stats associated with a particular
                     GS process that were taken during a prior sample()
<tr><td> &#149;&nbsp;&nbsp; 
    <td> get    <td> Get the single stat value associated with the given name
<tr><td> &#149;&nbsp;&nbsp; 
    <td> detach &nbsp;&nbsp; <td> Detach from the shared page cache
</table>

<p>
The following is a code snippet demonstrating a typical calling sequence for
using these functions:

<pre>
<code>

	// Initialize Pid tracking lists
	List&lt;Long&gt; newPids  = new ArrayList&lt;Long&gt;(maxProcs);
	List&lt;Long&gt; oldPids  = new ArrayList&lt;Long&gt;(maxProcs);
	List&lt;Long&gt; deadPids = new ArrayList&lt;Long&gt;(maxProcs);

	// Attach cache
	int maxProcs = GbjCacheStats.attach("StoneName");

	// Take first sample
	GbjCacheStats.sample(newPids, oldPids, deadPids);

        // Get stat from a process of interest
	GbjCacheStats myStats = GbjCacheStats.fetch(PidOfInterest);
        long stat = myStats.get("StatOfInterest");

	// Loop over time and update stat
        do { 
	    // sleep for some time interval..
	    Thread.currentThread().sleep(SleepTimeInMilliseconds);

	    // Take next sample
	    GbjCacheStats.sample(newPids, oldPids, deadPids);

	    // Update our stat info from new sample
            myStats.update();
            stat = myStats.get("StatOfInterest");

	} while (true);

	// when done, detach the cache
	GbjCacheStats.detach();

</code>
</pre>
	    

@version 3.1

*/

public class GbjCacheStats extends GbjGciCacheStats {

    /** A representation for various types of GS processes */
    public enum CacheType {
                    SPCMON("Spcmon"), 
                    STONE("Stone"), 
                    PGSVR("Pgsvr"), 
                    GEM("Gem"), 
                    STATMON("Statmon"),
                    APP("App");
                  //SYSTEM("System");
       private String typeName;
       CacheType(String name) { this.typeName = name; }
       /** Returns the number of different CacheTypes */
       protected static int length() { return 6; }
       /** Returns the name associated with this CacheType */
       protected String typeName() { return typeName; }
    }

    protected static CacheType cacheType(int index) {
        switch (index) {
            case 0: return CacheType.SPCMON;
            case 1: return CacheType.STONE;
            case 2: return CacheType.PGSVR;
            case 3: return CacheType.GEM;
            case 4: return CacheType.STATMON;
            case 5: return CacheType.APP;
          //case 6: return CacheType.SYSTEM;
        }
    throw new IllegalArgumentException( );
    }

    // Maximum number of stats for any type 
    // As of 7/1/2015, there were 310 stone stats.
    // Since it's always changing, make a little larger for padding.
    // GciInterface.GCSIstatNames() will print 
    // the gem stat size when in debug mode.
    private static final int MaxStatSize = 350; 

    /** Logger for debug trace messages */
    private static Logger log;

    private static String[][] statNames;

    private static Boolean initialized = false;

    private static String stoneName;

    private static int maxProcessCount;

    private static Boolean sampleTaken = false;

    private static long timeStamp;

    private static long currentProcIds[];  // live processes on current cycle
    private static long priorProcIds[];    // live processes on previous cycle
    private static long newProcIds[]; // processes that have just showed up this cycle
    private static long oldProcIds[]; // processes that appeared earlier and continue
    private static long deadProcIds[]; // processes that existed earlier and are now gone

    // cacheMap tracks cache stats keyed by processId
    private static Map<Long, GbjCacheStats> cacheMap = null;

    Map<String, Long> statMap = null;

    // static code for setting up prefs, logger,
    // and initializing GCSI cache system
    {
	// setup logging
	log = Logger.getLogger(GbjGciSession.class.getName());
	Level logLevel = GbjGciPreferences.logLevel("GbjCacheStats");
	log.setLevel(logLevel);
	Handler[] handlers = Logger.getLogger("").getHandlers();
	for (int i=0; i<handlers.length; i++) {
	    handlers[i].setLevel(logLevel);
	}

    }

    /** Establish a set of cache stats associated with this process ID 
	@param p long process ID
     */
    public GbjCacheStats(long p) {
	pid = p;
    }

    /** Return the type of cache stat: spcmon, stone, pgsvr, gem 
	@return String representing type of cache stat
     */
    public String cacheTypeString( ) { return cacheType(type).typeName(); }


    // returns maxProcessCount for this SPC - use for sizing pid arrays

    /** Attach to a particular shared page cache 
	@param stnName String name of stone
	@return the maximum number of processes that can 
	use this shared page cache
    */
    public static int attach(String stnName) {
	// GbjCacheStats creation: attach the cache

	if (!initialized) {
	    // setup GCSI cache system and stat names
	    GbjGciInterface.GCSIinitialize();
	    statNames = GbjGciInterface.GCSIstatNames();
	    initialized = true;
	}

	if (stoneName != null)
	    throw new GbjException("Already attached");
	maxProcessCount = GbjGciInterface.GCSIattach(stnName);
	stoneName = stnName;
	cacheMap = new HashMap<Long,GbjCacheStats>(maxProcessCount);
	currentProcIds = new long[maxProcessCount+1]; // +1 for length indicator
        priorProcIds = new long[maxProcessCount+1];
        newProcIds = new long[maxProcessCount+1];
        oldProcIds = new long[maxProcessCount+1];
        deadProcIds = new long[maxProcessCount+1];
        return maxProcessCount;
    }


    /** Detach from this shared page cache */
    public static void detach( ) {
	if (stoneName != null) {
	    GbjGciInterface.GCSIdetach();
	    stoneName = null;
	    cacheMap = null;
	    currentProcIds = null;
            priorProcIds = null;
            newProcIds = null;
            oldProcIds = null;
            deadProcIds = null;
            sampleTaken = false;
	}
    }

    // following private methods are for handling pidLists 
    // (arrays of longs with the length contained in the first element)

    private static int lengthPids(long[] pids) { return (int)pids[0]; }

    private static void initializePids(long[] pids) {
        pids[0] = 0;
    }

    // initialize pids1 with contents of pid2
    private static void initializePids(long[] pids1, long[] pids2) {
        int length = (int)pids2[0];
        for (int i=0; i<=length; i++) pids1[i] = pids2[i];
    }

    // returns index of pid in pidList, 0 if not in there
    private static int indexOfPid(long[] pids, long x) {
        int length = (int)pids[0];
        for (int i=1; i<=length; i++) {
            if (pids[i] == x) return i;
        }
        return 0;
    }

    private static boolean containsPid(long[] pids, long x) {
        return (indexOfPid(pids, x) != 0);
    }

    private static void addPid(long[] pids, long x) {
        if (!containsPid(pids, x)) { 
	    pids[0]++;
	    pids[(int)pids[0]] = x;
        }
    }

    // add those pids in pids1 that *are not* in pids2 to pids3
    private static void diffPids(long[] pids1, long[] pids2, long[] pids3) {
        int length = (int)pids1[0];
        for (int i=1; i<=length; i++) {
            if (!containsPid(pids2, pids1[i])) addPid(pids3, pids1[i]);
        }
    }

    // GCSIsampleStats occationally returns a -1 -- clean these out
    private static void cleanPids(long[] pids) {
        int offset = 0;
        int originalLength = (int)pids[0];
        int length = originalLength;
        for (int i=1; i<=originalLength; i++) {
            if (pids[i] < 0) {
                offset++;
                length--;
            } else { 
                pids[i+offset] = pids[i];
            }
        }
        pids[0] = length;
    }


    // return pids in a properly sized long[] array
    private static void returnPids(long[] pids, List<Long> results) {
        results.clear();
        int length = (int)pids[0];
        if (length > 0) {
            for (int i=1; i<=length; i++) {
                results.add(pids[i]);
            }
        }
    }


    // debug print method
    private static void printPids(String title, long pids[], List<Long> lpids) {
        System.out.print(title);
        int length = (int)pids[0];
        for (int i=1; i<=length; i++) System.out.print(" " + pids[i]);
        System.out.print(" / ");
        for (long pid : lpids) System.out.print(" " + pid);
        System.out.println("");
    }

    /** 
     Sample all the cache statistics associated with this stone/shared page cache.
     The arguments are lists used to return process ID information as follows:
     <p>
     newPids - processes that have appeared since the last sample <br>
     oldPids - processes which existed during the last sample <br>
     deadPids - processes which have disappeared since the last sample <br>
     <p>
     These lists should be initialized using:
     <p>
     new ArrayList&lt;Long&gt;(maxProcs)
     <p>
     Where maxProcs is the maximum number of processes returned by the 
     GbjGciStats.attach() call.
     <p>
     Note that the first time this method is called, *all* processes will 
     be returned as newPids with oldPids and deadPids empty.
     @param newPids List of long process ID's that just appeared
     @param oldPids List of long process ID's that were in last sample
     @param deadPids List of long process ID's that have disappeard
     */
    public static void sample(List<Long> newPids, 
                              List<Long> oldPids, 
                              List<Long> deadPids) {
	if (stoneName == null)
	    throw new GbjException("GbjCacheStats not attached");
	timeStamp = System.currentTimeMillis( );
	// save the last round of process id's
        initializePids(newProcIds);
        initializePids(oldProcIds);
        initializePids(deadProcIds);
        // reset priorPids with the earlier currentPids
        initializePids(priorProcIds, currentProcIds);
	GbjGciInterface.GCSIsampleStats(currentProcIds);
        cleanPids(currentProcIds);
        diffPids(currentProcIds, priorProcIds, newProcIds);
        diffPids(currentProcIds, newProcIds, oldProcIds);
        diffPids(priorProcIds, currentProcIds, deadProcIds);
        returnPids(newProcIds, newPids);
        returnPids(oldProcIds, oldPids);
        returnPids(deadProcIds, deadPids);
	sampleTaken = true;
    }

    /** Update (refresh) the information for this cache stat with the
        data taken from the last sample() */
    public void update( ) {
	// preliminary state checks
	if (stoneName == null)
	    throw new GbjException("GbjCacheStats not attached");
	if (!sampleTaken) 
	    throw new GbjException("GbjCacheStats sample not taken");
	// check that stats are in map
	if (this != cacheMap.get(pid))
            throw new GbjException("update: supplied stats not in cacheMap");
	GbjGciInterface.GCSIgetStats(pid, this);
	this.setup();
    }

    /**
    Return the GbjCacheStats associated with this process ID taken during a 
    prior sample( ).
    @param pid long process ID
    @return GbjCacheStats result
    */
    public static GbjCacheStats fetch(long pid) {
	// preliminary state checks
	if (stoneName == null)
	    throw new GbjException("GbjCacheStats not attached");
	if (!sampleTaken) 
	    throw new GbjException("GbjCacheStats sample not taken");
	// get old stats from map if present, add if not
	GbjCacheStats stats = cacheMap.get(pid);
	if (stats == null) {
	    stats = new GbjCacheStats(pid);
	    cacheMap.put(pid, stats);
	}
	stats.update( );
	return stats;
    }

    private void setup( ) {
	int i,j;
	if (statMap == null)
	    statMap = new HashMap<String,Long>(MaxStatSize);
	String[] names = statNames[type];
	i = 0;
	for (j=0; j < stats32.length; j++) {
	    statMap.put(names[i++], (long)stats32[j]);
	}
	for (j=0; j < stats64.length; j++) {
	    statMap.put(names[i++], stats64[j]);
	}
    }

   /** 
    Return the value of the stat with this given name 
    @param name String name of stat
    @return long value of stat
    @exception IllegalArgumentException thrown if stat name doesn't exist
   */
    public long get(String name) throws IllegalArgumentException {
        if (!statMap.containsKey(name))
            throw new IllegalArgumentException(name);
	return statMap.get(name);
    }

    /** Return the list of stat names associated with this stat 
	@return array of strings containing names of stats
     */
    public String[] statNames( ) {
	return statNames[type];
    }


}


