/*
 * GbjHypericPluginGenerator.java
 */

package com.gemstone.gbjstats;
import java.io.*;
import com.gemstone.gbjgci.*;
import com.gemstone.gbj.*;

/**

GbjHypericPluginGenerator generates an XML file that can be used by
the VMware Hyperic system monitoring tool to recognize and monitor
a GemStone environment that includes a GbxJmxCacheStatsMonitor program.

<p>

The VMware Hyperic system monitoring tool uses specialized "plugins" to
be able to recognize and manipulate its various servers and services.
These plugins are either a single XML file, or a jar file that contains
the XML file plus various code files that support the functionality
defined in the XML file.  GbjHypericPluginGenerator generates a compatible
XML file that works with a specific stone, monitored by a an instance
of GbxJmxCacheStatsMonitor running on a specific JMX port.

<p>

Prior to running GbjHypericPluginGenerator, make sure that $GEMSTONE is
set in the environment.

<p>

To run GbjHypericPluginGenerator, execute:

<p>

java -d64 com.gemstone.gbjstats.GbjHypericPluginGenerator &lt;stonename&gt; &lt;jmxport&gt;

<p>

Where &lt;stonename&gt; is the name of the associated stone and &lt;jmxport&gt;
is the JMX port being used by the GbjJmxCacheStatMonitor application 
monitoring this stone.

<p>

Execution of this code produces an output file with the name:

<p>

gs_&lt;version&gt;_&lt;stonename&gt;_&lt;jmxport&gt;-plugin.xml.

<p>

Where &lt;stonename&gt; and &lt;jmxport&gt; are as above and &lt;version&gt;
 is the version of the Gemstone system associated with $GEMSTONE.

<p>

You will then need to copy this file to the appropriate $HYPERIC directories
for the agent(s) and server(s) that will need to monitor and manage this
GemStone environment.  Refer to your Hyperic documentation for details.

<p>

Note that different plugins will need to be generated for each stone
(configured with the appropriate $GEMSTONE so that the version information
is correct).  Also, each separate stone will require a 
different GbjJmxCacheStatsMonitor application, each running on a different
JMX port.  For distributed GemStone environments, there will also need to
be a separate GbjJmxCacheStatsMonitor for each remote shared page cache.
GbjJmxCacheStatsMonitors on remote caches should all use the same JMX port 
as is assigned to the stone GbjJmxCacheStatsMonitor.

*/

public class GbjHypericPluginGenerator {

  private static String[][] statNames;

  private static String gciVersion;
  static String ServerTemplateFile = "GbjHypericServerTemplate.xml";
  static String ServiceTemplateFile = "GbjHypericServiceTemplate.xml";
  static int ServerTemplateSize = 10000;

  protected GbjHypericPluginGenerator( ) { super( ); }

  private static void substitute(StringBuilder context, 
                                 String str1, 
                                 String str2) {
    int index = 0;
    int count = 0;
    do {
      index = context.indexOf(str1, index);
      if (index >= 0) context.replace(index, index + str1.length(), str2);
    } while (index >= 0);
  }



  private static String getCommands( ) {
        // extract just the name field from command info
        GbjJmxCommandInfo[] commandInfo = GbjJmxCommandInfo.readCommandInfo();
        StringBuilder cmds = new StringBuilder();
        for (int i = 0; i < commandInfo.length; i++) {
            cmds.append(",");
            cmds.append(commandInfo[i].name);
        }
        return cmds.toString();
    }

  private static void generateSystemMetrics(StringBuilder template) {
   
     StringBuilder buf = new StringBuilder( );

    // retrieve commands from the GbjJmxSystemStats.dat file
    GbjJmxSystemStatInfo[] stats =
       GbjJmxSystemStatInfo.readSystemStatInfo();

    // generate metric details from stat info
    for (GbjJmxSystemStatInfo stat : stats) {
      buf.append("      <metric name=\"" + stat.name + "\"\n");
      buf.append("              template=\"${OBJECT_NAME}:" + stat.name + "\"\n");
      buf.append("              category=\"" + stat.category + "\"\n");
      buf.append("              indicator=\"true\"\n");
      buf.append("              collectionType=\"" + stat.type + "\"\n");
      buf.append("              interval=\"60000\"\n");
      if (stat.units != null)   // units is optional
      buf.append("              units=\"" + stat.units + "\"\n");
      if (stat.rate != null)    // rate is optional
      buf.append("              rate=\"" + stat.rate + "\"\n");
      buf.append("              />\n\n");
    }
    substitute(template, "###METRICS###", buf.toString());
  }


  private static void generateService(StringBuilder template,
                                      String name, 
                                      String[] statNames) {

    substitute(template, "###PTYPE###", name);

    StringBuilder buf = new StringBuilder( );

    // use a generic template for service-level stats
    // since we don't have a template file to spell out the details:
    //   category=PERFORMANCE
    //   indicator=true
    //   collectionType=dynamic
    //   interval=60000 (1 minute - best Hyperic can do)
    //
    for (String statName : statNames) {
      buf.append("      <metric name=\"" + statName + "\"\n");
      buf.append("              template=\"${OBJECT_NAME}:" + statName + "\"\n");
      buf.append("              category=\"UTILIZATION\"\n");
      buf.append("              indicator=\"true\"\n");
      buf.append("              collectionType=\"dynamic\"\n");
      buf.append("              interval=\"60000\"/>\n\n");
    }
    substitute(template, "###METRICS###", buf.toString());
  }

  private static void waitForDebug() {
     return ; // place to set jdb breakpoint
  }

  /* For simplicity, we declare "throws Exception".
     Real programs will usually want finer-grained exception handling. */

  /** Execute the GbjHypericPluginGenreator program. 
      @param args array of strings containing arguments
      @exception Exception  thrown for any problems
   */

  public static void main(String[] args) throws Exception {

    GbjSession.initialize(false, "GbjHypericPluginGenerator", 10000, 200, 1000);
    gciVersion = GbjGciInterface.gciVersion( );

    if (args.length != 2) {
        System.out.println("Usage: ");
        System.out.println(
"java com.gemstone.gbjstats.GbjHypericPluginGenerator <stonename> <portnum>");
        System.out.println("Produces XML file gs_" + 
                           gciVersion + 
                           "_<stonename>_<portnum>-plugin.xml");
        System.exit(0);
    }

    String stoneName = args[0];
    String portStr   = args[1];

    String fileName = "gs_" + 
                      gciVersion + "_" + 
                      stoneName + "_" + 
                      portStr + "-plugin.xml";

    System.out.println("Stone:    " + stoneName);
    System.out.println("Port:     " + portStr);
    System.out.println("Version:  " + gciVersion);
    System.out.println("XML file: " + fileName);

    GbjGciInterface.GCSIinitialize();
    statNames = GbjGciInterface.GCSIstatNames( );

    GbjJmxDataFile inputFile = null;
    int size = 0;
    char[] cbuf = new char[ServerTemplateSize];

    // get the server template from the template file
    try {
        inputFile = new GbjJmxDataFile(ServerTemplateFile);
    } catch (java.io.FileNotFoundException e) {
      System.err.println("Problem opening server template file");
      waitForDebug();
    }
    String serverTemplate = inputFile.readTemplate();
    inputFile.close( );
    StringBuilder serverBuf = new StringBuilder(serverTemplate);


    // get the service template from the template file
    try {
        inputFile = new GbjJmxDataFile(ServiceTemplateFile);
    } catch (java.io.FileNotFoundException e) {
      System.err.println("Problem opening service template file");
      waitForDebug();
    }
    String serviceTemplate = inputFile.readTemplate();
    inputFile.close( );

    substitute(serverBuf, "###STONE###", stoneName);
    substitute(serverBuf, "###PORT###",  portStr);
    substitute(serverBuf, "###VERSION###", gciVersion);
    // placeholder for system-level command handling
    substitute(serverBuf, "###COMMANDS###", getCommands());
    generateSystemMetrics(serverBuf);

    for (GbjCacheStats.CacheType type : GbjCacheStats.CacheType.values()) {
        int index = type.ordinal();
        StringBuilder serviceBuf = new StringBuilder(serviceTemplate);
        String name = type.typeName();
        generateService(serviceBuf, name, statNames[index]);
        String tag = "###SERVICE_" + name + "###";
        substitute(serverBuf, tag, serviceBuf.toString());
        }

    PrintStream file = new PrintStream(fileName);
    file.println(serverBuf.toString());
    file.close( );
    }

}

