/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjJmxCacheStats.java  
 *
 *========================================================================
 */

package com.gemstone.gbjstats;
import com.gemstone.gbjgci.*;
import com.gemstone.gbj.*;
import java.util.*;
import javax.management.*;
import javax.management.modelmbean.*;

/**

GbjJmxCacheStats is a wrapper around GbjCacheStats to allow access to shared
page cache statistics through java JMX-based monitoring tools and applications
via the DynamicMBean interface.  They are used extensively in the
GbjJmxCacheStatsMonitor program.

*/

public class GbjJmxCacheStats implements DynamicMBean {

    /** The cache stats wrappered by this instance */
    public GbjCacheStats stats;
    /** The GS session to be used for executing control commands */
    public GbjSession session;    // session for control commands
    /** The GbjJmxCacheStatsMonitor process updating these cache stats */
    public GbjJmxCacheStatsMonitor monitor; // monitor watching this stat
    static protected Runtime rt = Runtime.getRuntime( );

    /** Base constructor needed for JMX protocol */
    public GbjJmxCacheStats() { }

    private Descriptor buildDescriptor(GbjCacheStats s)  
    {  
       final Descriptor descriptor = new DescriptorSupport();  
       descriptor.setField("name", "GbjJmxCacheStats:" + stats.name);  
       descriptor.setField("descriptorType", "mbean");  
       return descriptor;  
    }  


    /** Create a new GbjJmxCacheStats object on a particular set of
        GbjCacheStats from a given GS session 
	@param s GbjCacheStats stats
	@param ses GbjSession
    */
    public GbjJmxCacheStats(GbjCacheStats s, GbjSession ses) {
       stats = s;
       session = ses;

    }

    /** Special-purpose GbjJmxCacheStats creator for use 
        with GbjJmxCacheStatsMonitor */
    protected GbjJmxCacheStats(GbjCacheStats s, 
                               GbjSession ses,
                               GbjJmxCacheStatsMonitor m) {
       this(s, ses);
       monitor = m;
    }

  /** Return the stat names associated with this set of cache stats */
  protected String[] statNames( ) {
    return stats.statNames();
  }

  /** Return the value of the named stat */
  protected long getStats(String name) {
    return stats.get(name);
  }

    // operations

    /** JMX operation to stop the session 
        associated with this GbjJmxCacheStat 
	@return String containing results of executing stop
    */
    public String stop() {
      if (stats.type == GbjGciCacheStats.TYPE_STONE) {
        try {
          rt.exec("stopstone -I " + stats.name);
        } catch (Exception e) { return "Problem stopping stone" + stats.name; }
        return "Stone shutdown";
      } else if (session == null) {
        return "Control session not running";
      } else if (stats.sid == session.getSessionId()) {
        return "Cannot stop JMX control session";
      } else {
        String result = session.printit("System stopSession: " + stats.sid);
        if (result.startsWith("System")) return "Session stopped";
        else return result;
      }
    }

    /** JMX operation to kill the process 
        associated with this GbjJmxCacheStat using kill -9 
	@return String containing results of executing kill
    */
    public String kill() {

      if (stats.sid == session.getSessionId()) {
        return "Cannot kill JMX control session";
      } else {
        try {
          rt.exec("kill -9 " + stats.pid);
        } catch (Exception e) { return "Problem executing kill command"; }
        return "GemStone process killed";
      }
    }

    /** JMX operation to enable tracing of these process-level cache stats
        from Hyperic 
	@return String containing results of executing enableHypericTrace
    */
    public String enableHypericTrace( ) {
        try {
            monitor.registerHypericCacheStats(stats, session, stats.pid);
        } catch (Exception e) { 
            return "Problem trying to enable Hyperic tracing";
        }
        return "Hyperic tracing enabled";
    }

// http://weblogs.java.net/blog/emcmanus/archive/2006/11/a_real_example.html
  // Methods defining the DynamicMBean interface

  // Obtain the value of a specific attribute of the Dynamic MBean.
  /** Support for JMX DynamicMBean interface 
      @param attribute String name of attribute to fetch
      @return Object result
      @exception AttributeNotFoundException  thrown if cant find attribute
   */
  public Object getAttribute(String attribute) throws AttributeNotFoundException {
    if (attribute.equals("Name")) {
//        System.out.println("Returning stat " + stats.name + 
//                           "/" + attribute);
	return stats.name;
    } else {
      try { return getStats(attribute); } catch (IllegalArgumentException e) {
        throw new AttributeNotFoundException(attribute);
      }
    }
  }

  // Get the values of several attributes of the Dynamic MBean.
  /** Support for JMX DynamicMBean interface 
      @param attributes array of Strings containing names of attributes
      @return AttributeList
   */
  public AttributeList getAttributes(String[] attributes) {
//    System.out.println("Returning stats for " + stats.name);
    AttributeList list = new AttributeList();
    for (String attribute : attributes) {
      try { 
// System.out.println("Looking for stat " + attribute + "...");
        long value = getStats(attribute);
//        System.out.println("   " + attribute + " / " + value);
        list.add(new Attribute(attribute, value));
      } catch (IllegalArgumentException e) { }
    }
    return list;
  }


  protected MBeanOperationInfo[] getMBeanOperationInfo( ) {
    MBeanOperationInfo[] opers = {

      new MBeanOperationInfo(
                    "stop",
                    "Stop the GemStone process",
                    null,   // no parameters
                    "String",
                    MBeanOperationInfo.ACTION),

      new MBeanOperationInfo(
                    "kill",
                    "Kill the GemStone process",
                    null,   // no parameters
                    "String",
                    MBeanOperationInfo.ACTION),

      new MBeanOperationInfo(
                    "enableHypericTrace",
                    "Enable tracing this GemStone process from Hyperic",
                    null,   // no parameters
                    "String",
                    MBeanOperationInfo.ACTION)
    };
    return opers;
  }

  private MBeanInfo mbeanInfo = null;

  // Provides the exposed attributes and actions of the Dynamic MBean using an MBeanInfo object.
  /** Support for JMX DynamicMBean interface 
      @return MBeanInfo
   */
  public MBeanInfo getMBeanInfo() {

    if (mbeanInfo != null) return mbeanInfo;   // cached result

    SortedSet<String> names = new TreeSet<String>();
        for (String statName : statNames()) {
            names.add(statName);
        }
        MBeanAttributeInfo[] attrs = new MBeanAttributeInfo[names.size() + 1];
        Iterator<String> it = names.iterator();
        attrs[0] = new MBeanAttributeInfo(
                   "Name",
                   "java.lang.String",
                   "CacheStat: " + "Name", // description
                   true,   // isReadable
                   false,  // isWriteable
                   false); // isIs
        for (int i = 1; i < attrs.length; i++) {
          String name = it.next();
          attrs[i] = new MBeanAttributeInfo(
                     name,
                     "java.lang.Long",
                     "CacheStat: " + name, // description
                     true,   // isReadable
                     false,   // isWritable
                     false); // isIs
        }

        mbeanInfo =  
           new MBeanInfo(
                this.getClass().getName(),
                "Property Manager MBean",
                attrs,
                null,  // constructors
                getMBeanOperationInfo(),
                null); // notifications
        return mbeanInfo;
  }

  // Allows an action to be invoked on the Dynamic MBean.
  /** Support for JMX DynamicMBean interface 
      @param actionName name of action
      @param params array of Object parameters
      @param signature array of Strings containing signatures
      @return Object result
      @exception UnsupportedOperationException ReflectionException 
      thrown on errors
   */
  public Object invoke(String actionName, Object[] params, String[] signature)
  throws UnsupportedOperationException, ReflectionException {

// System.out.println("Calling invoke on " + actionName);

    if (actionName.equals("stop")) return stop( );
    else if (actionName.equals("kill")) return kill( );
    else if (actionName.equals("enableHypericTrace")) 
        return enableHypericTrace( );
    else return "Invalid command";
  }

  // Set the value of a specific attribute of the Dynamic MBean.
  // none of our attributes are settable..
  /** Support for JMX DynamicMBean interface 
      @param attribute Attribute
   */
  public void setAttribute(Attribute attribute) 
  throws AttributeNotFoundException {
    throw new AttributeNotFoundException("Can't set attributes for GbjCacheStats");
  }

  // Sets the values of several attributes of the Dynamic MBean.
  /** Support for JMX DynamicMBean interface 
      @param attributes list of attributes
      @return AttributeList 
   */
  public AttributeList setAttributes(AttributeList attributes) {
    // I *want* to do this, but it won't let me..
    // throw new AttributeNotFoundException("Can't set attributes for GbjCacheStats");
    return attributes;
  }


}



