/*
 * GbjJmxCacheStatsMonitor.java
 */

package com.gemstone.gbjstats;

import java.util.*;
import java.io.*;
import java.lang.management.*;
import javax.management.*;
import javax.management.remote.*;
import com.gemstone.gbjgci.*;
import com.gemstone.gbj.*;


/** 

GbjJmxCacheStatsMonitor is a JMX-compatible application that can be used to
monitor and control a GemStone system from a JMX-compatible monitoring tool,
such as jconsole or VMware's Hyperic product.

<p>

GbjJmxCacheStatsMonitor operates like statmonitor, gathering samples of
the associated stone/cache's cache statistics at a regular sampling rate.
But instead of writing them to a file, it instead organizes them as a 
collection of GbjJmxCacheStats (an extension of JMX's DynamicMBeans) that 
are accessible to JMX applications.  
In addition, it provides a way to execute short GS Smalltalk code snippets 
for performing basic system management tasks. To do so, it starts up an
optional gem process (called the "control session" below) to execute the
GS Smalltalk code.

<h2> Executing GbjJmxCacheStatsMonitor </h2>

Before starting GbjJmxCacheStatsMonitor, you must have properly registered
the port you are planning to use.  This is done by executing:

<p>
unix&gt; rmigistry &lt;portNumber&gt;
<p>

To execute the GbjJmxCacheStatsMonitor, do the following:

<p>

java -d64 com.gemstone.gbjstats.GbjJmxCacheStatsMonitor &lt;portNumber&gt; [args]

<p>

where:
<table> 
<caption> " " </caption>
<tr> <td> portNumber:   <td> A valid unused port on this machine
<tr> <td>               <td> (Must have rmiregistry &lt;portNumber&gt; in effect)
<tr>
<tr> <td> args:
<tr> <td> &nbsp;&nbsp; -h
                <td>  -- print help text
<tr> <td> &nbsp;&nbsp; -s &lt;stone&gt;
                <td>  -- Stone name
<tr> <td>       <td> &nbsp;&nbsp;  (defaults to gs64stone)
<tr> <td> &nbsp;&nbsp; -u &lt;user&gt;
                <td>  -- GemStone user for control session
<tr> <td>       <td> &nbsp;&nbsp; if not supplied, control session is not started
<tr> <td>       <td> &nbsp;&nbsp; (User will be prompted for password)
<tr> <td> &nbsp;&nbsp; -g &lt;gemnetid&gt;&nbsp;&nbsp;
                <td>  -- GemNetId
<tr> <td>       <td> &nbsp;&nbsp; (defaults to gemnetobject)
<tr> <td> &nbsp;&nbsp; -i &lt;interval&gt; 
                <td> -- Sampling interval in milliseconds
<tr> <td>       <td> &nbsp;&nbsp; (defaults to 1000)
</table>

<p>

In addition, GbjJmxCacheStatsMonitor can also be used in conjunction with 
the VMware Hyperic system monitoring tool.  In this case, an additional
argument must be added to the command line:

<p>

java -d64 -Dinstallpath=&lt;JDK directory&gt; com.gemstone.gbjstats.GbjJmxCacheStatsMonitor...

<p>

Where &lt;JDK directory&gt; is the directory containing the JDK installation on
this machine.

<h2> Data Layout </h2>

Data from GbjJmxCacheStats will be presented in a layout determined by
the underlying JMX object name format (refer to java JMX documentation
for details).   These use a pattern consistenting of a top-level domain
name followed by a hierarchy of key/value pairs.

<p>

There are two domains defined by GbjJmxCacheStatsMonitor:  

<table>
<caption> " " </caption>
<tr> <td> &#149;&nbsp;&nbsp; 
     <td> GemStone 
     <td> for basic cache stat information for all GemStone processes, and 
<tr> <td> &#149;
     <td> Gemstone/Hyperic &nbsp;&nbsp;
     <td> for a more complex layout designed for access 
          from VMware's Hyperic system monitoring tool.
</table>

<p>

For the GemStone domain, the JMX registered object name format is:

<table>
<caption> " " </caption>
<tr> <td> &#149;&nbsp;&nbsp; 
     <td> GemStone &nbsp;&nbsp;
     <td> (The Domain)
<tr> <td> &#149;
     <td> stone
     <td> The name of the stone
<tr> <td> &#149;
     <td> process
     <td> A particular GemStone process
</table>

<p>

GemStone/Hyperic stats are available at two different levels:  A summary
"System" level and a "Process" level similar to the "GemStone" stats
defined above.

The format used for System-level GemStone/Hyperic stats is as follows:

<table> 
<caption> " " </caption>
<tr> <td> &#149;&nbsp;&nbsp; 
     <td> GemStone/Hyperic &nbsp;&nbsp;
     <td> (The Domain)
<tr> <td> &#149;
     <td> version
     <td> The GemStone version of the stone being monitored
<tr> <td> &#149;
     <td> stone
     <td> The name of the stone
</table>

<p>

Process-level GemStone/Hyperic stats extend from the System-level format
as follows:

<table>
<caption> " " </caption>
<tr> <td> &#149;&nbsp;&nbsp; 
     <td> GemStone/Hyperic &nbsp;&nbsp;
     <td> (The Domain)
<tr> <td> &#149;
     <td> version
     <td> The GemStone version of the stone being monitored
<tr> <td> &#149;
     <td> stone
     <td> The name of the stone
<tr> <td> &#149;
     <td> ptype
     <td> The type of GS process: stone, shpcmon, pgsvr, gem
<tr> <td> &#149;
     <td> process
     <td> The GS process
</table>

<p>

The actual System-level cache stats displayed are configurable by the
user, using a template file that will be described later.

<p>

By default, there are no Process-level GemStone/Hyperic stats displayed.
These need to be manually activated by the user via the "enableHypericTracing"
command (described below).

<p>

Working with a JMX monitoring application, you can navigate through
these hierarchies to find the GS process of interest.  Once you reach
this level, you can either display attributes or execute commands.  In
the jconsole tool, you can display attributes by selecting the "Attributes"
tab in the tool.

<h2> Executing Commands </h2>

Once you navigate through the registration name hierarchy to either the
System-level or Process-level statistics, you can find a list of commands
to execute appropriate for that process/system.  For the jconsole tool, these
are available under the "Operations" tab.

<p>

For Process-level statistics, there are three command available:

<table>
<caption> " " </caption>
<tr> <td> &#149;&nbsp;&nbsp; 
     <td> stop 
     <td> -- Stop the process using the GS Smalltalk stopSession command.
<tr> <td> &#149;&nbsp;&nbsp; 
     <td> kill 
     <td> -- Kill the process using the OS-level "kill -9" command.
<tr> <td> &#149;&nbsp;&nbsp; 
     <td> enableHypericTrace &nbsp; &nbsp;
     <td> -- enable tracing of this process under the GemStone/Hyperic domain.
</table>

<p>

For GemStone/Hyperic System-level statistics, the available commands
are configurable by the user.  By default, the following are available:

<ul>
<li>  Shutdown
<li>  Gslist
<li>  FileSizeReport
<li>  CurrentSessions
<li>  CheckPointStatus
<li>  SessionsRefOldestCr
<li>  GcConfigReport
<li>  ReclaimGcGemReport
<li>  GcVoteState
<li>  Execute
</ul>

Most of these command execute the appropriate GS Smalltalk method with the
same name on class System -- refer to the appropriate GS documentation for
details on these commands.  Shutdown will shutdown the stone and Gslist
execute the OS-level gslist command.  Execute can be used to execute any
GS Smalltalk code entered as an argument into the command.


<h2> Support Files </h2>

As mentioned earlier, the user can configure the selection of System stats
displayed or System-level commands available using 2 template files available
under $GBJ/server:

<ul>
<li> GbjJmxSystemStatsTemplate.dat
<li> GbjJmxCommandsTemplate.dat
</ul>

<h3> GbjJmxSystemStatsTemplate.dat </h3>

After the header information (indicated by lines that start with "#"), the
file contains a line for each statistic to be displayed, with 5 columns of
information.  Only the first column containing the name of the stat is
used by GbjJmxCacheStatsMonitor -- the others are used by the 
GbjHypericPluginGenerator and
deal with how Hyperic displays the information.  Refer to Hyperic
documentation for details.

<p>

Any statistic from either the stone or the shared page cache monitor
may be listed.  GbjJmxCacheStatsMonitor will first check for a stat by
that name in the stone, and if not found will then look for that stat
in the shared page cache monitor.


<h3> GbjJmxCommandsTemplate.dat </h3>

After the header information (indicated by lines that start with "#"),
the file contains clusters of 3 lines, which the 3 lines indicating:

<ul>
<li> The single-word short name to be used for the command
<li> A longer sentence that describes the command
<li> The GS Smalltalk code for the command
</ul>

GbjJmxCacheStatsMonitor only uses the first (single-word command name) and
last (GS Smalltalk code) lines.  The command description is only used
in the Hyperic interface.


 * @version 3.1.1
 */

public class GbjJmxCacheStatsMonitor {


// Internal classes for masking passwords from input
// Taken from:
// http://java.sun.com/developer/technicalArticles/Security/pwordmask/
// with permission

private static class MaskingThread extends Thread {
   private volatile boolean stop;
   private char echochar = '*';

  /**
   *@param prompt The prompt displayed to the user
   */
   public MaskingThread(String prompt) {
      System.out.print(prompt);
   }

  /**
   * Begin masking until asked to stop.
   */
   public void run() {

      int priority = Thread.currentThread().getPriority();
      Thread.currentThread().setPriority(Thread.MAX_PRIORITY);

      try {
         stop = true;
         while(stop) {
           System.out.print("\010" + echochar);
           try {
              // attempt masking at this rate
              Thread.currentThread().sleep(1);
           }catch (InterruptedException iex) {
              Thread.currentThread().interrupt();
              return;
           }
         }
      } finally { // restore the original priority
         Thread.currentThread().setPriority(priority);
      }
   }

  /**
   * Instruct the thread to stop masking.
   */
   public void stopMasking() {
      this.stop = false;
   }
}


private static class PasswordField {

  /**
   *@param input stream to be used (e.g. System.in)
   *@param prompt The prompt to display to the user.
   *@return The password as entered by the user.
   */

   public static final char[] getPassword(InputStream in, String prompt) throws IOException {
      MaskingThread maskingthread = new MaskingThread(prompt);
      Thread thread = new Thread(maskingthread);
      thread.start();
	
      char[] lineBuffer;
      char[] buf;
      int i;

      buf = lineBuffer = new char[128];

      int room = buf.length;
      int offset = 0;
      int c;

      loop:   while (true) {
         switch (c = in.read()) {
            case -1:
            case '\n':
               break loop;

            case '\r':
               int c2 = in.read();
               if ((c2 != '\n') && (c2 != -1)) {
                  if (!(in instanceof PushbackInputStream)) {
                     in = new PushbackInputStream(in);
                  }
                  ((PushbackInputStream)in).unread(c2);
                } else {
                  break loop;
                }

                default:
                   if (--room < 0) {
                      buf = new char[offset + 128];
                      room = buf.length - offset - 1;
                      System.arraycopy(lineBuffer, 0, buf, 0, offset);
                      Arrays.fill(lineBuffer, ' ');
                      lineBuffer = buf;
                   }
                   buf[offset++] = (char) c;
                   break;
         }
      }
      maskingthread.stopMasking();
      if (offset == 0) {
         return null;
      }
      char[] ret = new char[offset];
      System.arraycopy(buf, 0, ret, 0, offset);
      Arrays.fill(buf, ' ');
      return ret;
   }
}

   // slightly modified version of the original main() test program 
   // for the above class
   private static String getPassword(String user ) {
      char password[] = null;
      try {
         password = 
           PasswordField.getPassword(System.in, "Enter password for user " + user + ": ");
      } catch(IOException ioe) {
         ioe.printStackTrace();
      }
      return String.valueOf(password);
   }


// NOW WE START GbjJmxCacheStatsMonitor

    // configurable parameters
    static String GbjSessionName = "GbjJmxCacheStatsMonitor";
    static String StoneName = "gs64stone";
    static String UserName = "DataCurator";
    static String Password = "swordfish";
    static String GemNetName = "gemnetobject";
    static long PortNumber = 0;
    static long CycleTimeMilliseconds = 1000;
    static boolean ControlSession = false;

    static MBeanServer mbs;
    static private Map<Long,ObjectName> nameMap;
    static private Map<Long,GbjJmxCacheStats> statMap;
    static private Map<Long,ObjectName> hypericNameMap;
    static private Map<Long,GbjJmxCacheStats> hypericStatMap;

    protected GbjJmxCacheStatsMonitor( ) { super( ); }

    private static void printHelpText( ) {
        System.out.println("");
        System.out.println("java -d64 com.gemstone.gbjstats.GbjJmxCacheStatsMonitor <portNumber> [args]");
        System.out.println("");
        System.out.println("portNumber:");
        System.out.println("  A valid unused port on this machine");
        System.out.println("  (Must have rmiregistry <portNumber> in effect)");
        System.out.println("");
        System.out.println("args:");
        System.out.println("  -h            -- print help text");
        System.out.println("  -s <stone>    -- Stone name");
        System.out.println("                   (defaults to " + StoneName + ")");
        System.out.println("  -u <user>     -- GemStone user for control session");
        System.out.println("                   if not supplied, control session is not started");
        System.out.println("                   (User will be prompted for password)");
        System.out.println("  -g <gemnetid> -- GemNetId");
        System.out.println("                   (defaults to " + GemNetName + ")");
        System.out.println("  -i <interval> -- Sampling interval in milliseconds");
        System.out.println("                   (defaults to " + CycleTimeMilliseconds + ")");
        System.out.println("");
    }
       
    private static void parseArgs(String[] args) {
        if ((args.length == 0) || (args[0].equals("-h"))) { 
            printHelpText(); System.exit(0); }
        try { PortNumber = Long.valueOf(args[0]); }
        catch (NumberFormatException e) {
            System.out.println("First argument must be a port number");
            System.exit(1);
        }
        for (int i=1; i<args.length; i++) {
            if (args[i].equals("-s")) StoneName = args[++i];
            else if (args[i].equals("-u")) { 
                ControlSession = true; 
                UserName = args[++i];
                Password = getPassword(UserName);
            }
            else if (args[i].equals("-g")) GemNetName = args[++i];
            else if (args[i].equals("-i")) CycleTimeMilliseconds = Long.valueOf(args[++i]);
            else {
                System.out.println("Invalid command line argument: " + args[i]);
                System.exit(1);
            }
        }
    }

        
    private static void registerHypericSystemStats(GbjCacheStats stoneStats,
                                                   GbjCacheStats shrpcStats,
                                                   GbjSession session) {

        // setup system bean
        try {
  	    GbjJmxSystemStats systemMBean = 
                new GbjJmxSystemStats(stoneStats, shrpcStats, session);
            ObjectName mbeanSystemName =
                     new ObjectName(
                            "GemStone/Hyperic" + 
                            ":version=" + GbjGciInterface.GBJGCI_VERSION +
                            ",stone=" + StoneName);
            mbs.registerMBean(systemMBean, mbeanSystemName);
        } catch (Exception e) {
            System.err.println("ERROR: Problem in registerHypericSystemStats()");
            e.printStackTrace( );
        }
    }

    protected static void registerHypericCacheStats(
                           GbjCacheStats stats,
                           GbjSession session,
                           long pid) {
        try {
            // Create the stats instance and register/map it
            ObjectName mbeanName =
                new ObjectName(
                    "GemStone/Hyperic" + 
                    ":version=" + GbjGciInterface.GBJGCI_VERSION +
                    ",stone=" + StoneName + 
                    ",ptype=" + stats.cacheTypeString() + 
                    ",process="  + stats.name);
            GbjJmxCacheStats gjcStats =
                new GbjJmxCacheStats(stats, session);
	    mbs.registerMBean(gjcStats, mbeanName);
            hypericNameMap.put(pid, mbeanName);
            hypericStatMap.put(pid, gjcStats);
        } catch (Exception e) {
            System.err.println("ERROR: Problem in registerHypericCacheStats()");
            e.printStackTrace( );
        }
    }

    private static void registerCacheStats(
                           GbjCacheStats stats,
                           GbjSession session,
                           long pid) {

        try {
            // Create the stats instance and register/map it
            ObjectName mbeanName =
                new ObjectName(
                    "GemStone" + 
                 // ":version=" + GbjGciInterface.GBJGCI_VERSION +
                 // ":type=Process" +
                    ":stone=" + StoneName + 
                 // ",ptype=" + stats.cacheTypeString() + 
                    ",process="  + stats.name);
            GbjJmxCacheStats gjcStats =
                new GbjJmxCacheStats(stats, session);
            mbs.registerMBean(gjcStats, mbeanName);
            nameMap.put(pid, mbeanName);
            statMap.put(pid, gjcStats);
        } catch (Exception e) {
            System.err.println("ERROR: Problem in registerCacheStats()");
            e.printStackTrace( );
        }
    }

    private static void unregisterCacheStats(long pid) {

        try {
            mbs.unregisterMBean(nameMap.get(pid));
        } catch (Exception e) { 
            System.out.println("ERROR: Problem in unregisterCacheStats()");
            e.printStackTrace( );
        }
        nameMap.remove(pid);
        statMap.remove(pid);
        try { 
            mbs.unregisterMBean(hypericNameMap.get(pid));
            hypericNameMap.remove(pid);
            hypericStatMap.remove(pid);
        } catch (Exception e) { }
    }


    /** Execute the GbjJmxCacheStatsMonitor program. 
	@param args array of String containing arguments for progrem
     */
    public static void main(String[] args) {

        Boolean systemStatsRegistered = false;
        GbjCacheStats stoneStats = null;
        GbjCacheStats shrpcStats = null;

	// parse the command line arguments
        parseArgs(args);

        // Get the Platform MBean Server
        mbs = ManagementFactory.getPlatformMBeanServer();

        // create a RMI connector server so that operations 
        // can be performed on the MBeans remotely
        // remember to run "rmiregistry <portNum> on localhost 
        // to configure this...

        String urlString = "service:jmx:rmi:///jndi/rmi://localhost:" +
                         PortNumber + "/jmxrmi";

       JMXServiceURL url;
       JMXConnectorServer cs = null;
        try {
            url = new JMXServiceURL(urlString); 
            cs = 
              JMXConnectorServerFactory.newJMXConnectorServer(url, null, mbs); 
        } catch (Exception e) {
            System.out.println("ERROR: Problem with JMX connector setup");
            e.printStackTrace( );
            System.exit(1);
        }

        // Initialize GBJ
        GbjSession.initialize(false, GbjSessionName, 10000, 200, 1000);

        // Setup GBJ session if control desired
        // use manual transaction mode and setup sigabort handler
        GbjSession session = null;
        if (ControlSession) {
	    GbjParameters p = new GbjParameters();
	    p.serverName = StoneName;
	    p.userName = UserName;
	    p.password = Password;
	    p.gemnetName = GemNetName;
	    p.transactionMode = GbjParameters.ManualTransactions;
	    session = new GbjSession(p);
	    session.connect();
            session.execute(
              " Exception " +
              "   category: GemStoneError " + 
              "   number: 6009 " + 
              "   do: [:ex:cat:num:args | " +
              "     System abortTransaction. " + 
              "     System enableSignaledAbortError ]. " +
              " System enableSignaledAbortError. "
              );
        }

        // Configure process id tracking lists and maps
        int maxSessions = 0;
        try {
            maxSessions = GbjCacheStats.attach(StoneName);
        } catch (GbjGciException e) {
            System.out.println("ERROR: problem attaching to stone " + StoneName);
            System.exit(1);
        }
        List<Long> newPids = new ArrayList<Long>(maxSessions);
        List<Long> oldPids = new ArrayList<Long>(maxSessions);
        List<Long> deadPids = new ArrayList<Long>(maxSessions);
        nameMap = new HashMap<Long,ObjectName>(maxSessions);
        statMap = new HashMap<Long,GbjJmxCacheStats>(maxSessions);
        hypericNameMap = new HashMap<Long,ObjectName>(maxSessions);
        hypericStatMap = new HashMap<Long,GbjJmxCacheStats>(maxSessions);

/* SKIP FOR NOW...     
        try {
            // fork rmiregistry execution
            Runtime.getRuntime().exec("rmiregistry " + PortNumber);
        } catch (IOException e) {
            System.out.println("ERROR: Problem forking rmiregistry");
            System.out.println("Confirm that java executable rmiregistry");
            System.out.println("can be run by this user.");
            System.exit(1);
        }
*/
        try {
            cs.start(); 
        } catch (IOException e) {
            System.out.println("ERROR: Problem with port " + PortNumber);
            System.exit(1);
        }

	// loop refreshing stats
        while (true) {

            // check the time and sample the stats
            long startTime = System.currentTimeMillis();
            GbjCacheStats.sample(newPids, oldPids, deadPids);

	    // add stats for new sessions
            for (long pid : newPids) {
                GbjCacheStats stats = GbjCacheStats.fetch(pid);
                if (stats.type <= GbjGciCacheStats.TYPE_GEM) {
                    registerCacheStats(stats, session, pid);
                    if (stats.type == GbjGciCacheStats.TYPE_STONE) 
                        stoneStats = stats;
                    if (stats.type == GbjGciCacheStats.TYPE_SPCMONITOR)
                        shrpcStats = stats;
                    if (!systemStatsRegistered &&
                        (stoneStats != null) && 
                        (shrpcStats != null)) {
                        systemStatsRegistered = true;
                        registerHypericSystemStats(stoneStats, shrpcStats, 
                                                   session);
                    }
                }
            }

            // remove stats for dead sessions
            for (long pid : deadPids) {
                unregisterCacheStats(pid);
            }

            // update the on-going old sessions
            for (long pid : oldPids) {
                statMap.get(pid).stats.update();
            }

            // See how long the above took us, and sleep for the remainder of the cycle time
            long workTime = System.currentTimeMillis() - startTime;
            long sleepTime = CycleTimeMilliseconds - workTime;
            if (sleepTime < 0) 
                System.out.println(
                    "WARNING: Sample interval too short -- calculations take longer!");
            else {
                try {
                    Thread.sleep(CycleTimeMilliseconds - workTime);
                }  catch (Exception e) { }
            }
        }
    }
}
