/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjJmxDataFile.java  
 *
 *========================================================================
 */

// Stuff for reading GbjJmx data files

package com.gemstone.gbjstats;
import java.io.*;

/** Helper class used by GbjJmxSystemStatInfo and GbjJmxCommandInfo to
    open and read template data files held in $GBJ/server.
    <p>
    No public protocol available.
 */

public class GbjJmxDataFile {

    private BufferedReader reader;
    private String regex;
    private String[] tokens;
    private int nextToken;

    // files are expected to be in directory $GBJ/server
    protected GbjJmxDataFile(String fileName) 
           throws FileNotFoundException {
        String fullName = System.getenv("GBJ") + "/server/" + fileName;
        reader = new BufferedReader(new FileReader(fullName));
        regex = "\\s+";
        nextToken = 0;
        tokens = null;
    }

    protected GbjJmxDataFile(String fileName, String r)
           throws FileNotFoundException {
        reader = new BufferedReader(new FileReader(fileName));
        regex = r;
        nextToken = 0;
        tokens = null;
    }

    // read a line from the data file, skipping over blank and #comment lines
    // return null on EOF
    protected String readLine( ) {
        String line;
        do {
            try {
                line = reader.readLine( );
            } catch (IOException e) { return null; }
            if (line == null) { 
                return null; }
            if (line.length() == 0) continue;    // skip blank lines
            if (line.charAt(0) == '#') continue; // skip comment lines
            tokens = line.split(regex);
            nextToken = 0;
            return line;
        } while (true);  // we'll eventually hit EOF..
    }

    // read entire template file, after skipping over 
    // preliminary lines that start with "#"
    protected String readTemplate() {
        StringBuilder contents = new StringBuilder( );
        String line;
        // skip over beginning comment lines
        do {
            try {
                line = reader.readLine( );
            } catch (IOException e) { return null; }
            if (line == null) { 
                return null; }
        } while (line.charAt(0) == '#');
        // now read rest of file
        do {
            contents.append(line);
            try {
                line = reader.readLine( );
            } catch (IOException e) { break; }
            if (line == null) { break; }
            contents.append("\n");
        } while (true);  // we'll eventually hit EOF..
        return contents.toString();
    }

    protected void close( ) {
        try { reader.close( ); }
        catch (Exception e) { }
    }

    // return "words" from lastLine read (separated by blank space)
    protected String nextToken( ) {
        if (tokens == null) return null;
        if (nextToken >= tokens.length) return null; 
        return tokens[nextToken++];
    }

    /** Developer test/debug program 
	@param args array of Strings containing arguments for program
     */
    public static void main(String[] args) {
        GbjJmxDataFile file = null;
        try {
            file = new GbjJmxDataFile(args[0]);
        } catch (Exception e) { 
            System.out.println("Cant open file " + args[0]);
            e.printStackTrace();
            System.exit(1); }
        for (int i = 1; ; i++) {
            String line = null;
            try { line = file.readLine( ); 
            } catch (Exception e) { System.exit(1); }
            if (line == null) { break; }
            System.out.println("Line " + i + ": " + line);
            String token;
            for (int j = 1; ; j++) {
                if ((token = file.nextToken()) == null) break;
                System.out.println("   " + j + ": " + token);
            }
        }
        file.close();
        try {
            file = new GbjJmxDataFile(args[0]);
        } catch (Exception e) { 
            System.out.println("Cant open file " + args[0]);
            e.printStackTrace();
            System.exit(1); }
        System.out.println("readTemplate:");
        System.out.println(file.readTemplate());
        System.out.println("END");
        file.close();
    }
}



