/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 *      GbjJmxSystemStats.java  
 *
 *========================================================================
 */

package com.gemstone.gbjstats;
import com.gemstone.gbj.*;
import java.util.*;
import javax.management.*;

/** GbjJmxSystemStats provides System-level statistics for a GemStone system.
    <p>
    The actual statistics included is user-configurable via the
    GbjJmxSystemStats[Template].dat file.
 */
public class GbjJmxSystemStats extends GbjJmxCacheStats {

    private GbjJmxCommandInfo[] commandInfo = null;

    private GbjCacheStats shrpcStats;
  

    /** GbjJmxSystemStats constructor 
	@param s GbjCacheStats
	@param shrpc GbjCacheStats for shared page cache monitor
	@param session GbjSession
     */
    public GbjJmxSystemStats(GbjCacheStats s, 
                             GbjCacheStats shrpc,
                             GbjSession session) {

      super(s, session);
      shrpcStats = shrpc;
      commandInfo = GbjJmxCommandInfo.readCommandInfo();
    }

    protected static String[] readSystemStatInfo( ) {
        if (statNames == null) {
            // retrieve from the GbjJmxSystemStats.dat file
            GbjJmxSystemStatInfo[] statInfo = 
                GbjJmxSystemStatInfo.readSystemStatInfo( );
            statNames = new String[statInfo.length];
            for (int i = 0; i < statInfo.length; i++) 
               statNames[i] = statInfo[i].name;
        }
        return statNames;
    }



    // Operations

    /** JMX operation to execute a GS Smalltalk command. 
	@param command String command to execute
	@return String result from command execution
     */
    public String execute(String command) {
        if (session == null) {
            return "Control session not running";
        } else {
            try {
                return session.printit(command);
            } catch (Exception e) { 
                return "Error executing ST command: " + command;
            }
         }
     }

    /** JMX operation to shutdown the stone. 
	@return String result from stone shutdown
     */
    public String shutdown() {
        try {
          rt.exec("stopstone -I " + stats.name);
        } catch (Exception e) { return "Problem stopping stone" + stats.name; }
        return "Stone shutdown";
    }

  /** JMX DynamicMBean protocol for invoking an action.  
      @param actionName name of action to perform
      @param params array of parameters for action
      @param signature array of signatures for action
      @return Object result
      @exception UnsupportedOperationException ReflectionException
        thrown on errors
   */
  public Object invoke(String actionName, Object[] params, String[] signature)
  throws UnsupportedOperationException, ReflectionException {

// System.out.println("Calling invoke on " + actionName);

    if (actionName.equals("Shutdown")) return shutdown( );
    else if (actionName.equals("Execute")) {
        String cmd = (String)params[0];
        return execute(cmd);
    } else {
        for (int i=0; i<commandInfo.length; i++) {
            if (actionName.equals(commandInfo[i].name)) {
                return execute(commandInfo[i].code);
            }
        }
    }
    return "Invalid command";
  }


    // we'll pick up the stat names from the GbjJmxSystemStats.dat file
    private static String[] statNames;

    protected String[] statNames( ) {
        if (statNames == null) {
            // retrieve from the GbjJmxSystemStats.dat file
            GbjJmxDataFile file = null;
            try {
                file = new GbjJmxDataFile("GbjJmxSystemStats.dat");
            } catch (java.io.FileNotFoundException e) {
                try {
                    file = new GbjJmxDataFile("GbjJmxSystemStatsTemplate.dat");
                } catch (java.io.FileNotFoundException e2) {
                    System.err.println("Problem finding GbjJmxSystemStats.dat");
                }
            }
            ArrayList<String> stats = new ArrayList<String>( );
            while (file.readLine( ) != null) {
                String stat = file.nextToken();
                stats.add(stat);              
            }
            try { file.close( ); } catch (Exception e) { }
            statNames = new String[stats.size()];
            for (int i = 0; i < stats.size(); i++) 
               statNames[i] = stats.get(i);
        }
        return statNames;
    }

  private MBeanOperationInfo[] mbeanOperationInfo = null;

  protected MBeanOperationInfo[] getMBeanOperationInfo( ) {
    if (mbeanOperationInfo != null) return mbeanOperationInfo;

    MBeanOperationInfo[] opers = 
        new MBeanOperationInfo[commandInfo.length + 2];
    opers[0] =  new MBeanOperationInfo(
                    "Shutdown",
                    "Shutdown this GemStone database",
                    null,   // no parameters
                    "String",
                    MBeanOperationInfo.ACTION);

    opers[commandInfo.length + 1] =  
                new MBeanOperationInfo(
                    "Execute",
                    "Execute a smalltalk command on control session",
                    new MBeanParameterInfo[]{
                        new MBeanParameterInfo("command", 
                                               "java.lang.String", 
                                               "Command to execute") },
                    "String",
                    MBeanOperationInfo.ACTION);
    for (int i = 0; i < commandInfo.length; i++) {
        opers[i+1] = new MBeanOperationInfo(
                         commandInfo[i].name,
                         commandInfo[i].description,
                         null, // no parameters
                         "String",
                         MBeanOperationInfo.ACTION);
    };
    mbeanOperationInfo = opers;
    return opers;
  }

  protected long getStats(String name) {
    try { 
      return stats.get(name);
    } catch (IllegalArgumentException e) { 
        return shrpcStats.get(name); }
  }

    // operations (overriding GbjJmxCacheStats)

    // attributes (overriding GbjJmxCacheStats)

}



