/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 * Name - GbxToolsSupportImpl.java
 *
 *========================================================================
 */

package com.gemstone.tools;

import com.gemstone.gbjgci.GbjGciInterface;
import com.gemstone.gbj.GbjObject;
import com.gemstone.gbj.GbjCollection;
import com.gemstone.gbj.GbjIterator;
import com.gemstone.gbj.GbjSession;
import com.gemstone.gbj.GbjException;

import java.util.Vector;
import java.util.HashMap;   // GBJ30: adds a classInfo cache

class GbxToolsSupportImpl extends GbjObject implements GbxToolsSupport
{

private GbjObject serverObj; // added for v3.2.0

// GBJ30:  add classInfoTable cache to hold classInfo data
private HashMap<GbjObject, GbjObject[]> classInfoTable = new HashMap<GbjObject, GbjObject[]>();


public GbxToolsSupportImpl(GbjSession aSession) throws GbjException
{
  super();
  serverObj = aSession.doit("GbjToolsInterface new");
}

public GbjObject classFor_(GbjObject aClass) throws GbjException
{
	return serverObj.sendMsg("classFor:", aClass);
}

public GbjObject classNameFor_(GbjObject anObject) throws GbjException
{
	return serverObj.sendMsg("classNameFor:", anObject);
}

public String fileOutOfClass_(GbjObject aClass) throws GbjException
{
	return serverObj.sendMsg("fileOutClass:", aClass).stringValue();
}

public String fileOutOfMethod_inClass_(String aSelector, GbjObject aClass) throws GbjException
{
	return serverObj.sendMsg("fileOutMethod:", aSelector, "inClass:", aClass).stringValue();
}

public String fileOutOfDictionary_(GbjObject aDict) throws GbjException
{
	return serverObj.sendMsg("fileOutDict:", aDict).stringValue();
}

public int remoteSizeOf_(GbjObject anObject) throws GbjException
{
	return serverObj.sendMsg("sizeOf:", anObject).intValue();
}

public GbjObject classNamed_(String aClassName) throws GbjException
{
	return serverObj.sendMsg("classNamed:", aClassName);
}

public void moveClass_fromDict_toDict_(GbjObject aClass, GbjObject fromDict, GbjObject toDict) throws GbjException
{
    // GBJ30: clear the classInfo cache for the 2 affected dictionaries
    classInfoTable.put(fromDict, null);
    classInfoTable.put(toDict, null);

	serverObj.sendMsg("moveClass:", aClass, "fromDict:", fromDict, "toDict:", toDict);
}

public String [] dictionaryNames() throws GbjException
{
	GbjCollection aCollection;

	aCollection = (GbjCollection) serverObj.sendMsg("dictionaryNames");
	return stringContentsOf_(aCollection);
}

public boolean isDictionary_(GbjObject anObject) throws GbjException
{
    GbjObject dict = serverObj.getSession().objectNamed("SymbolDictionary");
    return serverObj.sendMsg("isObject:", anObject, "aKindOf:", dict).booleanValue();
}

public boolean collection_includes_(GbjObject anObject, GbjObject element) throws GbjException
{
	return serverObj.sendMsg("doesCollection:", anObject, "include:", element).booleanValue();
}

public void removeClass_fromDict_(GbjObject aClass, GbjObject aDict) throws GbjException
{
    // GBJ30: clear the classInfo cache for the affected dictionary
    classInfoTable.put(aDict, null);

	serverObj.sendMsg("removeClass:", aClass, "fromDict:", aDict);
}

public void addProtocol_toClass_(String newProtocol, GbjObject aClass) throws GbjException
{
	serverObj.sendMsg("addProtocol:", newProtocol, "toClass:", aClass);
}

public void renameProtocol_to_inClass_(String oldName, String newName, GbjObject aClass) throws GbjException
{
	    serverObj.sendMsg("renameProtocol:", oldName, "to:", newName, "class:", aClass);
}

public void removeProtocol_fromClass_(String aProtocol, GbjObject aClass) throws GbjException
{
	serverObj.sendMsg("removeProtocol:", aProtocol, "fromClass:", aClass);
}

public String [] selectorsForClass_(GbjObject aClass) throws GbjException
{
	GbjCollection aProxy;

	aProxy = (GbjCollection) serverObj.sendMsg("selectorsForClass:", aClass);
	return stringContentsOf_(aProxy);
}

public String protocolForSelector_inClass_(String aSelector, GbjObject aClass) throws GbjException
{
	return serverObj.sendMsg("protocolOfSelector:", aSelector, "inClass:", aClass).stringValue();
}

public GbxMethodHolder [] methodHoldersForImplementorsOf_(String selector) throws GbjException
{
	GbxMethodHolder [] methods;
	GbjObject [] anArray;
	GbjObject method;
	String aSelector;
	GbjObject aClass;
	String className;
	String printString;
	String category;
	int position;

	anArray = proxiesForMessage_arg_("implementorInfoForSelector:", selector);
	methods = new GbxMethodHolder[anArray.length / 6];
	for (int i = 0; i < anArray.length; i += 6) {
		method = anArray[i];
		aSelector = anArray[i + 1].stringValue();
		aClass = anArray[i + 2];
		className = anArray[i + 3].stringValue();
		printString = anArray[i + 4].stringValue();
		category = anArray[i + 5].stringValue();
		methods[i / 6] = new GbxMethodHolder(aSelector, method, aClass, className, printString, category, -1, null);
	}
	return methods;
}

public GbxMethodHolder [] methodHoldersForSendersOf_(String selector) throws GbjException
{
	GbxMethodHolder [] methods;
	GbjObject [] anArray;
	GbjObject method;
	String aSelector;
	GbjObject aClass;
	String className;
	String printString;
	String category;
	int position;

	anArray = proxiesForMessage_arg_("senderInfoForSelector:", selector);
	methods = new GbxMethodHolder[anArray.length / 7];
	for (int i = 0; i < anArray.length; i += 7) {
		method = anArray[i];
		aSelector = anArray[i + 1].stringValue();
		aClass = anArray[i + 2];
		className = anArray[i + 3].stringValue();
		printString = anArray[i + 4].stringValue();
		category = anArray[i + 5].stringValue();
		position = anArray[i + 6].intValue();
		methods[i / 7] = new GbxMethodHolder(aSelector, method, aClass, className, printString, category, position, null);
	}
	return methods;
}

public GbxMethodHolder [] methodHoldersForSubstring_(String substring) throws GbjException
{
	GbxMethodHolder [] methods;
	GbjObject [] anArray;
	GbjObject method;
	String aSelector;
	GbjObject aClass;
	String className;
	String printString;
	String category;
	int position;

	anArray = proxiesForMessage_arg_("substringSearch:", substring);
	methods = new GbxMethodHolder[anArray.length / 7];
	for (int i = 0; i < anArray.length; i += 7) {
		method = anArray[i];
		aSelector = anArray[i + 1].stringValue();
		aClass = anArray[i + 2];
		className = anArray[i + 3].stringValue();
		printString = anArray[i + 4].stringValue();
		category = anArray[i + 5].stringValue();
		position = anArray[i + 6].intValue();
		methods[i / 7] = new GbxMethodHolder(aSelector, method, aClass, className, printString, category, position, null);
	}
	return methods;
}

public GbxMethodHolder [] methodHoldersForReferencesTo_(String aSymbol) throws GbjException
{
	GbxMethodHolder [] methods;
	GbjObject [] anArray;
	GbjObject method;
	String aSelector;
	GbjObject aClass;
	String className;
	String printString;
	String category;
	int position;

	anArray = proxiesForMessage_arg_("referencesTo:", aSymbol);
	methods = new GbxMethodHolder[anArray.length / 7];
	for (int i = 0; i < anArray.length; i += 7) {
		method = anArray[i];
		aSelector = anArray[i + 1].stringValue();
		aClass = anArray[i + 2];
		className = anArray[i + 3].stringValue();
		printString = anArray[i + 4].stringValue();
		category = anArray[i + 5].stringValue();
		position = anArray[i + 6].intValue();
		methods[i / 7] = new GbxMethodHolder(aSelector, method, aClass, className, printString, category, position, null);
	}
	return methods;
}

public GbxMethodHolder [] methodHoldersForRefsToInstVar_inClass_(String ivName, GbjObject aProxy) throws GbjException
{
	GbxMethodHolder [] methods;
	GbjObject [] anArray;
	GbjObject method;
	String aSelector;
	GbjObject aClass;
	String className;
	String printString;
	String category;
	int position;

	anArray = proxiesForMessage_arg_message_arg_("referencesToInstVar:", ivName, "startingClass:", aProxy);
	methods = new GbxMethodHolder[anArray.length / 6];
	for (int i = 0; i < anArray.length; i += 6) {
		method = anArray[i];
		aSelector = anArray[i + 1].stringValue();
		aClass = anArray[i + 2];
		className = anArray[i + 3].stringValue();
		printString = anArray[i + 4].stringValue();
		category = anArray[i + 5].stringValue();
		methods[i / 6] = new GbxMethodHolder(aSelector, method, aClass, className, printString, category, -1, null);
	}
	return methods;
}

public GbxMethodHolder [] methodHoldersForRefsToClassVar_inClass_(String cvName, GbjObject aProxy) throws GbjException
{
	GbxMethodHolder [] methods;
	GbjObject [] anArray;
	GbjObject method;
	String aSelector;
	GbjObject aClass;
	String className;
	String printString;
	String category;
	int position;

	anArray = proxiesForMessage_arg_message_arg_("referencesToClassVar:", cvName, "startingClass:", aProxy);
	methods = new GbxMethodHolder[anArray.length / 6];
	for (int i = 0; i < anArray.length; i += 6) {
		method = anArray[i];
		aSelector = anArray[i + 1].stringValue();
		aClass = anArray[i + 2];
		className = anArray[i + 3].stringValue();
		printString = anArray[i + 4].stringValue();
		category = anArray[i + 5].stringValue();
		methods[i / 6] = new GbxMethodHolder(aSelector, method, aClass, className, printString, category, -1, null);
	}
	return methods;
}

public GbxMethodHolder [] methodHoldersForRefsToClass_(GbjObject aProxy) throws GbjException
{
	GbxMethodHolder [] methods;
	GbjObject [] anArray;
	GbjObject method;
	String aSelector;
	GbjObject aClass;
	String className;
	String printString;
	String category;
	int position;

	anArray = proxiesForMessage_arg_("referencesToClass:", aProxy);
	methods = new GbxMethodHolder[anArray.length / 6];
	for (int i = 0; i < anArray.length; i += 6) {
		method = anArray[i];
		aSelector = anArray[i + 1].stringValue();
		aClass = anArray[i + 2];
		className = anArray[i + 3].stringValue();
		printString = anArray[i + 4].stringValue();
		category = anArray[i + 5].stringValue();
		methods[i / 6] = new GbxMethodHolder(aSelector, method, aClass, className, printString, category, -1, null);
	}
	return methods;
}

public String [] messagesInMethod_inClass_(String aSelector, GbjObject aClass) throws GbjException
{
	GbjCollection selectorsProxy;

	selectorsProxy = (GbjCollection) serverObj.sendMsg("messagesIn:", aSelector, "class:", aClass);
	return stringContentsOf_(selectorsProxy);
}

public void moveMethod_to_inClass_(String aSelector, String newProtocol, GbjObject aClass) throws GbjException
{
	serverObj.sendMsg("moveMethod:", aSelector, "to:", newProtocol, "inClass:", aClass);
}

public void removeMethod_fromClass_(String aSelector, GbjObject aClass) throws GbjException
{
	serverObj.sendMsg("removeSelector:", aSelector, "fromClass:", aClass);
}


public GbjObject [] acceptClassText_oldClass_(String aString, GbjObject oldClass) throws GbjException
{
	GbjObject infoProxy;

	// GBJ30: clear the entire classInfo cache,
	// since we can't really tell what cached dictionary is affected
	classInfoTable = new HashMap<GbjObject, GbjObject[]>();

	infoProxy = serverObj.sendMsg("acceptClassText:", aString, "oldClass:", oldClass);
	return new GbjIterator(infoProxy).remainingElements();
}

public String commentForClass_(GbjObject aClass) throws GbjException
{
	return serverObj.sendMsg("commentForClass:", aClass).stringValue();
}


public GbxClassHolder [] classHoldersFor_(GbjObject anObject) throws GbjException
{
	GbxClassHolder [] holders;
	GbjObject [] classInfoProxies;
	GbjCollection classNamesProxy;  // 40314
	GbjObject classProxy, metaclassProxy;
	String className;
	String [] classNames;  // 40314
	int classVersion, classSize;
	int structSize = 5;

	// GBJ30: modified to use new classInfoTable

        // 40314: gather up the class name info using the more efficent
        // stringContentsOf_()  (which now uses packed strings)
	classNamesProxy = 
	   (GbjCollection)serverObj.sendMsg("classNamesForDictionary:", anObject);
	classNames = stringContentsOf_(classNamesProxy);

	classInfoProxies = classInfoTable.get(anObject);
	if (classInfoProxies == null) {
	    classInfoProxies = 
		proxiesForMessage_arg_("classInfoForDictionary:", anObject);
	    classInfoTable.put(anObject, classInfoProxies);
	}

	holders = new GbxClassHolder[classInfoProxies.length / structSize];
	for (int i = 0; i < classInfoProxies.length; i += structSize) {
		classProxy = classInfoProxies[i];
		// 40314: get get className from the earlier collection
		className = classNames[i / structSize ];
		classVersion = classInfoProxies[i + 2].intValue();
		classSize = classInfoProxies[i + 3].intValue();
		metaclassProxy = classInfoProxies[i + 4];
		holders[i / structSize] = new GbxClassHolder(className, classProxy, classVersion, classSize, metaclassProxy);
	}
	return holders;
}

public GbxDictionaryHolder [] dictionaryHolders() throws GbjException
{
	GbxDictionaryHolder [] holders;
	GbjObject dictsProxy;
	GbjObject [] dictProxies;
	GbjObject dictProxy;
	String dictName;

	dictProxies = proxiesForMessage_("dictionaryInfo");
	holders = new GbxDictionaryHolder[dictProxies.length / 2];
	for (int i = 0; i < dictProxies.length; i += 2) {
		dictProxy = dictProxies[i];
		dictName = dictProxies[i + 1].stringValue();
		holders[i / 2] = new GbxDictionaryHolder(dictName, dictProxy);
	}
	return holders;
}

public GbxMethodHolder [] methodHoldersFor_inProtocol_(GbjObject anObject, String aProtocol) throws GbjException
{
	GbjObject [] selectorProxies;
	GbxMethodHolder [] holders;
	String selectorName;

	selectorProxies = proxiesForMessage_arg_message_arg_("selectorsIn:", aProtocol, "class:", anObject);
	holders = new GbxMethodHolder[selectorProxies.length];
	for (int i = 0; i < selectorProxies.length; i++) {
		selectorName = selectorProxies[i].stringValue();
		holders[i] = new GbxMethodHolder(selectorName, null);
	}
	return holders;
}

public GbxMethodHolder [] methodHoldersFor_inProtocol_toLevel_(GbjObject anObject, String aProtocol, int aLevel) throws GbjException
{
	GbxMethodHolder [] methods;
	GbjObject [] anArray;
	GbjObject method;
	String aSelector;
	GbjObject aClass;
	String className;
	String printString;
	String category;

	anArray = proxiesForMessage_arg_message_arg_message_arg_("methodInfoForClass:", anObject, "inProtocol:", aProtocol, "toLevel:", Integer.valueOf(aLevel));
	methods = new GbxMethodHolder[anArray.length / 6];
	for (int i = 0; i < anArray.length; i += 6) {
		method = anArray[i];
		aSelector = anArray[i + 1].stringValue();
		aClass = anArray[i + 2];
		className = anArray[i + 3].stringValue();
		printString = anArray[i + 4].stringValue();
		category = anArray[i + 5].stringValue();
		methods[i / 6] = new GbxMethodHolder(aSelector, method, aClass, className, printString, category, -1, null);
	}
	return methods;
}

public String [] protocolHoldersFor_(GbjObject anObject) throws GbjException
{
	GbjObject [] protocolProxies;
	String [] holders;
	String protocolName;

	protocolProxies = proxiesForMessage_arg_("protocolNamesInClass:", anObject);
	holders = new String[protocolProxies.length];
	for (int i = 0; i < protocolProxies.length; i++) {
		protocolName = protocolProxies[i].stringValue();
		holders[i] = protocolName;
	}
	return holders;
}

public String [] protocolHoldersFor_toLevel_(GbjObject anObject, int aLevel) throws GbjException
{
	GbjObject protocolsProxy;
	GbjObject [] protocolProxies;
	String [] holders;
	String protocolName;

	protocolProxies = proxiesForMessage_arg_message_arg_("protocolNamesInClass:", anObject, "toLevel:", Integer.valueOf(aLevel));
	holders = new String[protocolProxies.length];
	for (int i = 0; i < protocolProxies.length; i++) {
		protocolName = protocolProxies[i].stringValue();
		holders[i] = protocolName;
	}
	return holders;
}

public GbxDictionaryHolder dictionaryHolderFor_(GbjObject anObject) throws GbjException
{
	//Return the symbol list dictionary for <aClassProxy>."

	GbjObject arr;
	GbjObject [] elements;
	GbjObject dictProxy;
	String dictName;

	elements = proxiesForMessage_arg_("dictionaryInfoForClass:", anObject);
	if (elements.length == 0) {
		return null;
	} else {
		dictName = elements[0].stringValue();
		dictProxy = elements[1];
		return new GbxDictionaryHolder(dictName, dictProxy);
	}
}

public GbjObject addDict_at_(String aString, int index) throws GbjException
{
	return serverObj.sendMsg("addDict:", aString, "at:", Integer.valueOf(index));
}

public boolean renameDictionary_as_(GbjObject aDictionary, String aString) throws GbjException
{
	return serverObj.sendMsg("renameDict:", aDictionary, "as:", aString).booleanValue();
}

public void removeDictionary_(GbjObject aProxy) throws GbjException
{
	serverObj.sendMsg("removeDict:", aProxy);
}

public String classDefinitionFor_(GbjObject anObject) throws GbjException
{
	return serverObj.sendMsg("definitionForClass:" , anObject).stringValue();
}

public String sourceCodeAt_inClass_ (String selector, GbjObject aClass) throws GbjException
{
	return serverObj.sendMsg("sourceCodeAt:", selector, "inClass:", aClass).stringValue();
}

public Vector compile_classified_class_(String source, String aProtocol, GbjObject aClass) throws GbjException
{
	Vector aVector = new Vector();
	GbjObject [] result;
	String errorText;
	int errorIndex;
	boolean noError;

	result = proxiesForMessage_arg_message_arg_message_arg_("compile:", source, "class:" , aClass, "category:", aProtocol);
	noError = result[0].booleanValue();
	aVector.addElement(Boolean.valueOf(noError));
	if (noError) {
		aVector.addElement(result[1]);
		aVector.addElement(result[2].stringValue());
		aVector.addElement(result[3].stringValue());
	} else {
		aVector.addElement(result[1].stringValue());
		aVector.addElement(Integer.valueOf(result[2].intValue()));
	}
	return aVector;
}

public GbxClassHolder [] classHierarchyHoldersFor_(GbjObject anObject) throws GbjException
{
	GbxClassHolder [] holders;
	GbjObject classInfoProxy;
	GbjObject [] classInfoProxies;
	GbjObject classProxy, metaclassProxy;
	String className;
	int classVersion, classSize, indentLevel;
	String ivNames;
	int structSize = 7;

	classInfoProxies = proxiesForMessage_arg_("hierarchyListForClass:", anObject);
	holders = new GbxClassHolder[classInfoProxies.length / structSize];
	for (int i = 0; i < classInfoProxies.length; i += structSize) {
		classProxy = classInfoProxies[i];
		className = classInfoProxies[i + 1].stringValue();
		classVersion = classInfoProxies[i + 2].intValue();
		classSize = classInfoProxies[i + 3].intValue();
		indentLevel = classInfoProxies[i + 4].intValue();
		ivNames = classInfoProxies[i + 5].stringValue();
		metaclassProxy = classInfoProxies[i + 6];
		holders[i / structSize] = new GbxClassHolder(className, classProxy, classVersion, classSize, indentLevel, ivNames, false, metaclassProxy);
	}
	return holders;
}

public GbjObject classHistoryFor_(GbjObject aClass) throws GbjException
{
	return serverObj.sendMsg("classHistoryFor:", aClass);
}

public GbxDictionaryHolder dictionaryForClass_(GbjObject aClassProxy) throws GbjException
{
	GbjObject [] info;

	info = proxiesForMessage_arg_("dictionaryInfoForClass:", aClassProxy);
	if (info.length == 0) {
		return null;
	} else {
		return new GbxDictionaryHolder(info[0].stringValue(), info[1]);
	}
}

public GbxClassHolder [] classVersionHoldersFor_(GbjObject anObject) throws GbjException
{
	GbxClassHolder [] holders;
	GbjObject [] classInfoProxies;
	GbjObject classProxy, metaclassProxy;
	String className;
	int classVersion, classSize;
	int structSize = 5;

	classInfoProxies = proxiesForMessage_arg_("classVersionInfoForHistory:", anObject);
	holders = new GbxClassHolder[classInfoProxies.length / structSize];
	for (int i = 0; i < classInfoProxies.length; i += structSize) {
		classProxy = classInfoProxies[i];
		className = classInfoProxies[i + 1].stringValue();
		classVersion = classInfoProxies[i + 2].intValue();
		classSize = classInfoProxies[i + 3].intValue();
		metaclassProxy = classInfoProxies[i + 4];
		holders[i / structSize] = new GbxClassHolder(className, classProxy, classVersion, classSize, metaclassProxy);
	}
	return holders;
}

public String sourceStringFor_(GbjObject anObject) throws GbjException
{
	return serverObj.sendMsg("sourceStringFor:", anObject).stringValue();
}

public GbxActivationHolder [] activationHoldersFor_maxLevels_(GbjObject aProcess, int maxLevels) throws GbjException
{
	GbxActivationHolder [] holders;
	GbjObject exceptionInfoProxy;
	GbjObject [] exceptionInfoProxies;
	String aPrintString;
	GbjObject aMethod;
	GbjObject aClass;
	String className;
	GbjObject receiver;
	String aSelector;
	String aProtocol;
	GbjObject args, argNames;
	String sourceString;
	int textPosition;
	int endPoint;

	exceptionInfoProxy = serverObj.sendMsg("stackInfoForProcess:", aProcess, "toLevel:", Integer.valueOf(maxLevels));
	exceptionInfoProxies = new GbjIterator(exceptionInfoProxy).remainingElements();
	holders = new GbxActivationHolder[exceptionInfoProxies.length / 11];
	for (int i = 0; i < exceptionInfoProxies.length; i += 11) {
		aPrintString = exceptionInfoProxies[i].stringValue();
		aMethod = exceptionInfoProxies[i + 1];
		aClass = exceptionInfoProxies[i + 2];
		className = exceptionInfoProxies[i + 3].stringValue();
		receiver = exceptionInfoProxies[i + 4];
		aSelector = exceptionInfoProxies[i + 5].stringValue();
		aProtocol = exceptionInfoProxies[i + 6].stringValue();
		textPosition = exceptionInfoProxies[i + 7].intValue();
		endPoint = exceptionInfoProxies[i + 8].intValue();
		argNames = exceptionInfoProxies[i + 9];
		args = exceptionInfoProxies[i + 10];
		holders[i / 11] = new GbxActivationHolder(aPrintString, aMethod, aPrintString, aClass, className, receiver, aSelector, aProtocol, textPosition, endPoint, argNames, args);
	}
	return holders;
}

public String printStringFor_(GbjObject anObject) throws GbjException
{
	return serverObj.sendMsg("printStringFor:", anObject).stringValue();
}

public GbxObjectHolder [] objectHoldersFor_maxObjects_(GbjObject anObject, int max) throws GbjException
{
	GbxObjectHolder [] holders;
	GbjObject objectInfoProxy;
	GbjObject [] objectInfoProxies;
	GbjObject objProxy;
	GbjObject slotObject;
	int namedSize, idxSize, slotIndex;
	int lastNamedIndex;
	String aString;
	String className;
	GbxObjectHolder selfHolder;

	objectInfoProxy = serverObj.sendMsg("objectInfoFor:", anObject, "max:", Integer.valueOf(max));
	objectInfoProxies = new GbjIterator(objectInfoProxy).remainingElements();
	holders = new GbxObjectHolder[objectInfoProxies.length / 5];
	slotIndex = objectInfoProxies[0].intValue();
	slotObject = objectInfoProxies[1];
	aString = objectInfoProxies[2].stringValue();
	objProxy =objectInfoProxies[3];
	className = objectInfoProxies[4].stringValue();
	holders[0] = new GbxObjectHolder(aString, objProxy, true, slotIndex, slotObject, className);
	selfHolder = holders[0];
	namedSize = objectInfoProxies[5].intValue();
	lastNamedIndex = (namedSize * 5) + 6;
	for (int i = 6; i < lastNamedIndex; i += 5) {
		slotIndex = objectInfoProxies[i].intValue();
		slotObject = objectInfoProxies[i + 1];
		aString = objectInfoProxies[i + 2].stringValue();
		objProxy = objectInfoProxies[i + 3];
		className = objectInfoProxies[i + 4].stringValue();
		holders[i / 5] = new GbxObjectHolder(aString, objProxy, true, slotIndex, slotObject, className);
	}
	idxSize = objectInfoProxies[lastNamedIndex].intValue();
	selfHolder.fetchedSize = idxSize;
	for (int i = lastNamedIndex + 1; i < objectInfoProxies.length - 1; i += 5) {
		slotIndex = objectInfoProxies[i].intValue();
		slotObject = objectInfoProxies[i + 1];
		aString = objectInfoProxies[i + 2].stringValue();
		objProxy = objectInfoProxies[i + 3];
		className = objectInfoProxies[i + 4].stringValue();
		holders[i / 5] = new GbxObjectHolder(aString, objProxy, false, slotIndex, slotObject, className);
	}
	selfHolder.remoteSize = objectInfoProxies[objectInfoProxies.length - 1].intValue();
	return holders;
}

public GbxObjectHolder [] objectHoldersForNames_values_(GbjObject names, GbjObject values) throws GbjException
{
	GbxObjectHolder [] holders;
	GbjObject objectInfoProxy;
	GbjObject [] objectInfoProxies;
	GbjObject objProxy;
	GbjObject slotObject;
	int namedSize, idxSize, slotIndex;
	int lastNamedIndex;
	String aString;
	String className;

	objectInfoProxy = serverObj.sendMsg("objectInfoForNames:", names, "values:", values);
	objectInfoProxies = new GbjIterator(objectInfoProxy).remainingElements();
	holders = new GbxObjectHolder[objectInfoProxies.length / 5];
	for (int i = 0; i < objectInfoProxies.length; i += 5) {
		slotIndex = objectInfoProxies[i].intValue();
		slotObject = objectInfoProxies[i + 1];
		aString = objectInfoProxies[i + 2].stringValue();
		objProxy = objectInfoProxies[i + 3];
		className = objectInfoProxies[i + 4].stringValue();
		holders[i / 5] = new GbxObjectHolder(aString, objProxy, true, slotIndex, slotObject, className);
	}
	return holders;
}

public GbjObject [] infoForClassNamed_(String aClassName) throws GbjException
{
	GbjObject destClass = null;

	destClass = serverObj.sendMsg("infoForClassNamed:", aClassName);
	return new GbjIterator(destClass).remainingElements();
}

public String nameOf_(GbjObject anObject) throws GbjException
{
	return serverObj.sendMsg("nameOf:", anObject).stringValue();
}

public GbjObject thisClass_(GbjObject aClass) throws GbjException
{
	return serverObj.sendMsg("thisClass:", aClass);
}

public GbjObject at_(GbjObject anObject, int index) throws GbjException
{
	return serverObj.sendMsg("object:", anObject, "at:", Integer.valueOf(index));
}

public void fileInString_(String aString) throws GbjException
{
	serverObj.sendMsg("fileInString:", aString);
}

public String [] symbolsMatching_(String sel) throws GbjException
{
	GbjCollection selectorCollectionProxy;

	selectorCollectionProxy = (GbjCollection) serverObj.sendMsg("symbolsMatching:", sel);
	return stringContentsOf_(selectorCollectionProxy);
}

public String [] classNamesMatching_(String aClassName) throws GbjException
{
	GbjCollection classCollectionProxy;

	classCollectionProxy = (GbjCollection) serverObj.sendMsg("classNamesMatching:", aClassName);
	return stringContentsOf_(classCollectionProxy);
}

public String [] stringContentsOf_(GbjCollection collectionProxy)
{
	// 40314: use packed strings to optimize transfer of collections of strings
	String s = serverObj.sendMsg("packStrings:", collectionProxy).stringValue();
	int size = (s.charAt(0) * 256) + s.charAt(1);
	String [] anArray = new String[size];
	int i1 = 2;
	for (int i=0;i<size;i++) {
	    int i2 = s.indexOf('\0',i1);
	    String ss = s.substring(i1,i2);
	    anArray[i] = ss;
	    i1 = i2 + 1;
	}
	return anArray;
}

public GbjObject evaluate_inContext_names_values_printing_(String aString, GbjObject context, GbjObject names, GbjObject values, boolean aBoolean) throws GbjException
{
	return serverObj.sendMsg("evaluate:", aString, "inContext:", context, "names:", names, 
              "values:", values, "printing:", Boolean.valueOf(aBoolean));
}

public String sourceCodeForMethod_(GbjObject aMethod) throws GbjException
{
	return serverObj.sendMsg("sourceStringFor:", aMethod).stringValue();
}

public void resetClassOrganizer() throws GbjException
{
	serverObj.sendMsg("resetClassOrganizer");
}

public GbjObject [] errorInfoFrom_(GbjObject errorArray) throws GbjException
{
	GbjObject infoProxy;

	infoProxy = serverObj.sendMsg("errorInfoFrom:", errorArray);
	return new GbjIterator(infoProxy).remainingElements();
}

public GbjObject instancesOf_(GbjObject aClass) throws GbjException
{
	return serverObj.sendMsg("instancesOf:", aClass);
}

public void migrateInstancesOf_to_(GbjObject fromClass, GbjObject toClass) throws GbjException
{
	serverObj.sendMsg("migrateInstancesOf:", fromClass, "to:", toClass);
}

public void removeFromClassHistory_(GbjObject aClass) throws GbjException
{
	serverObj.sendMsg("removeFromClassHistory:", aClass);
}

public boolean makeClass_aVersionOf_(GbjObject aClass, GbjObject targetClass) throws GbjException
{
	return serverObj.sendMsg("makeClass:", aClass, "aVersionOf:", targetClass).booleanValue();
}

public GbjObject transactionConflicts() throws GbjException
{
	return evaluate_inContext_names_values_printing_("System transactionConflicts", null, null, null, false);
}

public void storeObject_intoObject_atIndex_atKey_isNamedSlot_(GbjObject slotObject, GbjObject anObject, int slot, GbjObject aKey, boolean isNamedSlot) throws GbjException
{
	serverObj.sendMsg("storeObject:", slotObject, "intoObject:", anObject, "atIndex:", 
         Integer.valueOf(slot), "atKey:", aKey, "isNamedSlot:", Boolean.valueOf(isNamedSlot));
}

public String [] instVarNamesInClass_(GbjObject aClass) throws GbjException
{
	GbjCollection aCollection;
	aCollection = (GbjCollection) serverObj.sendMsg("instVarNamesInClass:", aClass);
	return stringContentsOf_(aCollection);
}

public String [] classVarNamesInClass_(GbjObject aClass) throws GbjException
{
	GbjCollection aCollection;
	aCollection = (GbjCollection) serverObj.sendMsg("classVarNamesInClass:", aClass);
	return stringContentsOf_(aCollection);
}

public GbjObject [] proxiesForMessage_(String msg) throws GbjException
{
	GbjObject infoProxy;

	infoProxy = serverObj.sendMsg(msg);
	return new GbjIterator(infoProxy).remainingElements();
}

public GbjObject [] proxiesForMessage_arg_(String msg, Object arg) throws GbjException
{
	GbjObject infoProxy;

	infoProxy = serverObj.sendMsg(msg, arg);
	return new GbjIterator(infoProxy).remainingElements();
}

public GbjObject [] proxiesForMessage_arg_message_arg_(String msg1, Object arg1, String msg2, Object arg2) throws GbjException
{
	GbjObject infoProxy;

	infoProxy = serverObj.sendMsg(msg1, arg1, msg2, arg2);
	return new GbjIterator(infoProxy).remainingElements();
}

public GbjObject [] proxiesForMessage_arg_message_arg_message_arg_(String msg1, Object arg1, String msg2, Object arg2, String msg3, Object arg3) throws GbjException
{
	GbjObject infoProxy;

	infoProxy = serverObj.sendMsg(msg1, arg1, msg2, arg2, msg3, arg3);
	return new GbjIterator(infoProxy).remainingElements();
}

public void trimStackOf_toLevel_(GbjObject aProcess, int aLevel) throws GbjException
{
	serverObj.sendMsg("trimStackOf:", aProcess, "toLevel:", Integer.valueOf(aLevel));
}

public GbjObject continueProcess_(GbjObject aProcess, GbjSession session) throws GbjException
{

	// 45662/46468
        GbjObject stResult = null;
	GbjObject jvResult = new GbjObject( );
	try {
            session.serverContinue(aProcess, false, 
	       stResult, jvResult, null, null);
	} catch (com.gemstone.gbjgci.GbjGciException e) {
	    throw new GbjException(session, e);
	}
	return jvResult;
}

public GbjObject stepOverInProcess_(GbjObject aProcess) throws GbjException
{
	return serverObj.sendMsg("stepOverInProcess:", aProcess);
}

public GbjObject stepOverInProcess_inFrame_(GbjObject aProcess, int aLevel) throws GbjException
{
	return serverObj.sendMsg("stepOverInProcess:", aProcess, "inFrame:", Integer.valueOf(aLevel));
}

public void stepIntoInProcess_(GbjObject aProcess) throws GbjException
{
	serverObj.sendMsg("stepIntoInProcess:", aProcess);
}

public void setBreakIn_at_(GbjObject aMethod, int position) throws GbjException
{
	serverObj.sendMsg("setBreakIn:", aMethod, "at:", Integer.valueOf(position));
}

public GbxMethodHolder [] methodHoldersForBreakpoints() throws GbjException
{
	GbxMethodHolder [] methods;
	GbjObject [] anArray;
	GbjObject method;
	String aSelector;
	GbjObject aClass;
	String className;
	String printString;
	String category;
	int position, endPosition, stepPoint;

	anArray = proxiesForMessage_("breakpointInfo");
	methods = new GbxMethodHolder[anArray.length / 9];
	for (int i = 0; i < anArray.length; i += 9) {
		method = anArray[i];
		aSelector = anArray[i + 1].stringValue();
		aClass = anArray[i + 2];
		className = anArray[i + 3].stringValue();
		printString = anArray[i + 4].stringValue();
		category = anArray[i + 5].stringValue();
		position = anArray[i + 6].intValue();
		endPosition = anArray[i + 7].intValue();
		stepPoint = anArray[i + 8].intValue();
		methods[i / 9] = new GbxMethodHolder(aSelector, method, aClass, className, printString, category, position, endPosition, stepPoint, null);
	}
	return methods;
}

public void removeAllBreaks() throws GbjException
{
	serverObj.sendMsg("removeAllBreaks");
}

public void enableAllBreaks() throws GbjException
{
	serverObj.sendMsg("enableAllBreaks");
}

public void disableAllBreaks() throws GbjException
{
	serverObj.sendMsg("disableAllBreaks");
}

public void removeBreakpoint_inMethod_(int stepPoint, GbjObject aMethod) throws GbjException
{
	serverObj.sendMsg("removeBreakpoint:", Integer.valueOf(stepPoint), "inMethod:", aMethod);
}

public void enableBreakpoint_inMethod_(int stepPoint, GbjObject aMethod) throws GbjException
{
	serverObj.sendMsg("enableBreakpoint:", Integer.valueOf(stepPoint), "inMethod:", aMethod);
}

public void disableBreakpoint_inMethod_(int stepPoint, GbjObject aMethod) throws GbjException
{
	serverObj.sendMsg("disableBreakpoint:", Integer.valueOf(stepPoint), "inMethod:", aMethod);
}

}
