/*=========================================================================
 * Copyright (C) GemTalk Systems 1986-2024.  All Rights Reserved.
 *
 * Name - TopazProcess.java
 *
 * Code for controlling a remote topaz process
 *
 *========================================================================
 */

package com.gemstone.tools;

import java.io.*;
import java.util.*;

public class TopazProcess {

    InputStream in;
    // InputStream err;
    OutputStream out;
    Process topaz;
    String header;
    String prompt = "";

    static final int bufsize = 200;

    // flag to enable debugging code
    public boolean debug = false;

    public TopazProcess(String GSdir, String args) throws IOException {

	// GSdir = $GEMSTONE directory
	// Replace with System.getEnv() when java 1.5 supported

	String executable;
	StringBuffer command;
	Runtime runtime;

	if (System.getProperty("os.name").startsWith("Windows"))
	    executable = GSdir + "\\bin\\topaz";
	else
	    executable = GSdir + "/bin/topaz";

	command = new StringBuffer(executable);
	    
	if (args != null) {
	    command.append(" ");
	    command.append(args);
	}

	runtime = Runtime.getRuntime();
	topaz = runtime.exec(command.toString());
	in = topaz.getInputStream();
	// err = topaz.getErrorStream();
	out = topaz.getOutputStream();
	header = receiveResponse();

    }

    // returns the last topaz prompt
    public String prompt() {
	return prompt;
    }

    // returns the header info
    public String header() {
	return header;
    }

    // write string to topaz, don't wait for response
    public void commandNoWait(String s) throws IOException {
	if (debug) System.out.println("DEBUG no wait command: " + s);
	prompt = "";
	out.write(s.getBytes());
	out.write('\n');
	out.flush();
    }

    // write string to topaz, wait for response
    public String command(String s, int secondsToWait) throws IOException {
	if (debug) System.out.println("DEBUG wait command: " + s);
	prompt = "";
	out.write(s.getBytes());
	out.write('\n');
	out.flush();
	if (s.equals("exit"))
	    return "";  // nothing to return on an exit command
	else
	    return receiveResponse(secondsToWait);
    }

    public String command(String s) throws IOException {
	return command(s, 999999);
    }

    // read until you see "topaz>" or "topaz X>" (where X is a number)
    // or until N seconds have passed..
    private String receiveResponse(int secondsToWait) throws IOException {
	StringBuffer result = new StringBuffer();
	char buffer[];
	int j;
	int seconds;
	int i = 0;
	prompt = "";
	buffer = new char[bufsize];

	// brief 100ms sleep to give time for output to catch up..
	try { Thread.sleep(100); } catch (InterruptedException e) { };

	for (seconds = 1; seconds <= secondsToWait; seconds++) {

	    if (debug) {
		System.out.println("Second " + seconds);
		System.out.println("Available chars: " +
				   in.available());
	    }

	    // now scan available characters..
	    while (in.available() > 0) {
		j = i;
		buffer[i++] = (char)in.read();
		if (debug) {
		    System.out.println("reading " + buffer[j] + 
				       "(" + (char)buffer[j] + ") index " + i);
		    System.out.println("DEBUG available: " + in.available());
		}

		if ((buffer[j] == '\n') || (j >= bufsize-1)) {
		    result.append(buffer, 0, j+1);
		    for (i = 0; i < j; i++) buffer[i] = 0;
		    i = 0;
		} else if (buffer[j] == '>') {
		    // look for prompt..
		    if (debug) System.out.println("Checking for prompt");
		    if (i > 10) continue;
		    if (buffer[0] != 't') continue;
		    if (buffer[1] != 'o') continue;
		    if (buffer[2] != 'p') continue;
		    if (buffer[3] != 'a') continue;
		    if (buffer[4] != 'z') continue;
		    // found the prompt
		    if (debug) 
			System.out.println("Found prompt, read space..");
		    buffer[i++] = (char)in.read();// include space after prompt
		    if (debug) System.out.println("trim buffer");
		    prompt = new String(buffer).trim();
		    if (debug) System.out.println("done");
		    if (debug) {
			System.out.println("DEBUG prompt: |" + prompt + "|");
			System.out.println("DEBUG response: " + result);
		    }
		    return result.toString();
		}
	    }

	    // sleep for a second..
	    if (seconds < secondsToWait) {
		try { 
		    Thread.sleep(1000);
		} catch (InterruptedException e) { /* don't worry about it */ }
	    }
        }
	return result.toString();
    }

    // read until you see "topaz>" or "topaz X>" (where X is a number)
    private String receiveResponse() throws IOException {
	return receiveResponse(999999);
    }

}







